import sys

from math import *
from vector import *
from tensor import *
from random_tensor import *


class TensorGridElement:
    """ A tensor which is random until given a value. """
    
    def __init__(self, point, tensor = None,
                 rand_mag_max = 1.0, rand_mag_min = 0.1):
        self.point = point
        self.tensor = tensor

        self.rand_mag_max = rand_mag_max
        self.rand_mag_min = rand_mag_min

    def pointStr(self, p):
        s = '(%f' % p[0]
        for x in p[1:]:
            s += ', %f' % x
        s += ')'

        return s

    def __str__(self):
        s  = self.pointStr(self.point)
        s += '\n\n'

        if self.tensor:
            s += str(self.tensor)
        else:
            s += str(randomTensor(len(self.point),
                                  mag_max = self.rand_mag_max,
                                  mag_min = self.rand_mag_min))
        
        return s

    def __repr__(self):
        return str(self)

class TensorGrid:
    def __init__(self, size, origin, delta,
                 bg_max = 1.0, bg_min = 0.1):
        self.dim = len(size)
        
        self.size = size
        self.origin = origin
        self.delta = delta

        self.elts = []
        
        for p in self.gridPoints(size, origin, delta):
            self.elts.append(TensorGridElement(p,
                                               rand_mag_max = bg_max,
                                               rand_mag_min = bg_min))
        
    def gridPoints(self, size, origin, delta):
        if len(size) == 0:
            yield []
            return
        
        for p in self.gridPoints(size[1:], origin[1:], delta[1:]):
            for i in range(size[0]):
                q = p + [i * delta[0] + origin[0]]
                yield q

    def elt(self, *i):
        """ Retrieve an element by its multidimensional index into the grid. """
        
        n = len(self.size)
        
        def offset(i, j, base, prod):
            if j < 0:
                return base
            return offset(i, j - 1, base + prod * i[j], prod * self.size[j])

        return self.elts[offset(i, n - 1, 0, 1)]

    def addLine(self, p1, p2, strength = 1.0, noise = 1e-1):
        """ Add tensor in voxels along the line p1-p2 in that direction. """
        
        def roundp(p):
            return tuple([int(round(x)) for x in p])
        
        # Find the vector between p1 and p2
        v = Vector(self.dim, [p2[i] - p1[i] for i in range(self.dim)])
        u = v.unitVector()

        # Find largest element of v, divide by that the get the increment
        i, mag = max(enumerate(v), key = lambda k: abs(k[1]))
        v /= abs(mag)
        steps = int(ceil(abs(p2[i] - p1[i])))

        p = p1
        for j in range(steps + 1):
            if j != 0:
                # Advance to next point
                p = [p[i] + v[i] for i in range(self.dim)]
            
            # Find grid position
            pi = roundp(p)
            in_bounds = True
            for k in range(self.dim):
                if pi[k] < 0 or pi[k] >= self.size[k]:
                    in_bounds = False
                    break

            if not in_bounds:
                continue
            
            # Build tensor
            T = dyadicTensor(u)
            T += randomTensor(self.dim) * noise

            # Place in grid
            elt = self.elt(*roundp(p))

            if elt.tensor == None:
                elt.tensor = T
            else:
                elt.tensor += T

    def addLineBundle(self, p1, p2, width, spacing = 1.0, **kwargs):
        if self.dim == 2:
            p1 = Vector(self.dim, init = p1)
            p2 = Vector(self.dim, init = p2)

            v = p2 - p1
            v = v.unitVector()
            v = Vector(v.dim(), init = (-v[1], v[0]))

            offset = -width/2
            while offset < width/2 + 1:
                self.addLine(p1 + v * offset, p2 + v * offset, **kwargs)
                offset += spacing
        else:
            p1 = Vector(self.dim, init = p1)
            p2 = Vector(self.dim, init = p2)

            v = p2 - p1
            v = v.unitVector()
            
            # TODO

    def normalize(self):
        """ Normalize tensors so principal eigenvalue is 1. """

        # TODO

        pass

    def __iter__(self):
        return self.elts.__iter__()

    def write(self, file = sys.stdout):
        file.write('dim %d\n\n' % self.dim)
        
        for T in self:
            file.write(str(T))
            file.write('\n\n')

