# Astxx, the Asterisk C++ API and Utility Library.
# Copyright (C) 2005, 2006  Matthew A. Nicholson
# Copyright (C) 2006  Tim Blechmann
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 2.1 as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# See http://www.scons.org/wiki/DoxygenBuilder
# Additional modifications (C) 2009 Maxwell Collins

import os, os.path
import subprocess
import glob
import copy
from fnmatch import fnmatch
from StringIO import StringIO


def DoxyfileParse(file_contents, verbose = False):
   """
   Parse a Doxygen source file and return a dictionary of all the values.
   Values will be strings and lists of strings.
   """
   data = {}

   import shlex
   lex = shlex.shlex(instream = file_contents, posix = True)
   lex.wordchars += "*+./-:"
   lex.whitespace = lex.whitespace.replace("\n", "")
   lex.escape = ""

   lineno = lex.lineno
   token = '\n'
   key = token   # the first token should be a key
   last_token = ""
   key_token = False
   next_key = False
   new_data = True

   def append_data(data, key, new_data, token):
      if new_data or len(data[key]) == 0:
         data[key].append(token)
      else:
         data[key][-1] += token

   while token:
      if token in ['\n']:
         if last_token not in ['\\']:
            key_token = True
      elif token in ['\\']:
         pass
      elif key_token:
         key = token
         key_token = False
      else:
         if token == "+=":
            if key not in data:
               data[key] = []
         elif token == "=":
            data[key] = []
         else:
            append_data( data, key, new_data, token )
            new_data = True

      last_token = token
      token = lex.get_token()

      if last_token == '\\' and token != '\n':
         new_data = False
         append_data( data, key, new_data, '\\' )

   # compress lists of len 1 into single strings
   for k, v in data.items():
      if len(v) == 0:
         data.pop(k)

      # items in the following list will be kept as lists and not converted to strings
      if k in ["INPUT", "FILE_PATTERNS", "EXCLUDE_PATTERNS"]:
         continue

      if len(v) == 1:
         data[k] = v[0]

   return data

class Doxyfile:
    LIST_ITEMS = ["INPUT", "FILE_PATTERNS", "EXCLUDE_PATTERNS"]

    def __init__(self, **values):
        # Get + parse default doxygen file
        proc = subprocess.Popen(['doxygen', '-g', '-'],
                                stdout = subprocess.PIPE)

        self.data = DoxyfileParse(proc.stdout.read())

        for i, v in values.items():
            self[i] = v

    def __getitem__(self, i):
        return self.data[i]

    def __setitem__(self, i, v):
        self.data[i] = self.sanitize(v)

    def __delitem__(self, i):
        del self.data[i]

    def sanitize(self, v):
        if self.isiterable(v):
            v = [self.sanitize(w) for w in self.flatten(v)]
            return v

        if type(v) == bool:
            if v:
                v = 'YES'
            else:
                v = 'NO'

        return str(v)

    def isiterable(self, v):
        if type(v) == list:
            return True
        elif type(v) == str:
            return False

        try:
            if '__iter__' in vars(v):
                return True
        except:
            pass

        return False

    def flatten(self, v):
        if self.isiterable(v):
            for w in v:
                for x in self.flatten(w):
                    yield x
        else:
            yield(v)

    def write(self, f):
        for i, v in self.data.items():
            f.write(i)
            f.write(' = ')

            if i in self.LIST_ITEMS:
                for w in v:
                    f.write(str(w))
                    f.write(' ')
            else:
                v = str(v)

                if ' ' in v or '\t' in v:
                    f.write('"')
                    f.write(v)
                    f.write('"')
                else:
                    f.write(v)

            f.write('\n')

    def __str__(self):
        buf = StringIO()
        self.write(buf)
        return buf.getvalue()

    def clone(self):
        return copy.deepcopy(self)

