import os, os.path

from source_list import SourceList

import static_library
from static_library import StaticLibrary

import program
from program import Program

import documentation
from documentation import Documentation

from tar_prefix import TarPrefix

class Project:
    def __init__(self, name, version, env, dir = '.'):
        self.name = name
        self.version = version

        self.env = env
        self.dir = env.Dir(dir)

        self.build_scripts = SourceList(env)
        self.data = SourceList(env)
        self.info_files = SourceList(env)

        self.targets = {}

    def build(self, **kwargs):
        for target in self.targets.values():
            target.build(**kwargs)

    def ext(self, s):
        return os.path.splitext(str(s))[1][1:]

    def targetFilter(classes):
        def filter(self):
            for target in self.targets.values():
                out = False
                for c in classes:
                    if isinstance(target, c):
                        out = True
                if out:
                    yield target

        return filter

    staticLibraries = targetFilter([static_library.StaticLibrary])
    programs = targetFilter([program.Program])
    libraries = targetFilter([static_library.StaticLibrary])
    binaries = targetFilter([program.Program, static_library.StaticLibrary])
    docs = targetFilter([documentation.Documentation])

    def CSources(self):
        for component in self.targets.values():
            for s in component.sources:
                if self.ext(s) == 'c':
                    yield s

    def allSources(self):
        for component in self.targets.values():
            for s in component.headers:
                yield s
            for s in component.sources:
                yield s

    def allFiles(self):
        for component in self.components():
            for f in component.allFiles():
                yield f

        for f in self.build_scripts:
            yield f

        for f in self.info_files:
            yield f

        for f in self.data:
            yield f

    def components(self):
        return self.targets.values()

    def srcTar(self):
        distdir = os.path.join(str(self.dir), 'dist')

        # Construct names
        tar_prefix = TarPrefix('%s-%s/' % (self.name, self.version))
        tar_name = '%s-src-%s.tar.bz2' % (self.name, self.version)
        tar_name = os.path.join(distdir, tar_name)

        # Collect sources
        srcs = list(self.allFiles())
        srcs += list([self.env.Dir(s) for s in self.docs()])

        # Create the tar
        tar = self.env.PrefixedTar(tar_name, srcs + [tar_prefix])

        return tar

    def binTar(self):
        distdir = os.path.join(str(self.dir), 'dist')

        # Construct names
        tar_prefix = TarPrefix('%s-%s/' % (self.name, self.version))
        tar_name = '%s-bin-%s.tar.bz2' % (self.name, self.version)
        tar_name = os.path.join(distdir, tar_name)

        # Collect binaries
        bins = [str(t) for t in self.binaries() if not t.internal]
        bins += self.info_files()
        bins += self.data()

        # Create the tar
        tar = self.env.PrefixedTar(tar_name, bins + [tar_prefix])

        return tar

    def StaticLibrary(self, name, **kwargs):
        libdir = os.path.join(str(self.dir), 'lib')

        lib = StaticLibrary(self.env, name, dir = libdir, **kwargs)
        self.targets[name] = lib

        return lib

    def Program(self, name, **kwargs):
        bindir = os.path.join(str(self.dir), 'bin')

        prog = Program(self.env, name, dir = bindir, **kwargs)
        self.targets[prog] = prog

        return prog

    def Documentation(self, name, doxyfile, **kwargs):
        docdir = os.path.join(str(self.dir), 'doc')

        doc = Documentation(self.env, name, doxyfile, dir = docdir, **kwargs)

        self.targets[name] = doc
        return doc
        
