from SCons.Script import *
import os, os.path

from nifti import NiftiImage
import numpy

def normalizeInput(env, input):
    if type(input) == str or type(input) == instance and 'path' in vars(input):
        input = [input]
    for i in range(len(input)):
        if type(input[0]) == str:
            input[0] = env.File(input[0])

    return input

def niiSuffix(nii, suffix):
    f = nii.path
    i = f.find('.')
    f = f[:i] + suffix + f[i:]
    return f

def niiRoot(nii):
    f = nii.path
    i = f.find('.')
    return f[:i]

def FractionalAnisotropy(source, target, env, for_signature):
    intermediate = niiSuffix(source[0], '_fa')
    return ['TVAnisotropy -in $SOURCE', Move('$TARGET', intermediate)]

segment_wm_suffixes = []
for i in range(3):
    segment_wm_suffixes.append('-seg-%d.nii.gz' % i)
    segment_wm_suffixes.append('-pve-%d.nii.gz' % i)
segment_wm_suffixes.append('-seg.nii.gz')
segment_wm_suffixes.append('-pveseg.nii.gz')
segment_wm_suffixes.append('-mixeltype.nii.gz')

def SegmentWM(source, target, env, for_signature):
    root = target[0].path

    for suffix in segment_wm_suffixes:
        if root[-len(suffix):] == suffix:
            root = root[:-len(suffix)]

    cmd = ['$FSL_FAST -n 3 -t 2 -g -v -o %s $SOURCE' % root]

    for suffix in segment_wm_suffixes:
        underscore_suffix = suffix.replace('-', '_')
        cmd.append(Move(root + suffix, root + underscore_suffix))

    return cmd

def SegmentWMEmitter(target, source, env):
    root = os.path.basename(target[0].path)
    target = [env.File(root + suffix) for suffix in segment_wm_suffixes]
    return target, source

def NiftiOr(source, target, env, for_signature):
    cmd = '$FSL_MATHS '
    cmd += ' -add '.join([s.path for s in source])
    cmd += ' $TARGET'

    cmd = [cmd]
    cmd.append('$FSL_MATHS $TARGET -bin $TARGET')

    return cmd

def voxIter(img):
    size = img.data.shape
    i = [0] * len(size)

    while True:
        # Iterate to next coordinate
        j = -1
        while i[j] == size[j]-1 and j > -len(size):
            i[j] = 0
            j -= 1
        i[j] += 1

        if i[0] == size[0]:
            break

        t = tuple(i)

        #print t, size
        yield t, img.data[t]

def NiftiBorder(source, target, env):
    source = NiftiImage(source = source[0].path, load = True)
    dims = source.data.shape

    target = target[0].path
    border = numpy.zeros(dims)

    # Build appropriate neighborhood
    neighborhood = []
    for i in range(len(dims)):
        j = [0]*len(dims)
        j[i] = 1
        neighborhood.append(j)

        j = [0]*len(dims)
        j[i] = -1
        neighborhood.append(j)
            
    # Mark border
    for i, v in voxIter(source):
        if v != 0:
            for n in neighborhood:
                j = [sum(x) for x in zip(i, n)]
                if source.data[tuple(j)] == 0:
                    border[i] = 1

    border = NiftiImage(source = border)
    border.header = source.header
    border.save(filename = target, filetype = 'NIFTI_GZ')
            
def NiftiRectAction(source, target, env):
    for s in source:
        if 'path' in vars(s):
            source = NiftiImage(source = s.path, load = True)
        else:
            min, max = s.value

    for i in range(len(max)):
        if max[i] < 0:
            max[i] = source.data.shape[i]


    out = numpy.zeros(source.data.shape)
    for i, v in voxIter(source):
        if v != 0:
            bounds = True
            for j in range(len(i)):
                if i[j] < min[j] or i[j] > max[j]:
                    bounds = False

            if bounds:
                out[i] = 1

    out = NiftiImage(source = out)
    out.header = source.header
    out.save(filename = target[0].path, filetype = 'NIFTI_GZ')

def NiftiRect(env, target, source, **kwargs):
    min = [0, 0, 0]
    max = [-1, -1, -1]

    for dim, value in kwargs.items():
        dim, limit = dim[0], dim[1:]
        dim = ['x', 'y', 'z'].index(dim)
        if limit == 'min':
            min[dim] = value
        else:
            max[dim] = value

    return env.NiftiRectBuilder(target, source + [env.Value((min, max))])

def generate(env):
    FABuilder = env.Builder(generator = FractionalAnisotropy)
    WMBuilder = env.Builder(generator = SegmentWM,
                            emitter = SegmentWMEmitter)
    OrBuilder = env.Builder(generator = NiftiOr)
    BorderBuilder = env.Builder(action = NiftiBorder, single_source = True)
    RectBuilder = env.Builder(action = NiftiRectAction)

    env.Append(BUILDERS = {'FractionalAnisotropy': FABuilder,
                           'SegmentWM': WMBuilder,
                           'NiftiOr': OrBuilder,
                           'NiftiBorder': BorderBuilder,
                           'NiftiRectBuilder': RectBuilder})

    env.AddMethod(NiftiRect, 'NiftiRect')

def exists(env):
    e = env.Detect('TVAnisotropy')
    return e
