# Loosely based on GCH Builder by Tim Blechmann,
# http://www.scons.org/wiki/GchBuilder?action=recall&rev=6

import os, os.path
import shlex

from SCons.Script import *

C_HEADER_SUFFIXES = set(['.h'])
CXX_HEADER_SUFFIXES = set(['.hh', '.H', '.hpp', '.hxx', '.h++'])
C_SOURCE_SUFFIXES = set(['.c'])
CXX_SOURCE_SUFFIXES = set(['.cc', '.C', '.cpp', '.cxx', '.c++'])

def command_factory(shared):
    def command(source, target, env, for_signature):
        source = source[0]
        target = target[0]

        if source.get_suffix() in C_HEADER_SUFFIXES:
            if shared:
                cmd = env['SHCCCOM']
                cmd = cmd.replace('$SHCFLAGS', '$SHCPCHFLAGS')
            else:
                cmd = env['CCCOM']
                cmd = cmd.replace('$CFLAGS', '$CPCHFLAGS')

        elif source.get_suffix() in CXX_HEADER_SUFFIXES:
            if shared:
                cmd = env['SHCXXCOM']
                cmd = cmd.replace('$SHCXXFLAGS', '$SHCXXPCHFLAGS')
            else:
                cmd = env['CXXCOM']
                cmd = cmd.replace('$CXXFLAGS', '$CXXPCHFLAGS')

        return cmd

    return command

def suffix(env, sources):
    return sources[0].get_suffix() + '.gch'

# Create emitter to append onto C
def emitter_factory(shared):
    def emitter(target, source, env):
        if shared:
            SCons.Defaults.SharedObjectEmitter(target, source, env)
        else:
            SCons.Defaults.StaticObjectEmitter(target, source, env)

        # Perform standard C scanning
        # TODO: breaks sconsign...?
        #scanner = CScanner
        #deps = scanner(source[0], env, scanner.path(env))
        #print [str(f) for f in deps]
        deps = []

        # Add dependency on any -include flags
        if source[0].get_suffix() in C_SOURCE_SUFFIXES:
            if shared:
                flags = env['SHCFLAGS'] + env['SHCCFLAGS']
                gchkey = 'SHGCH'
            else:
                flags = env['CFLAGS'] + env['CCFLAGS']
                gchkey = 'GCH'
        elif source[0].get_suffix() in CXX_SOURCE_SUFFIXES:
            if shared:
                flags = env['SHCXXFLAGS']
                gchkey = 'SHGCHXX'
            else:
                flags = env['CXXFLAGS']
                gchkey = 'GCHXX'

        for i in range(len(flags) - 1):
            if flags[i] == '-include':
                for dir in env['CPPPATH'] + ['.']:
                    file = os.path.join(str(dir), str(flags[i+1]))

                    if os.path.exists(file):
                        deps.append(env.File(file))
                        break

        # Add dependency on precompiled header
        deps = [str(f) for f in deps]

        if env.has_key(gchkey):
            for gch in env[gchkey]:
                if str(gch.sources[0]) in deps:
                    env.Depends(target, gch)

        return (target, source)

    return emitter

def generate(env):
    # Add precompiled header builders
    builder = env.Builder(generator = command_factory(False),
                          suffix = suffix,
                          single_source = True)
    shared_builder = env.Builder(generator = command_factory(True),
                                 suffix = suffix,
                                 single_source = True)

    env.Append(BUILDERS = {'PrecompiledHeader': builder})
    env.Append(BUILDERS = {'SharedPrecompiledHeader': shared_builder})

    # Add PCHflags
    env['SHCPCHFLAGS'] = ['$SHCFLAGS']
    env['CPCHFLAGS'] = ['$CFLAGS']
    env['SHCXXPCHFLAGS'] = ['$SHCXXFLAGS']
    env['CXXPCHFLAGS'] = ['$CXXFLAGS']

    # Add new emitters for C source files
    for suf in C_SOURCE_SUFFIXES.union(CXX_SOURCE_SUFFIXES):
        env['BUILDERS']['StaticObject'].add_emitter(suf, emitter_factory(False))
        env['BUILDERS']['SharedObject'].add_emitter(suf, emitter_factory(True))

    # Global lists of GCHs
    env['GCH'] = []
    env['SHGCHS'] = []
    env['GCHXX'] = []
    env['SHGCHXX'] = []

    # Automatic inclusion of GCHs
    def gchSource(gch_list):
        return [str(gch.sources[0]) for gch in gch_list]

    env['INCLUDEPREFIX'] = '-include '
    env['_gchsource'] = gchSource
    env['INCLUDESUFFIX'] = ''

    env.Append(CFLAGS = ['${_concat(INCLUDEPREFIX, GCH, INCLUDESUFFIX, __env__, _gchsource, TARGET, SOURCE)}'])
    env.Append(CXXFLAGS = ['${_concat(INCLUDEPREFIX, GCHXX, INCLUDESUFFIX, __env__, _gchsource, TARGET, SOURCE)}'])
    env.Append(SHCFLAGS = ['${_concat(INCLUDEPREFIX, SHGCH, INCLUDESUFFIX, __env__, _gchsource, TARGET, SOURCE)}'])
    env.Append(SHCXXFLAGS = ['${_concat(INCLUDEPREFIX, SHGCHXX, INCLUDESUFFIX, __env__, _gchsource, TARGET, SOURCE)}'])


def exists(env):
    return env.Detect('gcc') and env.Detect('g++')
