#include "bent_weight.h"
#include "util.h"
#include "tensor_eigens.h"

#include "monte_carlo.h"

const bent_weight_options_t bent_weight_options_defaults = {
  4.0, // sigma
  1e1, // angle_weight

  1024, // expectation_samples
};

// p position
// v unit direction
// s arc length
// kappa curvature
// vote unit direction at origin
static void osculating_circle(
  int dim,
  point_t p, point_t v,
  double *s, double *kappa, point_t vote)
{
	coord_t u[dim];
	double l, theta, dot;
	int i;

	l = point_l2_norm(dim, p);

  theta = acos(point_dot_product(dim, p, v) / l);
	theta = MIN(theta, M_PI - theta);

	// Arc length
	if(fabs(theta) < 1e-5 || fabs(theta - M_PI) < 1e-5) {
		*s = l;
	}
	else {
		*s = theta * l / sin(theta);
	}

	// curvature
	*kappa = 2 * sin(theta) / l;

	// Vote (reflect orientation across p)
	for(i = 0; i < dim; i++) {
    u[i] = p[i] / l;
	}
	dot = point_dot_product(dim, v, u);

	for(i = 0; i < dim; i++) {
		vote[i] = 2 * dot * u[i] - v[i];
	}
}

static double circle_weight(
  int dim, bent_weight_options_t *opts,
  point_t points[3], tensor_t *tensors[3],
  point_t directions[2])
{
	point_t p[2];
	tensor_t *T[2], *selfT;
	double sigma;
	double angle_weight;

  tensor_eigens_t *eig[2];

	double s[2], kappa[2];
	coord_t vote[2][dim];
	double c;
	double anisotropy[2], prob[2];

	double angle;
	double total_anisotropy, total_s, total_kappa, total_prob;

	double curve_weight;

	int i;

	// Reorganize arguments
	p[0] = points[0];
	p[1] = points[2];

	T[0] = tensors[0];
	T[1] = tensors[2];

	selfT = tensors[1];

	sigma = opts->sigma;
	angle_weight = opts->angle_weight;

	// Curvature weighting
	// from Mordohai06
	c = (-16 * log(0.1) * (sigma - 1)) / square(M_PI);

	for(i = 0; i < 2; i++) {
		// Find anisotropy of neighbor tensors
		eig[i] = tensor_eigens_get(T[i]);
		anisotropy[i] = tensor_eigenvalue(eig[i], 0) - tensor_eigenvalue(eig[i], 1);

		// Cast votes
		osculating_circle(dim, p[i], directions[i], &s[i], &kappa[i], vote[i]);

		// Find self probabilities
		prob[i] = tensor_mult2(selfT, vote[i]);
	}

	// Find angle between votes
	angle = acos(point_dot_product(dim, vote[0], vote[1]));
	
	// Totals
	total_anisotropy = anisotropy[0] + anisotropy[1];
	total_s = s[0] + s[1];
	total_kappa = kappa[0] + kappa[1];

	//printf("anisotropy = %2g\n", total_anisotropy);
	//printf("s = %2g\n", total_s);
	//printf("kappa = %2g\n", total_kappa);
	//printf("angle_weight * angle = %2g\n", angle_weight * angle);
	//printf("c = %2g\n", c);

	// Weight due to curves
	curve_weight = (square(total_s) + c * square(total_kappa)) / sigma;
	curve_weight -= total_anisotropy;
	curve_weight += angle_weight * angle;

	curve_weight = exp(-curve_weight);

	/*
	if(fabs(tensor_elt_get(&tensors[1], 1, 1)) > 1 && fabs(points[2][1] - points[0][1]) > 1.5) {
		printf("s = %2f, %2f\n", s[0], s[1]);
		printf("kappa = %2f, %2f\n", kappa[0], kappa[1]);
		printf("curve_weight = %2f\n", curve_weight);
		printf("prob = %2f\n", prob[0] * prob[1]);
	}
	*/

	return prob[0] * prob[1] * curve_weight;
}

// Weights assigned to contour going through 0-1-2
double bent_weight(
  int dim, weight_options_t *weight_opts,
  point_t points[3], tensor_t *tensors[3])
{
	double w;
	bent_weight_options_t *opts;

	coord_t direction_coords[2][dim];
  point_t directions[2];

	int i;

	opts = (bent_weight_options_t *)weight_opts->closure_data;

  // Find principal vectors
	for(i = 0; i < 2; i++) {
    directions[i] = direction_coords[i];
    tensor_principal_vector(tensors[2*i], directions[i]);
	}

	w = circle_weight(dim, opts, points, tensors, directions);

	return w;
}

typedef struct bent_weight_closure bent_weight_closure_t;
struct bent_weight_closure {
	bent_weight_options_t *opts;

  int dim;
	point_t points[3];
	tensor_t *tensors[3];
};

static double bent_weight_closure(void *closure_data, double *x)
{
	bent_weight_closure_t *closure;
	bent_weight_options_t *opts;
	point_t *points;
	tensor_t **tensors;
	double theta, phi;

	point_t directions[2];
	double weight, prob;

	tensor_eigens_t *eig;
	coord_t eigsum;

	int i, j, dim;

	// Unpack closure
	closure = (bent_weight_closure_t *)closure_data;
	opts = closure->opts;
  dim = closure->dim;
	points = closure->points;
	tensors = closure->tensors;

	// Calculate directions and find corresponding probability
  coord_t direction_coords[2][dim];

	prob = 1;
	for(i = 0; i < 2; i++) {
    directions[i] = direction_coords[i];

		if(dim == 2) {
			theta = x[i];

			directions[i][0] = cos(theta);
			directions[i][1] = sin(theta);
		}
		else {
			assert(dim == 3);

			theta = x[2*i];
			phi = x[2*i+1];

			directions[i][0] = cos(theta) * sin(phi);
			directions[i][1] = sin(theta) * sin(phi);
			directions[i][2] = cos(phi);
			
		}

		prob *= tensor_mult2(tensors[2*i], directions[i]);

		// Normalize probability
		eigsum = 0;
		eig = tensor_eigens_get(tensors[2*i]);
		for(j = 0; j < dim; j++) {
			eigsum += tensor_eigenvalue(eig, j);
		}
		prob /= eigsum;
	}

	weight = circle_weight(dim, opts, points, tensors, directions);

	/*
	if(weight > max_circle) {
		max_circle = weight;
		printf("(%f %f %f %f)\n", x[0], x[1], x[2], x[3]);
		printf("%2g\n", max_circle);
	}
	*/

	//printf("f = %2g * %2g = %2g\n", prob, weight, prob * weight);

	return prob * weight;
}

double bent_weight_expectation(
  int dim, weight_options_t *weight_opts,
  point_t points[3], tensor_t *tensors[3])
{
	double w;
	bent_weight_options_t *opts;
	bent_weight_closure_t closure;

	assert(dim == 2 || dim == 3);
	double min[2 * (dim - 1)], max[2 * (dim - 1)];

	int i;

	opts = (bent_weight_options_t *)weight_opts->closure_data;

  // Fill integration closure
	closure.opts = opts;

  closure.dim = dim;

  for(i = 0; i < 3; i++) {
    closure.points[i] = points[i];
    closure.tensors[i] = tensors[i];
  }

  // Integration bounds
	for(i = 0; i < 2; i++) {
		if(dim == 2) {
			min[i] = 0;
			max[i] = 2*M_PI;
		}
		else {
			assert(dim == 3);

			min[2*i] = 0;
			max[2*i] = 2 * M_PI;

			min[2*i+1] = 0;
			min[2*i+1] = M_PI;
		}
	}

  // Integrate!
	w = integrate_monte_carlo(
		2 * (dim - 1), &bent_weight_closure, min, max,
		opts->expectation_samples, &closure);

	//printf("w = %2g\n", w);

	return w;
}
