#include "bspline.h"

#include <gsl/gsl_bspline.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_statistics.h>

struct bspline {
	int dim, order;
	size_t ncoeffs;

	gsl_bspline_workspace **bw;
	gsl_vector **coeff;
  gsl_matrix **cov;

	gsl_vector *eval;

	double *yerr;
};

bspline_t *bspline_create(path_t *control_points, int order)
{
	bspline_t *b;
	int i, j, k, n, dim;
	size_t ncoeffs = 0, nbreak;
	double *t;
	gsl_vector *w = NULL, *y = NULL;
  gsl_matrix *X = NULL;
	gsl_multifit_linear_workspace *mw = NULL;
	point_t p, q;
  double chisq;

	dim = path_dim(control_points);
	n = path_length(control_points);
  nbreak = n/5;

	if(nbreak > n + 2 - order) {
		nbreak = n + 2 - order;
	}
	if(nbreak < 2) {
		return NULL;
	}

	// Spline structure
	b = (bspline_t *)malloc(sizeof(bspline_t));
  if(b == NULL) {
    return NULL;
  }
	b->dim = dim;
	b->order = order;
	
	b->yerr = (double *)malloc(sizeof(double) * dim);


	// Set up splines
  b->bw = (gsl_bspline_workspace **)malloc(sizeof(gsl_bspline_workspace *) * dim);
  b->coeff = (gsl_vector **)malloc(sizeof(gsl_vector *) * dim);
  b->cov = (gsl_matrix **)malloc(sizeof(gsl_matrix *) * dim);

	// Construct unit-speed parameterization of
	// a piecewise linear function between the control points.
	t = (double *)malloc(sizeof(double) * n);

	p = path_get(control_points, 0);
	t[0] = 0;

	for(i = 1; i < n; i++) {
		q = path_get(control_points, i);

		t[i] = t[i-1] + point_l2_dist(dim, p, q);

		p = q;
	}
	for(i = 1; i < n; i++) {
		t[i] /= t[n-1];
	}

  for(i = 0; i < dim; i++) {
    // Allocate spline for dimension i
    b->bw[i] = gsl_bspline_alloc(order, nbreak);
    if(i == 0) {
      ncoeffs = b->ncoeffs = gsl_bspline_ncoeffs(b->bw[0]);
      b->eval = gsl_vector_alloc(ncoeffs);

      y = gsl_vector_alloc(n);
      w = gsl_vector_alloc(n);
      X = gsl_matrix_alloc(n, ncoeffs);

      mw = gsl_multifit_linear_alloc(n, ncoeffs);
    }

    b->coeff[i] = gsl_vector_alloc(ncoeffs);
    b->cov[i] = gsl_matrix_alloc(ncoeffs, ncoeffs);

    gsl_bspline_knots_uniform(0, 1, b->bw[i]);

    // Data
    for(j = 0; j < n; j++) {
      p = path_get(control_points, j);
      gsl_vector_set(y, j, p[i]);
      gsl_vector_set(w, j, 1.0);
    }

    // Construct fit matrix
    for(j = 0; j < n; j++) {
      gsl_bspline_eval(t[j], b->eval, b->bw[i]);

      for(k = 0; k < ncoeffs; k++) {
        gsl_matrix_set(X, j, k, gsl_vector_get(b->eval, k));
      }
    }

    // Fit the splines to the data
    gsl_multifit_wlinear(X, w, y, b->coeff[i], b->cov[i], &chisq, mw);
  }


  // Cleanup
	gsl_multifit_linear_free(mw);
  gsl_matrix_free(X);
  gsl_vector_free(y);
  gsl_vector_free(w);
	free(t);

  return b;
}

void bspline_destroy(bspline_t *b)
{
  int i, dim;

	if(b == NULL) {
		return;
	}
	
  dim = b->dim;

  for(i = 0; i < dim; i++) {
    gsl_matrix_free(b->cov[i]);
    gsl_vector_free(b->coeff[i]);
    
    gsl_bspline_free(b->bw[i]);
  }
  free(b->cov);
  free(b->coeff);
  free(b->bw);

	free(b->yerr);
	free(b);
}

// Accessors
int bspline_dim(bspline_t *b)
{
	return b->dim;
}

// Calculation
void bspline_calculate(bspline_t *b, double t, point_t out)
{
	int i, dim;

	dim = b->dim;

  for(i = 0; i < dim; i++) {
    gsl_bspline_eval(t, b->eval, b->bw[i]);
    gsl_multifit_linear_est(b->eval, b->coeff[i], b->cov[i],
                            out + i, b->yerr + i);
  }
}

path_t *bspline_calculate_path(bspline_t *b, int N)
{
	int i;
	double t;
	point_t p;
	path_t *path;

	path = path_create(b->dim, N);

	for(i = 0; i < N; i++) {
		t = (double)i / (N - 1);
		p = path_get(path, i);

		bspline_calculate(b, t, p);
	}

	return path;
}
