#include "bundle.h"
#include "util.h"
#include "affine.h"
#include "util.h"

struct bundle {
	int dim, count;
};

static inline path_t **bundle_data(bundle_t *b)
{
	intptr_t p;

	p = (intptr_t)b;
	p += sizeof(bundle_t);

	return (path_t **)p;
}

bundle_t *bundle_create(int dim, int count)
{
	size_t s;
	bundle_t *b;
	int i;
	path_t **data;

	// Allocate
	s = sizeof(bundle_t *) + sizeof(path_t *) * count;
	b = (bundle_t *)malloc(s);

	b->dim = dim;
	b->count = count;

	// Zero out paths
	data = bundle_data(b);
	for(i = 0; i < count; i++) {
		data[i] = NULL;
	}

	return b;
}

bundle_t *bundle_copy(bundle_t *src)
{
	bundle_t *dst;
	int i, n;
	path_t *p;

	n = src->count;
	dst = bundle_create(src->dim, n);

	for(i = 0; i < n; i++) {
		p = bundle_get(src, i);
		p = path_copy(p);
		bundle_set(dst, i, p);
	}

	return dst;
}

bundle_t *bundle_resize(bundle_t *b, int new_count)
{
	size_t s;
	int i, old_count;

	old_count = b->count;

	// Free old
	for(i = new_count; i < old_count; i++) {
		bundle_remove(b, i);
	}

	// Reallocate
	s = sizeof(bundle_t) + sizeof(path_t *) * new_count;
	b = (bundle_t *)realloc(b, s);
	b->count = new_count;

	// Zero out new
	for(i = old_count; i < new_count; i++) {
		bundle_data(b)[i] = NULL;
	}

	return b;
}

void bundle_destroy(bundle_t *b)
{
	int i, n;

	if(b == NULL) {
		return;
	}

	n = b->count;
	for(i = 0; i < n; i++) {
		path_destroy(bundle_get(b, i));
	}

	free(b);
}

// Accessors
int bundle_dim(bundle_t *b)
{
	return b->dim;
}

int bundle_count(bundle_t *b)
{
	return b->count;
}

// Members
path_t *bundle_get(bundle_t *b, int i)
{
	return bundle_data(b)[i];
}

void bundle_set(bundle_t *b, int i, path_t *p)
{
	path_t **pp;

	assert(b->dim == path_dim(p));

	pp = bundle_data(b) + i;

	if(*pp != NULL) {
		path_destroy(*pp);
	}

	*pp = p;
}

void bundle_remove(bundle_t *b, int i)
{
	path_t **pp;

	pp = bundle_data(b) + i;

	if(*pp != NULL) {
		path_destroy(*pp);
		*pp = NULL;
	}
}

bundle_t *bundle_append(bundle_t *b, path_t *p)
{
  int n;

  n = b->count;
  b = bundle_resize(b, n + 1);

  bundle_set(b, n, p);

  return b;
}

bundle_t *bundle_merge(bundle_t *dst, bundle_t *src)
{
	int i, n, m;
	path_t **dst_data, **src_data;

	n = dst->count;
	m = src->count;

	dst = bundle_resize(dst, n + m);

	dst_data = bundle_data(dst);
	src_data = bundle_data(src);

	for(i = 0; i < m; i++) {
		if(src_data[i] == NULL) {
			dst_data[n+i] = NULL;
		}
		else {
			dst_data[n + i] = path_copy(src_data[i]);
		}
	}

	return dst;
}

bundle_t *bundle_create_uniform(path_t *core, double radius, int count)
{
	int i, j, dim, length;
	bundle_t *b;

	path_t *path;

	angle_t theta = 0;
	length_t r;

	point_t p, q;
	coord_t tangent[3], offset[3], normal_avg[3], normal[3], avg_factor = 0.9;
	length_t epsilon;

	affine_t *A;

	length = path_length(core);
	dim = path_dim(core);
	assert(dim == 2 || dim == 3);

	b = bundle_create(dim, count);
	A = affine_create(dim);

	for(i = 0; i < count; i++) {
		path = path_create(dim, length);

		// Sample an offset
		//r = frand(-radius, radius);
		if(dim == 2) {
			r = 2 * i * radius / (count - 1) - radius;
		}
		else {
			assert(dim == 3);

			r = sqrt(frand(0, 1)) * radius;
			theta = frand(0, 2 * M_PI);
		}

		// Offset each point from the core
		for(j = 0; j < length; j++) {
			// Tangent
			path_tangent(core, j, tangent);

			// Find the offset
			if(dim == 2) {
				offset[0] = -tangent[1];
				offset[1] = tangent[0];
			}
			else {
				assert(dim == 3);

				// Normal with some smoothing
				path_normal(core, j, normal);

				/*
				if(i == 0 && j == 0) {
					printf("normal = ");
					point_printf_fmt(dim, normal, "%g");
					fputc('\n', stdout);
				}
				//*/

				epsilon = point_linfty_norm(dim, tangent) / 16;
				if(point_linfty_norm(dim, normal) > epsilon) {
					unit_vector(dim, normal, normal);
					point_mul(dim, normal, avg_factor);
					point_add(dim, normal_avg, normal_avg, normal);
				}
				else if(j == 0) {
					// TODO... starts with straight line
					// may be all straight line...

					//normal_avg[0] = -tangent[1];
					//normal_avg[1] = tangent[0];
					//normal_avg[2] = 0;

					normal_avg[0] = 1;
					normal_avg[1] = normal_avg[2] = 0;
				}

				/*
				if(i == 0 && j == 0) {
					printf("normal = ");
					point_printf(dim, normal);
					fputc('\n', stdout);

					printf("l_\\infty = %lg, epsilon = %lg\n",
								 point_linfty_norm(dim, normal), epsilon);

					printf("normal_avg = ");
					point_printf(dim, normal_avg);
					fputc('\n', stdout);
				}
				//*/

				// Rotate normal_avg by theta around tangent
				affine_rotation3(A, theta, tangent);
				affine_transform(A, offset, normal_avg);
			}

			point_mul(dim, offset, r / point_l2_norm(dim, offset));

			// Apply the offset
			p = path_get(core, j);
			q = path_get(path, j);
			point_add(dim, q, p, offset);
		}

		bundle_set(b, i, path);
	}

	affine_destroy(A);

	return b;
}

void bundle_extent(bundle_t *b, point_t min, point_t max)
{
	int i, j, k, dim, count, length;
	path_t *path;
	point_t p;

	count = b->count;
	dim = b->dim;

	// Initialize to +/- infinity
	for(k = 0; k < dim; k++) {
		min[k] = 1.0/0.0;
		max[k] = -1.0/0.0;
	}

	// Iterate over each point
	for(i = 0; i < count; i++) {
		path = bundle_data(b)[i];

		if(path != NULL) {
			length = path_length(path);

			for(j = 0; j < length; j++) {
				p = path_get(path, j);

				for(k = 0; k < dim; k++) {
					min[k] = MIN(min[k], p[k]);
					max[k] = MAX(max[k], p[k]);
				}
			}
		}
	}
}
