#include "flow.h"

#include "mask.h"
#include "flat_graph.h"
#include "flat_line_graph.h"

#include "optimize.h"
#include "optimize_internal.h"

#include "util.h"

static bool is_incidence(int nnz, int *rowpos, int *colpos, double *val);

/**
 * Return a list of the indices of edges connected to a vertex in the mask
 */
static int *mask_edge_indices(flat_graph_t *FG, mask_t *M, int *count)
{
	int i, j, k, n, m;
	edge_t *e;
	int *selected;

	assert(flat_graph_vertex_count(FG) == (int)mask_size(M));

	n = flat_graph_edge_count(FG);

	// Get a count for allocation
	m = 0;
	for(i = 0; i < n; i++) {
		e = &(flat_graph_edge(FG, i)->edge);

		if(mask_get(M, e->u) || mask_get(M, e->v)) {
			m++;
		}
	}

	*count = m;
	
	// Allocate
	selected = malloc(sizeof(int) * m);

	// Create
	j = 0;
	for(i = 0; i < n; i++) {
		e = &(flat_graph_edge(FG, i)->edge);

		if(mask_get(M, e->u) || mask_get(M, e->v)) {
			assert(j < m);
			selected[j++] = i;
		}
	}

	return selected;
}

static char *create_name(const char *format, ...)
{
	va_list args;
	char buf[256];

	va_start(args, format);
	vsnprintf(buf, 256, format, args);
	va_end(args);

	return strdup(buf);
}

static int graph_edge_index(flat_graph_t *FG, edge_pair_t *ep, int a)
{
	int u, v, swap, i;

	u = ep->common_vertex;
	v = ep->neighbors[a];

	i = flat_graph_edge_index(FG, u, v);
	assert(i >= 0 && i < flat_graph_edge_count(FG));

	return i;
}

static optimize_problem_t *flow_problem_create(
  optimize_options_t *opts,
  line_graph_t *L,
	flat_line_graph_t *FL,
	mask_t *source, mask_t *sink, double min_flow)
{
	int vertex_count, edge_count, edge_pair_count;
	int nrows, ncols;
	int *source_edges, *sink_edges, source_size, sink_size;

	graph_t *G;
	flat_graph_t *FG;

	optimize_problem_t *prob;

	int a, i, j, k, d, n, m, row;
	double w;
	edge_t *e;
	edge_pair_t *ep;
	flat_edge_pair_t *FLep;

	int beta_col_offset = 0, beta_row_offset = 0;

	int Aelt;
	int *Ai, *Aj, Annz;
	double *Av;

	int Qelt;
	int *Qi, *Qj, Qnnz;
	double *Qv;

	// ------------- Setup -------------

	// Create data structures
	G = line_graph_unline(L);
	FG = flat_graph_create(G);

	source_edges = mask_edge_indices(FG, source, &source_size);
	sink_edges = mask_edge_indices(FG, sink, &sink_size);

	// Count elements
	vertex_count = line_graph_vertex_count(L);
	edge_count = graph_edge_count(G);
	edge_pair_count = line_graph_total_edge_pair_count(L);

	// Find problem size
	// Flow problem:
	ncols = 2 * edge_pair_count + source_size + sink_size;
	nrows = edge_count + 2;
	Annz = 4 * edge_pair_count + 2*(source_size + sink_size);
	Qnnz = 0;

	// Branching penalty:
	if(opts->lambda != 0) {
		beta_col_offset = ncols;
		beta_row_offset = nrows;

		ncols += vertex_count;
		nrows += vertex_count;
		Annz += 2*edge_pair_count + vertex_count;
		//Qnnz += vertex_count;
	}

	prob = optimize_problem_create(opts, ncols, nrows, Annz, Qnnz);

	// ---------------------- Fill columns --------------------

	// --- Flow problem ---
	// Find minimum weight
	double wmin = INFINITY, wmax = -INFINITY;
	for(i = 0; i < edge_pair_count; i++) {
		ep = &(flat_line_graph_flat_edge_pair(FL, i)->edge_pair);
		w = ep->weight;
		//w = -log(ep->weight);

		wmin = MIN(wmin, w);
		wmax = MAX(wmax, w);
	}

	printf("w in [%lf, %lf]\n", wmin, wmax);

	// Add line graph edge flow variables
	for(i = 0; i < edge_pair_count; i++) {
		ep = &(flat_line_graph_flat_edge_pair(FL, i)->edge_pair);
		w = ep->weight;
		w = (w - wmin) / (wmax - wmin) + (wmax - wmin) * 1e-10;
		//w = 1;

		for(a = 0; a < 2; a++) {
			j = 2*i + a;

			prob->obj[j] = w;

			assert(prob->obj[j] > 0);

			prob->lb[j] = 0;
			prob->ub[j] = 1;

			prob->ctype[j] = opts->lambda == 0 ? 'C' : 'B';

			if(prob->colnames != NULL) {
				prob->colnames[j] = create_name(
          "%d_%d_%d",
					ep->neighbors[0],
					ep->common_vertex,
					ep->neighbors[1]); 
			}
		}
	}

	// Dummy flows from source/sink nodes
	n = 2 * edge_pair_count + source_size + sink_size;
	for(i = 2 * edge_pair_count; i < n; i++) {
		prob->obj[i] = 0;

		prob->lb[i] = 0;
		prob->ub[i] = INFINITY;

		prob->ctype[i] = 'C';
	}

	if(prob->colnames != NULL) {
		for(i = 0; i < source_size; i++) {
			j = 2 * edge_pair_count + i;
			e = &(flat_graph_edge(FG, source_edges[i])->edge);

			prob->colnames[j] = create_name("source_%d_%d", e->u, e->v);
		}

		for(i = 0; i < sink_size; i++) {
			j = 2 * edge_pair_count + source_size + i;
			e = &(flat_graph_edge(FG, sink_edges[i])->edge);

			prob->colnames[j] = create_name("sink_%d_%d", e->u, e->v);
		}
	}

	// --- Branching constraints ---
	if(opts->lambda != 0) {
		for(i = 0; i < vertex_count; i++) {
			j = beta_col_offset + i;

			prob->obj[j] = opts->lambda;

			prob->lb[j] = 1;
			prob->ub[j] = INFINITY;

			prob->ctype[j] = 'C';
		}

		if(prob->colnames != NULL) {
			for(i = 0; i < vertex_count; i++) {
				prob->colnames[beta_col_offset + i] = create_name("beta_%d", i);
			}
		}
	}

	// ---------------------- Quadratic Matrix --------------------
	Qelt = 0;
	Qi = prob->Q->i;
	Qj = prob->Q->j;
	Qv = prob->Q->v;

	// --- Branching constraints ---
	/*
	if(opts->lambda != 0) {
		for(i = 0; i < vertex_count; i++) {
			Qi[Qelt] = beta_col_offset + i;
			Qj[Qelt] = beta_col_offset + i;
			Qv[Qelt] = opts->lambda;
			Qelt++;
		}
	}
	*/

	// Some checks
	assert(Qelt == prob->Q->nnz);

	// ---------------------- Constraint Matrix --------------------

	Ai = prob->A->i;
	Aj = prob->A->j;
	Av = prob->A->v;
	Aelt = 0;

	// --- Flow problem ---

	// Edge incidence constraints
	for(i = 0; i < edge_pair_count; i++) {
		ep = &(flat_line_graph_flat_edge_pair(FL, i)->edge_pair);

		for(a = 0; a < 2; a++) {
			Ai[Aelt] = graph_edge_index(FG, ep, a);
			Aj[Aelt] = 2 * i + a;
			Av[Aelt] = 1;
			Aelt++;

			Ai[Aelt] = graph_edge_index(FG, ep, (a + 1) % 2);
			Aj[Aelt] = 2 * i + a;
			Av[Aelt] = -1;
			Aelt++;
		}
	}

	row = 0;
	for(row = 0; row < edge_count; row++) {
		prob->sense[row] = 'E';
		prob->rhs[row] = 0;
	}

	// Incidence constraints for dummy source/sink nodes
	prob->sense[row] = 'G';
	prob->rhs[row] = min_flow;
	row++;

	prob->sense[row] = 'L';
	prob->rhs[row] = -min_flow;
	row++;

	for(i = 0; i < source_size; i++) {
		j = 2*edge_pair_count + i;

		Ai[Aelt] = edge_count;
		Aj[Aelt] = j;
		Av[Aelt] = 1;
		Aelt++;

		Ai[Aelt] = source_edges[i];
		Aj[Aelt] = j;
		Av[Aelt] = -1;
		Aelt++;
	}

	for(i = 0; i < sink_size; i++) {
		j = 2*edge_pair_count + source_size + i;

		Ai[Aelt] = sink_edges[i];
		Aj[Aelt] = j;
		Av[Aelt] = 1;
		Aelt++;
		
		Ai[Aelt] = edge_count+1;
		Aj[Aelt] = j;
		Av[Aelt] = -1;
		Aelt++;
	}

	// Row names
	if(prob->rownames != NULL) {
		for(i = 0; i < edge_count; i++) {
			e = &(flat_graph_edge(FG, i)->edge);
			prob->rownames[i] = create_name("edge_%d_%d", e->u, e->v);
		}
		prob->rownames[edge_count] = strdup("source");
		prob->rownames[edge_count+1] = strdup("sink");
	}

	// --- Branching constraints ---
	if(opts->lambda != 0) {
		for(i = 0; i < vertex_count; i++) {
			// Subtract out the flow around this vertex
			d = flat_line_graph_edge_pair_count(FL, i);
			for(j = 0; j < d; j++) {
				FLep = flat_line_graph_edge_pair(FL, i, j);

				for(a = 0; a < 2; a++) {
					Ai[Aelt] = row;
					Aj[Aelt] = 2 * FLep->index + a;
					Av[Aelt] = -1;
					Aelt++;
				}
			}

			// Add beta
			Ai[Aelt] = row;
			Aj[Aelt] = beta_col_offset + i;
			Av[Aelt] = 1;
			Aelt++;

			prob->sense[row] = 'G';
			prob->rhs[row] = 0;
			row++;
		}
	}

	//printf("Aelt = %d, Annz = %d,%d\n", Aelt, Annz, prob->A->nnz);

	// Some checks
	assert(Aelt == prob->A->nnz);
	assert(row == prob->nrows);
	if(opts->lambda == 0) {
		assert(is_incidence(Annz, Ai, Aj, Av));
	}

	// Cleanup
	flat_graph_destroy(FG);
	graph_destroy(G);

	return prob;
}


static bool is_incidence(int Annz, int *rowpos, int *colpos, double *val)
{
	int i, ncols;
	int *plus1, *minus1;

	for(i = 0; i < Annz; i++) {
		ncols = MAX(ncols, colpos[i]);
	}

	plus1 = (int *)calloc(ncols, sizeof(int));
	minus1 = (int *)calloc(ncols, sizeof(int));

	for(i = 0; i < Annz; i++) {
		if(val[i] == 1) {
			plus1[colpos[i]]++;
		}
		else if(val[i] == -1) {
			minus1[colpos[i]]++;
		}
		else {
			fprintf(stderr, "non +/- 1 at (%d, %d)\n", rowpos[i], colpos[i]);
			return false;
		}
	}

	for(i = 0; i < ncols; i++) {
		if(!((plus1[i] == 1 && minus1[i] == 1) || (plus1[i] == 0 && minus1[i] == 0))) {
			fprintf(stderr, "non-unimodular column %d: %d %d\n", i, plus1[i], minus1[i]);
			return false;
		}
	}

	free(minus1);
	free(plus1);

	return true;
}

static line_graph_t *flow_solution(
	optimize_options_t *opts,
  flat_line_graph_t *FL,

	double *val)
{
	int i, j, vertex_count, edge_pair_count;
	double w;
	edge_pair_t ep;
	line_graph_t *H;
	int nselected;

	int beta_col_offset, branch_count;

	vertex_count = flat_line_graph_vertex_count(FL);
	edge_pair_count = flat_line_graph_total_edge_pair_count(FL);

	beta_col_offset = 2*edge_pair_count;

	H = line_graph_create(vertex_count);

	nselected = 0;

	for(i = 0; i < edge_pair_count; i++) {
		ep = flat_line_graph_flat_edge_pair(FL, i)->edge_pair;

		// Give the edge a weight equal to the total flow in both directions
		ep.weight = val[2*i] + val[2*i + 1];

		if(ep.weight > 0) {
			nselected++;
			//printf("w[%d -> %d] = %lf\n", ep.neighbors[0], ep.neighbors[1], ep.weight);
		}

		line_graph_add_edge_pair(H, &ep);
	}

	if(opts->verbose) {
		printf("%d nonzero weights in solution\n", nselected);

		if(opts->lambda != 0) {
			branch_count = 0;
			for(i = 0; i < vertex_count; i++) {
				j = beta_col_offset + i;
				if(val[j] > 1) {
					branch_count++;
					//printf("beta_%d = %lf\n", i, val[j]);
				}
			}

			printf("Branching at %d nodes\n", branch_count);
		}
	}

	return H;
}

static void flow_print(
  optimize_options_t *opts,
	line_graph_t *L,
	mask_t *source, mask_t *sink, double min_flow)
{
	int vertex_count, edge_pair_count, source_size, sink_size;

	vertex_count = line_graph_vertex_count(L);
	edge_pair_count = line_graph_total_edge_pair_count(L);

	source_size = mask_count_true(source);
	sink_size = mask_count_true(sink);

	printf("Problem Size:\n");
	printf("  %d vertices\n", vertex_count);
	printf("  %d line graph edges\n", edge_pair_count);
	printf("  %d/%d source/sink nodes\n", source_size, sink_size);
	printf("Constants:\n");
	printf("  Min flow = %.0lf\n", min_flow);
	printf("  lambda = %lg\n", opts->lambda);
}

line_graph_t *flow_tractography(
  optimize_options_t *opts,
	line_graph_t *L, mask_t *source, mask_t *sink, double min_flow)
{
	optimize_problem_t *prob;
	optimize_solution_t *sol;
	flat_line_graph_t *FL;
	line_graph_t *H;

	if(opts->verbose) {
		flow_print(opts, L, source, sink, min_flow);
	}

	// Setup
	if(opts->verbose) {
		printf("Setting up problem...\n");
	}
	FL = flat_line_graph_create(L);
	prob = flow_problem_create(opts, L, FL, source, sink, min_flow);

	// Solve
	if(opts->verbose) {
		printf("Beginning optimization...\n");
	}
	sol = optimize(opts, prob);

	if(opts->verbose) {
		printf("Extracting solution...\n");
	}
	H = flow_solution(opts, FL, sol->solution);

	// Cleanup
	optimize_solution_destroy(sol);
	optimize_problem_destroy(prob);
	flat_line_graph_destroy(FL);

	return H;
}
