#include "grid_vectors.h"
#include "util.h"
#include "weight_vector.h"

struct grid_vector_set
{
  int dim, point_count, pair_count;
  coord_t delta;

  coord_t *points;
};

static int grid_vector_set_contains_canonical(grid_vector_set_t *, point_t v[2]);

static void grid_points(int dim, length_t delta, length_t dist, int *len, coord_t **grid)
{
  int count, i, j, k, l, m, n;
  point_t p;
  length_t norm;

  // Initial overestimate
  n = iceil(2 * dist / delta) + 1;
  count = 1;
  for(i = 0; i < dim; i++) {
    count *= n;
  }

  // printf("n = %d\n", n);

  // Allocate the corresponding amount of space
  *grid = malloc(sizeof(coord_t) * dim * count);
  if(*grid == NULL) {
    return;
  }

  m = 0;
  for(i = 0; i < count; i++) {
    // Grab next point
    assert(m < count);
    p = *grid + dim * m;

    // Fill in point
		k = i;

		for(j = 0; j < dim; j++) {
			l = k % n;
			k /= n;

			p[j] = delta * (coord_t)(l - n/2);
		}

    //point_printf(dim, p);
    //printf("\n");

    // Add to set if it's within the circle
    norm = point_l2_norm(dim, p);
    if(norm < (1 + 1e-6) * dist && norm > 1e-6 * delta) {
      m++;
    }
	}

  // Set size
  *grid = realloc(*grid, sizeof(coord_t) * dim * m);
  *len = m;
}

grid_vector_set_t *grid_vector_set(int dim, length_t delta, length_t dist)
{
  int i, j, k, m, n;
  grid_vector_set_t *S;
  point_t grid, p[2], q[2];
  coord_t normalized[2][dim];

  // Initialize set structure
  S = malloc(sizeof(grid_vector_set_t));
  if(S == NULL) {
    return NULL;
  }

  S->dim = dim;
  S->delta = delta;

  // Find the set of points we're finding pairs of...
  grid_points(dim, delta, dist, &m, &grid);
  if(grid == NULL) {
    free(S);
    return NULL;
  }

  //for(i = 0; i < m; i++) {
  //  point_printf(dim, grid + dim * i);
  //  printf("\n");
  //}

  // Overestimate pair count
  n = isquare(m);
  S->points = malloc(sizeof(coord_t) * 2 * n * dim);
  if(S->points == NULL) {
    free(S->points);
    free(S);
    return NULL;
  }
  
  // Setup to normalize
  for(i = 0; i < 2; i++) {
    q[i] = normalized[i];
  }

  // Consider all pairs
  S->pair_count = 0;
  for(i = 0; i < m; i++) {
    p[0] = grid + i * dim;

    for(j = i + 1; j < m; j++) {
      p[1] = grid + j * dim;

      //point_printf(dim, p[0]);
      //printf(", ");
      //point_printf(dim, p[1]);
      //printf("\n  --> ");

      // Normalize p to q
      canonical_point_pair(dim, p, q);

      // Add normalized pair to set if not present
      if(!grid_vector_set_contains_canonical(S, q)) {
        assert(S->pair_count < n);

        for(k = 0; k < 2; k++) {
          point_assign(dim, S->points + (2 * S->pair_count + k) * dim, q[k]);
        }
        S->pair_count++;

        //point_printf(dim, q[0]);
        //printf(", ");
        //point_printf(dim, q[1]);
        //printf("\n\n");
      }
    }
  }
  

  // Cleanup
  S->points = (coord_t *)realloc(S->points, sizeof(coord_t) * 2 * S->pair_count* dim);
  free(grid);

  return S;
}

void grid_vector_set_destroy(grid_vector_set_t *S)
{
  if(S != NULL) {
    free(S->points);
    free(S);
  }
}

int grid_vector_set_dim(grid_vector_set_t *S)
{
  return S->dim;
}

int grid_vector_set_pair_count(grid_vector_set_t *S)
{
  return S->pair_count;
}

void grid_vector_set_get(grid_vector_set_t *S, int i, point_t v[2])
{
  int j, dim;

  assert(i < S->pair_count);
  
  dim = S->dim;
  for(j = 0; j < 2; j++) {
    v[j] = S->points + S->dim * (2*i + j);
  }
}

// Check if l_\infty norm of difference is greater then epsilon
static int canonical_point_pair_equal(
  int dim,
  point_t u[2], point_t v[2],
  coord_t epsilon)
{
  int i, j;

  // Don't actually need to compare first point (always e_1)
  for(i = 0; i < 2; i++) {
    for(j = 0; j < dim; j++) {
      if(fabs(u[i][j] - v[i][j]) > epsilon) {
        return 0;
      }
    }
  }

  return 1;
}

int grid_vector_set_contains_canonical(grid_vector_set_t *S, point_t u[2])
{
  point_t v[2];
  coord_t epsilon;
  int i, n, dim;

  dim = S->dim;
  n = S->pair_count;

  epsilon = 1e-2 * S->delta;

  for(i = 0; i < n; i++) {
    grid_vector_set_get(S, i, v);

    if(canonical_point_pair_equal(dim, u, v, epsilon)) {
      return 1;
    }
  }

  return 0;
}
