#include "line_graph_struct.h"
#include "kd_tree.h"
#include "util.h"
#include "point.h"
#include "neighborhood.h"


static void set_edge_pair_count(line_graph_t *G, vertex_entry_t *v, int n)
{
	G->edge_pair_count += n - v->edge_pair_count;
	
	v->edge_pair_count = n;
	v->edge_pairs = realloc(v->edge_pairs, n * sizeof(edge_pair_t));
}

line_graph_t *line_graph_create(int vertex_count)
{
	size_t s;
	line_graph_t *G;
	int i;
	vertex_entry_t *v;

	s = sizeof(line_graph_t);
	s += vertex_count * sizeof(vertex_entry_t);

	G = (line_graph_t *)malloc(s);

	// Initialize G
	G->vertex_count = vertex_count;
	G->edge_pair_count = 0;

	for(i = 0; i < vertex_count; i++) {
		v = get_vertex_entry(G, i);

		v->edge_pair_count = 0;
		v->edge_pairs = NULL;
	}

	return G;
}

void line_graph_destroy(line_graph_t *G)
{
	intptr_t p;
	vertex_entry_t *vertex_entries;
  int i;

  if(G == NULL) {
    return;
  }

  // Clean up vertex lists
	p = (intptr_t)G;
	p += sizeof(line_graph_t);
	vertex_entries = (vertex_entry_t *)p;

  for(i = 0; i < G->vertex_count; i++) {
    free(vertex_entries[i].edge_pairs);
  }

	free(G);
}

int line_graph_vertex_count(line_graph_t *G)
{
	return G->vertex_count;
}

int line_graph_total_edge_pair_count(line_graph_t *G)
{
	return G->edge_pair_count;
}

int line_graph_edge_pair_count(line_graph_t *G, int vertex)
{
	return get_vertex_entry(G, vertex)->edge_pair_count;
}

edge_pair_t *line_graph_edge_pair(line_graph_t *G, int vertex, int i)
{
	vertex_entry_t *vertex_entry;

	vertex_entry = get_vertex_entry(G, vertex);
	
	assert(i < vertex_entry->edge_pair_count);

	return vertex_entry->edge_pairs + i;
}

edge_pair_t *line_graph_add_edge_pair(line_graph_t *G, edge_pair_t *new)
{
  vertex_entry_t *v;
  int pair_count;
  edge_pair_t *ep;

  v = get_vertex_entry(G, new->common_vertex);
  pair_count = v->edge_pair_count;

  // Allocate additional edge pair at the end
  set_edge_pair_count(G, v, pair_count + 1);
  ep = v->edge_pairs + pair_count;

  // copy
  *ep = *new;

  return ep;
}

void line_graph_vertex_set_neighbors(
  line_graph_t *G, int vertex,
  int neighbor_count, int *neighbors)
{
	vertex_entry_t *v;
	int pair_count, i, j, k;
	edge_pair_t *ep;

	v = get_vertex_entry(G, vertex);


	// Allocate new edge pairs
	pair_count = (neighbor_count * (neighbor_count - 1)) / 2;
	set_edge_pair_count(G, v, pair_count);

	// Initialize edge pairs
	k = 0;
	for(i = 0; i < neighbor_count; i++) {
		for(j = i+1; j < neighbor_count; j++) {
			assert(neighbors[i] != vertex && neighbors[j] != vertex);
			assert(neighbors[i] != neighbors[j]);

			ep = &(v->edge_pairs[k++]);

			ep->common_vertex = vertex;
			ep->neighbors[0] = neighbors[i];
			ep->neighbors[1] = neighbors[j];
			ep->weight = 0;
		}
	}
}

line_graph_t *line_graph_create_from_neighborhoods(
  int dim, int point_count,
  point_t (*get_point)(void *closure_data, int i), void *closure_data,
  double neighborhood_distance)
{
	graph_t *G;
	line_graph_t *L;

	G = graph_create_from_neighborhoods(dim, point_count,
																			get_point, closure_data,
																			neighborhood_distance);

	L = line_graph_create_from_graph(G);

	graph_destroy(G);

	return L;
}

line_graph_t *line_graph_create_from_graph(graph_t *G)
{
	line_graph_t *L;
	int i, j, N, d, old_d, *neighbors;

	N = graph_vertex_count(G);

	L = line_graph_create(N);

	old_d = 0;
	neighbors = NULL;

	for(i = 0; i < N; i++) {
		d = graph_vertex_degree(G, i);

		if(old_d < d) {
			neighbors = (int *)realloc(neighbors, d * sizeof(int));
		}

		// Get neighborhood from graph
		for(j = 0; j < d; j++) {
			neighbors[j] = graph_vertex_neighbor(G, i, j);
		}

		// Set in line graph
		line_graph_vertex_set_neighbors(L, i, d, neighbors);

		old_d = d;
	}

	free(neighbors);

	return L;
}

static void endpoint_iter(line_graph_t *L,
                          void (*f)(void *data, edge_pair_t *, int v),
                          void *data)
{
  int a, b, i, j, k, n, d1, d2, neighbor;
  edge_pair_t *ep, *ep2;
  bool endpoint;

  n = line_graph_vertex_count(L);
  for(i = 0; i < n; i++) {
    d1 = line_graph_edge_pair_count(L, i);

    for(j = 0; j < d1; j++) {
      ep = line_graph_edge_pair(L, i, j);

      for(a = 0; a < 2; a++) {
        neighbor = ep->neighbors[a];

        endpoint = true;

        d2 = line_graph_edge_pair_count(L, neighbor);
        for(k = 0; k < d2; k++) {
          ep2 = line_graph_edge_pair(L, neighbor, k);

          assert(ep2->common_vertex == neighbor);

          for(b = 0; b < 2; b++) {
            if(ep2->neighbors[b] == i) {
              endpoint = false;
            }
          }

          if(!endpoint) {
            break;
          }
        }

        if(endpoint) {
          (*f)(data, ep, neighbor);
        }
      }
    }
  }
}

static void endpoint_count_closure(void *data, edge_pair_t *ep, int v)
{
  (*(int *)data)++;
}

static int endpoint_count(line_graph_t *L)
{
  int count;

  count = 0;
  endpoint_iter(L, &endpoint_count_closure, &count);

  return count;
}

typedef struct {
  line_graph_t *L;

  int path_count;
  int **paths;

} trace_t;

typedef struct node node_t;
struct node {
  int v;
  int len, start;
  node_t *prev;
};

static bool ll_contains(node_t *tail, int v)
{
  return v == tail->v || (tail->prev != NULL && ll_contains(tail->prev, v));
}

static void ll_path_add(trace_t *t, node_t *tail)
{
  node_t *n;
  int *path, len, i;

  if(tail->start < tail->v) {
    return;
  }

  len = tail->len;

  // Create flat list with this path
  path = malloc(sizeof(int) * (len + 1));
  i = 0;
  for(n = tail; n != NULL; n = n->prev) {
    assert(i < len);
    path[i++] = n->v;
  }
  assert(i == len);
  path[len] = -1;

  // Add path
  t->path_count++;
  t->paths = realloc(t->paths, sizeof(int *) * (t->path_count + 1));
  t->paths[t->path_count - 1] = path;
  t->paths[t->path_count] = NULL;
}

static void ll_cycle_add(trace_t *t, node_t *tail, int v)
{
	// TODO
	// Need some way of removing multiple cycles
}

static node_t *node_add(node_t *prev, int v)
{
  node_t *n;

  n = (node_t *)malloc(sizeof(node_t));
  n->v = v;
  n->prev = prev;
  if(prev == NULL) {
    n->len = 1;
    n->start = v;
  }
  else {
    n->len = prev->len + 1;
    n->start = prev->start;
  }

  return n;
}

static int trace_iter(trace_t *t, edge_pair_t *ep, int v, node_t *prev)
{
  int a, b, i, d, nextv;
  node_t *n, *n1, *n2;
  edge_pair_t *nextep;
  bool end;

  // Create a node for this point
  n = node_add(prev, v);

  // Find the next point
  if(ep->neighbors[0] == v) {
    a = 0; b = 1;
  }
  else {
    assert(ep->neighbors[1] == v);
    a = 1; b = 0;
  }
  nextv = ep->neighbors[b];

  /*
  for(i = 0; i < n->len - 1; i++) {
    printf("  ");
  }
  printf("%d %d %d\n", ep->neighbors[a], ep->common_vertex, ep->neighbors[b]);
  */

	// Cycle detection
	if(prev != NULL && ll_contains(prev, v)) {
    ll_cycle_add(t, n, v);
	}
	else {
		// Find the next edge pair
		d = line_graph_edge_pair_count(t->L, nextv);
		end = true;
		for(i = 0; i < d; i++) {
			nextep = line_graph_edge_pair(t->L, nextv, i);
    
			if(nextep->neighbors[0] == ep->common_vertex
				 || nextep->neighbors[1] == ep->common_vertex) {
				trace_iter(t, nextep, ep->common_vertex, n);
				end = false;
			}
		}
  
		if(end) {
			// Add the last two vertices and add the path
			n1 = node_add(n, ep->common_vertex);
			n2 = node_add(n1, ep->neighbors[b]);

			ll_path_add(t, n2);

			free(n2);
			free(n1);
		}
	}

  // Cleanup
  free(n);

  return 0;
}

static void trace(void *data, edge_pair_t *ep, int v)
{
  trace_t *t;

  t = (trace_t *)data;

  trace_iter(t, ep, v, NULL);
}

int **line_graph_paths(line_graph_t *L)
{
  trace_t t;

  t.L = L;

  t.path_count = 0;

  t.paths = malloc(sizeof(int *));
  t.paths[0] = NULL;

  endpoint_iter(L, &trace, &t);

  return t.paths;
}

line_graph_t *line_graph_threshold(line_graph_t *in, double threshold)
{
	line_graph_t *out;
	int i, j, n, d;
	edge_pair_t *ep;

	n = line_graph_vertex_count(in);
	out = line_graph_create(n);

	for(i = 0; i < n; i++) {
		d = line_graph_edge_pair_count(in, i);

		for(j = 0; j < d; j++) {
			ep = line_graph_edge_pair(in, i, j);

			if(ep->weight >= threshold) {
				line_graph_add_edge_pair(out, ep);
			}
		}
	}

	return out;
}

graph_t *line_graph_unline(line_graph_t *L)
{
	graph_t *G;
	edge_pair_t *ep;
	edge_t e;
	int i, j, k, a, n, d, u, v;
	bool present;

	e.weight = 0;

	n = line_graph_vertex_count(L);
	G = graph_create(n);

	for(i = 0; i < n; i++) {
		d = line_graph_edge_pair_count(L, i);

		for(j = 0; j < d; j++) {
			ep = line_graph_edge_pair(L, i, j);

			for(a = 0; a < 2; a++) {
				u = ep->common_vertex;
				v = ep->neighbors[a];

				if(!graph_edge_present(G, u, v)) {
					e.u = u; e.v = v;
					graph_edge_add(G, &e);
				}
			}
		}
	}

	return G;
}
