#include "mask.h"

typedef intptr_t group_t;

#define GROUP_SIZE (8 * sizeof(group_t))

struct mask {
	int n;
};

static inline group_t *data(mask_t *m)
{
	return (group_t *)((intptr_t)m + sizeof(mask_t));
}

static size_t group_count(int n)
{
	if(n % GROUP_SIZE == 0) {
		return n / GROUP_SIZE;
	}
	else {
		return n / GROUP_SIZE + 1;
	}
}

static size_t data_size(size_t n)
{
	return sizeof(mask_t) + group_count(n) * sizeof(group_t);
}

mask_t *mask_create(size_t n)
{
	mask_t *m;
	size_t groups, i;
	group_t *g;

	m = (mask_t *)malloc(data_size(n));
	if(m == NULL) {
		return NULL;
	}

	m->n = n;

	return m;
}

mask_t *mask_copy(mask_t *src)
{
	mask_t *dst;

	dst = mask_create(src->n);

	memcpy(data(dst), data(src), group_count(src->n) * sizeof(group_t));

	return dst;
}

mask_t *mask_resize(mask_t *m, size_t n)
{
	size_t old_groups, groups, i;
	group_t *g;

	old_groups = group_count(m->n);
	groups = group_count(n);

	m = (mask_t *)realloc(m, data_size(n));
	m->n = n;

	return m;
}

void mask_destroy(mask_t *m)
{
	free(m);
}

size_t mask_size(mask_t *m)
{
	return m->n;
}

bool mask_get(mask_t *m, int j)
{
	group_t i, g;

	assert(j < m->n);

	i = (group_t)j;

	g = data(m)[i / GROUP_SIZE];
	g &= ((group_t)1 << (i % GROUP_SIZE));

	return g != 0;
}

void mask_set(mask_t *m, int j, bool v)
{
	group_t i, *g;

	assert(j < m->n);

	i = (group_t)j;

	g = data(m) + i / GROUP_SIZE;
	if(v) {
		*g |= ((group_t)1 << (i % GROUP_SIZE));
	}
	else {
		*g &= (~((group_t)1 << (i % GROUP_SIZE)));
	}

	assert(mask_get(m, i) == v);
}

void mask_not(mask_t *m)
{
	int i, n;
	group_t *g;

	n = group_count(m->n);
	g = data(m);

	for(i = 0; i < n; i++) {
		g[i] = ~(g[i]);
	}
}

void mask_and(mask_t *a, mask_t *b)
{
	int i, n;
	group_t *ga, *gb;

	assert(a->n == b->n);

	n = group_count(a->n);
	ga = data(a);
	gb = data(b);

	for(i = 0; i < n; i++) {
		ga[i] &= gb[i];
	}
}

void mask_or(mask_t *a, mask_t *b)
{
	int i, n;
	group_t *ga, *gb;

	assert(a->n == b->n);

	n = group_count(a->n);
	ga = data(a);
	gb = data(b);

	for(i = 0; i < n; i++) {
		ga[i] |= gb[i];
	}
}


size_t mask_count_true(mask_t *m)
{
	size_t count;
	int i, n;

	n = m->n;

	count = 0;
	for(i = 0; i < n; i++) {
		if(mask_get(m, i)) {
			count++;
		}
	}

	return count;
}

void mask_print(mask_t *m)
{
	int i, n;

	n = m->n;

	for(i = 0; i < n; i++) {
		fputc(mask_get(m, i) ? '*' : '-', stdout);

		if(i % 8 == 0 && i != 0) {
			fputc(' ', stdout);
		}
	}
	fputc('\n', stdout);
}

void mask_nonzero_indices(mask_t *m, int *i)
{
	int j, k, n;

	n = m->n;
	k = 0;
	for(j = 0; j < n; j++) {
		if(mask_get(m, j)) {
			i[k++] = j;
		}
	}
}

mask_t *mask_subset_indices(mask_t *subset, mask_t *in)
{
	int i, j, n;
	mask_t *out;

	n = mask_size(subset);
	assert(mask_size(in) == n);

	out = mask_create(mask_count_true(subset));

	j = 0;
	for(i = 0; i < n; i++) {
		if(!mask_get(subset, i)) {
			continue;
		}

		mask_set(out, j++, mask_get(in, i));
	}

	return out;
}

mask_t *mask_superset_indices(mask_t *subset, mask_t *in)
{
	int i, j, n;
	mask_t *out;

	n = mask_size(subset);
	out = mask_create(n);

	j = 0;
	for(i = 0; i < n; i++) {
		mask_set(out, i, false);

		if(!mask_get(subset, i)) {
			continue;
		}

		mask_set(out, i, mask_get(in, j++));
	}

	return out;
}
