#include "min_spline_weight.h"

#include "util.h"

#include "splines/spline.h"
#include "monte_carlo.h"
#include "grid.h"

#if DIM != 2
// Need to expand spherical coordinates
#error Unsupported dimensionality
#endif

void min_spline_weight_options_default(min_spline_weight_options_t *opts)
{
  opts->spline_options = minimize_combined_options_default;

  opts->spline_options.minimization_options.max_iter = 8;
  opts->spline_options.minimization_options.start_precision.N = 16;
  opts->spline_options.minimization_options.verbose = 0;

  

  opts->sigma = 4.0;
  opts->expectation_samples = 1024;
}

typedef struct min_spline_weight_closure min_spline_weight_closure_t;
struct min_spline_weight_closure {
  min_spline_weight_options_t *opts;

  point_t *points;
  tensor_t *tensors;
};

static double min_spline_weight_angle(void *closure_data, double *orientations)
{
  min_spline_weight_closure_t *closure;
  min_spline_weight_options_t *opts;
  point_t *points;
  tensor_t *tensors;

  point_t u[3];

  spline_t *spline;
  double *p, *m;

  double weight, prob;

  int i, j;

  // Unpack closure
  closure = (min_spline_weight_closure_t *)closure_data;
  opts = closure->opts;
  points = closure->points;
  tensors = closure->tensors;

  // Create unit vectors in the given directions
  for(i = 0; i < 3; i++) {
    u[i][0] = cos(orientations[i]);
    u[i][1] = sin(orientations[i]);
  }

  // Create a spline with the given orientations and points
  spline = spline_create(3);

  for(i = 0; i < 3; i++) {
    point_assign(spline_p(spline, i), points[i]);
    point_assign(spline_m(spline, i), u[i]);
  }

  // Minimize weight
  weight = spline_magnitudes_min_combined(spline, &opts->spline_options);
  printf("weight = %2f\n", weight);

  // Find probability density of this orientation
  prob = 1;
  for(i = 0; i < 3; i++) {
    prob *= tensor_mult2(&tensors[i], u[i]);
  }
  printf("prob = %2f\n", prob);

  spline_destroy(spline);

  printf("f = %2f\n", prob * weight);

  return prob * weight;
}

double min_spline_weight(
  weight_options_t *weight_opts,
  point_t points[3], tensor_t tensors[3])
{
  min_spline_weight_options_t *opts;
  double sigma, c;
  min_spline_weight_closure_t closure;

  double min[3], max[3];
  int samples[3];
  double E;

  int i;

  opts = (min_spline_weight_options_t *)weight_opts->closure_data;

  sigma = opts->sigma;

	// Curvature weighting
	c = (-16 * log(0.1) * (sigma - 1)) / square(M_PI);
  opts->spline_options.curvature_factor = c;

  // Perform monte carlo integration to find expected spline weight
  closure.opts = opts;
  closure.points = points;
  closure.tensors = tensors;

  for(i = 0; i < 3; i++) {
    min[i] = -M_PI/2;
    max[i] = M_PI/2;
    samples[i] = 8;
  }

  //E = integrate_monte_carlo(3, &min_spline_weight_angle, min, max, opts->expectation_samples, &closure);
  E = integrate_grid(3, &min_spline_weight_angle, min, max, samples, &closure);

  return E;
}
