#include "neighborhood.h"
#include "kd_tree.h"
#include "util.h"

typedef struct
{
	int vertex;

	int allocated, size;
	int *neighbors;

} neighborhood_t;

static void add_neighbor(void *closure_data, spatial_tree_entry_t *e)
{
	neighborhood_t *neighborhood;
	int neighbor;

	neighborhood = (neighborhood_t *)closure_data;
	neighbor = *(int *)e->data;

	if(neighbor == neighborhood->vertex) {
		return;
	}

	// Increase size as necessary
	if(neighborhood->allocated == neighborhood->size) {
		neighborhood->allocated += 32;
		
		neighborhood->neighbors = (int *)realloc(
      neighborhood->neighbors,
      neighborhood->allocated * sizeof(int));
	}

	neighborhood->neighbors[neighborhood->size++] = neighbor;
}

graph_t *graph_create_from_neighborhoods(
  int dim, int point_count,
	point_t (*get_point)(void *closure_data, int i), void *closure_data,
  double neighborhood_distance)
{
	graph_t *G;

  spatial_tree_options_t kd_tree_opts;
	spatial_tree_t *kd_tree;

	int i, j, *integers;
	point_t p;
	neighborhood_t neighborhood;

	edge_t e;

	// Allocate line graph
	G = graph_create(point_count);

  kd_tree_opts.max_bin_size = 8;
  kd_tree_opts.dim = dim;

	// Fill a kd-tree with the points
	integers = (int *)malloc(sizeof(int) * point_count);
	kd_tree = spatial_tree_create(&kd_tree_opts);

	for(i = 0; i < point_count; i++) {
		integers[i] = i;
		p = get_point(closure_data, i);

		spatial_tree_add(kd_tree, p, &(integers[i]));
	}

	// For each point get the neighborhood and add to the line graph
	neighborhood.allocated = 0;
	neighborhood.neighbors = NULL;

	for(i = 0; i < point_count; i++) {
		p = get_point(closure_data, i);

		neighborhood.vertex = i;
		neighborhood.size = 0;

		// Retrieve neighborhood
		spatial_tree_neighborhood(kd_tree, p, neighborhood_distance,
															&add_neighbor, &neighborhood);

		// Add edges to graph
		for(j = 0; j < neighborhood.size; j++) {
			e.u = i;
			e.v = neighborhood.neighbors[j];
			e.weight = 0.0;

			// Avoid adding edges twice
			if(e.u < e.v) {
				graph_edge_add(G, &e);
			}
		}
	}

	// Cleanup
	spatial_tree_destroy(kd_tree);
	free(integers);

	return G;
}

