#include "nifti.h"
#include "affine.h"
#include "util.h"
#include "indices.h"

/////////////////////// Identifying NIfTI filenames ///////////////////////

static inline const char *extension(const char *filename, const char *end, int *length)
{
	const char *ext;

	for(ext = end; *ext != '.' && ext > filename; ext--);
	*length = end - ext;

	return ext;
}

static inline int is_nifti(const char *filename, const char *end)
{
	const char *ext;
	int len;

	ext = extension(filename, end, &len);

	if(strncmp(ext, ".nii", len) == 0) {
		return 1;
	}
	else if(strncmp(ext, ".gz", len) == 0) {
		return is_nifti(filename, end - len);
	}
	else {
		return 0;
	}
}

int nifti_filename(const char *filename)
{
	return is_nifti(filename, filename + strlen(filename));
}

//////////////////////////// Loading data ////////////////////////////
static void nifti_stderr(void *closure, const char *msg)
{
	fputs(msg, stderr);
	fputc('\n', stderr);
}

const nifti_load_options_t nifti_load_options_default = {
  &nifti_stderr, // error_closure
  NULL, // closure_data

  'L', // element order
};

static void format_error(
  void (*error)(void *closure, const char *msg), void *closure,
	const char *format, ...)
{
  va_list format_args;
	char msg[1024];

	if(error == NULL) {
		return;
	}

  va_start(format_args, format);

	vsnprintf(msg, 1024, format, format_args);
	msg[1023] = '\0';

	printf("format = \"%s\"\n", format);
	printf("msg = \"%s\"\n", msg);

	(*error)(closure, msg);

  va_end(format_args);
}

static void swap_byte_order(void *p, int n)
{
  int i;
  char *s, b;

  s = (char *)p;

  for(i = 0; i < n/2; i++) {
    b = s[i];
    s[i] = s[n-1-i];
    s[n-1-i] = b;
  }
}

static inline tensor_elt_t brick_value(
  struct nifti_1_header *hdr, void *brick,
  int nbyper, int i)
{
	intptr_t p;

	p = (intptr_t)brick;
	p += nbyper * i;

	switch(hdr->datatype) {
	case DT_INT8:
		return (tensor_elt_t)(*(int8_t *)p);
	case DT_UINT8:
		return (tensor_elt_t)(*(uint8_t *)p);

	case DT_INT16:
		return (tensor_elt_t)(*(int16_t *)p);
	case DT_UINT16:
		return (tensor_elt_t)(*(uint16_t *)p);

	case DT_INT32:
		return (tensor_elt_t)(*(int32_t *)p);
	case DT_UINT32:
		return (tensor_elt_t)(*(uint32_t *)p);

	case DT_INT64:
		return (tensor_elt_t)(*(int64_t *)p);
	case DT_UINT64:
		return (tensor_elt_t)(*(uint64_t *)p);

	case DT_FLOAT:
		assert(nbyper == sizeof(float));
		return (tensor_elt_t)(*(float *)p);

	case DT_DOUBLE:
		return (tensor_elt_t)(*(double *)p);

	default:
		return 0;
	}
}

static int nifti_sub2ind(int size[3], int i, int j, int k)
{
	return i + j * size[0] + k * size[0] * size[1];
	//return i * size[1] + j + k * size[0] * size[1];
}

tensor_field_t *nifti_load_tensor_field(
  const char *filename,
  const nifti_load_options_t *opts,	
	struct nifti_1_header *out_hdr)
{
	nifti_image *nim;
	nifti_brick_list bricks;
	struct nifti_1_header hdr;

	tensor_field_t *tf;
	
	int size[3];
	coord_t origin[3];
	coord_t delta[3];

	int sub[3];

	int i, j, k, n, m, d;
	int nbyper, swapsize;
	affine_t *A;

	point_t p;
	tensor_t *T;
	tensor_elt_t v[6];

  printf("Loading \"%s\"\n", filename);

	// Initialize
	nim = NULL;
	tf = NULL;
	A = NULL;

#define ERROR(format, ...)																							\
  {																																			\
		if(nim != NULL) {																										\
	    nifti_free_NBL(&bricks);																					\
      nifti_image_free(nim);																						\
    }																																		\
		affine_destroy(A);																									\
    tensor_field_destroy(tf);																						\
																																				\
    format_error(opts->error_closure, opts->closure_data,								\
	               format, ## __VA_ARGS__);																\
																																				\
    return NULL;																												\
  }

	//nifti_set_debug_level(3);

	// Read in file
	nim = nifti_image_read_bricks(filename, 0, NULL, &bricks);
	if(nim == NULL) {
		ERROR("Could not open NIfTI file \"%s\"", filename);
	}

	// Check type
	if(nim->intent_code != NIFTI_INTENT_SYMMATRIX) {
		ERROR("NIfTI file does not contain tensor data, has intent \"%s\"",
					nim->intent_name);

		// TODO: NIFTI_INTENT_GENMATRIX?

	}

	// Get real header
	hdr = nifti_convert_nim2nhdr(nim);

	// Check dimensionality
	// TODO: byte order
	if(nim->ndim != 5) {
		ERROR("Expected 5D volume\n");
	}

	if(nim->dim[4] != 1) {
		ERROR("No support for time series at this point");
	}

	if(nim->dim[5] != 6) {
		ERROR("Wrong number of values at each voxel to express a matrix");
	}

	if(bricks.nbricks != 6) {
		ERROR("Unexpected number of volumes: %d", bricks.nbricks);
	}

	nifti_datatype_sizes(hdr.datatype, &nbyper, &swapsize);

	// Build sizes
	n = 1;
	for(i = 0; i < 3; i++) {
		size[i] = nim->dim[i+1];
		origin[i] = 0;
		delta[i] = nim->pixdim[i+1];

		n *= size[i];
	}

	// Convert nifti coordinate system to general affine transformation
	A = affine_create(3);
	affine_identity(A);
	// TODO: normalize units to MM

	/*
	// TODO: make an option to choose between sform_code, qform_code
	if(nim->sform_code > 0) {
		for(j = 0; j < 4; j++) {
			affine_matrix_set(A, 0, j, hdr.srow_x[j]);
		}
		for(j = 0; j < 4; j++) {
			affine_matrix_set(A, 1, j, hdr.srow_y[j]);
		}
		for(j = 0; j < 4; j++) {
			affine_matrix_set(A, 2, j, hdr.srow_z[j]);
		}

		printf("sform\n");
	}
	else if(nim->qform_code > 0) {
		ERROR("qform_code only not yet supported"); // TODO
		// http://dijkstra.biostat.wisc.edu/doc/nifticlib-1.1.0/nifti1_8h-source.html
		// http://dijkstra.biostat.wisc.edu/doc/nifticlib-1.1.0/nifti1__io_8c.html#4b183639b2e242ac32210fe3437f60d2
	}
	else {
	*/
		affine_scale(A, delta);

		/*
		printf("pixdim only\n");
	}
		*/

	// Create tensor field
	//printf("count = %d\n", n);
	tf = tensor_field_create_free(3, n);

	// Load data
	for(i = 0; i < size[0]; i++) {
		for(j = 0; j < size[1]; j++) {
			for(k = 0; k < size[2]; k++) {
				// Load the given tensor from the file
				m = nifti_sub2ind(size, i, j, k);

				sub[0] = i;
				sub[1] = j;
				sub[2] = k;
				n = sub2ind(3, size, sub);

				//printf("m = %d, n = %d\n", m, n);

				p = tensor_field_point(tf, n);
				T = tensor_field_tensor(tf, n);

				// Set p according to the 3D coordinate system in the nifti file
				p[0] = (coord_t)i;
				p[1] = (coord_t)j;
				p[2] = (coord_t)k;

				affine_transform(A, p, p);

				// Load information into tensor
				for(d = 0; d < 6; d++) {
					v[d] = brick_value(&hdr, bricks.bricks[d], nbyper, m);
				}

				tensor_elt_set(T, 0, 0, v[0]);
				tensor_elt_set(T, 1, 0, v[1]);

				if(opts->element_order == 'L') {
					tensor_elt_set(T, 1, 1, v[2]);
					tensor_elt_set(T, 2, 0, v[3]);
				}
				else if(opts->element_order == 'U') {
					tensor_elt_set(T, 0, 2, v[2]);
					tensor_elt_set(T, 1, 1, v[3]);
				}
				else {
					ERROR("Invalid element order option: '%c'", opts->element_order);
				}

				tensor_elt_set(T, 1, 2, v[4]);
				tensor_elt_set(T, 2, 2, v[5]);

				/*
				if(fabs(v[0]) > 1e-5) {
					printf("i = %d, j = %d, k = %d\n", i, j, k);
					tensor_print(T);
					getc(stdin);
				}
				//*/

				/*
				if(i == 37 && j == 84 && k == 6) {
					printf("i = %d, j = %d, k = %d\n", i, j, k);
					tensor_print(T);
				}
				//*/
			}
		}
	}

	nifti_free_NBL(&bricks);
	nifti_image_free(nim);

	if(out_hdr != NULL) {
		*out_hdr = hdr;
	}

	return tf;

#undef ERROR
}

mask_t *nifti_load_mask(
  const char *filename,
  const nifti_load_options_t *opts,
	struct nifti_1_header *out_hdr)
{
	nifti_image *nim;
	nifti_brick_list bricks;
	struct nifti_1_header hdr;

	int i, j, k, n, m;
	int size[3], sub[3];
	int nbyper, swapsize;

	mask_t *mask;
	bool v;
	
	// Initialize
	nim = NULL;
	mask = NULL;

#define ERROR(format, ...)																							\
  {																																			\
		if(nim != NULL) {																										\
	    nifti_free_NBL(&bricks);																					\
      nifti_image_free(nim);																						\
    }																																		\
		mask_destroy(mask);																									\
																																				\
    format_error(opts->error_closure, opts->closure_data,								\
	               format, ## __VA_ARGS__);																\
																																				\
    return NULL;																												\
  }

	// Read in file
	nim = nifti_image_read_bricks(filename, 0, NULL, &bricks);
	if(nim == NULL) {
		ERROR("Could not open NIfTI file \"%s\"", filename);
	}

	// Get real header
	hdr = nifti_convert_nim2nhdr(nim);
	nifti_datatype_sizes(hdr.datatype, &nbyper, &swapsize);

	//printf("Intent: %d, \"%s\"\n", hdr.intent_code, hdr.intent_name);

	// Check dimensionality
	// TODO: byte order
	if(nim->ndim != 3) {
		ERROR("Expected 3D volume\n");
	}

	if(bricks.nbricks != 1) {
		ERROR("Unexpected number of volumes: %d", bricks.nbricks);
	}

	// Build sizes
	n = 1;
	for(i = 0; i < 3; i++) {
		size[i] = nim->dim[i+1];
		n *= size[i];
	}

	mask = mask_create(n);
	assert(mask != NULL);

	// Load data
	for(i = 0; i < size[0]; i++) {
		for(j = 0; j < size[1]; j++) {
			for(k = 0; k < size[2]; k++) {
				// Load the mask value from file
				m = nifti_sub2ind(size, i, j, k);
				v = brick_value(&hdr, bricks.bricks[0], nbyper, m) != 0;

				// Set in mask vector
				sub[0] = i;
				sub[1] = j;
				sub[2] = k;
				n = sub2ind(3, size, sub);
				mask_set(mask, n, v);
			}
		}
	}

	if(out_hdr != NULL) {
		*out_hdr = hdr;
	}

	return mask;
}

void nifti_write_mask(
	const char *filename, mask_t *mask,
	int size[3], float voxdim[3],
	const nifti_load_options_t *opts)
{
	nifti_image *nim;

	int dims[8];

	int i, j, k, m, n, sub[3];


	// Create dims
	dims[0] = 3;
	for(i = 0; i < 3; i++) {
		dims[i+1] = size[i];
	}
	for(i = 3; i < 7; i++) {
		dims[i+1] = 0;
	}

	//for(i = 0; i < 8; i++) {
	//	printf("dims[%d] = %d\n", i, dims[i]);
	//}

	// Set up nifti file
	nim = nifti_make_new_nim(dims, DT_UINT8, 1);	
	
	// Set filename
	nifti_set_filenames(nim, filename, 0, 1);

	// Set data
	for(i = 0; i < size[0]; i++) {
		for(j = 0; j < size[1]; j++) {
			for(k = 0; k < size[2]; k++) {
				m = nifti_sub2ind(size, i, j, k);

				sub[0] = i;
				sub[1] = j;
				sub[2] = k;
				n = sub2ind(3, size, sub);

				((uint8_t *)nim->data)[m] = mask_get(mask, n) ? 1 : 0;
			}
		}
	}

	// Write file
	//nifti_image_write(nim);
  nifti_image_write_hdr_img(nim, 1, "wb+");
}
