#include "optimize.h"
#include "optimize_internal.h"

// Default
#if defined(WITH_CPLEX)
#define DEFAULT_SOLVER SOLVER_CPLEX
#elif defined(WITH_GLPK)
#define DEFAULT_SOLVER SOLVER_GLPK
#else
#error No optimization being built
#endif

// Includes
#ifdef WITH_CPLEX
#include "optimize_cplex.h"
#endif

#ifdef WITH_GLPK
#include "optimize_glpk.h"
#endif

static const char *solver_name(solver_t solver)
{
	switch(solver) {

	case SOLVER_CPLEX:
		return "CPLEX";
	case SOLVER_GLPK:
		return "GLPK";

	case SOLVER_NULL:
  	return "no solver";
	default:
		return "unknown solver";
	}
}

static const char *model_name(model_t model)
{
	switch(model) {
	case MODEL_ILP1:
		return "ILP1";
	case MODEL_QP1:
		return "QP1";

	default:
		return "unknown model";
	}
}

line_graph_t *optimal_contours(optimize_options_t *opts, line_graph_t *G, mask_t *endpoints)
{
	solver_t solver;

	solver = opts->solver;
	if(solver == SOLVER_DEFAULT) {
		solver = DEFAULT_SOLVER;
	}

	if(opts->verbose) {
		printf("Optimizing %s model with %s backend...\n",
					 model_name(opts->model), solver_name(solver));
	}

	switch(solver) {

		// Dispatch to true solver
#ifdef WITH_CPLEX
	case SOLVER_CPLEX:
		return optimal_contours_cplex(opts, G, endpoints);
#endif

#ifdef WITH_GLPK
	case SOLVER_GLPK:
		return optimal_contours_glpk(opts, G, endpoints);
#endif

	default:
		fprintf(stderr, "Solver \"%s\" not supported\n", solver_name(opts->solver));
		return NULL;
	}
}

const optimize_options_t optimize_options_defaults = {
  1.0,  // lambda
  0.0,  // mu
  SOLVER_DEFAULT, // solver
	DEFAULT_THREAD_COUNT, // threads
	MODEL_ILP1, // model

	false // verbose
};

size_t optimize_constraints_nnz(
  optimize_options_t *opts,
	line_graph_t *G, flat_line_graph_t *FG,
	mask_t *endpoints)
{
	int i, j, m, n, a;
	int neighbor;
	size_t nnz;

  edge_pair_t *ep;
  flat_edge_pair_t *neighbor_ep;

	n = flat_line_graph_total_edge_pair_count(FG);

	nnz = 0;
	for(i = 0; i < n; i++) {
    // Grab the corresponding edge pair
    ep = &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);

		// ----- endpoint constraints -----
		for(a = 0; a < 2; a++) {
			neighbor = ep->neighbors[a];
			if(endpoints != NULL && mask_get(endpoints, neighbor)) {
				continue;
			}

			nnz += 2;

			m = flat_line_graph_edge_pair_count(FG, neighbor);
			
			for(j = 0; j < m; j++) {
				neighbor_ep = flat_line_graph_edge_pair(FG, neighbor, j);

				if(neighbor_ep->edge_pair.neighbors[0] == ep->common_vertex
					 || neighbor_ep->edge_pair.neighbors[1] == ep->common_vertex) {
					nnz++;
				}
			}
		}
	}

	// part constraints
	nnz += flat_line_graph_total_edge_pair_count(FG);

	return nnz;
}

void optimize_model_unhandled(optimize_options_t *opts)
{
	fprintf(stderr, "%s not implemented by %s backend",
					model_name(opts->model),
					solver_name(opts->solver));
	exit(1);
}


optimize_problem_t *optimize_problem_create(
  optimize_options_t *opts,
  int ncols, int nrows, int Annz, int Qnnz)
{
	optimize_problem_t *prob;

	prob = malloc(sizeof(optimize_problem_t));

	prob->ncols = ncols;
	prob->nrows = nrows;

	prob->minimize = true;

	prob->obj = (double *)malloc(sizeof(double) * ncols);
	prob->lb = (double *)malloc(sizeof(double) * ncols);
	prob->ub = (double *)malloc(sizeof(double) * ncols);
	prob->ctype = (char *)malloc(sizeof(char) * ncols);

	prob->sense = (char *)malloc(sizeof(char) * nrows);
	prob->rhs = (double *)malloc(sizeof(double) * nrows);

	prob->A = matrix_list_create(Annz, nrows, ncols);
	prob->Q = matrix_list_create(Qnnz, ncols, ncols);

	if(opts->verbose) {
		prob->colnames = (char **)malloc(sizeof(char *) * ncols);
		prob->rownames = (char **)malloc(sizeof(char *) * nrows);
	}
	else {
		prob->colnames = prob->rownames = NULL;
	}

	return prob;
}

static void free_name_list(int n, char **names)
{
	int i;

	if(names == NULL) {
		return;
	}
	
	for(i = 0; i < n; i++) {
		free(names[i]);
	}

	free(names);
}

void optimize_problem_destroy(optimize_problem_t *prob)
{
	free_name_list(prob->nrows, prob->rownames);
	free_name_list(prob->ncols, prob->colnames);

	// cols
	free(prob->obj); free(prob->lb); free(prob->ub); free(prob->ctype);

	// matrix
	matrix_list_destroy(prob->A);
	matrix_list_destroy(prob->Q);
 
	// rhs
	free(prob->sense); free(prob->rhs);

	free(prob);
}

optimize_solution_t *optimize_solution_create(optimize_problem_t *prob)
{
	optimize_solution_t *sol;

	sol = (optimize_solution_t *)malloc(sizeof(optimize_solution_t));
	sol->status = OPTIMIZE_NONE;
	sol->solution = (double *)malloc(sizeof(double) * prob->ncols);

	return sol;
}

void optimize_solution_destroy(optimize_solution_t *sol)
{
	free(sol->solution);
	free(sol);
}

optimize_solution_t *optimize(optimize_options_t *opts, optimize_problem_t *prob)
{
	solver_t solver;

	solver = opts->solver;
	if(solver == SOLVER_DEFAULT) {
		solver = DEFAULT_SOLVER;
	}

	if(opts->verbose) {
		printf("Optimizing with %s backend...\n", solver_name(solver));
	}

	switch(solver) {
		// Dispatch to true solver
#ifdef WITH_CPLEX
	case SOLVER_CPLEX:
		return optimize_cplex(opts, prob);
#endif

		/*
#ifdef WITH_GLPK
	case SOLVER_GLPK:
		return optimal_contours_glpk(opts, G, endpoints);
#endif
		*/

	default:
		fprintf(stderr, "Solver \"%s\" not supported\n", solver_name(opts->solver));
		return NULL;
	}
}
