#include "optimize_cplex.h"
#include "flat_line_graph.h"
#include "optimize_internal.h"

#include "util.h"
#include "flow.h"
#include "graph.h"
#include "flat_graph.h"

#include <ilcplex/cplex.h>

static void print_error(const char *error, CPXENVptr env, int status);

static int initialize_problem(
	optimize_options_t *opts,
	line_graph_t *G, flat_line_graph_t *FG,
	mask_t *endpoints,
  CPXENVptr env, CPXLPptr prob)
{
	int i, j, k, N, E, d, maxd, a;

  edge_pair_t *ep;
  flat_edge_pair_t *Fep, *neighbor_ep;

	int neighbor;

	// Column properties
	int cols;
	double *obj;
	double *lb, *ub;
	char *ctype;

	// Quadratic matrix
	int *qbeg, *qind, *qcnt;
	double *qval;

	// Constraint matrix
	int nnz;
	char *sense;
	int *Abeg, *Aind, Aelt;
	double *Aval;

	int status;

	// Set up a bunch of arrays to be fed to cplex
	N = flat_line_graph_vertex_count(FG);
  E = flat_line_graph_total_edge_pair_count(FG);
	maxd = flat_line_graph_max_pair_degree(FG);

	///////////////////// Allocate columns /////////////////////
	cols = E + N + E;

	obj = (double *)malloc(cols * sizeof(double));
	lb = (double *)malloc(cols * sizeof(double));
	ub = (double *)malloc(cols * sizeof(double));
	ctype = (char *)malloc(cols * sizeof(char));
	// TODO: check memory

	// Add alphas
	for(i = 0; i < E; i++) {
    // Grab the corresponding edge pair
    ep = &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);

		obj[i] = -ep->weight;
		
		lb[i] = 0;
		ub[i] = 1;

		ctype[i] = CPX_BINARY;
	}

	// Add betas
	for(i = E; i < N + E; i++) {
		if(opts->model == MODEL_ILP1) {
			obj[i] = opts->lambda;
		}
		else if(opts->model == MODEL_QP1) {
			obj[i] = 0;
		}
		else {
			optimize_model_unhandled(opts);
		}

		lb[i] = 0;
	  ub[i] = CPX_INFBOUND;

		ctype[i] = CPX_CONTINUOUS;
	}

	// Add gammas
	for(i = N + E; i < N + E + E; i++) {
		obj[i] = opts->mu;
		
		lb[i] = 0;
		ub[i] = CPX_INFBOUND;

		ctype[i] = CPX_CONTINUOUS;
	}

	// Add all columns
	if(opts->model == MODEL_ILP1) {
		status =  CPXnewcols(env, prob, cols, obj, lb, ub, ctype, NULL);
	}
	else if(opts->model == MODEL_QP1) {
		status =  CPXnewcols(env, prob, cols, obj, lb, ub, NULL, NULL);
	}
	else {
    status = 0;
		optimize_model_unhandled(opts);
	}

	free(ctype); free(ub); free(lb); free(obj);
	
	if(status) {
		print_error("Could not allocate columns in CPLEX", env, status);
		return status;
	}

	///////////////////// Quadratic beta objective /////////////////////
	if(opts->model == MODEL_ILP1) {
	}
	else if(opts->model == MODEL_QP1) {
		qbeg = (int *)malloc(cols * sizeof(int));
		qcnt = (int *)malloc(cols * sizeof(int));
		qind = (int *)malloc(N * sizeof(int));
		qval = (double *)malloc(N * sizeof(double));
		// TODO: check memory

		for(i = 0; i < E; i++) {
			// alphas
			qbeg[i] = 0;
			qcnt[i] = 0;

			// gammas
			qbeg[N + E + i] = N;
			qcnt[N + E + i] = 0;
		}

		for(i = 0; i < N; i++) {
			qbeg[E + i] = i;
			qcnt[E + i] = 1;

			qind[i] = E + i;
			qval[i] = opts->lambda;
		}

		status = CPXcopyquad(env, prob, qbeg, qcnt, qind, qval);

		free(qval); free(qind); free(qcnt); free(qbeg);

		if(status) {
			print_error("Could not copy quadratic coefficients to CPLEX", env, status);
			return status;
		}
	}
	else {
		optimize_model_unhandled(opts);
	}

	////////////////////// Constraint Matrix /////////////////

	// ------ beta constraints ------
	nnz = N + E;

	sense = (char *)malloc(sizeof(char) * N);
	Abeg = (int *)malloc(sizeof(int) * N);
	Aind = (int *)malloc(sizeof(int) * nnz);
	Aval = (double *)malloc(sizeof(double) * nnz);

	Aelt = 0;

	for(i = 0; i < N; i++) {
		sense[i] = 'G';

		Abeg[i] = Aelt;
		
		// Subtract out the alphas around this vertex
		d = flat_line_graph_edge_pair_count(FG, i);
		for(j = 0; j < d; j++) {
			Fep = flat_line_graph_edge_pair(FG, i, j);

			assert(Aelt < nnz);
			Aind[Aelt] = Fep->index;
			Aval[Aelt] = -1;
			Aelt++;
		}

		Aind[Aelt] = E + i;
		Aval[Aelt] = 1;
		Aelt++;
	}
	assert(Aelt == nnz);

	status = CPXaddrows(env, prob, 0, N, nnz, NULL, sense,
											Abeg, Aind, Aval,
											NULL, NULL);

	free(Aval); free(Aind); free(Abeg); free(sense);


	if(status) {
		print_error("Could not copy rows to CPLEX", env, status);
		return status;
	}

	// --------- gamma constraints --------

	sense = (char *)malloc(sizeof(char));
	sense[0] = 'G';

	Abeg = (int *)malloc(sizeof(int));
	Abeg[0] = 0;

	Aind = (int *)malloc(sizeof(int) * maxd);
	Aval = (double *)malloc(sizeof(double) * maxd);

	for(i = 0; i < E; i++) {
    // Grab the ith edge pair
    ep = &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);

		for(a = 0; a < 2; a++) {
			neighbor = ep->neighbors[a];
			if(endpoints != NULL && mask_get(endpoints, neighbor)) {
				continue;
			}

			d = flat_line_graph_edge_pair_count(FG, neighbor);

			// -\alpha_{ijk} term
			assert(0 < maxd);
			Aind[0] = i;
			Aval[0] = -1;
			Aelt = 1;

			for(j = 0; j < d; j++) {
				neighbor_ep = flat_line_graph_edge_pair(FG, neighbor, j);

				assert(neighbor_ep->edge_pair.common_vertex == neighbor);

				// predicate whether this edge pair shares an edge
				// with edge pair i
				if(neighbor_ep->edge_pair.neighbors[0] == ep->common_vertex
					 || neighbor_ep->edge_pair.neighbors[1] == ep->common_vertex) {

					// +\alpha_{kjl} or +\alpha_{lik} term
					assert(Aelt < maxd);
					Aind[Aelt] = neighbor_ep->index;
					Aval[Aelt] = 1;
					Aelt++;
				}
			}

			status = CPXaddrows(env, prob, 0, 1, Aelt, NULL, sense,
													Abeg, Aind, Aval,
													NULL, NULL);
			if(status) {
				free(Aval); free(Aind); free(Abeg); free(sense);
				print_error("Could not copy rows to CPLEX", env, status);
				return status;
			}
		}
	}

	free(Aval); free(Aind); free(Abeg); free(sense);

	// CPXchgobjsen: page 469 of API reference
	CPXchgobjsen(env, prob, CPX_MIN);

	printf("Problem is %dx%d\n",
				 CPXgetnumrows(env, prob), CPXgetnumcols(env, prob));


	return status;
}

static bool ismip(optimize_options_t *opts, optimize_problem_t *prob)
{
	int i;

	for(i = 0; i < prob->ncols; i++) {
		if(prob->ctype[i] != 'C') {
			return true;
		}
	}
	
	return false;
}

static bool isqp(optimize_options_t *opts, optimize_problem_t *prob)
{
	return prob->Q->nnz > 0;
}

static int initialize_problem_struct(
  CPXENVptr env, CPXLPptr prob,
	optimize_options_t *opts,
	optimize_problem_t *in)
{
	int i, j, n, status;
	bool continuous;
	char *ctype;
	matrix_rows_t *Q;

	if(ismip(opts, in)) {
		ctype = in->ctype;
	}
	else {
		ctype = NULL;
	}

	status = CPXnewrows(env, prob, in->nrows, in->rhs, in->sense, NULL, in->rownames);
	status = status || CPXnewcols(env, prob, in->ncols, in->obj, in->lb, in->ub, ctype, in->colnames);
	status = status || CPXchgcoeflist(env, prob, in->A->nnz, in->A->i, in->A->j, in->A->v);

	if(!status && isqp(opts, in)) {
		Q = matrix_list_to_rows(in->Q);
		status = status || CPXcopyquad(env, prob, Q->beg, Q->cnt, Q->ind, Q->val);
		matrix_rows_destroy(Q);
	}

	if(status) {
		print_error("Could initialize problem in CPLEX", env, status);
	}

	// Set objective sense
	if(in->minimize) {
		CPXchgobjsen(env, prob, CPX_MIN);
	}
	else {
		CPXchgobjsen(env, prob, CPX_MAX);
	}

	return status;
}

static optimize_solution_t *get_solution(
  CPXENVptr env, CPXLPptr prob,
	optimize_options_t *opts, optimize_problem_t *inprob)
{
	optimize_solution_t *sol;
	int status;
	char buffer[4096];

	// Allocation solution
	sol = optimize_solution_create(inprob);

	// Get solution status
	status = CPXgetstat(env, prob);

	if(opts->verbose) {
		CPXgetstatstring(env, status, buffer);
		printf("Solution status: %s\n", buffer);
	}

	// Get solution status
	if((status & CPX_STAT_OPTIMAL) != 0) {
		sol->status = OPTIMIZE_OPTIMAL;
	}
	else if((status & CPX_STAT_INFEASIBLE) != 0) {
		sol->status = OPTIMIZE_INFEASIBLE;
	}
	else if((status & CPX_STAT_UNBOUNDED) != 0) {
		sol->status = OPTIMIZE_UNBOUNDED;
	}
	else {
		sol->status = OPTIMIZE_ERROR;
	}

	// Objective value
	status = CPXgetobjval(env, prob, &sol->objective);
	if(status) {
		print_error("Could not retrieve objective value from CPLEX", env, status);
		optimize_solution_destroy(sol);
		return NULL;
	}
	printf("Objective value = %g\n", (float)sol->objective);

	// Solution vector
	status = CPXgetx(env, prob, sol->solution, 0, inprob->ncols-1);
	if(status) {
		print_error("Could not retrieve solution from CPLEX", env, status);
		optimize_solution_destroy(sol);
		return NULL;
	}

	return sol;
}

optimize_solution_t *optimize_cplex(
  optimize_options_t *opts,
	optimize_problem_t *inprob)
{
	CPXENVptr env;
  CPXLPptr prob;
	int status;
	optimize_solution_t *sol;

	sol = NULL;

	// Initialize CPLEX
	status = 0;
	env = CPXopenCPLEX(&status);

	if(env == NULL) {
		print_error("Could not start CPLEX", env, status);
		return NULL;
	}

	status = CPXsetintparam(env, CPX_PARAM_PARALLELMODE, CPX_PARALLEL_AUTO);
	//status = status || CPXsetintparam(env, CPX_PARAM_THREADS, DEFAULT_THREAD_COUNT);
	status = status || CPXsetintparam(env, CPX_PARAM_NETFIND, 1);
	status = status || CPXsetintparam(env, CPX_PARAM_LPMETHOD, CPX_ALG_BARRIER);

	// Allow CPLEX to output messages
	if(opts->verbose) {
		status = status || CPXsetintparam(env, CPX_PARAM_SCRIND, CPX_ON);
		status = status || CPXsetintparam(env, CPX_PARAM_MIPDISPLAY, 4);
	}
	
	if(status) {
		print_error("Could not set parameters", env, status);
		return NULL;
	}

	// Create CPLEX problem
	prob = CPXcreateprob(env, &status, "the nameless one");
	if(status) {
		print_error("Could not create CPLEX problem", env, status);
	}
	else {
		status = initialize_problem_struct(env, prob, opts, inprob);

		if(!status) {
			if(ismip(opts, inprob)) {
				printf("Performing MIP optimization...\n");
				status = CPXmipopt(env, prob);
			}
			else if(isqp(opts, inprob)) {
				printf("Performing QP optimization...\n");
				status = CPXqpopt(env, prob);
			}
			else {
				printf("Performing LP Optimization...\n");
				status = CPXlpopt(env, prob);
			}

			if(status) {
				print_error("Could not solve problem", env, status);
			}
			else {
				sol = get_solution(env, prob, opts, inprob);
			}
		}

		// Free CPLEX problem
		status = CPXfreeprob (env, &prob);
		if(status) {
			print_error("Could not free CPLEX problem", env, status);
		}
	}

	// Shut down cplex
	status = CPXcloseCPLEX(&env);
	if(status) {
		print_error("Could not close CPLEX", env, status);
		return sol;
	}

	return sol;
}


static line_graph_t *extract_solution(
  optimize_options_t *opts, line_graph_t *G, flat_line_graph_t *FG,
  CPXENVptr env, CPXLPptr prob)
{
	int i, n, E, endpoint_count;
	double *x, alpha;

	line_graph_t *H;
	edge_pair_t ep;

	int status;

	H = NULL;

	E = line_graph_total_edge_pair_count(G);

	x = malloc(E * sizeof(double));

	// Call CPXgetx (p179 of API reference)
	status = CPXgetx(env, prob, x, 0, E-1);
	if(status) {
		print_error("Could not retrieve solution from CPLEX", env, status);
	}
	else {
		// Construct H
		H = line_graph_create(flat_line_graph_vertex_count(FG));

		for(i = 0; i < E; i++) {
			alpha = x[i];

			ep = flat_line_graph_flat_edge_pair(FG, i)->edge_pair;
			ep.weight = alpha;

			line_graph_add_edge_pair(H, &ep);
		}


		// Also, can count endpoints
		status = CPXgetx(env, prob, x, E, 2*E-1);
		endpoint_count = 0;
		for(i = 0; i < E; i++) {
			if(x[i] > 0.5) {
				endpoint_count++;
			}
		}
		printf("%d endpoints\n", endpoint_count);

		free(x);
	}


	return H;
}

line_graph_t *optimal_contours_cplex(optimize_options_t *opts, line_graph_t *G, mask_t *endpoints)
{
	CPXENVptr env;
  CPXLPptr prob;
	int status;
	double objval;

  flat_line_graph_t *FG;
	line_graph_t *H;

	char buffer[4096];

	H = NULL;

	// Initialize CPLEX
	status = 0;
	env = CPXopenCPLEX(&status);

	if(env == NULL) {
		print_error("Could not start CPLEX", env, status);
		return NULL;
	}

	CPXsetintparam(env, CPX_PARAM_PARALLELMODE, CPX_PARALLEL_AUTO);
	CPXsetintparam(env, CPX_PARAM_THREADS, DEFAULT_THREAD_COUNT);

	// Allow CPLEX to output messages
	if(opts->verbose) {
		status = CPXsetintparam(env, CPX_PARAM_SCRIND, CPX_ON);
	}
	else {
		status = 0;
	}
	if(status) {
		print_error("Could not set CPX_PARAM_SCRIND", env, status);
	}
	else {

		// Create CPLEX problem
		prob = CPXcreateprob(env, &status, "optimal line graph");
		if(status) {
			print_error("Could not create CPLEX problem", env, status);
		}
		else {
			// Initialize problem
			FG = flat_line_graph_create(G);
			switch(opts->model) {
				case MODEL_ILP1:
				case MODEL_QP1:
					status = initialize_problem(opts, G, FG, endpoints, env, prob);
					break;

				default:
					optimize_model_unhandled(opts);
					break;
			}

			if(!status) {

				// Perform optimization
				switch(opts->model) {
				case MODEL_ILP1:
					printf("Performing MIP optimization...\n");
					status = CPXmipopt(env, prob);
					break;

				case MODEL_QP1:
					printf("Performing QP optimization...\n");
					status = CPXqpopt(env, prob);
					break;

				default:
					optimize_model_unhandled(opts);
					break;
				}


				//CPXsolwrite(env, prob, "/dev/stdout");

				if(status) {
					print_error("Could not solve problem", env, status);
				}
				else {
					// Get solution status
					status = CPXgetstat(env, prob);

					if(opts->verbose) {
						CPXgetstatstring(env, status, buffer);
						printf("Solution status: %s\n", buffer);
					}

					if((status & CPX_STAT_OPTIMAL) != 0) {
						// Extract solution value
						H = extract_solution(opts, G, FG, env, prob);

						CPXgetobjval(env, prob, &objval);
						printf("Objective value = %2g\n", objval);
					}
				}
			}

			flat_line_graph_destroy(FG);
		}

		// Free CPLEX problem
		status = CPXfreeprob (env, &prob);
		if(status) {
			print_error("Could not free CPLEX problem", env, status);
		}
	}

	// Shut down cplex
	status = CPXcloseCPLEX(&env);
	if(status) {
		print_error("Could not close CPLEX", env, status);
		return NULL;
	}

	return H;
}

static void print_error(const char *error, CPXENVptr env, int status)
{
	char cplex_error[4096];

	CPXgeterrorstring(env, status, cplex_error);
	fprintf(stderr, "%s: %s\n", error, cplex_error);
}
