#include "optimize_glpk.h"
#include "flat_line_graph.h"
#include "optimize_internal.h"

#include <glpk.h>

line_graph_t *optimal_contours_glpk(optimize_options_t *opts, line_graph_t *G, mask_t *endpoints)
{
	int i, j, k, n, m, a, b, d;

  glp_smcp lp_options;
  glp_iocp ilp_options;
	glp_prob *mip;
  int status;

  flat_line_graph_t *FG;
  edge_pair_t *ep;
  flat_edge_pair_t *neighbor_ep;

	line_graph_t *H;

  int *rowj;
  double *rowv;
  int neighbor;

  double alpha;
	int endpoint_count;

	// This backend only handles the basic ILP
	if(opts->model != MODEL_ILP1) {
		optimize_model_unhandled(opts);
	}

  FG = flat_line_graph_create(G);
  n = flat_line_graph_total_edge_pair_count(FG);
  d = flat_line_graph_max_pair_degree(FG);

  // Variables to set up a row
  rowj = malloc(sizeof(int) * (3+d));
  rowv = malloc(sizeof(double) * (3+d));
  
  rowv[1] = -1;
  for(k = 1; k < d+2; k++) {
    rowv[k+1] = 1;
  }

  // --------------- Initialize Problem ---------------
	mip = glp_create_prob();
	glp_set_prob_name(mip, "Optimal Countours Subgraph");

	glp_set_obj_dir(mip, GLP_MIN);

	// Set size
	glp_add_rows(mip, 2*n);
	glp_add_cols(mip, 2*n);

  printf("%dx%d\n", glp_get_num_rows(mip), glp_get_num_cols(mip));

	for(i = 0; i < n; i++) {
	  // ----- pair selection indicator variables -----

    // Grab the corresponding edge pair
    ep = &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);

		// binary
		glp_set_col_kind(mip, i+1, GLP_BV);

		// Objective coefficient lambda - weight
		glp_set_obj_coef(mip, i+1, opts->lambda - ep->weight);

		// ----- endpoint variables ----

		// continuous
		glp_set_col_kind(mip, n+i+1, GLP_CV);

		// lower-bounded by 0
		glp_set_col_bnds(mip, n+i+1, GLP_LO, 0, 0);

		// Objective coefficient of mu
		glp_set_obj_coef(mip, n+i+1, opts->mu);

		// ----- endpoint constraints -----
		for(a = 0; a < 2; a++) {
			neighbor = ep->neighbors[a];
			if(endpoints != NULL && mask_get(endpoints, neighbor)) {
				continue;
			}

			m = flat_line_graph_edge_pair_count(FG, neighbor);

			rowj[0+1] = i+1;
			rowj[1+1] = n+i+1;

			k = 0;
			for(j = 0; j < m; j++) {
				neighbor_ep = flat_line_graph_edge_pair(FG, neighbor, j);

				assert(neighbor_ep->edge_pair.common_vertex == neighbor);

				// predicate whether this edge pair shares an edge
				// with edge pair i
				if(neighbor_ep->edge_pair.neighbors[0] == ep->common_vertex
					 || neighbor_ep->edge_pair.neighbors[1] == ep->common_vertex) {

					assert(k < d);
					rowj[2 + (k++) + 1] = neighbor_ep->index + 1;
				}
			}

			glp_set_row_bnds(mip, 2*(i+1) - a, GLP_LO, 0, 0);
			glp_set_mat_row(mip, 2*(i+1) - a, 2+k, rowj, rowv);
		}
	}

	//glp_write_lp(mip, NULL, "problem.txt");

  // --------------- Solve ---------------
  // Choose basis
  glp_adv_basis(mip, 0);

  glp_init_smcp(&lp_options);
  glp_init_iocp(&ilp_options);
  lp_options.meth = GLP_DUALP; // start with dual simplex

  status = glp_simplex(mip, &lp_options);
  assert(status == 0);

  status = glp_intopt(mip, &ilp_options);
  assert(status == 0);

  // Construct line graph
  H = line_graph_create(flat_line_graph_vertex_count(FG));
	endpoint_count = 0;
	for(i = 0; i < n; i++) {
    alpha = glp_mip_col_val(mip, i+1);

		//printf("z = %2g\n", glp_get_col_prim(mip, n+i+1));

    if(alpha < 0.5) {
      continue;
    }

    ep = &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);

    line_graph_add_edge_pair(H, ep);
		endpoint_count += glp_mip_col_val(mip, n+i+1) > 0.5 ? 1 : 0;
  }
	printf("endpoints = %d\n", endpoint_count);

	// Cleanup
	glp_delete_prob(mip);
  flat_line_graph_destroy(FG);

  free(rowj);
  free(rowv);

	return H;
}
