#include "sparse_matrix.h"

matrix_list_t *matrix_list_create(int nnz, int rows, int cols)
{
	matrix_list_t *A;

	A = (matrix_list_t *)malloc(sizeof(matrix_list_t));

	A->rows = rows;
	A->cols = cols;
	A->nnz = nnz;
	A->i = (int *)malloc(sizeof(int) * nnz);
	A->j = (int *)malloc(sizeof(int) * nnz);
	A->v = (double *)malloc(sizeof(double) * nnz);
	
	return A;
}

void matrix_list_destroy(matrix_list_t *A)
{
	if(A == NULL) {
		return;
	}

	free(A->v);
	free(A->j);
	free(A->i);

	free(A);
}

void matrix_list_set_submatrix(matrix_list_t *dst, matrix_list_t *src,
															 int i_offset, int j_offset, int elt_offset)
{
	int *srci, *srcj, srcnnz;
	double *srcv;

	int *dsti, *dstj;
	double *dstv;

	int k;

	srci = src->i;
	srcj = src->j;
	srcv = src->v;
	srcnnz = src->nnz;

	dsti = dst->i;
	dstj = dst->j;
	dstv = dst->v;

	for(k = 0; k < srcnnz; k++)  {
		dsti[k + elt_offset] = srci[k] + i_offset;
		dstj[k + elt_offset] = srcj[k] + j_offset;
		dstv[k + elt_offset] = srcv[k];
	}
}

matrix_rows_t *matrix_rows_create(int rows, int cols, int nnz)
{
	matrix_rows_t *A;

	A = (matrix_rows_t *)malloc(sizeof(matrix_rows_t));

	A->rows = rows;
	A->cols = cols;
	A->nnz = nnz;
	A->beg = (int *)malloc(sizeof(int) * rows);
	A->cnt = (int *)malloc(sizeof(int) * rows);
	A->ind = (int *)malloc(sizeof(int) * nnz);
	A->val = (double *)malloc(sizeof(double) * nnz);
	
	return A;
}

void matrix_rows_destroy(matrix_rows_t *A)
{
	if(A == NULL) {
		return;
	}

	free(A->val);
	free(A->ind);
	free(A->cnt);
	free(A->beg);

	free(A);
}

typedef struct {
	int i, j;
	double v;
} matrix_entry_t;

/**
 * Comparison to sort matrix entries by index
 */
static int matrix_entry_rowmajor_compare(const void *p1, const void *p2)
{
	matrix_entry_t *e1, *e2;

	e1 = (matrix_entry_t *)p1;
	e2 = (matrix_entry_t *)p2;

	if(e1->i < e2->i) {
		return -1;
	}
	else if(e1->i > e2->i) {
		return 1;
	}
	else {
		if(e1->j < e2->j) {
			return -1;
		}
		else if(e1->j > e2->j) {
			return 1;
		}
		else {
			return 0;
		}
	}
}

matrix_rows_t *matrix_list_to_rows(matrix_list_t *A)
{
	int rows, cols, nnz, i, j, k;
	matrix_entry_t *entries;

	int *Ai, *Aj;
	double *Av;

	matrix_rows_t *B;

	int *Bbeg, *Bcnt, *Bind;
	double *Bval;

	Ai = A->i;
	Aj = A->j;
	Av = A->v;
	rows = A->rows;
	cols = A->cols;
	nnz = A->nnz;

	// Get sorted list of the entries of A
	entries = (matrix_entry_t *)malloc(sizeof(matrix_entry_t) * nnz);
	for(k = 0; k < nnz; k++) {
		entries[k].i = Ai[k];
		entries[k].j = Aj[k];
		entries[k].v = Av[k];
	}

	qsort(entries, nnz, sizeof(matrix_entry_t), &matrix_entry_rowmajor_compare);

	// Place into B
	B = matrix_rows_create(rows, cols, nnz);

	Bbeg = B->beg;
	Bcnt = B->cnt;
	Bind = B->ind;
	Bval = B->val;

	// Zero out row sizes
	memset(Bcnt, 0, sizeof(int) * rows);

	i = 0;
	for(k = 0; k < nnz; k++) {
		for(; i < entries[k].i; i++) {
			Bbeg[i+1] = k;
		}

		Bcnt[i]++;

		assert(k == Bbeg[i] + Bcnt[i] - 1);

		Bind[k] = entries[k].j;
		Bval[k] = entries[k].v;
	}

	free(entries);

	assert(Bbeg[rows-1] + Bcnt[rows-1] == nnz);

	return B;
}
