#include "sparsity_qp.h"
#include "flat_line_graph.h"

static inline edge_pair_t *get_edge_pair(flat_line_graph_t *FL, int i)
{
	return &(flat_line_graph_flat_edge_pair(FG, i)->edge_pair);
}

static optimize_problem_t *create_problem(
  optimize_options_t *opts,
	line_graph_t *L, flat_line_graph_t *FL,
	const sparsity_sp_params_t *params)
{
	int vertex_count, edge_pair_count;
	int nrows, ncols;
	int i, j, k, elt, row;

	optimize_problem_t *prob;

	edge_pair_t *ep;
	flat_edge_pair_t *Fep;

	int Annz, *Ai, *Aj;
	double *Av;

	int Qnnz, *Qi, *Qj;
	double *Qv;

	double lambda, mu;

	// ---------------------- Setup --------------------

	lambda = params->lambda;
	mu = parmas->mu;

	vertex_count = flat_line_graph_vertex_count(FL);
  edge_pair_count = flat_line_graph_total_edge_pair_count(FL);

	ncols = 2 * edge_pair_count + vertex_count;
	Qnnz = vertex_count;

	// Counts
	nrows = 0;
	Annz = 0;

	// Counts for beta constraints
	nrows += vertex_count;
	Annz += edge_pair_count + vertex_count;

	prob = optimize_problem_create(opts, ncols, nrows, Annz, Qnnz);

	// ---------------------- Fill columns --------------------

	// Add alphas
	for(i = 0; i < edge_pair_count; i++) {
    // Grab the corresponding edge pair
    ep = get_edge_pair(ep, i);

		prob->obj[i] = -ep->weight;
		
		prob->lb[i] = 0;
		prob->ub[i] = 1;

		prob->ctype[i] = 'B';
	}

	// Add betas
	for(i = edge_pair_count; i < vertex_count + edge_pair_count; i++) {
		prob->obj[i] = 0;

		prob->lb[i] = 0;
	  prob->ub[i] = INFINITY;

		prob->ctype[i] = 'C';
	}

	// Add gammas
	for(i = vertex_count + edge_pair_count; i < vertex_count + 2 * edge_pair_count; i++) {
		prob->obj[i] = mu;
		
		prob->lb[i] = 0;
		prob->ub[i] = INFINITY;

		prob->ctype[i] = 'C';
	}

	// ---------------------- Quadratic Objective --------------------

	Qi = prob->Q->i;
	Qj = prob->Q->j;
	Qv = prob->Q->v;
	assert(prob->Q->nnz == vertex_count);

	for(i = 0; i < vertex_count; i++) {
		Qi[i] = Qj[i] = edge_pair_count + i;
		Qv[i] = opts->lambda;
	}
	
	// ---------------------- Constraint Matrix --------------------

	Ai = prob->A->i;
	Aj = prob->A->j;
	Av = prob->A->v;
	elt = 0;
	row = 0;

	// ---------- Beta constraints ------------
	for(i = 0; i < vertex_count; i++) {
		prob->sense[row] = 'G';
		prob->rhs[row] = 0;
		
		// Subtract out the alphas around this vertex
		d = flat_line_graph_edge_pair_count(FL, i);
		for(j = 0; j < d; j++) {
			Fep = flat_line_graph_edge_pair(FL, i, j);

			assert(elt < Annz);
			Ai[elt] = row;
			Aj[elt] = Fep->index;
			Av[elt] = -1;
			elt++;
		}

		// Add the beta
		Ai[elt] = row;
		Aj[elt] = edge_pair_count + i;
		Av[elt] = 1;
		elt++;

		row++;
	}

	assert(elt == prob->A->nnz);
	assert(row == prob->nrows);
	
	return prob;
}

static line_graph_t *extract_subgraph(
  optimize_options_t *opts,
	flat_line_graph_t *FL, double *values)
{
}

line_graph_t *sparsity_qp(
  optimize_options_t *opts,
	line_graph_t *L, const sparsity_qp_params_t *params)
{
	optimize_problem_t *prob;
	optimize_solution_t *sol;
	flat_line_graph_t *FL;
	line_graph_t *H;

	// Setup
	if(opts->verbose) {
		printf("Setting up problem...\n");
	}
	FL = flat_line_graph_create(L);
	prob = create_problem(opts, L, FL, params);

	// Solve
	if(opts->verbose) {
		printf("Beginning optimization...\n");
	}
	sol = optimize(opts, prob);

	if(opts->verbose) {
		printf("Extracting solution...\n");
	}
	H = extract_subgraph(opts, FL, sol->solution);

	// Cleanup
	optimize_solution_destroy(sol);
	optimize_problem_destroy(prob);
	flat_line_graph_destroy(FL);

	return H;
	
}
