#include "tend.h"
#include "kd_tree.h"
#include "tensor_eigens.h"
#include "util.h"

const tend_opts_t tend_opts_default = {
  0.25, // f
  0.25, // g
	0.2, // step
	M_PI/3, // curvature_limit
	0.4 // FA_limit
};

// TODO: termination mask

static void normalize_step(const tend_opts_t *opts, int dim, point_t p)
{
	double norm;

	norm = point_l2_norm(dim, p);
	point_mul(dim, p, opts->step / norm);
}

static path_t *tract(
  int dim,
	spatial_tree_t *voxels, const tend_opts_t *opts,
	point_t start_pos, point_t start_dir,
	point_t min, point_t max)
{
	path_t *tract;
	spatial_tree_entry_t *nn;

	coord_t angle, FA;
	tensor_eigens_t *eig;
	tensor_t *D;
	point_t e1;
	length_t length2;

	coord_t pos[dim], v_in[dim], v_out[dim], Dv[dim];

	int i, j;

	tract = path_create(dim, 0);
	point_assign(dim, pos, start_pos);
	point_assign(dim, v_in, start_dir);
	normalize_step(opts, dim, v_in);

	length2 = square(opts->step);

	//printf("v_in = "); point_printf(dim, v_in); fputc('\n', stdout);

	// Tracing
	while(true) {
		tract = path_append(tract, pos);

		// Find next tracing direction
		spatial_tree_nearest_neighbor(voxels, pos, 1, &nn);
		D = (tensor_t *)nn->data;
		assert(tensor_dim(D) == dim);
		eig = tensor_eigens_get(D);

		tensor_mult(Dv, D, v_in);
		e1 = tensor_eigenvector(eig, 0);
		if(point_dot_product(dim, v_in, e1) < 0) {
			point_mul(dim, e1, -1);
		}
		point_mul(dim, e1, opts->step);

		for(j = 0; j < dim; j++) {
			v_out[j] = (1 - opts->g) * v_in[j] + opts->g * Dv[j];
			v_out[j] = opts->f * e1[j] + (1 - opts->f) * v_out[j];
		}
		normalize_step(opts, dim, v_out);

		//printf("v_out = "); point_printf(dim, v_out); fputc('\n', stdout);

		// Curvature termination
		angle = acos(point_dot_product(dim, v_in, v_out) / length2);
		if(angle > opts->curvature_limit) {
			//printf("Curvature termination\n");
			return tract;
		}

		// Update
		point_add(dim, pos, pos, v_out);
		point_assign(dim, v_in, v_out);

		// FA termination
		FA = tensor_fractional_anisotropy(eig);
		if(FA < opts->FA_limit) {
			//printf("FA termination\n");
			tract = path_append(tract, pos);
			return tract;
		}

		// Bounds termination
		for(j = 0; j < dim; j++) {
			if(pos[j] < min[j] || pos[j] > max[j]) {
				//printf("Bounds termination\n");
				tract = path_append(tract, pos);
				return tract;
			}
		}
	}
}

bundle_t *tend(tensor_field_t *tf, mask_t *roi, int N, const tend_opts_t *opts)
{
	int i, j, n, dim;
	int roi_size;
	int *roi_indices;
	spatial_tree_options_t treeopts;
	spatial_tree_t *voxels;
	path_t *forward, *backward;
	bundle_t *tracts;
	spatial_tree_entry_t *nn;
	tensor_t *D;

	dim = tensor_field_dim(tf);
	coord_t pos[dim], e1[dim], min[dim], max[dim];

	// Find the indices of ROI points
	if(roi != NULL) {
		roi_size = (int)mask_count_true(roi);
		roi_indices = (int *)malloc(sizeof(int) * roi_size);
		mask_nonzero_indices(roi, roi_indices);
	}
	else {
		roi_size = -1;
		roi_indices = NULL;
	}

	// Generate kD tree to find the nearest neighbor of the tensor field
	treeopts.max_bin_size = 8;
	treeopts.dim = dim;
	voxels = spatial_tree_create(&treeopts);
	if(voxels == NULL) {
		return NULL;
	}

	n = tensor_field_count(tf);
	for(i = 0; i < n; i++) {
		spatial_tree_add(voxels,
										 tensor_field_point(tf, i),
										 tensor_field_tensor(tf, i));
	}

	// Rectangular extent of the tensor field
	tensor_field_extent(tf, min, max);

	// Perform tracings
	tracts = bundle_create(dim, N);
	for(i = 0; i < N; i++) {
		// Sample point
		if(roi == NULL) {
			j = rand() % n;
		}
		else {
			j = roi_indices[rand() % roi_size];
		}
		point_assign(dim, pos, tensor_field_point(tf, j));
		// TODO: neighborhood of point

		// Find initial tracing direction
		spatial_tree_nearest_neighbor(voxels, pos, 1, &nn);
		D = (tensor_t *)nn->data;
		point_assign(dim, e1, tensor_eigenvector(tensor_eigens_get(D), 0));

		// Trace forward
		forward = tract(dim, voxels, opts, pos, e1, min, max);

		// Trace backwards
		point_mul(dim, e1, -1);
		backward = tract(dim, voxels, opts, pos, e1, min, max);
    path_reverse(backward);
    backward = path_resize(backward, path_length(backward) - 1);

		// Add tract
		bundle_set(tracts, i, path_add(backward, forward));
		//bundle_set(tracts, i, backward);

		// Cleanup
		path_destroy(backward);
		path_destroy(forward);
	}

	// Cleanup
	spatial_tree_destroy(voxels);
	free(roi_indices);

	return tracts;
}
