#include "tensor_internal.h"
#include "tensor_eigens.h"

static size_t tensor_data_size(int dim)
{
	return dim * (dim + 1) / 2 * sizeof(tensor_elt_t);
}


static tensor_eigens_t *tensor_eigens(tensor_t *T)
{
	intptr_t p;

	p = (intptr_t)T;
	p += sizeof(tensor_t);
	p += tensor_data_size(T->dim);

	return (tensor_eigens_t *)p;
}

size_t tensor_size(int dim)
{
	size_t s;

	s = sizeof(tensor_t);
	s += tensor_data_size(dim);
	s += tensor_eigens_size(dim);

  return s;
}

tensor_t *tensor_create(int dim)
{
	void *p;

	p = malloc(tensor_size(dim));
	if(p == NULL) {
		return NULL;
	}

	return tensor_create_at(dim, p);
}

tensor_t *tensor_create_at(int dim, void *location)
{
  tensor_t *T;

  T = (tensor_t *)location;


  T->dim = dim;
	T->eigens_updated = 0;
	tensor_eigens_create_at(dim, tensor_eigens(T));

  return T;
}

tensor_t *tensor_copy(tensor_t *src)
{
  tensor_t *dst;

	dst = (tensor_t *)malloc(tensor_size(src->dim));
	if(dst == NULL) {
		return NULL;
	}

	tensor_assign(dst, src);

  return dst;
}

void tensor_destroy(tensor_t *T)
{
  free(T);
}


void tensor_assign(tensor_t *dst, tensor_t *src)
{
	tensor_create_at(src->dim, dst);

	assert(dst->dim == src->dim);

  memcpy(tensor_data(dst), tensor_data(src), tensor_data_size(src->dim));
	if(src->eigens_updated) {
		dst->eigens_updated = 1;
		tensor_eigens_assign(tensor_eigens(dst), tensor_eigens(src));
	}
}

int tensor_dim(tensor_t *T)
{
  return T->dim;
}

static tensor_elt_t *tensor_elt(tensor_t *T, int i, int j)
{
	int k, n;

	assert(i < T->dim && j < T->dim);

  // Swap so j <= i
	if(i < j) {
		k = i;
		i = j;
		j = k;
	}

  n = T->dim;

  // Index in lower-triangular order
  k = i + j*(2*n - (j + 1))/2;

	return tensor_data(T) + k;
}

tensor_elt_t tensor_elt_get(tensor_t *T, int i, int j)
{
	return *tensor_elt(T, i, j);
}

void tensor_elt_set(tensor_t *T, int i, int j, tensor_elt_t v)
{
	*tensor_elt(T, i, j) = v;
	T->eigens_updated = 0;
}


void tensor_mult(point_t prod, tensor_t *Tstruct, point_t v)
{
  int i, j, k, n;
  tensor_elt_t *T; 

  n = Tstruct->dim;
  T = tensor_data(Tstruct);

  k = 0;

  point_zero(n, prod);

  for(i = 0; i < n; i++) {
    prod[i] += (coord_t)T[k++] * v[i];

    for(j = i+1; j < n; j++) {
      prod[i] += (coord_t)T[k] * v[j];
      prod[j] += (coord_t)T[k] * v[i];
      
      k++;
    }
  }
}

coord_t tensor_mult2(tensor_t *Tstruct, point_t v)
{
	int i, j, k, n;
  tensor_elt_t *T;
  coord_t p;

  n = Tstruct->dim;
  T = tensor_data(Tstruct);

	k = 0;
	p = 0;
	for(i = 0; i < n; i++) {
		p += T[k++] * v[i] * v[i];

		for(j = i+1; j < n; j++) {
			p += 2 * T[k++] * v[i] * v[j];
		}
	}

	return p;
}

void tensor_dyadic(tensor_t *Tstruct, point_t v)
{
	int i, j, k, n;
	tensor_elt_t *T;
	
	n = Tstruct->dim;
	T = tensor_data(Tstruct);

	k = 0;
	for(i = 0; i < n; i++) {
		for(j = i; j < n; j++) {
			T[k++] = v[i] * v[j];
		}
	}

	Tstruct->eigens_updated = 0;
}

void tensor_scale(tensor_t *Tstruct, tensor_elt_t s)
{
	int i, n, dim;
	tensor_elt_t *T;

	dim = Tstruct->dim;
	n = dim * (dim + 1) / 2;
	T = tensor_data(Tstruct);
	
	for(i = 0; i < n; i++) {
		T[i] *= s;
	}

	Tstruct->eigens_updated = 0;
}

void tensor_add(tensor_t *astruct, tensor_t *bstruct)
{
	int i, n, dim;
	tensor_elt_t *a, *b;

	dim = astruct->dim;
	assert(dim == bstruct->dim);

	n = dim * (dim + 1) / 2;
	a = tensor_data(astruct);
	b = tensor_data(bstruct);

	for(i = 0; i < n; i++) {
		a[i] += b[i];
	}

	astruct->eigens_updated = 0;
}

gsl_matrix *tensor_as_gsl_matrix(tensor_t *T)
{
  gsl_matrix *A;
	int i, j, k, n;
  
  n = T->dim;

  A = gsl_matrix_alloc(n, n);

	k = 0;
	for(i = 0; i < n; i++) {
    gsl_matrix_set(A, i, i, (double)tensor_data(T)[k++]);

		for(j = i+1; j < n; j++) {
      gsl_matrix_set(A, i, j, (double)tensor_data(T)[k]);
      gsl_matrix_set(A, j, i, (double)tensor_data(T)[k]);
      k++;
    }
  }

  return A;
}

tensor_eigens_t *tensor_eigens_get(tensor_t *T)
{
	if(!T->eigens_updated) {
		tensor_eigens_calc(T, tensor_eigens(T));
		T->eigens_updated = 1;
	}

	return tensor_eigens(T);
}

void tensor_normalize(tensor_t *T)
{
	tensor_eigens_t *eig;
	coord_t lambda1;

	// Get eigensystem
	eig = tensor_eigens_get(T);
	lambda1 = tensor_eigenvalue(eig, 0);

	// Normalize
	tensor_scale(T, 1.0/lambda1);
}

void tensor_zero(tensor_t *T)
{
	int i, n;
	int dim;
	tensor_elt_t *data;
	data = tensor_data(T);

	dim = T->dim;
	n = dim * (dim + 1) / 2;

	for(i = 0; i < n; i++) {
		data[i] = 0;
	}

	T->eigens_updated = 0;
}

// Sets a tensor to the identity
void tensor_identity(tensor_t *T)
{
	int i, dim;

	tensor_zero(T);
	dim = T->dim;

	for(i = 0; i < dim; i++) {
		*tensor_elt(T, i, i) = 1.0;
	}
}

tensor_elt_t tensor_trace(tensor_t *T)
{
	int i, dim;
	tensor_elt_t trace;

	dim = T->dim;

	trace = 0;
	for(i = 0; i < dim; i++) {
		trace += *tensor_elt(T, i, i);
	}

	return trace;
}

// See Bihan et al 2001
tensor_elt_t tensor_mean_diffusivity(tensor_t *T)
{
	return tensor_trace(T) / ((tensor_elt_t)T->dim);
}

void tensor_print(tensor_t *T)
{
	tensor_fprintf_fmt(stdout, T, "% .2f", 1);
}

void tensor_printf(tensor_t *T, int lower)
{
	tensor_fprintf_fmt(stdout, T, "% .2f", lower);
}

void tensor_fprintf(FILE *fp, tensor_t *T, int lower)
{
	tensor_fprintf_fmt(fp, T, "% .2f", lower);
}

int tensor_snprintf(size_t buflen, char *buf, tensor_t *T, int lower)
{
	return tensor_snprintf_fmt(buflen, buf, T, "% .2f", lower);
}

void tensor_printf_fmt(tensor_t *T, const char *coord_fmt, int lower)
{
	tensor_fprintf_fmt(stdout, T, coord_fmt, lower);
}

void tensor_fprintf_fmt(FILE *fp, tensor_t *T, const char *coord_fmt, int lower)
{
	int i, j, n;

  n = T->dim;

	for(i = 0; i < n; i++) {
		fputc(i == 0 ? '[' : ' ', fp);

		for(j = lower ? 0 : i; j < n; j++) {
			fprintf(fp, coord_fmt, (float)tensor_elt_get(T, i, j));

			if(j != n-1) {
				fputc(',', fp);
				fputc(' ', fp);
			}
		}

		fputc(i == n-1 ? ']': ';', fp);
		fputc('\n', fp);
	}
}

int tensor_snprintf_fmt(
  size_t buflen, char *buf,
  tensor_t *T,
  const char *coord_fmt, int lower)
{
	// TODO

	return 0;
}
