#include "tensor_eigens.h"
#include "tensor_internal.h"

#include <gsl/gsl_eigen.h>
#include <f2c.h>

#include "util.h"

struct tensor_eigens {
  int dim;
};

static size_t data_size(int dim)
{
	return sizeof(coord_t) * dim * (dim + 1);
}

static coord_t *data(tensor_eigens_t *eig)
{
	return (coord_t *)((intptr_t)eig + sizeof(tensor_eigens_t));
}

size_t tensor_eigens_size(int dim)
{
	size_t s;

  s  = sizeof(tensor_eigens_t);
  s += data_size(dim);

	return s;
}

tensor_eigens_t *tensor_eigens_create(int dim)
{
	void *p;

  p = (tensor_eigens_t *)malloc(tensor_eigens_size(dim));
	if(p == NULL) {
		return NULL;
	}

	return tensor_eigens_create_at(dim, p);
}

tensor_eigens_t *tensor_eigens_create_at(int dim, void *location)
{
  tensor_eigens_t *e;
  intptr_t p;

  e = (tensor_eigens_t *)location;
  e->dim = dim;

  return e;
}

tensor_eigens_t *tensor_eigens_copy(tensor_eigens_t *src)
{
  tensor_eigens_t *dst;

	dst = (tensor_eigens_t *)malloc(tensor_eigens_size(src->dim));
	if(dst == NULL) {
		return NULL;
	}

	tensor_eigens_assign(dst, src);

  return dst;
}

void tensor_eigens_destroy(tensor_eigens_t *e)
{
  free(e);
}

void tensor_eigens_assign(tensor_eigens_t *dst, tensor_eigens_t *src)
{
	tensor_eigens_create_at(src->dim, dst);
	memcpy(data(dst), data(src), data_size(src->dim));
}

////////////////////////// Accessors /////////////////////

coord_t tensor_eigenvalue(tensor_eigens_t *e, int i)
{
	assert(i < e->dim);
	
  return data(e)[e->dim - 1 - i];
}

point_t tensor_eigenvector(tensor_eigens_t *e, int i)
{
	assert(i < e->dim);

  return data(e) + e->dim * (e->dim - 1 - i + 1);
}

// LAPACK routine for eigendecomposition 
// http://netlib.org/lapack/lug/node48.html
// http://www.netlib.org/lapack/double/dspev.f
extern int dspev_(
  char *jobz, char *uplo, integer *n, doublereal *ap, 
  doublereal *w, doublereal *z, integer *ldz, doublereal *work, 
	integer *info);

// TODO: can take advantage of dspevx... maybe
// http://www.netlib.org/lapack/double/dspevx.f

// TODO: profile that dsyev isn't any faster
// http://www.netlib.org/lapack/lapack-3.1.1/SRC/dsyev.f

// http://localhost/doc/liblapack-doc/lug/node32.html#tabdriveseig

void tensor_eigens_calc_lapack(tensor_t *T, tensor_eigens_t *eigs)
{
  assert(T->dim == eigs->dim);

  integer n = T->dim;
  int i, m = n * (n + 1) / 2;

	char jobz, uplo;
	integer ldz, info;

	double *w, *z;
	double AP[m], work[3*n];

	tensor_elt_t *Tdata;

	Tdata = tensor_data(T);
  for(i = 0; i < m; i++) {
    AP[i] = (double)Tdata[i];
  }

	jobz = 'V';
	uplo = 'L';

	w = data(eigs);
	z = data(eigs) + n;

	ldz = T->dim;

	dspev_(&jobz, &uplo, &n, AP, w, z, &ldz, work, &info);

	assert(info == 0);
}


void tensor_eigens_calc_gsl(tensor_t *T, tensor_eigens_t *eigs)
{
  assert(T->dim == eigs->dim);

  gsl_matrix *A, *vecs;
  gsl_eigen_symmv_workspace *work;
  gsl_vector *vals;
  int i, j, k, d, n;

  coord_t swap;
  coord_t *lambda, *V, *W;

  // Perform GSL eigendecomposition
  n = T->dim;
  A = tensor_as_gsl_matrix(T);

  work = gsl_eigen_symmv_alloc(n);
  vals = gsl_vector_alloc(n);
  vecs = gsl_matrix_alloc(n, n);
  
  gsl_eigen_symmv(A, vals, vecs, work);

  // Copy into output structure
  lambda = data(eigs);
  
  for(i = 0; i < n; i++) {
    lambda[i] = gsl_vector_get(vals, i);

    V = data(eigs) + (i + 1) * n;
    for(j = 0; j < n; j++) {
      V[j] = gsl_matrix_get(vecs, j, i);
    }

    /*
    if(lambda[i] < 0) {
      lambda[i] *= -1;
      for(j = 0; j < n; j++) {
        V[j] *= -1;
      }
    }
    */
  }

  // Order lambda by an insertion sort
  for(i = 0; i < n; i++) {
    // Find maximum
    k = i;
    for(j = i+1; j < n; j++) {
      if(lambda[j] < lambda[k]) {
        k = j;
      }
    }

    // Swap
    swap = lambda[i];
    lambda[i] = lambda[k];
    lambda[k] = swap;

    V = data(eigs) + (i + 1) * n;
    W = data(eigs) + (k + 1) * n;
    for(d = 0; d < n; d++) {
      swap = V[d];
      V[d] = W[d];
      W[d] = swap;
    }
  }
  

  // Cleanup
  gsl_matrix_free(vecs);
  gsl_vector_free(vals);
  gsl_eigen_symmv_free(work);

  gsl_matrix_free(A);
}

void tensor_principal_vector(tensor_t *T, point_t v)
{
	int dim;
	tensor_eigens_t *eigs;

	dim = T->dim;
	eigs = tensor_eigens_get(T);

	point_assign(dim, v, tensor_eigenvector(eigs, 0));
	point_mul(dim, v, tensor_eigenvalue(eigs, 0));
}

void tensor_eigens_print(tensor_eigens_t *e)
{
  int i, n;

  n = e->dim;

  for(i = 0; i < n; i++) {
    printf("% .2f, ", tensor_eigenvalue(e, i));
    point_printf(n, tensor_eigenvector(e, i));
    printf("\n");
  }
}

// Find average of eigenvalues
static coord_t avg_eigenvalue(tensor_eigens_t *e)
{
	int i, dim;
	coord_t avg;

	dim = e->dim;

 	avg = 0;
	for(i = 0; i < dim; i++) {
		avg += tensor_eigenvalue(e, i);
	}
	avg /= (coord_t)dim;

	return avg;
}

// Basser & Pierpaoli 1996
coord_t tensor_relative_anisotropy(tensor_eigens_t *e)
{
	coord_t avg, numer;
	int i, dim;

	dim = e->dim;

	avg = avg_eigenvalue(e);

	numer = 0;
	for(i = 0; i < dim; i++) {
		numer += square(tensor_eigenvalue(e, i) - avg);
	}

	// TODO: constants dimension-dependent?
	return sqrt(numer / (3 * avg));
}

// See Bihan et al 2001 (or Basser & Pierpaoli 1996?)
coord_t tensor_fractional_anisotropy(tensor_eigens_t *e)
{
	coord_t avg, lambda_i, numer, denom;
	int i, dim;

	dim = e->dim;

	avg = avg_eigenvalue(e);

	// Find the numerator and denominator of FA
	numer = denom = 0;
	for(i = 0; i < dim; i++) {
		lambda_i = tensor_eigenvalue(e, i);

		numer += square(lambda_i - avg);
		denom += square(lambda_i);
	}

	// TODO: constants dimension-dependent?
	return sqrt((3*numer) / (2*denom));
}

// See Bihan et al 2001
coord_t tensor_relative_volume_ratio(tensor_eigens_t *e)
{
	coord_t avg, prod;
	int i, dim;

	dim = e->dim;

	avg = avg_eigenvalue(e);

	prod = 1;
	for(i = 0; i < dim; i++) {
		prod *= tensor_eigenvalue(e, i) / avg;
	}
	
	return avg;
}

/*
#include "timing.h"

#define TRIALS (1024 * 1024)

int main(int argc, char **argv)
{
  tensor_t **T;
  tensor_eigens_t *eig;
  int i, j;
  struct timeval timer;
  double duration;

  srand(time(NULL));

  // Create tensors
  T = (tensor_t **)malloc(TRIALS * sizeof(tensor_t *));
  for(i = 0; i < TRIALS; i++) {
    T[i] = tensor_create(3);
    for(j = 0; j < 6; j++) {
      T[i]->data[j] = frand(0, 1);
    }
  }

  eig = tensor_eigens_create(3);



  printf("\nGSL: \n");
  timer_start(&timer);
  for(i = 0; i < TRIALS; i++) {
    tensor_eigens_gsl(T[i], eig);
  }
  duration = timer_interval(&timer);

  time_print(stdout, duration);
  printf(" (");
  time_print(stdout, duration / TRIALS);
  printf(")\n\n");

  printf("\nLapack: \n");
  timer_start(&timer);
  for(i = 0; i < TRIALS; i++) {
    tensor_eigens_lapack(T[i], eig);
  }
  duration = timer_interval(&timer);

  time_print(stdout, duration);
  printf(" (");
  time_print(stdout, duration / TRIALS);
  printf(")\n\n");




  tensor_eigens_destroy(eig);
  for(i = 0; i < TRIALS; i++) {
    tensor_destroy(T[i]);
  }
  free(T);

  return 0;
}
*/
