#include "tensor_field.h"
#include "util.h"

struct tensor_field
{
	int dim, count;
  size_t pair_size;
};


static inline size_t pair_size(int dim)
{
  return dim * sizeof(coord_t) + tensor_size(dim);
}

static inline intptr_t pair_data(tensor_field_t *tf, int i)
{
  return (intptr_t)tf + sizeof(tensor_field_t) + i * tf->pair_size;
}


// Construct/destruct
tensor_field_t *tensor_field_create_grid(
  int dim,
  point_t origin, point_t spacing,
  int *size)
{
	int count, i, j, k, l;
	tensor_field_t *tf;
	point_t p;

	// Create the tensor field structure
	count = 1;
	for(i = 0; i < dim; i++) {
		count *= size[i];
	}

	tf = tensor_field_create_free(dim, count);

	// Fill in point locations according to a grid pattern
	for(i = 0; i < count; i++) {
    p = tensor_field_point(tf, i);

		k = i;

		for(j = 0; j < dim; j++) {
			l = k % size[j];
			k /= size[j];

			p[j] = origin[j] + spacing[j] * (coord_t)l;
		}
	}

	return tf;
}

tensor_field_t *tensor_field_create_free(int dim, int count)
{
	int i;
	size_t s;
	tensor_field_t *tf;

	// Allocate
	s = sizeof(tensor_field_t);
	s += count * pair_size(dim);

	tf = (tensor_field_t *)malloc(s);

	// Initialize
  tf->dim = dim;
	tf->count = count;
  tf->pair_size = pair_size(dim);

	for(i = 0; i < count; i++) {
		tensor_create_at(dim, tensor_field_tensor(tf, i));
	}
	
	return tf;
}

tensor_field_t *tensor_field_resize(tensor_field_t *tf, int new_count)
{
	size_t s;
	int i, dim, old_count;

	old_count = tf->count;

	s = sizeof(tensor_field_t);
	s += new_count * tf->pair_size;

	tf = (tensor_field_t *)realloc(tf, s);
	tf->count = new_count;

	dim = tf->dim;
	for(i = old_count; i < new_count; i++) {
		tensor_create_at(dim, tensor_field_tensor(tf, i));
	}

	return tf;
}

tensor_field_t *tensor_field_clone(tensor_field_t *tf)
{
	tensor_field_t *copy;

  copy = tensor_field_create_free(tf->dim, tf->count);

  memcpy((void *)((intptr_t)copy + sizeof(tensor_field_t)),
				 (const void *)((intptr_t)tf + sizeof(tensor_field_t)),
				 tf->count * tf->pair_size);

  return copy;
}

void tensor_field_destroy(tensor_field_t *tf)
{
	free(tf);
}

// Retreiving data
int tensor_field_dim(tensor_field_t *tf)
{
  return tf->dim;
}

int tensor_field_count(tensor_field_t *tf)
{
	return tf->count;
}

point_t tensor_field_point(tensor_field_t *tf, int i)
{
  intptr_t p;
  coord_t *q;

	assert(i < tf->count);

  p = pair_data(tf, i);
  q = (coord_t *)p;

	return q;
}

tensor_t *tensor_field_tensor(tensor_field_t *tf, int i)
{
	intptr_t p;

	assert(i < tf->count);

	p = pair_data(tf, i);
	p += sizeof(coord_t) * tf->dim;

	return (tensor_t *)p;
}

tensor_field_t *tensor_field_add(tensor_field_t *tf, point_t p, tensor_t *T)
{
  int i, dim;
  tensor_t *newT;

  dim = tf->dim;
  i = tf->count;

  // Resize
  tf = tensor_field_resize(tf, tf->count + 1);

  // Assign appended point
  point_assign(dim, tensor_field_point(tf, i), p);

  // Assign appended tensor
  newT = tensor_field_tensor(tf, i);
  tensor_assign(newT, T);

	tensor_elt_get(tensor_field_tensor(tf, i), 0, 0);

  return tf;
}

void tensor_field_extent(tensor_field_t *tf, point_t min, point_t max)
{
  int i, j, n, dim;
  point_t p;

  dim = tensor_field_dim(tf);;
  n = tensor_field_count(tf);
  if(n <= 0) {
    return;
  }

  // Initialize extent
  p = tensor_field_point(tf, 0);
  point_assign(dim, min, p);
  point_assign(dim, max, p);

  for(i = 1; i < n; i++) {
    p = tensor_field_point(tf, i);

    for(j = 0; j < dim; j++) {
      min[j] = MIN(min[j], p[j]);
      max[j] = MAX(max[j], p[j]);
    }
  }
}
