#include <pthread.h>

#include "weight_file.h"
#include "weight_vector_internal.h"

#include "io.h"
#include "util.h"
#include "parallel.h"

struct weight_file {
	FILE *fp;
	int dim;
};

weight_file_t *weight_file_create(FILE *fp, int dim)
{
	weight_file_t *f;

	// Alocate structure
	f = malloc(sizeof(weight_file_t));
	f->fp = fp;
	f->dim = dim;

	// Write file
	write_int32(fp, (int32_t)dim);

	return f;
}

weight_file_t *weight_file_open(FILE *fp)
{
	weight_file_t *f;

	// Alocate structure
	f = malloc(sizeof(weight_file_t));
	f->fp = fp;

	// Read header
	f->dim = (int)read_int32(fp);

	return f;
}

void weight_file_close(weight_file_t *f)
{
	fclose(f->fp);
	free(f);
}

int weight_file_dimensionality(weight_file_t *f)
{
	return f->dim;
}

void weight_file_write(weight_file_t *f, weight_vector_t *w)
{
	int i;

	assert(f->dim == w->dim);

	// angle_count
	write_int32(f->fp, (int32_t)w->angle_count);

	// pos
	for(i = 0; i < 2; i++) {
		write_doubles(f->fp, w->dim, w->pos[i]);
	}

	// weights
	write_doubles(f->fp, weight_vector_total_angles(w), w->weights);
}

weight_vector_t *weight_file_read(weight_file_t *f)
{
	int i, dim, angle_count;
	weight_vector_t *w;

	// Read angle count
	dim = f->dim;
	angle_count = (int)read_int32(f->fp);
	if(feof(f->fp)) {
		return NULL;
	}

	// Allocate
	w = weight_vector_create(dim, angle_count);

	// Fill
	for(i = 0; i < 2; i++) {
		read_doubles(f->fp, dim, w->pos[i]);
	}

	// Weights
	read_doubles(f->fp, weight_vector_total_angles(w), w->weights);

	return w;
}

///////////////////// Loading into memory /////////////////////

weight_vector_array_t *weight_vector_array_from_file(weight_file_t *wf)
{
	weight_vector_t *v;
	weight_vector_array_t *a;

	a = weight_vector_array_create(0);
	while((v = weight_file_read(wf)) != NULL) {
		a = weight_vector_array_resize(a, a->length + 1);

		a->w[a->length - 1] = v;
	}

	return a;
}

weight_vector_array_t *weight_vector_array_from_file_name(const char *filename)
{
	weight_vector_array_t *wa;
	weight_file_t *wf;
	FILE *fp;

	fp = fopen(filename, "rb");
	wf = weight_file_open(fp);

	wa = weight_vector_array_from_file(wf);

	weight_file_close(wf);

	return wa;
}

void weight_vector_array_to_file(weight_file_t *wf, weight_vector_array_t *wa)
{
	int i, n;

	n = wa->length;
	for(i = 0; i < n; i++) {
		weight_file_write(wf, wa->w[i]);
	}
}

void weight_vector_array_to_file_name(const char *filename, weight_vector_array_t *wa)
{
	weight_file_t *wf;
	int dim;
	FILE *fp;

	assert(wa->length > 0);

	fp = fopen(filename, "wb+");
	assert(fp != NULL);

	dim = weight_vector_dim(wa->w[0]);
	assert(dim == 2 || dim == 3);

	wf = weight_file_create(fp, dim);
	assert(wf != NULL);

	weight_vector_array_to_file(wf, wa);

	weight_file_close(wf);
}
