#include "weight_vector_internal.h"
#include "affine.h"

#include "util.h"
#include "indices.h"
#include "parallel.h"
#include "progress.h"

#include <pthread.h>

static affine_t *canonical_transform(int dim, point_t p[2], point_t normed[2])
{
  int i;
  length_t norm[2];
	affine_t *A;
  angle_t a1, a2, a3;
  coord_t xy[3], x[3], intermediate[3];

	coord_t flip[dim];

  // Order the points by their norm
  for(i = 0; i < 2; i++) {
    norm[i] = point_l2_norm(dim, p[i]);
  }

  if(norm[0] < norm[1]) {
    point_assign(dim, normed[0], p[1]);
    point_assign(dim, normed[1], p[0]);
  }
  else {
    point_assign(dim, normed[0], p[0]);
    point_assign(dim, normed[1], p[1]);
  }

  // Rotate first point to x-axis
  A = affine_create(dim);

  if(dim == 2) {
    a1 = -atan2(normed[0][1], normed[0][0]);

    affine_rotation2(A, a1);
  }
  else {
    assert(dim == 3);

    x[0] = 1; x[1] = x[2] = 0;
    //z[0] = z[1] = 0; z[2] = 1;

    // Rotate to x-y plane
    a1 = atan2(normed[0][2], normed[0][1]);
    affine_rotation3(A, a1, x);
    affine_transform(A, xy, normed[0]);

    // Rotate onto x
    a2 = atan2(xy[1], xy[0]);
    affine_pre_rotate2(A, a2);

    // Rotate second vector onto x-y plane
    affine_transform(A, intermediate, normed[1]);
    a3 = -atan2(intermediate[2], intermediate[1]);
    affine_pre_rotate3(A, a3, x);
  }

	// Apply
	affine_transform(A, flip, normed[1]);

  // Flip second point to positive y,z
	flip[0] = 1;
  for(i = 1; i < dim; i++) {
		if(flip[i] < 0) {
			flip[i] = -1;
		}
		else {
			flip[i] = 1;
		}
  }

	affine_pre_scale(A, flip);

	affine_transform(A, normed[0], normed[0]);
	affine_transform(A, normed[1], normed[1]);

	return A;
}

void canonical_point_pair(int dim, point_t p[2], point_t normed[2])
{
  affine_t *A;

	A = canonical_transform(dim, p, normed);
  affine_destroy(A);
}

////////////////////////////// Weight Vector /////////////////////

weight_vector_t *weight_vector_create(int dim, int angle_count)
{
  int i;
	size_t s;
	weight_vector_t *w;
	intptr_t p;

  assert(dim == 2 || dim == 3);

	// Allocate
	s = sizeof(weight_vector_t);
	s += sizeof(coord_t) * dim * 2;
  s += sizeof(int) * 3;
	s += sizeof(double) * icube(angle_count);

	w = (weight_vector_t *)malloc(s);
  if(w == NULL) {
    return NULL;
  }

	// Fill
	w->dim = dim;
	w->angle_count = angle_count;

	p = (intptr_t)w;

	p += sizeof(weight_vector_t);
	w->pos[0] = (point_t)p;

	p += sizeof(coord_t) * dim;
	w->pos[1] = (point_t)p;

	p += sizeof(coord_t) * dim;
  w->size = (int *)p;
  for(i = 0; i < 3; i++) {
    w->size[i] = angle_count;
  }

  p += sizeof(int) * 3;
	w->weights = (double *)p;

	return w;
}

void weight_vector_destroy(weight_vector_t *w)
{
  free(w);
}

int weight_vector_dim(weight_vector_t *w)
{
  return w->dim;
}

point_t weight_vector_pos(weight_vector_t *w, int i)
{
  return w->pos[i];
}

int weight_vector_angle_count(weight_vector_t *w)
{
  return w->angle_count;
}

int weight_vector_total_angles(weight_vector_t *w)
{
  int total_angles;

  assert(w->dim == 2 || w->dim == 3);
  total_angles = icube(w->angle_count);

  return total_angles;
}

double *weight_vector_elt(weight_vector_t *w, int *i)
{
  int j, ind, dim;

  dim = w->dim;
  assert(dim == 2 || dim == 3);

  ind = sub2ind(3, w->size, i);

  return w->weights + ind;
}

static inline angle_t angle2d(int i, int angle_count)
{
	angle_t angle;

	angle = M_PI * (angle_t)i / angle_count;

	return angle;
}

static spherical_orientation_t SPHERE_ANGLES[] = {
  //  theta    phi
  {       0,        0},
  {       0,   M_PI/2},
  {  M_PI/2,   M_PI/2},

	{  M_PI/4,   M_PI/4},
  {3*M_PI/4,   M_PI/4},
	{3*M_PI/4, 3*M_PI/4}
};

static inline spherical_orientation_t angle3d(int i, int angle_count)
{
	spherical_orientation_t angle;

	switch(angle_count) {
	case 3:
	case 6:
		return SPHERE_ANGLES[i];
	default:
		fprintf(stderr, "Cannot handle angle_count = %d\n", angle_count);
		exit(1);
	}

	return angle;
}

void weight_vector_unit_vectors(weight_vector_t *w, int ind, point_t p[3])
{
	angle_t a2;
	spherical_orientation_t a3;
	int sub[3], angle_count;
	int i, dim;
	
	dim = weight_vector_dim(w);
  angle_count = w->angle_count;

  ind2sub(3, w->size, ind, sub);

	if(dim == 2) {
		for(i = 0; i < 3; i++) {
			a2 = angle2d(sub[i], angle_count);

			p[i][0] = cos(a2);
			p[i][1] = sin(a2);
		}
	}
	else {
		assert(dim == 3);

		for(i = 0; i < 3; i++) {
			a3 = angle3d(sub[i], angle_count);

			p[i][0] = cos(a3.theta) * sin(a3.phi);
			p[i][1] = sin(a3.theta) * sin(a3.phi);
			p[i][2] = cos(a3.phi);
		}
	}
}

///////////////////////// Arrays of Weight Vectors ////////////////////
static size_t wva_size(int len)
{
	return sizeof(weight_vector_array_t) + sizeof(weight_vector_t *) * len;
}

weight_vector_array_t *weight_vector_array_create(int len)
{
	weight_vector_array_t *a;
	int i;
	intptr_t p;

	a = (weight_vector_array_t *)malloc(wva_size(len));
	if(a == NULL) {
		return NULL;
	}

	a->length = len;

	p = (intptr_t)a;
	p += sizeof(weight_vector_array_t);
	a->w = (weight_vector_t **)p;

	// Initialize pointers to zero
	for(i = 0; i < len; i++) {
		a->w[i] = NULL;
	}

	return a;
}

weight_vector_array_t *weight_vector_array_resize(weight_vector_array_t *a, int new_len)
{
	int i, old_len;
	intptr_t p;

	old_len = a->length;

	// Delete entries no longer covered
	for(i = old_len - 1; i >= new_len; i--) {
		weight_vector_destroy(a->w[i]);
	}

	// Change storage size
	a = (weight_vector_array_t *)realloc(a, wva_size(new_len));
	if(a == NULL) {
		return NULL;
	}

	p = (intptr_t)a;
	p += sizeof(weight_vector_array_t);
	a->w = (weight_vector_t **)p;

	// Zero out new entries
	for(i = old_len; i < new_len; i++) {
		a->w[i] = NULL;
	}

	a->length = new_len;

	return a;
}

weight_vector_t *weight_vector_array_get(
  weight_vector_array_t *a,
	length_t tol, int dim, point_t pos[2])
{
	int i, n;
	weight_vector_t *v;
	length_t dist;

	if(a == NULL) {
		return NULL;
	}

	n = a->length;
	for(i = 0; i < n; i++) {
		v = a->w[i];
		if(v == NULL) {
			continue;
		}

		assert(v->dim == dim);

		dist = MAX(point_linfty_dist(dim, pos[0], v->pos[0]),
							 point_linfty_dist(dim, pos[1], v->pos[1]));

		if(dist < tol) {
			return v;
		}
	}

	return NULL;
}

void weight_vector_array_destroy(weight_vector_array_t *a)
{
	int i, n;

	if(a == NULL) {
		return;
	}
	
	// Zero out vectors
	n = a->length;
	for(i = 0; i < n; i++) {
		weight_vector_destroy(a->w[i]);
	}

	free(a);
}

weight_vector_array_t *weight_vector_array_append(weight_vector_array_t *a, weight_vector_t *v)
{
	int n;

	n = a->length;

	a = weight_vector_array_resize(a, n + 1);
	a->w[n] = v;

	return a;
}

///////////////////////// Generating Arrays ////////////////////

weight_vector_array_t *weight_vector_array_for_line_graph(
  line_graph_t *L,
  tensor_field_t *tf,
  length_t tol,
  int angle_count)
{
  int i, j, k, n, d, dim;
  edge_pair_t *ep;
  
  point_t p[2], normed[2], q[3];

  weight_vector_t *wv;
  weight_vector_array_t *wva;

  // Stack storage
  dim = tensor_field_dim(tf);
  coord_t coords[4][dim];
  for(i = 0; i < 2; i++) {
    p[i] = coords[i];
    normed[i] = coords[2+i];
  }

  // Initialize
  wva = weight_vector_array_create(0);

  // Iterate over edge pairs
  n = line_graph_vertex_count(L);
  for(i = 0; i < n; i++) {
    d = line_graph_edge_pair_count(L, i);

    for(j = 0; j < d; j++) {
      ep = line_graph_edge_pair(L, i, j);

      // Extract triplet
      q[0] = tensor_field_point(tf, ep->neighbors[0]);
      q[1] = tensor_field_point(tf, ep->common_vertex);
      q[2] = tensor_field_point(tf, ep->neighbors[1]);

      // Zero center (put in p)
      for(k = 0; k < 2; k++) {
        point_sub(dim, p[k], q[2*k], q[1]);
      }

      // Find canonical points
      canonical_point_pair(dim, p, normed);

      // Find if it's already in the weight vector
      wv = weight_vector_array_get(wva, tol, dim, normed);
      if(wv != NULL) {
        continue;
      }

      // Add a new zero weight vector
      wv = weight_vector_create(dim, angle_count);
      assert(wv != NULL);
      for(k = 0; k < 2; k++) {
        point_assign(dim, wv->pos[k], normed[k]);
      }

			wva = weight_vector_array_append(wva, wv);
    }
  }

  return wva;
}

typedef struct {
  weight_vector_array_t *wva;

  double (*func)(void *data, int dim, point_t pos[3], point_t direction[3]);
	void *func_data;

	time_t start;
	int progress, total;
	pthread_mutex_t lock;

} wva_calc_t;

static int calcvector(int job_i, void *closure)
{
	wva_calc_t *data;
  weight_vector_t *wv;

	int i, j, k, n, m, dim, angle_count, total_angles;

	point_t points[3];
	point_t directions[3];
  angle_t angles2[3];
  spherical_orientation_t angles3[3];
	double w;
  
	// Unpack job
	data = (wva_calc_t *)closure;
  assert(job_i < data->wva->length);
  wv = data->wva->w[job_i];

  dim = wv->dim;
	angle_count = wv->angle_count;
	total_angles = weight_vector_total_angles(wv);

	// Copy points
	coord_t coords[6][dim];
	for(i = 0; i < 3; i++) {
		points[i] = coords[i];
		directions[i] = coords[3+i];
	}

	point_zero(dim, points[1]);
	for(i = 0; i < 2; i++) {
		point_assign(dim, points[2*i], wv->pos[i]);
	}

	// Iterate over possible angles
	for(i = 0; i < total_angles; i++) {
		// Generate weight for the given angle
		weight_vector_unit_vectors(wv, i, directions);
    w = (*data->func)(data->func_data, dim, points, directions);   
    wv->weights[i] = w;

		// Write out progress
		pthread_mutex_lock(&data->lock);
		data->progress++;
		if(data->progress % 8 == 0) {
			progress(data->start, data->progress, data->total, 1);
		}
		pthread_mutex_unlock(&data->lock);
	}

  return 0;
}

int weight_vector_array_total_weights(weight_vector_array_t *wva)
{
	int i, total;

	total = 0;
	for(i = 0; i < wva->length; i++) {
		total += weight_vector_total_angles(wva->w[i]);
	}

	return total;
}


void weight_vector_array_calculate(
  weight_vector_array_t *wva,
  int thread_count,
  double (*weight_closure)(void *data, int dim, point_t pos[3], point_t direction[3]), void *closure_data)
{
	wva_calc_t data;

	// Initialize closure data
  data.wva = wva;

	data.func = weight_closure;
	data.func_data = closure_data;

	data.progress = 0;
	data.total = weight_vector_array_total_weights(wva);
	pthread_mutex_init(&data.lock, NULL);
	data.start = time(NULL);

	// Run!
	parallel_execute(wva->length, thread_count, &calcvector, &data);

	// Cleanup
	pthread_mutex_destroy(&data.lock);
}

/////////////////////// Calculating Graph Weights ////////////////////

const weight_vector_calc_opts_t weight_vector_calc_opts_defaults = {
	NULL, // array
	1e-4 // tolerance
};

double weight_vector_calc(
  int dim, weight_options_t *wopts,
	point_t points[3], tensor_t *tensors[3])
{
	int i, j;

	weight_vector_calc_opts_t *opts;
	weight_vector_array_t *a;
	length_t tol;

	point_t pos[2], normed[2], u[3];
	affine_t *A;
	weight_vector_t *wv;

	double w, prob, total_prob, I;
	int angle_count, total_angles;

	coord_t v[dim];
	coord_t normed_coords[2][dim];
	coord_t u_coords[3][dim];

	// Unpack closure
	opts = (weight_vector_calc_opts_t *)wopts->closure_data;
	a = opts->weights;
	tol = opts->tol;

	// Assert center point is 0
	for(i = 0; i < dim; i++) {
		assert(points[1][i] == 0);
	}

	// Find weight vector
	// and transformation into canonical space
	for(i = 0; i < 2; i++) {
		pos[i] = points[2*i];
		normed[i] = normed_coords[i];
	}
	A = canonical_transform(dim, pos, normed);
	wv = weight_vector_array_get(a, tol, dim, normed);
  if(wv == NULL) {
    // Offsets for the line graph not in the weight vector array
    fputs("Cannot find weight vector for:", stderr);
    point_fprintf(stderr, dim, normed[0]);
    fputs(", ", stderr);
    point_fprintf(stderr, dim, normed[1]);
    fputc('\n', stderr);
    exit(1);
  }
	
  assert(dim == weight_vector_dim(wv));

	// Set up dot product iteration
	total_angles = weight_vector_total_angles(wv);
	for(i = 0; i < 3; i++) {
		u[i] = u_coords[i];
	}
	I = 0;
	total_prob = 0;

	// Iterate over possible angles
	for(i = 0; i < total_angles; i++) {
		// Weight for the given angle
		w = wv->weights[i];
		if(w != w) { // NaN -> 0
			continue;
		}

		// Get unit vectors in the appropriate directions
		weight_vector_unit_vectors(wv, i, u);

		// Find probability from tensors
		prob = 1;
		for(j = 0; j < 3; j++) {
			// Transform unit vector away from the canonical space
			affine_inverse_transform(A, v, u[j]);
			prob *= tensor_mult2(tensors[j], v);
		}


    I += w * prob;
		total_prob += prob;
	}

  // Multiply by area of each piece
  angle_count = weight_vector_angle_count(wv);
  if(dim == 2) {
    I *= pow(2 * M_PI / (double)angle_count, 3);
  }
  else {
    assert(dim == 3);
    I *= pow(4 * M_PI / (double)angle_count, 3);
  }

	// Normalize probability
	I /= total_prob;

	affine_destroy(A);

	return I;
}
