#include "weights.h"

#include "parallel.h"
#include "progress.h"
#include <pthread.h>

typedef struct {
	line_graph_t *G;
	tensor_field_t *tf;
	mask_t *endpoints;

	weight_options_t *opts;

	pthread_mutex_t lock, print_lock;	
	int progress, total;
	time_t start;

} weight_closure_t;

static void display_progress(weight_closure_t *w, int decimals)
{
	pthread_mutex_lock(&w->print_lock);

	progress(w->start, w->progress, w->total, decimals);

	pthread_mutex_unlock(&w->print_lock);
}

int weight_closure(int v, void *closure_data)
{
	weight_closure_t *data;

	line_graph_t *G;
	tensor_field_t *tf;
	mask_t *endpoints;
	weight_options_t *opts;
	pthread_mutex_t *lock;
	int progress;

	int i, j, k, e, m, d, dim;
	edge_pair_t *ep;
  tensor_t *tensors[3];
	point_t points[3];

	// Unpack closure
	data = (weight_closure_t *)closure_data;
	endpoints = data->endpoints;
	G = data->G;
	tf = data->tf;
	opts = data->opts;

	lock = &data->lock;

	m = line_graph_edge_pair_count(G, v);

	// Filter out endpoints
	if(endpoints != NULL && mask_get(endpoints, v)) {
		for(e = 0; e < m; e++) {
			ep = line_graph_edge_pair(G, v, e);
			
			ep->weight = 0;
		}
		return 1;
	}

	// Storage for translated points
	dim = tensor_field_dim(tf);
  coord_t coords[3][dim];
  for(i = 0; i < 3; i++) {
    points[i] = coords[i];
  }

	for(e = 0; e < m; e++) {
		ep = line_graph_edge_pair(G, v, e);
			
		// Retrieve points + tensors
		k = ep->common_vertex;
		i = ep->neighbors[0];
		j = ep->neighbors[1];

		point_assign(dim, coords[0], tensor_field_point(tf, i));
		point_assign(dim, coords[1], tensor_field_point(tf, k));
		point_assign(dim, coords[2], tensor_field_point(tf, j));

		tensors[0] = tensor_field_tensor(tf, i);
		tensors[1] = tensor_field_tensor(tf, k);
		tensors[2] = tensor_field_tensor(tf, j);

		// Zero points[1] to origin
		point_sub(dim, coords[0], coords[0], coords[1]);
		point_sub(dim, coords[2], coords[2], coords[1]);
		point_zero(dim, coords[1]);
			
		// Calculate weight
		ep->weight = (*opts->weight)(dim, opts, points, tensors);

		if(opts->verbose) {
			// TODO: move lock inwards
			pthread_mutex_lock(lock);
			progress = data->progress++;
			pthread_mutex_unlock(lock);

			// TODO: programmatically make it expand out to an appropriate decimal
			if(data->total > 100000) {
				if(progress % (data->total / 10000) == 0) {
					display_progress(data, 2);
				}
			}
			else if(data->total > 10000) {
				if(progress % (data->total / 1000) == 0) {
					display_progress(data, 1);
				}
			}
			else if(data->total > 100) {
				if(progress % (data->total / 100) == 0) {
					display_progress(data, 0);
				}
			}
		}
	}

	return 1;
}

void weighted_line_graph(weight_options_t *opts, tensor_field_t *tf, mask_t *endpoints, line_graph_t *G)
{
	int n;
	weight_closure_t data;

	n = tensor_field_count(tf);

	// Initialize weight closure
	data.G = G;
	data.tf = tf;
	data.endpoints = endpoints;
	data.opts = opts;

	data.progress = 0;
	data.total = line_graph_total_edge_pair_count(G);
	data.start = time(NULL);

  pthread_mutex_init(&data.lock, NULL);
  pthread_mutex_init(&data.print_lock, NULL);

	// Find weights
	assert(line_graph_vertex_count(G) == n);

	if(opts->verbose) {
		printf("Calculating %d weights...\n", data.total);
	}
	parallel_execute(n, -1, weight_closure, &data);

	if(opts->verbose) {
		printf("\n");
	}

	// Cleanup
  pthread_mutex_destroy(&data.lock);
}

static point_t get_point(void *closure_data, int i)
{
	return tensor_field_point((tensor_field_t *)closure_data, i);
}

line_graph_t *weighted_neighborhood_line_graph(weight_options_t *opts, tensor_field_t *tf, mask_t *endpoints)
{
	line_graph_t *G;

	int n, dim;

  dim = tensor_field_dim(tf);
	n = tensor_field_count(tf);

	// Create line graph + find the edges
  if(opts->verbose) {
    printf("Constructing line graph...\n");
  }
	G = line_graph_create_from_neighborhoods(dim, n, &get_point, tf, opts->neighborhood);

	weighted_line_graph(opts, tf, endpoints, G);

	return G;
}

////////// Defaults ///////////
#include "bent_weight.h"

const weight_options_t weight_options_defaults = {
	1.0, // distance
	&bent_weight_expectation, // weight function
  (void *)(&bent_weight_options_defaults), // closure data
  0 // verbose
};
