#include "binary.h"

#define STRIDE_MULTIPLE GROUP_SIZE

image_binary_t *image_binary_create(
  unsigned int width,
	unsigned int height)
{
	image_binary_t *I;
	unsigned int stride;
	size_t s;
	intptr_t p;

	// Make the stride the least multiple of STRIDE_MULTIPLE greater than width
	if(width % GROUP_SIZE == 0) {
		stride = width / GROUP_SIZE;
	}
	else {
		stride = width / GROUP_SIZE + 1;
	}

	// Find the size of the image struct and it's pixel data
	s = sizeof(image_binary_t);
	s += height * stride * sizeof(bit_group_t);

	// Allocate space for all of these
	I = (image_binary_t *)malloc(s);
	if(I == NULL) {
		fprintf(stderr, "malloc failure\n");
		return NULL;
	}

	// Find the start of pixel data
	p = (intptr_t)I + sizeof(image_binary_t);

	// Initialize image struct
	I->width = width;
	I->stride = stride;
	I->height = height;
	I->pixels = (bit_group_t *)p;

	return I;
}

image_binary_t *image_binary_condition(
  int width, int height,
	int (*condition)(void *data, int i, int j), void *data)
{
	unsigned int i, j, k, stride;
	image_binary_t *bin;
	bit_group_t pixel_group;

	bin = image_binary_create(width, height);
	stride = bin->stride;

	for(i = 0; i < height; i++) {
		for(j = 0; j < width / GROUP_SIZE; j++) {
			// Accumulate all of the pixels for a given word
			pixel_group = 0;
			for(k = 0; k < GROUP_SIZE; k++) {
				if((*condition)(data, i, j * GROUP_SIZE + k)) {
					pixel_group |= bit_mask(k);
				}
			}

			bin->pixels[i * stride + j] = pixel_group;
		}
	}

	return bin;
}

void image_binary_destroy(image_binary_t *bin)
{
	free(bin);
}

typedef struct {
	image_grey_t *grey;
	uint8_t tau;
} threshold_data_t;

static int threshold_condition(void *p, int i, int j)
{
	threshold_data_t *data;
	image_grey_t *grey;
	uint8_t g, tau;

	data = (threshold_data_t *)p;

	grey = data->grey;
	g = *grey_pixel(grey, i, j);

	tau = data->tau;

	return g >= tau;
}

image_binary_t *image_threshold(image_grey_t *grey, uint8_t tau)
{
	threshold_data_t data;

	data.grey = grey;
	data.tau = tau;

	return image_binary_condition(
	  grey->width, grey->height,
		&threshold_condition, &data);
}

/*
 * Move a binary image back to greyscale
 */
image_grey_t *image_binary_to_grey(image_binary_t *bin)
{
	unsigned int i, j, width, height;
	image_grey_t *grey;

	width = bin->width;
	height = bin->height;

	grey = image_grey_create(width, height);

	for(i = 0; i < height; i++) {
		for(j = 0; j < width; j++) {
			*grey_pixel(grey, i, j) = binary_get_pixel(bin, i, j) ? 255 : 0;
		}
	}

	return grey;
}
