#include "io.h"

#include "jpeg.h"
#include "png.h"

typedef enum {
	UNKNOWN,
	PNG,
	JPEG,
} image_type_t;

typedef struct {
	const char *extension;
	image_type_t type;
} known_extension_t;

static known_extension_t KNOWN_EXTENSIONS[] = {
	{"jpg", JPEG},
  {"jpeg", JPEG},
  {"png", PNG},
	{NULL, UNKNOWN}
};

static int ext_equal(const char *ext, const char *cmp)
{
	for(; *ext != '\0' && *cmp != '\0'; ext++, cmp++) {
		if(tolower(*ext) != tolower(*cmp)) {
			return 0;
		}
	}

	return 1;
}

static image_type_t type_from_extension(const char *filename)
{
	const char *ext;
	int i;

	// Find extension
	ext = filename + strlen(filename);
	for(; *ext != '.' && ext != filename; ext--);

	if(*ext != '.') {
		return UNKNOWN;
	}
	ext++;

	for(i = 0; KNOWN_EXTENSIONS[i].extension != NULL; i++) {
		if(ext_equal(ext, KNOWN_EXTENSIONS[i].extension)) {
			return KNOWN_EXTENSIONS[i].type;
		}
	}

	return UNKNOWN;
}

image_rgb_t *image_rgb_load(const char *filename)
{
	image_type_t type;

	type = type_from_extension(filename);

	switch(type) {
	case JPEG:
		return jpeg_load(filename);
	case PNG:
		return png_load(filename);
	case UNKNOWN:
	default:
		return NULL;
	}
}

void image_rgb_write(const char *filename, image_rgb_t *I)
{
	image_type_t type;

	type = type_from_extension(filename);

	
	switch(type) {
	case JPEG:
		jpeg_write(filename, I);
		break;
	case PNG:
		png_write(filename, I);
		break;
	case UNKNOWN:
	default:
		break;
	}
}

image_grey_t *image_grey_load(const char *filename)
{
	image_rgb_t *rgb;
	image_grey_t *grey;

	rgb = image_rgb_load(filename);

	grey = image_rgb_to_grey(rgb);
	image_rgb_destroy(rgb);

	return grey;
}

void image_grey_write(const char *filename, image_grey_t *grey)
{
	image_rgb_t *rgb;

	rgb = image_grey_to_rgb(grey);

	image_rgb_write(filename, rgb);

	image_rgb_destroy(rgb);
}

image_float_t *image_float_load(const char *filename)
{
	image_grey_t *G;
	image_float_t *F;

	G = image_grey_load(filename);

	F = image_grey_to_float(G);
	image_grey_destroy(G);

	return F;
}

void image_float_write(const char *filename, image_float_t *F)
{
	image_grey_t *G;

	G = image_float_to_grey(F);
	
	image_grey_write(filename, G);

	image_grey_destroy(G);
}

void image_binary_write(const char *filename, image_binary_t *B)
{
	image_grey_t *G;
	
	G = image_binary_to_grey(B);

	image_grey_write(filename, G);

	image_grey_destroy(G);
}
