#include "line_graph_io.h"
#include "tensor_field_io.h"
#include "mask_io.h"
#include "weight_file.h"

#include "weights.h"
#include "rect_endpoints.h"
#include "nifti.h"

#include "cli.h"

static tensor_field_t *tf = NULL;
static line_graph_t *G = NULL;
static weight_vector_array_t *wva = NULL;
static char out_graph_path[PATH_MAX + 1];

static bool verbose = false;

static weight_options_t weight_opts;
static weight_vector_calc_opts_t weight_vector_opts;

static bool use_rect_endpoints = false;
static mask_t *endpoints = NULL;


static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_rect_endpoints(void *closure_data, const argdef_t *def, const char *value);
static int set_endpoints(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_weight_vector(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename);


// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	// Endpoints
  {'\0', "rect-endpoints", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Use the rectangular bounds of the tensor field as an endpoint set.",
	 &set_rect_endpoints},

  {'\0', "endpoints", 0,
   "file", ARG_VALUE_ONE, NULL,
   "Use the endpoint set from a file.",
	 &set_endpoints},

	{0}
};

static const arggroup_t groups[] = {
	{"Endpoint Options", 1},
	
	{0}
};


static const paramdef_t params[] = {
  {"tensor-field", "File containing the tensor field.", ARG_REQUIRED,
	 &set_tensor_field},
	
	{"unweighted", "Unweighted line graph over the tensor field.", ARG_REQUIRED,
	 &set_line_graph},

	{"weight-file", "File containing proposal curve weights.", ARG_REQUIRED,
	 &set_weight_vector},

	{"weighted", "File to output the weighted line graph.", ARG_REQUIRED,
	 &set_out_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "apply_weight_file",
  "Calculate the weights for a line graph based on the weights in the given file.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	// Initialize options to defaults
	weight_opts = weight_options_defaults;
	weight_opts.weight = &weight_vector_calc;
	weight_opts.closure_data = &weight_vector_opts;
	weight_vector_opts= weight_vector_calc_opts_defaults;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Print out arguments
	if(verbose) {
	}

	if(use_rect_endpoints) {
		endpoints = rect_endpoints(tf);
	}

	// Calculate weights
	weighted_line_graph(&weight_opts, tf, endpoints, G);
	fputc('\n', stdout);

	// Output
	line_graph_write_file_name(out_graph_path, G);

	// Cleanup
	mask_destroy(endpoints);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	weight_opts.verbose = 1;
	return 0;
}

static int set_rect_endpoints(void *closure_data, const argdef_t *def, const char *value)
{
	use_rect_endpoints = true;
	return 0;
}

static int set_endpoints(void *closure_data, const argdef_t *def, const char *filename)
{
	if(nifti_filename(filename)) {
		endpoints = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		endpoints = mask_from_file_name(filename);
	}

	if(endpoints == NULL) {
		fprintf(stderr, "Error reading input endpoints mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	G = line_graph_read_file_name(filename);
	if(G == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_weight_vector(void *closure_data, const paramdef_t *def, const char *filename)
{
	wva = weight_vector_array_from_file_name(filename);
	if(wva == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
	}
	weight_vector_opts.weights = wva;

	return 0;
}

static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_graph_path, filename, PATH_MAX);
	out_graph_path[PATH_MAX] = '\0';

	return 0;
}
