#include "line_graph_io.h"
#include "tensor_field_io.h"

#include "weights.h"
#include "bent_weight.h"

#include "nifti.h"
#include "mask_io.h"

#include "cli.h"

static int verbose = 0;
static tensor_field_t *tf = NULL;
static char out_graph_path[PATH_MAX + 1];
static bent_weight_options_t bent_opts;
static weight_options_t weight_opts;
static mask_t *endpoints = NULL;


static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_neighborhood(void *closure_data, const argdef_t *def, const char *value);
static int set_expectation_samples(void *closure_data, const argdef_t *def, const char *value);

static int set_endpoints(void *closure_data, const argdef_t *def, const char *filename);

static int set_angle_weight(void *closure_data, const argdef_t *def, const char *value);
static int set_sigma(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	{'\0', "neighborhood", 0,
	 "distance", ARG_VALUE_ONE, NULL,
	 "The maximum distance within which two tensor are potentially connected.",
	 &set_neighborhood},

	{'\0', "expectation-samples", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The number of points to sample when integrating an expected weight with Monte Carlo methods.",
	 &set_expectation_samples},

	{'\0', "endpoints", 0,
	 "mask-file", ARG_VALUE_ONE, NULL,
	 "Mask of free endpoints.",
	 &set_endpoints},

	// Bent weight options
  {'\0', "angle-weight", 0,
   "weight", ARG_VALUE_ONE, NULL,
   "Weight to place on the angle difference in the bent weight",
	 &set_angle_weight},
	
	{'\0', "sigma", 0,
	 "distance", ARG_VALUE_ONE, NULL,
	 "Neighborhood constant for bent weight",
	 &set_sigma},

	{0}
};


static const arggroup_t groups[] = {
	{"Bent weight options", 4},

	{0}
};


static const paramdef_t params[] = {
  {"tensor-field", "File containing the tensor field.", ARG_REQUIRED,
	 &set_tensor_field},
	
	{"line-graph", "File to output the weighted line graph.", ARG_REQUIRED,
	 &set_line_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "contours",
  "Building a weighted line graph between tensors.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	line_graph_t *G;
	
	// Initialize options to defaults
	bent_opts = bent_weight_options_defaults;
	weight_opts = weight_options_defaults;
	weight_opts.closure_data = &bent_opts;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Print out arguments
	if(verbose) {
		printf("\nLine graph options:\n");
		printf("  neighborhood        = %g\n", weight_opts.neighborhood);
		printf("  expectation-samples = %d\n", bent_opts.expectation_samples);

		printf("\nUsing Expected Bent Weight with options:\n");
		printf("  angle-weight        = %g\n", bent_opts.angle_weight);
		printf("  sigma               = %g\n", bent_opts.sigma);

		printf("\n");
	}

	// Calculate weights
	G = weighted_neighborhood_line_graph(&weight_opts, tf, endpoints);

	if(G == NULL) {
		fprintf(stderr, "Error in calculating weights\n");
		return 1;
	}
	else {
		// Output
		line_graph_write_file_name(out_graph_path, G);
	}

	line_graph_destroy(G);
	tensor_field_destroy(tf);
	
	return 0;
}

static float float_arg(const argdef_t *def, const char *value)
{
	int s;
	float v;

	s = sscanf(value, "%g", &v);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return v;
}

static unsigned int unsigned_int_arg(const argdef_t *def, const char *value)
{
	int s;
	unsigned int u;

	s = sscanf(value, "%u", &u);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return u;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = weight_opts.verbose = 1;
	return 0;
}

static int set_neighborhood(void *closure_data, const argdef_t *def, const char *value)
{
	weight_opts.neighborhood = float_arg(def, value);
	return 0;
}

static int set_expectation_samples(void *closure_data, const argdef_t *def, const char *value)
{
	bent_opts.expectation_samples = unsigned_int_arg(def, value);
	return 0;
}


static int set_angle_weight(void *closure_data, const argdef_t *def, const char *value)
{
	bent_opts.angle_weight = float_arg(def, value);
	return 0;
}

static int set_sigma(void *closure_data, const argdef_t *def, const char *value)
{
	bent_opts.sigma = float_arg(def, value);
	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_graph_path, filename, PATH_MAX);
	out_graph_path[PATH_MAX] = '\0';

	return 0;
}

static int set_endpoints(void *closure_data, const argdef_t *def, const char *filename)
{
	if(nifti_filename(filename)) {
		endpoints = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		endpoints = mask_from_file_name(filename);
	}

	if(endpoints == NULL) {
		fprintf(stderr, "Error reading input endpoints mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return 0;
}
