#include "bundle_io.h"
#include "bspline.h"

#include "cli.h"
#include "cli_util.h"

static bool verbose = false;
static bundle_t *paths = NULL;
static char out_path[PATH_MAX+1];
static int order = 4;
static int N = 512;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_N(void *closure_data, const argdef_t *def, const char *value);

static int set_order(void *closure_data, const argdef_t *def, const char *value);


static int set_paths(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

  {'N', "output-count", 0,
   "integer", ARG_VALUE_ONE, NULL,
	 "Number of points at which to output the spline.",
	 &set_N},

  {'O', "order", 0,
   "integer", ARG_VALUE_ONE, NULL,
	 "The order of the generated b-splines.",
	 &set_order},

	{0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"in-file", "Input paths file.", ARG_REQUIRED,
	 &set_paths},

  {"out-file", "Output paths file.", ARG_REQUIRED,
	 &set_out_file},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "bsplinify",
  "Convert paths to smoothed splines.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
  int i, n;
  path_t *path;
  bspline_t *spline;
	bundle_t *out;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Perform bsplinify
	assert(paths != NULL);
	n = bundle_count(paths);
	out = bundle_create(bundle_dim(paths), n);

	for(i = 0; i < n; i++) {
		path = bundle_get(paths, i);
		spline = bspline_create(path, order);

		if(spline == NULL) {
			path = path_copy(path);
		}
		else {
			path = bspline_calculate_path(spline, N);
		}

		bundle_set(out, i, path);

		bspline_destroy(spline);
	}

	// Output
	bundle_write_file_name(out_path, out);

	// Cleanup
	bundle_destroy(out);
	bundle_destroy(paths);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int set_N(void *closure_data, const argdef_t *def, const char *value)
{
	N = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_order(void *closure_data, const argdef_t *def, const char *value)
{
	order = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_paths(void *closure_data, const paramdef_t *def, const char *filename)
{
	paths = bundle_read_file_name(filename);
	if(paths == NULL) {
		fprintf(stderr, "Error reading input path file \"%s\"\n", filename);
		exit(EX_NOINPUT);
	}

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
