#include "bundle_io.h"
#include "splines/catmull_rom.h"

#include "cli.h"
#include "cli_util.h"

static bool verbose = false;
static bundle_t *paths = NULL;
static char out_path[PATH_MAX+1];

static double tau = 1.0;
static int N = 512;


static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_tau(void *closure_data, const argdef_t *def, const char *value);
static int set_N(void *closure_data, const argdef_t *def, const char *value);
static int add_bundle(void *closure_data, const argdef_t *def, const char *value);

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

	{'p', "paths", ARG_MANY,
	 "path-file", ARG_VALUE_ONE, NULL,
	 "Add new core paths.",
	 &add_bundle},

	{'t', "tau", 0,
	 "number", ARG_VALUE_ONE, NULL,
	 "tau parameter of Catmull-Rom spline.",
	 &set_tau},

  {'N', "interpolation-count", 0,
   "integer", ARG_VALUE_ONE, NULL,
	 "Number of points to interpolate the spline.",
	 &set_N},

	{0}
};

static const arggroup_t groups[] = {
  {"Spline Options", 2},

	{0}
};

static const paramdef_t params[] = {
  {"out-file", "Output paths file.", ARG_REQUIRED,
	 &set_out_file},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "catmull_romify",
  "Convert paths to Catmull-Rom splines with those control points.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};


static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
  int i, n;
  path_t *path;
  spline_t *spline;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Output job
	if(paths == NULL) {
		printf("No paths given\n");
		return 0;
	}

	n = bundle_count(paths);
	if(verbose) {
		printf("Generating %d tubes:\n", n);
		printf("  tau = %lf\n", tau);
		printf("  N = %d\n", N);
	}

  // Create interpolating splines
	for(i = 0; i < n; i++) {
		path = bundle_get(paths, i);

    // Create spline
    spline = catmull_rom_spline(path, tau);

    // Interpolate
    bundle_set(paths, i, spline_path(spline, N));

    spline_destroy(spline);
	}

	// Output
	if(verbose) {
		printf("Writing output...\n");
	}
	bundle_write_file_name(out_path, paths);

	bundle_destroy(paths);

  return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int set_tau(void *closure_data, const argdef_t *def, const char *value)
{
	tau = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_N(void *closure_data, const argdef_t *def, const char *value)
{
	N = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int add_bundle(void *closure_data, const argdef_t *def, const char *filename)
{
	bundle_t *new_bundle;

	new_bundle = bundle_read_file_name(filename);
	if(new_bundle == NULL) {
		fprintf(stderr, "Error reading input path file \"%s\"\n", filename);
		exit(EX_NOINPUT);
	}

	if(paths == NULL) {
		paths = new_bundle;
	}
	else {
		paths = bundle_merge(paths, new_bundle);
		bundle_destroy(new_bundle);
	}

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
