#include "tensor_field_io.h"
#include "line_graph_io.h"
#include "mask_io.h"

#include "nifti.h"

#include "rect_endpoints.h"

#include "optimize.h"
#include "cli.h"
#include "cli_util.h"

// Set up options, input/output
static tensor_field_t *tf = NULL;

static line_graph_t *in_graph = NULL;
static char out_graph_path[PATH_MAX + 1];
static optimize_options_t optim_opts;
static int verbose = 0;
static bool use_rect_endpoints = false;

static mask_t *endpoints = NULL;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_lambda(void *closure_data, const argdef_t *def, const char *value);
static int set_mu(void *closure_data, const argdef_t *def, const char *value);

static int set_solver(void *closure_data, const argdef_t *def, const char *value);

static int set_model(void *closure_data, const argdef_t *def, const char *value);

static int set_endpoints(void *closure_data, const argdef_t *def, const char *filename);
static int set_rect_endpoints(void *closure_data, const argdef_t *def, const char *filename);


static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	// Optimization options
	{'\0', "lambda", 0,
	 "lambda", ARG_VALUE_ONE, NULL,
	 "Penalty per tensor triplet.",
	 &set_lambda},

	{'\0', "mu", 0,
	 "mu", ARG_VALUE_ONE, NULL,
	 "Penalty per endpoint.",
	 &set_mu},

	{'\0', "solver", 0,
	 "name", ARG_VALUE_ONE, NULL,
	 "Use the given solver for MIP solution instead of default. (glpk, cplex)",
	 &set_solver},

	{'M', "model", ARG_REQUIRED,
	 "name", ARG_VALUE_ONE, NULL,
	 "Use the given model. (qp, ilp)",
	 &set_model},

	{'\0', "endpoints", 0,
	 "mask-file", ARG_VALUE_ONE, NULL,
	 "Mask of free endpoints.",
	 &set_endpoints},

	{'\0', "rect-endpoints", 0,
	 NULL, ARG_VALUE_OPTIONAL, NULL,
	 "Set the endpoints to be those that lie on the rectilinear box containing all points in the tensor field.",
	 &set_rect_endpoints},

  {0}
};

static const arggroup_t groups[] = {
	{"Optimization options",  1},
	{"ROI options",  5},

	{0}
};

static const paramdef_t params[] = {
  {"tensor-field", "File containing the tensor field.", ARG_REQUIRED,
	 &set_tensor_field},
	
	{"in-line-graph", "File to read input weighted line graph.", ARG_REQUIRED,
	 &set_in_graph},

	{"out-line-graph", "File to output the optimal line graph.", ARG_REQUIRED,
	 &set_out_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "contours",
  "Finding optimal contours with a line graph representation",
  "Curvature-LP",
  "0.1",
  COPYRIGHT
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	line_graph_t *H;
	int n;

	// Initialize options to defaults
	optim_opts = optimize_options_defaults;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Print out arguments
	if(verbose) {
		printf("\nOptimization options:\n");
		printf("  lambda              = %g\n", optim_opts.lambda);
		printf("  mu                  = %g\n", optim_opts.mu);

		printf("\n");
	}

	if(use_rect_endpoints) {
		endpoints = rect_endpoints(tf);
	}

	// Check counts
	n = tensor_field_count(tf);
	if(verbose) {
		printf("Contours over %d tensors\n", n);
	}

	if(endpoints != NULL && mask_size(endpoints) != n) {
		fprintf(stderr, "Bad mask size: %lu\n", (unsigned long int)mask_size(endpoints));
		return 1;
	}

	if(line_graph_vertex_count(in_graph) != n) {
		fprintf(stderr, "Bad vertex count: %d\n", line_graph_vertex_count(in_graph));
		return 1;
	}

	// Optimize
	if(verbose) {
		printf("Finding optimal contours...\n");
	}
  H = optimal_contours(&optim_opts, in_graph, endpoints);

	if(H == NULL) {
		fprintf(stderr, "Error in calculating optimal line graph\n");
	}
	else {
		// Output
		line_graph_write_file_name(out_graph_path, H);
	}

	// Cleanup
  line_graph_destroy(in_graph);
	line_graph_destroy(H);
	tensor_field_destroy(tf);
	mask_destroy(endpoints);

	return 0;

	/*
	bent_weight_options_t bent_opts = {4, 1e1, 4*1024};
	weight_options_t weight_opts = {1.5, &bent_weight_expectation, &bent_opts, 1};

	bent_opts.angle_weight = 1e2;

	optimize_options_t opt_opts;
	opt_opts.lambda = 2e-3;
	opt_opts.mu = 0;
	opt_opts.endpoints = endpoints;
	opt_opts.solver = SOLVER_DEFAULT;
	*/
}


static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = 1;
	optim_opts.verbose = true;
	return 0;
}

static int set_solver(void *closure_data, const argdef_t *def, const char *value)
{
	if(strcmp(value, "glpk") == 0) {
		optim_opts.solver = SOLVER_GLPK;
	}
	else if(strcmp(value, "cplex") == 0) {
		optim_opts.solver = SOLVER_CPLEX;
	}
	else {
		fprintf(stderr, "Invalid value for solver: \"%s\"\n", value);
		exit(EX_USAGE);
	}

	return 0;
}


static int set_model(void *closure_data, const argdef_t *def, const char *value)
{
	if(strcmp(value, "qp") == 0) {
		optim_opts.model = MODEL_QP1;
	}
	else if(strcmp(value, "ilp") == 0) {
		optim_opts.model = MODEL_ILP1;
	}
	else {
		fprintf(stderr, "Invalid value for model: \"%s\"\n", value);
		exit(EX_USAGE);
	}

	return 0;
}

static int set_lambda(void *closure_data, const argdef_t *def, const char *value)
{
	optim_opts.lambda = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_mu(void *closure_data, const argdef_t *def, const char *value)
{
	optim_opts.mu = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}


static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	in_graph = line_graph_read_file_name(filename);
	if(in_graph == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_graph_path, filename, PATH_MAX);
	out_graph_path[PATH_MAX] = '\0';

	return 0;
}

static mask_t *load_mask(const char *filename)
{
	mask_t *mask;

	if(nifti_filename(filename)) {
		mask = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		mask = mask_from_file_name(filename);
	}

	if(mask == NULL) {
		fprintf(stderr, "Error reading input mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return mask;
}

static int set_endpoints(void *closure_data, const argdef_t *def, const char *filename)
{
	endpoints = load_mask(filename);
	return 0;
}

static int set_rect_endpoints(void *closure_data, const argdef_t *def, const char *filename)
{
	use_rect_endpoints = true;
	return 0;
}
