#include "cairo.h"
#include "context_internal.h"

#include "util.h"

typedef struct cairo_transform cairo_transform_t;

typedef struct {
	draw_context_t super;
	cairo_t *cr;

	int stack_size;
	cairo_transform_t **stack;

	int mark_count;
	int *mark;

} cairo_context_t;

static inline cairo_context_t *downcast(draw_context_t *super);

//////////////////////////// Transforms ////////////////////////////
struct cairo_transform {
	union {
		coord_t scale[2];
		coord_t translate[2];
		angle_t rotate;
	};

	enum {
		TRANSFORM_SCALE,
		TRANSFORM_TRANSLATE,
		TRANSFORM_ROTATE2
	} type;

};

static void transform_apply(cairo_t *cr, cairo_transform_t *T)
{
	point_t s;
	coord_t ls;

	switch(T->type) {
	case TRANSFORM_SCALE:
		s = T->scale;
		cairo_scale(cr, s[0], s[1]);

		ls = MIN(s[0], s[1]);		
		break;
	case TRANSFORM_TRANSLATE:
		//printf("  Translate (%f, %f)\n", (float)T->translate[0], (float)T->translate[1]);
    cairo_translate(cr, T->translate[0], T->translate[1]);
		break;
	case TRANSFORM_ROTATE2:
		//printf("  Rotate %f\n", T->rotate);
		cairo_rotate(cr, T->rotate);
		break;
	}
}

static void transform_apply_reverse(cairo_t *cr, cairo_transform_t *T)
{
	point_t s;
	coord_t ls;

	switch(T->type) {
	case TRANSFORM_SCALE:
		s = T->scale;
		cairo_scale(cr, 1/s[0], 1/s[1]);

		ls = MIN(s[0], s[1]);
		break;
	case TRANSFORM_TRANSLATE:
		//printf("  Translate (%f, %f)\n", -(float)T->translate[0], -(float)T->translate[1]);
    cairo_translate(cr, -T->translate[0], -T->translate[1]);
		break;
	case TRANSFORM_ROTATE2:
		//printf("  Rotate -%f\n", T->rotate);
		cairo_rotate(cr, -T->rotate);
		break;
	}
}

static void push_transform(draw_context_t *super)
{
	cairo_context_t *c;
	int n;

	c = downcast(super);

	// Push a mark onto the stack
	c->mark_count++;
	c->mark = realloc(c->mark, sizeof(int) * c->mark_count);

	c->mark[c->mark_count - 1] = c->stack_size;
}

static void pop_transform(draw_context_t *super)
{
	cairo_context_t *c;
	cairo_t *cr;
	int i, n, m;

	c = downcast(super);
	cr = c->cr;

	// Pop from the mark stack
	m = c->mark[c->mark_count - 1];
	c->mark_count--;
	c->mark = realloc(c->mark, sizeof(int) * c->mark_count);

	// Undo the end of the transform stack
	n = c->stack_size;
	for(i = n-1; i >= m; i--) {
		transform_apply_reverse(cr, c->stack[i]);
		free(c->stack[i]);
	}

	// Pop the transform stack
	c->stack = realloc(c->stack, sizeof(cairo_transform_t *) * m);
	c->stack_size = m;
}

static void add_transform(draw_context_t *super, cairo_transform_t *T)
{
	cairo_context_t *c;
	int n;

	c = downcast(super);

	// Push onto stack
	n = ++c->stack_size;
	c->stack = realloc(c->stack, sizeof(cairo_transform_t *) * n);
	c->stack[n - 1] = malloc(sizeof(cairo_transform_t));
	*(c->stack[n - 1]) = *T;

	// Apply
	transform_apply(c->cr, T);
}

static void scale(draw_context_t *super, point_t s)
{
	cairo_transform_t T;

	point_assign(2, T.scale, s);
	T.type = TRANSFORM_SCALE;

	add_transform(super, &T);
}

static void scale_symm(draw_context_t *super, coord_t s)
{
	cairo_transform_t T;

	T.scale[0] = s;
	T.scale[1] = s;
	T.type = TRANSFORM_SCALE;

	add_transform(super, &T);
}

static void translate(draw_context_t *super, point_t t)
{
	cairo_transform_t T;

	point_assign(2, T.translate, t);
	T.type = TRANSFORM_TRANSLATE;

	add_transform(super, &T);
}

static void rotate2(draw_context_t *super, angle_t theta)
{
	cairo_transform_t T;

	T.rotate = theta;
	T.type = TRANSFORM_ROTATE2;

	add_transform(super, &T);
}

static void undo_all(cairo_context_t *c)
{
	int i;
	cairo_t *cr;

	cr = c->cr;
	for(i = c->stack_size - 1; i >= 0; i--) {
		transform_apply_reverse(cr, c->stack[i]);
	}

}

static void redo_all(cairo_context_t *c)
{
	int i, n;
	cairo_t *cr;

	cr = c->cr;
	n = c->stack_size;
	for(i = 0; i < n; i++) {
		transform_apply(cr, c->stack[i]);
	}
}

///////////////////////////// Drawing /////////////////////////////
static void circle(draw_context_t *super)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;

	cairo_arc(cr, 0, 0, 1, 0, 2 * M_PI);

	undo_all(c);
	cairo_stroke(cr);
	redo_all(c);
}

static void line_begin(draw_context_t *super, point_t p)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;

	cairo_move_to(cr, p[0], p[1]);
}

static void line_add(draw_context_t *super, point_t p)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;

	cairo_line_to(cr, p[0], p[1]);
}

static void line_end(draw_context_t *super)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;

	undo_all(c);
	cairo_stroke(cr);
	redo_all(c);
}

///////////////////////////// Other /////////////////////////////

static void color(draw_context_t *super, float r, float g, float b)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;
	
	cairo_set_source_rgb(cr, r, g, b);
}

void color_alpha(draw_context_t *super, float r, float g, float b, float a)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;
	
	cairo_set_source_rgba(cr, r, g, b, a);
}

static void line_width(draw_context_t *super, float w)
{
	cairo_context_t *c;
	cairo_t *cr;

	c = downcast(super);
	cr = c->cr;

	undo_all(c);
  cairo_set_line_width(cr, w);
	redo_all(c);
}

static void finish(draw_context_t *super)
{
	cairo_context_t *c;
	cairo_t *cr;
	int i, n;

	c = downcast(super);
	cr = c->cr;
	
	cairo_destroy(cr);

	// Cleanup
	n = c->stack_size;
	for(i = 0; i < n; i++) {
		free(c->stack[i]);
	}

	free(c->stack);
	free(c->mark);
}

//////////////////////////// Class Setup ////////////////////////////

static const draw_context_vtable_t cairo_vtable = {
	&push_transform,
	&pop_transform,
	
	&scale,
	&scale_symm,
	&translate,
	&rotate2,

	&circle,
	
	&line_begin,
	&line_add,
	&line_end,

	&color,
	&color_alpha,
	&line_width,

	&finish,
};

static inline cairo_context_t *downcast(draw_context_t *super)
{
	assert(super->vtable == &cairo_vtable);
	
	return (cairo_context_t *)super;
}

static void robust_scale_symm(cairo_t *cr, coord_t s)
{
  cairo_scale(cr, s, s);
  cairo_set_line_width(cr, cairo_get_line_width(cr) * s);
}

draw_context_t *draw_context_cairo(
  cairo_surface_t *surface,
	int width, int height,
	point_t min, point_t max)
{
	cairo_context_t *c;
	cairo_t *cr;

	double start_ratio, end_ratio, scale;
	coord_t size[2];
  double inner_width, inner_height;

	c = (cairo_context_t *)malloc(sizeof(cairo_context_t));
	c->super.vtable = &cairo_vtable;

	c->cr = cr = cairo_create(surface);

	// Transform stack
	c->stack_size = 0;
	c->stack = NULL;

	c->mark_count = 0;
	c->mark = NULL;

	// Clear background
  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_paint(cr);

	/////// Normalize space //////

  // Flip y
  cairo_scale(cr, 1, -1);
  cairo_translate(cr, 0, -height);

  // Calculate size, aspect ratios
  point_sub(2, size, max, min);
  start_ratio = (double)width/height;
  end_ratio = size[0]/size[1];

  if(end_ratio < start_ratio) {
    // height is limiting dimension
    scale = height / size[1];

    inner_width = end_ratio * height;
    cairo_translate(cr, (width - inner_width) / 2, 0);
  }
  else {
    // width is limiting dimension
    scale = width / size[0];

    inner_height = width / end_ratio;
    cairo_translate(cr, 0, (height - inner_height) / 2);
  }

  robust_scale_symm(cr, scale);
  cairo_translate(cr, -min[0], -min[1]);


	// Return as parent type
	assert((intptr_t)(&c->super) == (intptr_t)c);
	return &c->super;
}

void draw_cairo(
  cairo_surface_t *surface, int width, int height,
  const draw_options_t *opts,
  tensor_field_t *tf, line_graph_t *G, bundle_t *paths)
{
	draw_context_t *context;
	coord_t min[2], max[2];

	// Figure out our space
	draw_space(opts, tf, paths, min, max);

	// Create drawing context
	context = draw_context_cairo(surface, width, height, min, max);

	// Draw
	draw(context, opts, tf, G, paths);

	// Cleanup
	draw_context_finish(context);
}
