#include "draw.h"

#include "tensor_eigens.h"
#include "util.h"

#define TENSOR_GRAY (0.7)
#define GRAPH_GRAY (0.0)
#define PATH_GRAY (0.0)

const draw_options_t draw_default_options = {
  5e-2, // line_width
	4e-2, // tract_line_width
  1.0, // tensor_scaling
  {1.0, 1.0, 1.0}, // background_color
  {TENSOR_GRAY, TENSOR_GRAY, TENSOR_GRAY}, // tensor_color
	//{0,0,0},
  {1.0, GRAPH_GRAY, GRAPH_GRAY}, // line_graph_color
	{PATH_GRAY, PATH_GRAY, 1.0}, // path_color
	false, // shade_graph
};

const draw_options_t draw_grayscale_default_options = {
	1.0, // line_width
	0.5, // tract_line_width
	1.0, // tensor_scaling
	{1.0, 1.0, 1.0}, // background_color
	{TENSOR_GRAY, TENSOR_GRAY, TENSOR_GRAY}, // tensor_color
	{GRAPH_GRAY, GRAPH_GRAY, GRAPH_GRAY}, // line_graph_color
	{PATH_GRAY, PATH_GRAY, PATH_GRAY}, // path_color 
	false, // shade_graph
};

void draw_tensor(
  draw_context_t *context,
	const draw_options_t *opts,
	point_t p,
	tensor_t *T)
{
	tensor_eigens_t *eig;

  point_t V;
  coord_t lambda0, lambda1;
  coord_t theta;

	coord_t s[2];
	coord_t x[2];

  // Unpack eigensystem
	eig = tensor_eigens_get(T);
	
  V = tensor_eigenvector(eig, 0);
  lambda0 = tensor_eigenvalue(eig, 0);
  lambda1 = tensor_eigenvalue(eig, 1);

  // Find ellipse/line orientation
  theta = atan2(V[1], V[0]);

	// Initial transform
	draw_context_push_transform(context);

	draw_context_translate(context, p);
	draw_context_scale_symm(context, opts->tensor_scaling);
	draw_context_rotate2(context, theta);


	//draw_context_circle(context);

	///*
  if(lambda0 < 1e-4) {
    // TODO?
  }
  else if(lambda1 < 1e-4) {
    // Draw line
		s[0] = -lambda0;
		s[1] = 0;
		draw_context_line_begin(context, s);

		s[0] = lambda0;
		draw_context_line_add(context, s);

		draw_context_line_end(context);
  }
  else {
		s[0] = lambda0;
		s[1] = lambda1;

		draw_context_scale(context, s);
		draw_context_circle(context);

		x[0] = -1; x[1] = 0;
		draw_context_line_begin(context, x);
		x[0] = 1;
		draw_context_line_add(context, x);
		draw_context_line_end(context);
  }
	//*/


	draw_context_pop_transform(context);
}

void line_graph_weight_limits(line_graph_t *L, double *min, double *max)
{
	int i, j, n, m;
	edge_pair_t *ep;

	*min = 1.0/0.0;
	*max = -1.0/0.0;

	n = line_graph_vertex_count(L);
	for(i = 0; i < n; i++) {
		m = line_graph_edge_pair_count(L, i);
		for(j = 0; j < m; j++) {
			ep = line_graph_edge_pair(L, i, j);

			*min = MIN(*min, ep->weight);
			*max = MAX(*max, ep->weight);
		}
	}
}

void draw(draw_context_t *context,
					const draw_options_t *opts,
					tensor_field_t *tf,
					line_graph_t *G,
					bundle_t *paths)
{
	int i, j, k, n, m;

	point_t p, q;
	tensor_t *T;

	double weight_min, weight_max, w;
	edge_pair_t *ep;

	path_t *path;

	// Setup
	draw_context_line_width(context, opts->line_width);

  // Draw tensor field
	if(tf != NULL) {
		draw_context_color(context,
      opts->tensor_color[0],
      opts->tensor_color[1],
      opts->tensor_color[2]);
		
		n = tensor_field_count(tf);
		for(i = 0; i < n; i++) {
			p = tensor_field_point(tf, i);
			T = tensor_field_tensor(tf, i);

			draw_tensor(context, opts, p, T);
		}
	}

	// Draw line graph
	if(G != NULL) {
    draw_context_color(context,
      opts->line_graph_color[0],
      opts->line_graph_color[1],
      opts->line_graph_color[2]);
		
		line_graph_weight_limits(G, &weight_min, &weight_max);

		n = line_graph_vertex_count(G);
		for(i = 0; i < n; i++) {
			p = tensor_field_point(tf, i);

			m = line_graph_edge_pair_count(G, i);
			for(j = 0; j < m; j++) {
				ep = line_graph_edge_pair(G, i, j);

				w = ep->weight;
				w = (w - weight_min) / (weight_max - weight_min);

				if(opts->shade_graph && weight_min != weight_max) {
					draw_context_color_alpha(context,
						opts->line_graph_color[0],
						opts->line_graph_color[1],
						opts->line_graph_color[2],
						w);
				}

				// Draw line
				q = tensor_field_point(tf, ep->neighbors[0]);
				draw_context_line_begin(context, q);

				draw_context_line_add(context, p);

				q = tensor_field_point(tf, ep->neighbors[1]);
				draw_context_line_add(context, q);

				draw_context_line_end(context);
			}
		}
	}

	// Draw paths
	if(paths != NULL) {
		draw_context_line_width(context, opts->tract_line_width);

    draw_context_color(context,
      opts->path_color[0],
      opts->path_color[1],
      opts->path_color[2]);

		n = bundle_count(paths);
		for(i = 0; i < n; i++) {
			path = bundle_get(paths, i);

			if(path == NULL) {
				continue;
			}

			m = path_length(path);
			if(m < 1) {
				continue;
			}

			draw_context_line_begin(context, path_get(path, 0));
			for(j = 1; j < m; j++) {
				draw_context_line_add(context, path_get(path, j));
			}
			draw_context_line_end(context);
		}
	}
}

void tf_extent(const draw_options_t *opts,
							 tensor_field_t *tf,
							 point_t min, point_t max)
{
	tensor_eigens_t *eigs;
	coord_t max_lambda;
	int i, n, dim;

	dim = tensor_field_dim(tf);
	n = tensor_field_count(tf);

  tensor_field_extent(tf, min, max);
	
  // Leave room for tensors
  max_lambda = 0;
  for(i = 0; i < n; i++) {
		eigs = tensor_eigens_get(tensor_field_tensor(tf, i));
    max_lambda = MAX(max_lambda, tensor_eigenvalue(eigs, 0));
  }
  for(i = 0; i < dim; i++) {
    min[i] -= max_lambda * opts->tensor_scaling;
    max[i] += max_lambda * opts->tensor_scaling;
  }
}

void draw_space(const draw_options_t *opts,
								tensor_field_t *tf, bundle_t *paths,
								point_t min, point_t max)
{
	int i, dim;

	// Get dimensionality
	if(tf != NULL) {
		dim = tensor_field_dim(tf);

		if(paths != NULL) {
			assert(dim == bundle_dim(paths));
		}
	}
	else {
		assert(paths != NULL);

		dim = bundle_dim(paths);
	}

	coord_t tf_min[dim], tf_max[dim], bundle_min[dim], bundle_max[dim];

	// Get extent
	if(tf != NULL) {
		if(paths != NULL) {
			tf_extent(opts, tf, tf_min, tf_max);
			bundle_extent(paths, bundle_min, bundle_max);

			for(i = 0; i < dim; i++) {
				min[i] = MIN(tf_min[i], bundle_min[i]);
				max[i] = MAX(tf_max[i], bundle_max[i]);
			}
		}
		else {
			tf_extent(opts, tf, min, max);
		}
	}
	else {
		assert(paths != NULL);

		bundle_extent(paths, min, max);
	}
}
