#include "edge_tensors.h"
#include "point.h"
#include "image/sobel.h"
#include "image/interpolate.h"
#include "image/blur.h"
#include "image/io.h"
#include "image/binary.h"
#include "util.h"

static const point_t zero = {};

#if DIM == 2
static const point_t ones = {1, 1};
#elif DIM == 3
static const point_t ones = {1, 1, 1};
#else
#error Unsupported dimension
#endif

static image_float_t *gradient_magnitude(image_float_t *I_dx, image_float_t *I_dy)
{
	int i, j;
	image_float_t *g;
	double dx, dy;

	assert(I_dx->width == I_dy->width && I_dx->height == I_dy->height);

	g = image_float_create(I_dx->width, I_dx->height);

	for(i = 0; i < I_dx->height; i++) {
		for(j = 0; j < I_dx->width; j++) {
			dx = *float_pixel(I_dx, i, j);
			dy = *float_pixel(I_dy, i, j);

			*float_pixel(g, i, j) = sqrt(square(dx) + square(dy));
		}
	}

	return g;
}

// Non-maximum supression
static int gradient_maximum(
  image_float_t *I_g, image_float_t *I_dx, image_float_t *I_dy,
  int i, int j,
  double scale)
{
	double dx, dy, g;
	double u[2];

	// Get gradient at (i, j)
	dx = *float_pixel(I_dx, i, j);
	dy = *float_pixel(I_dy, i, j);
	g = *float_pixel(I_g, i, j);

	// Get unit vector in (dx, dy) direction
	u[0] = dx / g * scale;
	u[1] = dy / g * scale;

	/*
	if(i < 4) {
	printf("(i, j) = (%d, %d), u = (%2.2f %2.2f): g = %2.2f, g(+u) = %2.2f, g(-u) = %2.2f \n",
				 i, j,
				 u[0], u[1],
				 g,
				 interpolate(I_g, i + u[1], j + u[0]),
				 interpolate(I_g, i - u[1], j - u[0]));

	printf("%d\n", g >= interpolate(I_g, i + u[1], j + u[0])
				 && g >= interpolate(I_g, i - u[1], j - u[0]));
	}
	*/

	return g > interpolate(I_g, i + u[1], j + u[0])
		&& g > interpolate(I_g, i - u[1], j - u[0]);
}

// Create a tensor field corresponding to the probabilities of
// edges passing through each point in the image
// Note: caller must free *endpoints
tensor_field_t *edge_tensors(
  image_float_t *I,
  double scale, double g_threshold, double blur_sigma,
  int **endpoints)
{
	int width, height;
	image_float_t *blur, *dx, *dy, *dxx, *dxy, *dyy, *g;
	image_binary_t *gmax;

	tensor_field_t *tf;
	double *p;
	tensor_t *T;
	int i, j, k, n;

	width = I->width; height = I->height;

	// Blur image
	if(blur_sigma == 0) {
		blur = image_float_clone(I);
	}
	else {
		blur = image_float_create(width, height);
		gaussian_blur(I, blur_sigma, blur);
	}

	//image_float_write("blur.png", blur);

	// Calculate derivatives
	dx = image_float_create(width, height);
	dy = image_float_create(width, height);
	dxx = image_float_create(width, height);
	dxy = image_float_create(width, height);
	dyy = image_float_create(width, height);

	gmax = image_binary_create(width, height);

	sobelx(blur, dx);
	sobely(blur, dy);

	sobelx(dx, dxx);
	sobely(dx, dxy);
	sobely(dy, dyy);

	g = gradient_magnitude(dx, dy);

	image_float_write("grad.png", g);
	//image_float_write("dx.png", dx);
	//image_float_write("dy.png", dy);

	// Create grid tensor field
  tf = tensor_field_create_free(width * height);

	if(endpoints != NULL) {
		*endpoints = malloc(sizeof(int) * width * height);
	}

	n = 0;
	for(i = 0; i < height; i++) {
		for(j = 0; j < width; j++) {
			binary_set_pixel(gmax, i, j, 0);

			if(*float_pixel(g, i, j) < g_threshold) {
				continue;
			}

			if(!gradient_maximum(g, dx, dy, i, j, blur_sigma)) {
				continue;
			}

			binary_set_pixel(gmax, i, j, 1);

			// Write hessian tensor
			p = tensor_field_point(tf, n);
			T = tensor_field_tensor(tf, n);

			p[0] = (double)j;
			p[1] = (double)(height - 1 - i);

			tensor_elt_set(T, 0, 0, *float_pixel(dxx, i, j) * scale);
			tensor_elt_set(T, 0, 1, *float_pixel(dxy, i, j) * scale);
			tensor_elt_set(T, 1, 1, *float_pixel(dyy, i, j) * scale);

      /*
			tensor_elt_set(T, 0, 0, *float_pixel(dxx, i, j) * scale);
			tensor_elt_set(T, 0, 1, -*float_pixel(dxy, i, j) * scale);
			tensor_elt_set(T, 1, 1, *float_pixel(dyy, i, j) * scale);
      */

			if(endpoints != NULL) {
				(*endpoints)[n] = i == 0 || i == height - 1 || j == 0 || j == width - 1;
			}

			n++;
		}
	}
	tf = tensor_field_resize(tf, n);

	image_binary_write("gmax.png", gmax);

	// Cleanup
	image_float_destroy(g);
	image_float_destroy(dyy);
	image_float_destroy(dxy);
	image_float_destroy(dxx);
	image_float_destroy(dy);
	image_float_destroy(dx);
	image_float_destroy(blur);

	return tf;
}
