#include "line_graph_io.h"
#include "mask_io.h"
#include "nifti.h"

#include "cli.h"
#include "cli_util.h"

#include "flow.h"

static bool verbose = false;
static optimize_options_t optim_opts;

static line_graph_t *in_graph = NULL;
static char out_graph_path[PATH_MAX + 1];

static mask_t *source = NULL;
static mask_t *sink = NULL;

static double min_flow = 1;


static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_solver(void *closure_data, const argdef_t *def, const char *value);
static int set_min_flow(void *closure_data, const argdef_t *def, const char *value);
static int set_lambda(void *closure_data, const argdef_t *def, const char *value);


static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_source(void *closure_data, const paramdef_t *def, const char *filename);
static int set_sink(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	{'\0', "solver", 0,
	 "name", ARG_VALUE_ONE, NULL,
	 "Use the given solver for MIP solution instead of default. (glpk, cplex)",
	 &set_solver},

	{'\0', "min-flow", 0,
	 "quantity", ARG_VALUE_ONE, NULL,
	 "The minimum amount of flow that must move from source to sink. \
    Roughly corresponds to the minimum number of tracts recovered. \
    (default = 1)",
	 &set_min_flow},

	{'l', "lambda", 0,
	 "value", ARG_VALUE_ONE, NULL,
	 "The penalty placed on the square of the branching factor. \
    (default = 0.0)",
	 &set_lambda},

	{0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
	{"in-line-graph", "File to read input weighted line graph.", ARG_REQUIRED,
	 &set_in_graph},

  {"source", "File containing mask for source set.", ARG_REQUIRED,
	 &set_source},
	
  {"sink", "File containing mask for sink set.", ARG_REQUIRED,
	 &set_sink},

	{"out-line-graph", "File to output the optimal line graph.", ARG_REQUIRED,
	 &set_out_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "flow",
  "Finding optimal contours between a source and a sink set using minimum cost flow.",
  "Curvature-LP",
  "0.1",
  COPYRIGHT
};

static int argc;
static char **argv;


int main(int _argc, char **_argv)
{
	line_graph_t *H;
	int status = 0;

	// Initialize options to defaults
	optim_opts = optimize_options_defaults;
	optim_opts.lambda = 0;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	H = flow_tractography(&optim_opts, in_graph, source, sink, min_flow);

	if(H == NULL) {
		fprintf(stderr, "Error in calculating optimal line graph\n");
		status = 1;
	}
	else {
		// Output
		line_graph_write_file_name(out_graph_path, H);
	}

	// Cleanup
  line_graph_destroy(in_graph);
	line_graph_destroy(H);
	mask_destroy(sink);
	mask_destroy(source);

	return status;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = optim_opts.verbose = true;
	return 0;
}

static int set_solver(void *closure_data, const argdef_t *def, const char *value)
{
	if(strcmp(value, "glpk") == 0) {
		optim_opts.solver = SOLVER_GLPK;
	}
	else if(strcmp(value, "cplex") == 0) {
		optim_opts.solver = SOLVER_CPLEX;
	}
	else {
		fprintf(stderr, "Invalid value for solver: \"%s\"\n", value);
		exit(EX_USAGE);
	}

	return 0;
}

static int set_min_flow(void *closure_data, const argdef_t *def, const char *value)
{
	min_flow = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_lambda(void *closure_data, const argdef_t *def, const char *value)
{
	optim_opts.lambda = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	in_graph = line_graph_read_file_name(filename);
	if(in_graph == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}


static mask_t *load_mask(const char *filename)
{
	mask_t *mask;

	if(nifti_filename(filename)) {
		mask = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		mask = mask_from_file_name(filename);
	}

	if(mask == NULL) {
		fprintf(stderr, "Error reading input mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return mask;
}

static int set_source(void *closure_data, const paramdef_t *def, const char *filename)
{
	source = load_mask(filename);
	return 0;
}

static int set_sink(void *closure_data, const paramdef_t *def, const char *filename)
{
	sink = load_mask(filename);
	return 0;
}

static int set_out_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_graph_path, filename, PATH_MAX);
	out_graph_path[PATH_MAX] = '\0';

	return 0;
}
