#include "splines/min_combined.h"
#include "util.h"

#include "line_graph_io.h"
#include "tensor_field_io.h"

#include "weight_vector.h"
#include "weight_file.h"

#include "cli.h"
#include "cli_util.h"

#include <gsl/gsl_multimin.h>

static int verbose = 0;
static tensor_field_t *tf = NULL;
static line_graph_t *L = NULL;
static int angle_count = 8;
static char out_weights_path[PATH_MAX + 1];
static int min_error_count = 0;

static minimize_combined_options_t min_spline_opts;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);


static int set_curvature_factor(void *closure_data, const argdef_t *def, const char *value);
static int set_angle_divisions(void *closure_data, const argdef_t *def, const char *value);


static int set_start_intervals(void *closure_data, const argdef_t *def, const char *value);
static int set_max_intervals(void *closure_data, const argdef_t *def, const char *value);
static int set_max_iters(void *closure_data, const argdef_t *def, const char *value);
static int set_min_tolerance(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_weights_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   "level", ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	{'a', "angle-divisions", 0,
	 "intervals", ARG_VALUE_ONE, NULL,
	 "The number of intervals to divide [0, pi) in the weight file",
	 &set_angle_divisions},

	{'C', "curvature-factor", 0,
	 "number", ARG_VALUE_ONE, NULL,
	 "Constant to multiply to curvature in the saliency function.",
	 &set_curvature_factor},

	{'\0', "start-intervals", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The starting number of intervals on which to perform the composite integral estimator in spline optimizations.",
	 &set_start_intervals},

	{'I', "max-intervals", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The maximum number of intervals on which to perform the composite integral estimator in spline optimizations.",
	 &set_max_intervals},

	{'i', "max-iters", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The maximum number of gradient descent iterations in spline optimization.",
	 &set_max_iters},

	{'t', "min-tolerance", 0,
	 "tau", ARG_VALUE_ONE, NULL,
	 "The tolerance used in finding a minimal spline.",
	 &set_min_tolerance},

  {0}
};

static const arggroup_t groups[] = {
	{"Spline Options", 2},

	{0}
};


static const paramdef_t params[] = {
  {"tensor-field", "File containing the tensor field.", ARG_REQUIRED,
	 &set_tensor_field},
	
	{"line-graph", "The line graph over the tensors for which we may need weights.", ARG_REQUIRED,
	 &set_line_graph},

	{"weights-file", "File to output spline weights.", ARG_REQUIRED,
	 &set_weights_file},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "spline_weights_generate",
  "Generating weights for the proposal splines between tensors.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static void min_error(minimize_options_t *opts, int status, const char *msg)
{
	min_error_count += 1;

	if(opts->verbose) {
		fputs(msg, stderr);
		fputc('\n', stderr);
	}
}

static double spline_weight_closure(void *data, int dim, point_t pos[3], point_t directions[3])
{
  int i;
  spline_t *s;
  point_t m;
  double w;

  // Initialize spline
  s = spline_create(dim, 3);
  
  for(i = 0; i < 3; i++) {
    point_assign(dim, spline_p(s, i), pos[i]);
		point_assign(dim, spline_m(s, i), directions[i]);
  }

  // Minimize
  spline_magnitudes_min_combined(s, &min_spline_opts);
  
  w = min_combined_objective(s, &min_spline_opts);
  //w = exp(-w);

  /*
  for(i = 0; i < 3; i++) {
    point_printf(dim, pos[i]);
    fputs(", ", stdout);
    point_printf(dim, directions[i]);
    fputc('\n', stdout);
  }
  printf("w = %lg\n\n", w);
  //*/

  return w;
}

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
  int i;
  weight_vector_array_t *wva;
  weight_vector_t *wv;

  // Set arguments to defaults
  min_spline_opts = minimize_combined_options_default;
	min_spline_opts.minimization_options.error = &min_error;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Display options
	if(verbose) {
		printf("Spline Options:\n");
		printf("  Minimization Options:\n");
		printf("    Maximum iterations: %d\n", min_spline_opts.minimization_options.max_iter);
		printf("    Tolerance: %e\n", (float)min_spline_opts.minimization_options.tol);
		printf("  Integration Options:\n");
		printf("    Starting intervals: %d\n", min_spline_opts.minimization_options.start_precision.N);
		printf("    Maximum intervals: %d\n", min_spline_opts.max_integration_intervals);
		printf("  Weight Parameters:\n");
		printf("    %f * TC^%f + length^%f\n",
					 (float)min_spline_opts.curvature_factor,
					 (float)min_spline_opts.TC_power,
					 (float)min_spline_opts.arc_length_power);

    printf("Weight Options:\n");
    printf("  Angle divisions: %d\n", angle_count);
	}

  // Calculate weights
  wva = weight_vector_array_for_line_graph(L, tf, 1e-4, angle_count); // TODO: tolerance option
  if(wva == NULL) {
    fprintf(stderr, "Error in calculating canonical point set.");
    return 1;
  }

  // Display the canonical points
  if(verbose > 0) {
    printf("Calculating weights for point pairs:\n");
    for(i = 0; i < wva->length; i++) {
      wv = wva->w[i];
      point_printf(weight_vector_dim(wv), weight_vector_pos(wv, 0));
      fputs(", ", stdout);
      point_printf(weight_vector_dim(wv), weight_vector_pos(wv, 1));
      fputc('\n', stdout);
    }

    printf("For a total of %d weights\n", weight_vector_array_total_weights(wva));
  }

  // Find the spline weights
	weight_vector_array_calculate(wva, -1, &spline_weight_closure, NULL);
	fputc('\n', stdout);

	weight_vector_array_to_file_name(out_weights_path, wva);

	if(verbose > 0) {
		printf("Observed %d minimization errors.\n", min_error_count);
	}
	
  return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	if(value == NULL) {
		verbose = 1;
	}
	else {
		verbose = cli_uint_arg(argc, argv, &cli, def, value);
	}

	if(verbose >= 2) {
		min_spline_opts.minimization_options.verbose = 1;
	}

	return 0;
}

static int set_angle_divisions(void *closure_data, const argdef_t *def, const char *value)
{
	angle_count = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_curvature_factor(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.curvature_factor = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_start_intervals(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.start_precision.N = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_max_intervals(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.max_integration_intervals = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_max_iters(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.max_iter = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}


static int set_min_tolerance(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.tol = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input line graph
  L = line_graph_read_file_name(filename);
  if(L == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_weights_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_weights_path, filename, PATH_MAX);
	out_weights_path[PATH_MAX] = '\0';

  return 0;
}
