#include "bundle_io.h"
#include "line_graph_io.h"
#include "tensor_field_io.h"

#include "cli.h"

static bool verbose = false;
static tensor_field_t *tf = NULL;
static line_graph_t *L = NULL;
static char out_path[PATH_MAX+1];

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

  {0}
};


static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"tensor-field", "Input tensor field.", ARG_REQUIRED,
	 &set_tensor_field},

  {"graph", "Input line graph.", ARG_REQUIRED,
	 &set_graph},

  {"out-file", "Output paths file.", ARG_REQUIRED,
	 &set_out_file},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "graph-paths",
  "Create the endpoint-to-endpoint paths corresponding to a line graph.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
  bundle_t *paths;
  path_t *path;
  point_t p, q;
  int **path_vertices, *v, i, j, n, m, dim;
  
  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

  // Construct path
  path_vertices = line_graph_paths(L);
  
  for(n = 0; path_vertices[n] != NULL; n++);

  dim = tensor_field_dim(tf);
  paths = bundle_create(dim, n);

  if(verbose) {
    printf("Found %d paths\n", n);
  }

  for(i = 0; i < n; i++) {
    v = path_vertices[i];
    for(m = 0; v[m] != -1; m++);

    path = path_create(dim, m);
    for(j = 0; j < m; j++) {
      p = tensor_field_point(tf, v[j]);
      q = path_get(path, j);

      point_assign(dim, q, p);
    }

    bundle_set(paths, i, path);
  }

	// Output
	if(verbose) {
		printf("Writing output...\n");
	}
	bundle_write_file_name(out_path, paths);

	bundle_destroy(paths);
  tensor_field_destroy(tf);
  line_graph_destroy(L);

  return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}


static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input line graph
  L = line_graph_read_file_name(filename);
  if(L == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
