#include "bundle_io.h"
#include "tractfile/trk.h"
#include "tractfile/nifti.h"

#include "cli.h"

static bool verbose = false;
static bool flip[3] = {false, false, false};
static bundle_t *paths = NULL;
static trk_hdr_t hdr;
static char out_path[PATH_MAX+1];

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_flipx(void *closure_data, const argdef_t *def, const char *value);
static int set_flipy(void *closure_data, const argdef_t *def, const char *value);
static int set_flipz(void *closure_data, const argdef_t *def, const char *value);
static int add_paths(void *closure_data, const argdef_t *def, const char *filename);


static int set_coordinates(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

	{'p', "paths", ARG_REQUIRED | ARG_MANY,
	 "file", ARG_VALUE_ONE, NULL,
	 "Add a group of paths to the .trk file.",
	 &add_paths},

	{'\0', "flipx", 0,
	 NULL, ARG_VALUE_OPTIONAL, NULL,
	 "Perform x flip.",
	 &set_flipx},

	{'\0', "flipy", 0,
	 NULL, ARG_VALUE_OPTIONAL, NULL,
	 "Perform y flip.",
	 &set_flipy},

	{'\0', "flipz", 0,
	 NULL, ARG_VALUE_OPTIONAL, NULL,
	 "Perform z flip.",
	 &set_flipz},

  {0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"nifti", "NIfTI file from which to read coordinates.", ARG_REQUIRED,
	 &set_coordinates},

  {"trk", "Output .trk file.", ARG_REQUIRED,
	 &set_out_path},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "paths-to-trk",
  "Convert paths to trackvis .trk files.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};


static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	int i, j, k, n, length;
	path_t *path;
	point_t p;
	tract_file_t *out;

	trk_hdr_init(&hdr);

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Get information
	n = bundle_count(paths);
	
  if(verbose) {
    printf("Converting %d paths\n", n);
  }

	// Open .trk file
	out = trk_open_write(out_path, &hdr);

	// Copy tracts
	for(i = 0; i < n; i++) {
		path = bundle_get(paths, i);
		assert(path_dim(path) == 3);

		length = path_length(path);
		for(j = 0; j < length; j++) {
			p = path_get(path, j);

			for(k = 0; k < 3; k++) {
				if(flip[k]) {
					p[k] = hdr.voxel_size[k] * hdr.dim[k] - p[k];
				}
			}
		}

		tract_file_append(out, path);
	}

	// Cleanup
	tract_file_close(out);
	bundle_destroy(paths);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int set_flipx(void *closure_data, const argdef_t *def, const char *value)
{
	flip[0] = true;
	return 0;
}

static int set_flipy(void *closure_data, const argdef_t *def, const char *value)
{
	flip[1] = true;
	return 0;
}

static int set_flipz(void *closure_data, const argdef_t *def, const char *value)
{
	flip[2] = true;
	return 0;
}

static int set_coordinates(void *closure_data, const paramdef_t *def, const char *filename)
{
	nifti_1_header nifti;

	nifti_set_header(&nifti, filename);
	nifti_to_trk(&hdr, &nifti);

	return 0;
}

static int add_paths(void *closure_data, const argdef_t *def, const char *filename)
{
	bundle_t *new_paths;

	new_paths = bundle_read_file_name(filename);
	if(new_paths == NULL) {
		fprintf(stderr, "Error reading input path file \"%s\"\n", filename);
		exit(EX_NOINPUT);
	}

	if(paths == NULL) {
		paths = new_paths;
	}
	else {
		paths = bundle_merge(paths, new_paths);
		bundle_destroy(new_paths);
	}

	return 0;
}

static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
