#include "line_graph_io.h"
#include "tensor_field_io.h"
#include "bundle_io.h"

#include "draw/ext.h"

#include "cli.h"
#include "cli_util.h"

static bool verbose = false;

static tensor_field_t *tf = NULL;
static line_graph_t *G = NULL;
static bundle_t *paths = NULL;

static draw_options_t draw_opts;
static char out_path[PATH_MAX+1];

static int width = 0, height = 0;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_shade_graph(void *closure_data, const argdef_t *def, const char *value);
static int add_bundle(void *closure_data, const argdef_t *def, const char *value);
static int set_tensor_scaling(void *closure_data, const argdef_t *def, const char *value);

static int set_dim(void *closure_data, const argdef_t *def, const char *value);
static int set_grayscale(void *closure_data, const argdef_t *def, const char *value);


static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);
static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_graph(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

	{'p', "paths", ARG_MANY,
	 "path-file", ARG_VALUE_ONE, NULL,
	 "Plot subvoxel paths.",
	 &add_bundle},

	{'\0', "tensor-scaling", 0,
	 "scale", ARG_VALUE_ONE, NULL,
	 "Scale tensors in plot by some quantity.",
	 &set_tensor_scaling},

  {'\0', "shade-graph", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_shade_graph},

	{'w', "width", ARG_REQUIRED,
	 "integer", ARG_VALUE_ONE, NULL,
	 "Width of the output image, in format-dependent units.",
	 &set_dim},

	{'h', "height", ARG_REQUIRED,
	 "integer", ARG_VALUE_ONE, NULL,
	 "Height of the output image, in format-dependent units.",
	 &set_dim},

	{'\0', "grayscale", 0,
	 NULL, ARG_VALUE_ONE, NULL,
	 "Set the image to output in grayscale.",
	 &set_grayscale},
	

	{0}
};

static const arggroup_t groups[] = {
	{"Tensor Field Options", 2},
	{"Image Options", 4},

	{0}
};

static const paramdef_t params[] = {
  {"out-file", "Output image file.", ARG_REQUIRED,
	 &set_out_file},

  {"tensor-field", "File containing the tensor field.", 0,
	 &set_tensor_field},
	
	{"line-graph", "File to read input line graph.", 0,
	 &set_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "plot",
  "Creating plots of tractography results.  Supports 2D output in PNG, PDF.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	int dim;

	// Defaults
	draw_opts = draw_default_options;
	draw_opts.tensor_scaling = 4e-1;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Find dimensions
	if(tf != NULL) {
		dim = tensor_field_dim(tf);
	}
	else if(paths != NULL) {
		dim = bundle_dim(paths);
	}
	else {
		printf("Nothing to plot\n");
		return 0;
	}

	// Some checks
  if(G != NULL && tensor_field_count(tf) != line_graph_vertex_count(G)) {
    fprintf(stderr, "Mismatch between tensor field and line graph.\n");
    return 1;
  }

	if(paths != NULL && dim != bundle_dim(paths)) {
		fprintf(stderr, "Dimensionality mismatch.\n");
		return 1;
	}

	// Draw!
  if(dim == 2) {
		if(width == 0) {
			width = 640;
		}
		if(height == 0) {
			height = 640;
		}

    draw_for_ext(out_path, width, height,
								 &draw_opts,
								 tf, G, paths);
  }
  else {
    UNSUPPORTED_DIMENSIONALITY(dim);
  }

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int set_shade_graph(void *closure_data, const argdef_t *def, const char *value)
{
	draw_opts.shade_graph = true;
	return 0;
}

static int add_bundle(void *closure_data, const argdef_t *def, const char *filename)
{
	bundle_t *new_bundle;

	new_bundle = bundle_read_file_name(filename);
	if(new_bundle == NULL) {
		fprintf(stderr, "Error reading input path file \"%s\"\n", filename);
		exit(EX_NOINPUT);
	}

	if(paths == NULL) {
		paths = new_bundle;
	}
	else {
		paths = bundle_merge(paths, new_bundle);
		bundle_destroy(new_bundle);
	}

	return 0;
}

static int set_tensor_scaling(void *closure_data, const argdef_t *def, const char *value)
{
	draw_opts.tensor_scaling = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_dim(void *closure_data, const argdef_t *def, const char *value)
{
	int d;

	d = cli_uint_arg(argc, argv, &cli, def, value);

	if(def->short_arg == 'h') {
		if(width == 0) {
			width = d;
		}
		height = d;
	}
	else if(def->short_arg == 'w') {
		if(height == 0) {
			height = d;
		}
		width = d;
	}
	
	return 0;
}

static void copy_color(float *dst, const float *src)
{
	int i;

	for(i = 0; i < 3; i++) {
		dst[i] = src[i];
	}
}

static int set_grayscale(void *closure_data, const argdef_t *def, const char *value)
{
	copy_color(draw_opts.background_color, draw_grayscale_default_options.background_color);
	copy_color(draw_opts.tensor_color, draw_grayscale_default_options.tensor_color);
	copy_color(draw_opts.line_graph_color, draw_grayscale_default_options.line_graph_color);
	copy_color(draw_opts.path_color, draw_grayscale_default_options.path_color);

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	G = line_graph_read_file_name(filename);
	if(G == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}
