#include "splines/spline.h"
#include "splines/min_combined.h"

#include "cli.h"
#include "cli_util.h"

#include "util.h"
#include "indices.h"
#include "parallel.h"
#include "progress.h"

#include "draw/context.h"
#include "draw/cairo.h"

#include <cairo.h>
#include <cairo-pdf.h>

#include <gsl/gsl_multimin.h>
#include <pthread.h>

static int verbose = 0;
static int angle_count = 8;
static char out_path[PATH_MAX + 1];
static int min_error_count = 0;

//static draw_options_t draw_opts;
static bool grayscale = false;
static int width = 0, height = 0;

static minimize_combined_options_t min_spline_opts;

typedef struct {
	int n;
	spline_t **splines;
	double *weights;

	int progress;
	time_t started;
	pthread_mutex_t mutex;
} weighted_splines_t;


int weighted_spline_closure(int i, void *closure_data)
{
	int j, size[3], sub[3];
	coord_t pos[3][2] = {
		{-1, 0},
		{0, 0},
		{1, 0}
	};
	double angle;
	point_t p, m;
	weighted_splines_t *w;

	w = (weighted_splines_t *)closure_data;

	// Find what spline we're at
	for(j = 0; j < 3; j++) {
		size[j] = angle_count/2;
	}
	ind2sub(2, size, i, sub);

	// Construct the spline
	w->splines[i] = spline_create(2, 3);

	for(j = 0; j < 3; j++) {
			angle = (M_PI * i) / angle_count;
			
			p = spline_p(w->splines[i], j);
			m = spline_m(w->splines[i], j);

			point_assign(2, p, pos[j]);
			m[0] = cos(angle);
			m[1] = sin(angle);
		}

	// Minimize the spline
	w->weights[i] = spline_magnitudes_min_combined(w->splines[i], &min_spline_opts);

	pthread_mutex_lock(&w->mutex);
	progress(w->started, ++w->progress, w->n, 1);
	pthread_mutex_unlock(&w->mutex);

	return 0;
}

static void weight_color(draw_context_t *context, float w)
{
	float r, g, b;
	float hue, x, intpart;
	int hue_interval;

	assert(w >= 0 && w <= 1);

	if(grayscale) {
		draw_context_color_alpha(context, 0, 0, 0, 1-w);
		return;
	}

	hue = w * 360;
	hue_interval = (int)hue/60;
	
	x = 1 - fabsf(modff(hue/180, &intpart) - 1);

	switch(hue_interval) {
	case 0:
		r = 1;
		g = x;
		b = 0;
		break;

	case 1:
		r = x;
		g = 1;
		b = 0;
		break;

	case 2:
		r = 0;
		g = 1;
		b = x;
		break;

	case 3:
		r = 0;
		g = x;
		b = 1;
		break;

	case 4:
		r = x;
		g = 0;
		b = 1;
		break;

	case 5:
	case 6:
		r = 1;
		g = 0;
		b = x;
		break;

	default:
		fprintf(stderr, "h' = %d\n", hue_interval);
		assert(0);
	}

	draw_context_color(context, r, g, b);
}

static int spline_weight_map()
{
	int i, j, k, total_angles;
	path_t *path;
	point_t p;
	weighted_splines_t w;

	// Get sizes
	w.n = total_angles = icube(angle_count/2);

	// Generate splines
	w.splines = (spline_t **)malloc(sizeof(spline_t *) * total_angles);
	w.weights = (double *)malloc(sizeof(double) * total_angles);
	w.started = time(NULL);
	w.progress = 0;
	pthread_mutex_init(&w.mutex, NULL);

	parallel_execute(total_angles, -1, &weighted_spline_closure, &w);
	fputc('\n', stdout);

	// Output to file
	FILE *spline_file = fopen("splines.txt", "w+");
	FILE *weights_file = fopen("weights.txt", "w+");
	int N = 512;
	for(i = 0; i < total_angles; i++) {
		path = spline_path(w.splines[i], N);

		for(j = 0; j < 2; j++) {
			for(k = 0; k < path_length(path); k++) {
				p = path_get(path, k);
				fprintf(spline_file, "%.10e ", p[j]);
			}
			fputc('\n', spline_file);
		}

		path_destroy(path);

		fprintf(weights_file, "%.10e\n", w.weights[i]);
	}
	fclose(weights_file);
	fclose(spline_file);

	// Scale weights
	double wmin = 1.0/0.0, wmax = -1.0/0.0;
	for(i = 0; i < total_angles; i++) {
		assert(w.weights[i] > 0);
		w.weights[i] = log(w.weights[i]);

		wmin = MIN(wmin, w.weights[i]);
		wmax = MAX(wmax, w.weights[i]);
	}

	for(i = 0; i < total_angles; i++) {
		w.weights[i] = (w.weights[i] - wmin) / (wmax - wmin);
	}

	// Perform plotting
	cairo_surface_t *surface = cairo_pdf_surface_create(out_path, width, height);

	printf("Drawing...\n");

	// Find limits
	draw_context_t *context;
	coord_t min[2] = {INFINITY, INFINITY}, max[2] = {-INFINITY, -INFINITY};
	for(i = 0; i < total_angles; i++) {
		path = spline_path(w.splines[i], 512);

		for(j = 0; j < path_length(path); j++) {
			p = path_get(path, j);
			for(k = 0; k < 2; k++) {
				min[k] = MIN(min[k], p[k]);
				max[k] = MAX(max[k], p[k]);
			}
		}

		path_destroy(path);
	}

	// Create drawing context
	context = draw_context_cairo(surface, width, height, min, max);

	draw_context_line_width(context, 2e-3);

	for(i = 0; i < total_angles; i++) {
		path = spline_path(w.splines[i], 512);
		
		weight_color(context, (float)(w.weights[i]));

		draw_context_line_begin(context, path_get(path, 0));
		for(j = 1; j < path_length(path); j++) {
			draw_context_line_add(context, path_get(path, j));
		}
		draw_context_line_end(context);

		path_destroy(path);
	}

	// Cleanup
	draw_context_finish(context);

  // Cleanup
  cairo_surface_destroy(surface);


	// Cleanup
	pthread_mutex_destroy(&w.mutex);
	for(i = 0; i < total_angles; i++) {
		spline_destroy(w.splines[i]);
	}
	free(w.weights);
	free(w.splines);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_curvature_factor(void *closure_data, const argdef_t *def, const char *value);
static int set_angle_divisions(void *closure_data, const argdef_t *def, const char *value);


static int set_start_intervals(void *closure_data, const argdef_t *def, const char *value);
static int set_max_intervals(void *closure_data, const argdef_t *def, const char *value);
static int set_max_iters(void *closure_data, const argdef_t *def, const char *value);
static int set_min_tolerance(void *closure_data, const argdef_t *def, const char *value);

static int set_dim(void *closure_data, const argdef_t *def, const char *value);
static int set_grayscale(void *closure_data, const argdef_t *def, const char *value);

static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename);


// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   "level", ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	{'a', "angle-divisions", 0,
	 "intervals", ARG_VALUE_ONE, NULL,
	 "The number of intervals to divide [0, pi) in the weight file",
	 &set_angle_divisions},

	{'C', "curvature-factor", 0,
	 "number", ARG_VALUE_ONE, NULL,
	 "Constant to multiply to curvature in the saliency function.",
	 &set_curvature_factor},

	{'\0', "start-intervals", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The starting number of intervals on which to perform the composite integral estimator in spline optimizations.",
	 &set_start_intervals},

	{'I', "max-intervals", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The maximum number of intervals on which to perform the composite integral estimator in spline optimizations.",
	 &set_max_intervals},

	{'i', "max-iters", 0,
	 "count", ARG_VALUE_ONE, NULL,
	 "The maximum number of gradient descent iterations in spline optimization.",
	 &set_max_iters},

	{'t', "min-tolerance", 0,
	 "tau", ARG_VALUE_ONE, NULL,
	 "The tolerance used in finding a minimal spline.",
	 &set_min_tolerance},

	{'w', "width", ARG_REQUIRED,
	 "integer", ARG_VALUE_ONE, NULL,
	 "Width of the output image, in format-dependent units.",
	 &set_dim},

	{'h', "height", ARG_REQUIRED,
	 "integer", ARG_VALUE_ONE, NULL,
	 "Height of the output image, in format-dependent units.",
	 &set_dim},

	{'\0', "grayscale", 0,
	 NULL, ARG_VALUE_ONE, NULL,
	 "Set the image to output in grayscale.",
	 &set_grayscale},

  {0}
};

static const arggroup_t groups[] = {
	{"Spline Options", 2},
	{"Image", 7},

	{0}
};

static const paramdef_t params[] = {
	{"out-file", "File to output the plot.", ARG_REQUIRED,
	 &set_out_path},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "spline-weight-map",
  "Generating weights for the proposal splines between tensors.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};



static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
  // Set arguments to defaults
  min_spline_opts = minimize_combined_options_default;


  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Display options
	if(verbose) {
		printf("Spline Options:\n");
		printf("  Minimization Options:\n");
		printf("    Maximum iterations: %d\n", min_spline_opts.minimization_options.max_iter);
		printf("    Tolerance: %e\n", (float)min_spline_opts.minimization_options.tol);
		printf("  Integration Options:\n");
		printf("    Starting intervals: %d\n", min_spline_opts.minimization_options.start_precision.N);
		printf("    Maximum intervals: %d\n", min_spline_opts.max_integration_intervals);
		printf("  Weight Parameters:\n");
		printf("    %f * TC^%f + length^%f\n",
					 (float)min_spline_opts.curvature_factor,
					 (float)min_spline_opts.TC_power,
					 (float)min_spline_opts.arc_length_power);

    printf("Weight Options:\n");
    printf("  Angle divisions: %d\n", angle_count);
	}

	return spline_weight_map();
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	if(value == NULL) {
		verbose = 1;
	}
	else {
		verbose = cli_uint_arg(argc, argv, &cli, def, value);
	}

	if(verbose >= 2) {
		min_spline_opts.minimization_options.verbose = 1;
	}

	return 0;
}

static int set_angle_divisions(void *closure_data, const argdef_t *def, const char *value)
{
	angle_count = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_curvature_factor(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.curvature_factor = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_start_intervals(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.start_precision.N = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_max_intervals(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.max_integration_intervals = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_max_iters(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.max_iter = cli_uint_arg(argc, argv, &cli, def, value);
	return 0;
}


static int set_min_tolerance(void *closure_data, const argdef_t *def, const char *value)
{
	min_spline_opts.minimization_options.tol = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_dim(void *closure_data, const argdef_t *def, const char *value)
{
	int d;

	d = cli_uint_arg(argc, argv, &cli, def, value);

	if(def->short_arg == 'h') {
		if(width == 0) {
			width = d;
		}
		height = d;
	}
	else if(def->short_arg == 'w') {
		if(height == 0) {
			height = d;
		}
		width = d;
	}
	
	return 0;
}

static int set_grayscale(void *closure_data, const argdef_t *def, const char *value)
{
	grayscale = true;
	return 0;
}

static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

  return 0;
}
