#include "tend.h"

#include "tensor_field_io.h"
#include "mask_io.h"
#include "bundle_io.h"
#include "nifti.h"

#include "cli.h"
#include "cli_util.h"

static bool verbose = false;

static tensor_field_t *tf = NULL;
static mask_t *roi = NULL;
static int N = 0;
static tend_opts_t tend_opts;

static char out_path[PATH_MAX+1];

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_roi(void *closure_data, const argdef_t *def, const char *filename);

static int set_f(void *closure_data, const argdef_t *def, const char *value);
static int set_g(void *closure_data, const argdef_t *def, const char *value);
static int set_curvature_limit(void *closure_data, const argdef_t *def, const char *value);
static int set_FA_limit(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_tracts(void *closure_data, const paramdef_t *def, const char *value);
static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

  {'r', "roi", 0,
   "mask-file", ARG_VALUE_ONE, NULL,
	 "ROI in which to start tracts.",
	 &set_roi},

  {'f', "eigenvector-factor", 0,
   "number", ARG_VALUE_ONE, NULL,
	 "Weight to place on the principal eigenvector in update step.",
	 &set_f},

  {'g', "TEND-factor", 0,
   "number", ARG_VALUE_ONE, NULL,
	 "Weight to place on the TEND orientation in update step.",
	 &set_f},

  {'k', "curvature-limit", 0,
   "number", ARG_VALUE_ONE, NULL,
	 "Curvature limit above which to terminate the tract.",
	 &set_curvature_limit},

  {'F', "FA-limit", 0,
   "number", ARG_VALUE_ONE, NULL,
	 "Fractional Anisotropy (FA) limit below which to terminate the tract.",
	 &set_FA_limit},

	{0}
};

static const arggroup_t groups[] = {
	{"Update Options", 2},
  {"Termination Options", 4},
	{0}
};

static const paramdef_t params[] = {
  {"tensor-field", "Input tensor field file.", ARG_REQUIRED,
	 &set_tensor_field},

  {"N", "Number of tracts to recover.", ARG_REQUIRED,
	 &set_tracts},

  {"out-file", "Output paths file.", ARG_REQUIRED,
	 &set_out_file},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "tend",
  "Perform TEND tractography. \
   Based on \"White Matter Tractography Uisng Diffusion Tensor Deflection\" \
   by Lazar et al (2003).",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	bundle_t *tracts;

	srand(time(NULL));
	//srand(42);

	// Handle arguments
	tend_opts = tend_opts_default;
	
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Output options
	if(verbose) {
		// TODO
	}

	// Perform TEND
	tracts = tend(tf, roi, N, &tend_opts);

	// Output
	bundle_write_file_name(out_path, tracts);

	// Cleanup
	bundle_destroy(tracts);
	tensor_field_destroy(tf);
	mask_destroy(roi);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static mask_t *load_mask(const char *filename)
{
	mask_t *mask;

	if(nifti_filename(filename)) {
		mask = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		mask = mask_from_file_name(filename);
	}

	if(mask == NULL) {
		fprintf(stderr, "Error reading input mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return mask;
}

static int set_roi(void *closure_data, const argdef_t *def, const char *filename)
{
	roi = load_mask(filename);
	return 0;
}

static int set_f(void *closure_data, const argdef_t *def, const char *value)
{
	tend_opts.f = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_g(void *closure_data, const argdef_t *def, const char *value)
{
	tend_opts.g = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_curvature_limit(void *closure_data, const argdef_t *def, const char *value)
{
	tend_opts.curvature_limit = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_FA_limit(void *closure_data, const argdef_t *def, const char *value)
{
	tend_opts.FA_limit = cli_float_arg(argc, argv, &cli, def, value);
	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_tracts(void *closure_data, const paramdef_t *def, const char *value)
{
	int s;

	s = sscanf(value, "%u", &N);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->name, value);
		exit(EX_USAGE);
	}

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
