#include "tensor_field_io.h"
#include "tensor_field.h"
#include "tensor_internal.h"

//////////////////////////////// Read //////////////////////////////

typedef enum state state_t;
enum state {
  STATE_BEGIN,

  STATE_COMMAND_ARGUMENT,
  STATE_POINT_BODY,
  STATE_POINT_END,
  STATE_TENSOR_BODY,
  STATE_END,

  STATE_COMMAND,
  STATE_NUMBER
};

typedef enum token_type token_type_t;
enum token_type {
  TOKEN_NUMBER,
  TOKEN_SYMBOL,
  TOKEN_COMMAND
};

typedef enum command command_t;
enum command {
  COMMAND_NONE = -1,
  COMMAND_DIM = 0,
};

static const char *command_strings[] = {
  "dim",
};

#define BUFFER_LEN (256)

typedef struct token token_t;
struct token {
  token_type_t type;

  union {
    double num;
    char symbol;
    command_t command;
  };
};

struct tensor_field_parser
{
  tensor_field_t *tf;

  state_t tokenizer_state;
  char buffer[BUFFER_LEN];
  int i;

  state_t parser_state;
  command_t command;
  int j;

  int dim;

  point_t p;
  tensor_t *T;
};

tensor_field_parser_t *tensor_field_parser_create()
{
  tensor_field_parser_t *parser;

  parser = (tensor_field_parser_t *)malloc(sizeof(tensor_field_parser_t));

  parser->tokenizer_state = STATE_BEGIN;
  parser->i = 0;

  parser->parser_state = STATE_BEGIN;
  parser->j = 0;

  parser->dim = 0;
  parser->p = NULL;
  parser->T = NULL;

  parser->tf = NULL;

  return parser;
}

void tensor_field_parser_destroy(tensor_field_parser_t *parser)
{
  free(parser->p);
  tensor_destroy(parser->T);
  //tensor_field_destroy(parser->tf);
  free(parser);
}

static int set_dim(tensor_field_parser_t *parser, int dim)
{
  if(parser->dim != 0) {
    return 0;
  }

  parser->dim = dim;
  parser->p = (point_t)malloc(sizeof(coord_t *) * dim);
  parser->T = tensor_create(dim);
  parser->tf = tensor_field_create_free(dim, 0);

  if(parser->p == NULL || parser->T == NULL) {
    return 0;
  }

  return 1;
}

static int add_token(tensor_field_parser_t *parser, token_t *token)
{
  switch(parser->parser_state) {
  case STATE_BEGIN:
    if(token->type == TOKEN_SYMBOL) {
      if(token->symbol == '(') {
        parser->parser_state = STATE_POINT_BODY;
        parser->j = 0;
      }
      else if(token->symbol == '\0') {
        parser->parser_state = STATE_END;
      }
      else {
        return 0;
      }
    }
    else if(token->type == TOKEN_COMMAND) {
      parser->parser_state = STATE_COMMAND_ARGUMENT;
      parser->command = token->command;
    }
    else {
      return 0;
    }

    break;

  case STATE_COMMAND_ARGUMENT:
    if(parser->command == COMMAND_DIM) {
      if(token->type == TOKEN_NUMBER) {
        if(!set_dim(parser, (int)token->num)) {
          return 0;
        }
      }
      else {
        return 0;
      }
    }
    else {
      return 0;
    }

    parser->parser_state = STATE_BEGIN;
    break;

  case STATE_POINT_BODY:
    if(token->type == TOKEN_NUMBER) {
      if(parser->j >= parser->dim) {
        return 0;
      }

      assert(parser->p != NULL);
      parser->p[parser->j++] = token->num;
    }
    else if(token->type == TOKEN_SYMBOL) {
      if(token->symbol == ',') {
      }
      else if(token->symbol == ')') {
        if(parser->j == parser->dim) {
          parser->parser_state = STATE_POINT_END;
        }
        else {
          return 0;
        }
      }
      else {
        return 0;
      }
    }
    else {
      return 0;
    }
    break;

  case STATE_POINT_END:
    if(token->type == TOKEN_SYMBOL && token->symbol == '[') {
      parser->parser_state = STATE_TENSOR_BODY;
      parser->j = 0;
    }
    else {
      return 0;
    }
    break;

  case STATE_TENSOR_BODY:
    if(token->type == TOKEN_SYMBOL) {
      if(token->symbol == ';' || token->symbol == ',') {
      }
      else if(token->symbol == ']') {
        if(parser->j != (parser->dim * (parser->dim + 1) / 2)) {
          return 0;
        }


        // Add the finished pair to field
        assert(parser->tf != NULL && parser->p != NULL && parser->T != NULL);
        parser->tf = tensor_field_add(parser->tf, parser->p, parser->T);

        parser->parser_state = STATE_BEGIN;
      }
      else {
        return 0;
      }
    }
    else if(token->type == TOKEN_NUMBER) {
      if(parser->j >= (parser->dim * (parser->dim + 1) / 2)) {
        return 0;
      }

      assert(parser->T != NULL);
      tensor_data(parser->T)[parser->j++] = token->num;
    }
    else {
      return 0;
    }
    break;

  case STATE_END:
    return 0;
    break;

  default:
    return 0;
  }

  return 1;
}

static int string_same(const char *s1, const char *s2)
{
  for(; *s1 != '\0' && *s2 != '\0' && tolower(*s1) == tolower(*s2);
      s1++, s2++);


  return *s1 == '\0' && *s2 == '\0';
}

static command_t command(const char *s)
{
  if(string_same(s, "dim")) {
    return COMMAND_DIM;
  }
  else {
    return COMMAND_NONE;
  }
}

static int add_character(tensor_field_parser_t *parser, char c)
{
  token_t token;
  float num;

  if(parser->tokenizer_state == STATE_BEGIN) {
    if(isspace(c)) {
      // ignore whitespace
    }
    else if(isdigit(c) || c == '-' || c == '+') {
      // start number
      parser->tokenizer_state = STATE_NUMBER;
      parser->buffer[0] = c;
      parser->i = 1;
    }
    else if(isalpha(c)) {
      // start command
      parser->tokenizer_state = STATE_COMMAND;
      parser->buffer[0] = c;
      parser->i = 1;
    }
    else {
      // output symbol
      token.type = TOKEN_SYMBOL;
      token.symbol = c;
      return add_token(parser, &token);
    }
  }
  else if(parser->tokenizer_state == STATE_COMMAND) {
    if(isalnum(c)) {
      if(parser->i >= BUFFER_LEN-1) {
        return 0;
      }
      parser->buffer[parser->i++] = c;
    }
    else {
      // output command token
      token.type = TOKEN_COMMAND;

      parser->buffer[parser->i] = '\0';
      token.command = command(parser->buffer);

      if(token.command == COMMAND_NONE) {
        return 0;
      }

      if(add_token(parser, &token) == 0) {
        return 0;
      }
      parser->tokenizer_state = STATE_BEGIN;

      return add_character(parser, c);
    }
  }
  else if(parser->tokenizer_state == STATE_NUMBER) {
    if(isdigit(c) || c == '.' || c == 'e' || c == 'E' || c == '+' || c == '-') {
      if(parser->i >= BUFFER_LEN-1) {
        return 0;
      }
      parser->buffer[parser->i++] = c;
    }
    else {
      // output number token
      parser->buffer[parser->i] = '\0';
      sscanf(parser->buffer, "%g", &num);
      
      token.type = TOKEN_NUMBER;
      token.num = num;

      if(add_token(parser, &token) == 0) {
        return 0;
      }
      parser->tokenizer_state = STATE_BEGIN;

      return add_character(parser, c);
    }
  }

  return 1;
}

int tensor_field_parser_feed(
  tensor_field_parser_t *parser,
  int length, const char *data)
{
  int i, status;

  for(i = 0; i < length; i++) {
    status = add_character(parser, data[i]);

    if(status == 0) {
      return 0;
    }
  }

  return 1;
}

// Note: deallocates parser, caller must free result
tensor_field_t *tensor_field_parser_end(tensor_field_parser_t *parser)
{
  tensor_field_t *tf;

  if(add_character(parser, '\0') == 0) {
    return NULL;
  }

  tf = parser->tf;
  parser->tf = NULL;

  return tf;
}

tensor_field_t *tensor_field_from_file(FILE *f)
{
  tensor_field_t *tf;
  tensor_field_parser_t *parser;
  char buffer[4096];
  int n;

  parser = tensor_field_parser_create();

  while(!feof(f)) {
    n = fread(buffer, 1, 4096, f);
    if(tensor_field_parser_feed(parser, n, buffer) == 0) {
      fprintf(stderr, "Parsing error\n");
      tensor_field_parser_destroy(parser);
      return NULL;
    }
  }

  // TODO: check errors with ferror

  tf = tensor_field_parser_end(parser);
  if(tf == NULL) {
    fprintf(stderr, "Parsing error\n");
  }

  tensor_field_parser_destroy(parser);

  return tf;
}

tensor_field_t *tensor_field_from_file_name(const char *path)
{
  tensor_field_t *tf;
  FILE *fp;

  fp = fopen(path, "r");
  if(fp == NULL) {
    return NULL;
  }

  tf = tensor_field_from_file(fp);

  fclose(fp);

  return tf;
}


//////////////////////////////// Write //////////////////////////////

void tensor_field_write(FILE *fp, tensor_field_t *tf)
{
	int i, n, dim;
	
	point_t p;
	tensor_t *T;

	dim = tensor_field_dim(tf);
	n = tensor_field_count(tf);

	fprintf(fp, "dim %d\n\n", dim);

	for(i = 0; i < n; i++) {
		p = tensor_field_point(tf, i);
		T = tensor_field_tensor(tf, i);

		point_fprintf_fmt(fp, dim, p, "%12g");
		fputs("\n", fp);
		tensor_fprintf_fmt(fp, T, "%12g", 0);
		fputs("\n\n", fp);
	}
}

void tensor_field_write_file_name(const char *filename, tensor_field_t *tf)
{
	FILE *fp;

	fp = fopen(filename, "w+");
	if(fp == NULL) {
		return; // TODO: error
	}

	tensor_field_write(fp, tf);

	fclose(fp);
}
