#include "tensor_field_io.h"
#include "nifti.h"
#include "tensor_eigens.h"
#include "neighborhood.h"
#include "boundary.h"
#include "mask_io.h"

#include "cli.h"

static int argc;
static char **argv;

static int verbose = 0;
static struct nifti_1_header nifti_hdr;
static tensor_field_t *tf = NULL;
static char out_path[PATH_MAX + 1];

static float min[3], max[3];
static float lambda = 0;
static mask_t *mask = 0;
static char mask_path[PATH_MAX + 1];

static length_t neighborhood = -1;
static bool write_boundary = false;
static char boundary_path[PATH_MAX + 1];

static bool write_test_subset = false;
static char test_subset_path[PATH_MAX + 1];

static bool write_test_boundary = false;
static char test_boundary_path[PATH_MAX + 1];

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_rect(void *closure_data, const argdef_t *def, const char *value);
static int set_min_lambda(void *closure_data, const argdef_t *def, const char *value);
static int set_mask(void *closure_data, const argdef_t *def, const char *filename);

static int set_neighborhood(void *closure_data, const argdef_t *def, const char *filename);
static int set_boundary(void *closure_data, const argdef_t *def, const char *filename);

static int set_test_subset(void *closure_data, const argdef_t *def, const char *filename);
static int set_test_boundary(void *closure_data, const argdef_t *def, const char *filename);

static int set_in(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

  {'\0', "rect", 0,
   "xmin,ymin,zmin-xmax,ymax,zmax", ARG_VALUE_ONE, NULL,
   "Select out a rectilinear block of tensors",
	 &set_rect},

  {'\0', "lambda", 0,
   "value", ARG_VALUE_ONE, NULL,
   "Select out tensors with |lambda_1| greater than a threshold.",
	 &set_min_lambda},

  {'m', "mask", 0,
   "file", ARG_VALUE_ONE, NULL,
   "A NIfTI or ASCII file containing a mask over the tensor field.",
	 &set_mask},

  // Boundary options
  {'\0', "neighborhood", 0,
   "distance", ARG_VALUE_ONE, NULL,
   "Distance between two tensors for them to be neighbors.",
	 &set_neighborhood},

  {'\0', "boundary", 0,
   "filename", ARG_VALUE_ONE, NULL,
   "File to output the mask for the boundary of the subset.",
	 &set_boundary},

	// Test output options
	{'\0', "test-subset", 0,
	 "filename", ARG_VALUE_ONE, NULL,
	 "Output nifti file of the subset we're selecting.",
	 &set_test_subset},

	{'\0', "test-boundary", 0,
	 "filename", ARG_VALUE_ONE, NULL,
	 "Output nifti file of the boundary of the subset we're selecting.",
	 &set_test_boundary},

  {0}
};


static const arggroup_t groups[] = {
	{"Selection options", 1},

	{"Boundary options", 4},

	{"Test output options", 6},

	{0}
};


static const paramdef_t params[] = {
  {"in", "Input tensor field.", ARG_REQUIRED,
	 &set_in},
	
	{"out", "Output (subset) tensor field.", ARG_REQUIRED,
	 &set_out},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "contours",
  "Select a subset of the tensors in a tensor field.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};


static point_t get_point(void *closure_data, int i)
{
	return tensor_field_point((tensor_field_t *)closure_data, i);
}

int main(int _argc, char **_argv)
{
	tensor_field_t *subset;
	int i, j, m, n, dim;
	bool accept;

	graph_t *G;
	point_t p;
	tensor_t *T;
	tensor_eigens_t *eig;

	mask_t *result, *not_result, *bound_mask, *subset_bound_mask;
	int *index_map;

	int size[3];
	float voxdim[3];

	// Initialize arguments to defaults
	for(i = 0; i < 3; i++) {
		min[i] = -1e100; // TODO
		max[i] = 1e100;
	}

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	printf("Parsed arguments\n");

	// Get properties of input tensor
	n = tensor_field_count(tf);
	dim = tensor_field_dim(tf);

	// Print out arguments
	if(verbose) {
		printf("Selecting tensors:\n");
		printf("  Within: [%g, %g]", min[0], max[0]);
		for(j = 1; j < dim; j++) {
			printf(", [%g, %g]", min[j], max[j]);
		}
		printf("\n");
		printf("  With principal eigenvalue greater than %g\n", lambda);
		if(mask != NULL) {
			printf("  Applying mask \"%s\"\n", mask_path);
		}

		// TODO
	}

	////////////// Find subset //////////////
	result = mask_create(n);

	for(i = 0; i < n; i++) {
		// Grab ith tensor
		p = tensor_field_point(tf, i);
		T = tensor_field_tensor(tf, i);

		eig = tensor_eigens_get(T);

		accept = true;

		// Check tensor properties
		if(mask != NULL && !mask_get(mask, i)) {
			accept = false;
		}

		if(fabs(tensor_eigenvalue(eig, 0)) < lambda) {
			accept = false;
		}

		for(j = 0; j < dim; j++) {
			if((float)p[j] < min[j] || (float)p[j] > max[j]) {
				accept = false;
			}
		}

		mask_set(result, i, accept);
	}

	////////////// Boundary //////////////
	if(verbose) {
		fputs("Calculating boundary...\n", stdout);
	}

	// Find the inverse of the mask
	not_result = mask_copy(result);
	mask_not(not_result);
	
	G = graph_create_from_neighborhoods(dim, n, &get_point, tf, neighborhood);
	bound_mask = boundary(G, not_result);

	// Add boundary to the tensors we want
	mask_or(result, bound_mask);

	// Create subset tensor field
	m = mask_count_true(result);
	subset = tensor_field_create_free(dim, m);
	index_map = (int *)malloc(m * sizeof(int));

	if(verbose) {
		printf("Selected %d tensors\n", m);
	}

	j = 0;
	for(i = 0; i < n; i++) {
		if(!mask_get(result, i)) {
			continue;
		}

		// Grab ith tensor
		p = tensor_field_point(tf, i);
		T = tensor_field_tensor(tf, i);
		
		assert(j < m);
		point_assign(dim, tensor_field_point(subset, j), p);
		tensor_assign(tensor_field_tensor(subset, j), T);
		index_map[j] = i;
		j++;
	}

	// Write subset
	if(verbose) {
		printf("Writing subset tensor field...\n");
	}
	tensor_field_write_file_name(out_path, subset);

	// Write boundary
	if(write_boundary) {
		// Map boundary mask to subset indices
		subset_bound_mask = mask_create(m);
		for(j = 0; j < m; j++) {
			mask_set(subset_bound_mask, j, mask_get(bound_mask, index_map[j]));
		}

		if(verbose) {
			printf("Writing boundary to \"%s\"...\n", boundary_path);
		}
		mask_write_file_name(boundary_path, subset_bound_mask);
	}

	// Write test output
	for(i = 0; i < 3; i++) {
		size[i] = nifti_hdr.dim[i+1];
		voxdim[i] = nifti_hdr.pixdim[i+1];
	}

	// Test ouptu
 if(write_test_subset) {
	 nifti_write_mask(test_subset_path, result, size, voxdim, &nifti_load_options_default);
 }

 if(write_test_boundary) {
	 nifti_write_mask(test_boundary_path, bound_mask, size, voxdim, &nifti_load_options_default);
 }

	// Cleanup
	// TODO

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = 1;

	return 0;
}

//int min[3] = {}, max[3] = {INT_MAX, INT_MAX, INT_MAX};

static int set_rect(void *closure_data, const argdef_t *def, const char *value)
{
	int s;

	s = sscanf(value, "%g,%g,%g-%g,%g,%g",
						 min, min + 1, min + 2,
						 max, max + 1, max + 2);

	if(s != 6) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return 0;
}

static float float_arg(const argdef_t *def, const char *value)
{
	int s;
	float v;

	s = sscanf(value, "%g", &v);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return v;
}

static int set_min_lambda(void *closure_data, const argdef_t *def, const char *value)
{
	lambda = float_arg(def, value);

	return 0;
}

static int set_in(void *closure_data, const paramdef_t *def, const char *filename)
{
	if(nifti_filename(filename)) {
		tf = nifti_load_tensor_field(filename, &nifti_load_options_default, &nifti_hdr);
	}
	else {
		tf = tensor_field_from_file_name(filename);
	}

	if(tf == NULL) {
		fprintf(stderr, "Error reading input tensor field\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_out(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}

static int set_mask(void *closure_data, const argdef_t *def, const char *filename)
{
	strncpy(mask_path, filename, PATH_MAX);
	mask_path[PATH_MAX] = '\0';

	if(nifti_filename(filename)) {
		mask = nifti_load_mask(filename, &nifti_load_options_default, NULL);
	}
	else {
		mask = mask_from_file_name(filename);
	}

	if(mask == NULL) {
		fprintf(stderr, "Error reading input mask file \"%s\"\n", filename);
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_neighborhood(void *closure_data, const argdef_t *def, const char *filename)
{
	neighborhood = (length_t)float_arg(def, filename);

	return 0;
}

static int set_boundary(void *closure_data, const argdef_t *def, const char *filename)
{
	write_boundary = true;

	// Copy to output file
	strncpy(boundary_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}


static int set_test_subset(void *closure_data, const argdef_t *def, const char *filename)
{
	write_test_subset = true;

	strncpy(test_subset_path, filename, PATH_MAX);
	test_subset_path[PATH_MAX] = '\0';

	return 0;
}

static int set_test_boundary(void *closure_data, const argdef_t *def, const char *filename)
{
	write_test_boundary = true;

	strncpy(test_boundary_path, filename, PATH_MAX);
	test_boundary_path[PATH_MAX] = '\0';

	return 0;
}
