#include "tensor_canvas.hh"
#include "gl_templates.hh"
#include "model/glmatrix.hh"

extern "C" {
#include "tensor_eigens.h"
}

static const int DEFAULT_SPHERE_PRECISION = 64;
static const float DEFAULT_TENSOR_SCALE = 4e-1;
//static const float DEFAULT_TENSOR_SCALE = 8e1;

static void InitMaterials()
{
    static const GLfloat ambient[4] = {0.1f, 0.1f, 0.1f, 1.0f};
    static const GLfloat diffuse[4] = {0.5f, 1.0f, 1.0f, 1.0f};
    static const GLfloat position0[4] = {0.0f, 0.0f, 20.0f, 0.0f};
    static const GLfloat position1[4] = {0.0f, 0.0f, -20.0f, 0.0f};
    static const GLfloat front_mat_shininess[1] = {60.0f};
    static const GLfloat front_mat_specular[4] = {0.2f, 0.2f, 0.2f, 1.0f};
    static const GLfloat front_mat_diffuse[4] = {0.5f, 0.28f, 0.38f, 1.0f};
    /*
    static const GLfloat back_mat_shininess[1] = {60.0f};
    static const GLfloat back_mat_specular[4] = {0.5f, 0.5f, 0.2f, 1.0f};
    static const GLfloat back_mat_diffuse[4] = {1.0f, 1.0f, 0.2f, 1.0f};
    */
    static const GLfloat lmodel_ambient[4] = {1.0f, 1.0f, 1.0f, 1.0f};
    static const GLfloat lmodel_twoside[1] = {GL_FALSE};

    /*
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position0);
    glEnable(GL_LIGHT0);
    */

    glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT1, GL_POSITION, position1);
    glEnable(GL_LIGHT1);

    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
    glEnable(GL_LIGHTING);

    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);
}


BEGIN_EVENT_TABLE(TensorCanvas, ModelCanvas)
END_EVENT_TABLE()

TensorCanvas::TensorCanvas(
    wxWindow *parent,
    const wxPoint &pos, const wxSize &size,
    const wxString &name,
    tensor_field_t *init_tf, line_graph_t *init_G,
    bool init_draw_tensor_glyphs, 
    bundle_t *init_paths,
    float init_tensor_scale) :

  ModelCanvas(parent, pos, size, name,
              init_tf == NULL ? 2 : tensor_field_dim(init_tf))
{

  if(init_tensor_scale < 0) {
    tensor_scale = DEFAULT_TENSOR_SCALE;
  }
  else {
    tensor_scale = init_tensor_scale;
  }

  // Initialize tensor field
  tf = NULL;
  if(init_tf != NULL) {
    setTensorField(init_tf);
  }

  draw_tensor_glyphs = init_draw_tensor_glyphs;

  // Initialize line graph
  if(init_G == NULL) {
    G = NULL;
  }
  else {
    setLineGraph(init_G);
  }

  // Initialize paths
  paths = init_paths;

  // Initialize OpenGL
  use_display_lists = true; // TODO: determine this
  field_list_valid = false;

  quad = gluNewQuadric();

  if(use_display_lists) {
    lists = glGenLists(3);
  }

  setSpherePrecision(DEFAULT_SPHERE_PRECISION);
}

TensorCanvas::~TensorCanvas()
{
  if(use_display_lists) {
    glDeleteLists(lists, 3);
  }
  gluDeleteQuadric(quad);

  //line_graph_destroy(G);
  tensor_field_destroy(tf);
}


void TensorCanvas::setTensorField(tensor_field_t *new_tf)
{
  // Set tf
  tensor_field_destroy(tf);
  tf = tensor_field_clone(new_tf);

  // Dimensionality
  int dim = tensor_field_dim(tf);
  assert(dim == 2 || dim == 3);
  SetDimensionality(dim);

  // Scaling
  tensor_field_extent(tf, min, max);

  // Find the maximimum eigenvalue
  int n = tensor_field_count(tf);
  max_lambda = 0;
  for(int i = 0; i < n; i++) {
    tensor_eigens_t *eigs = tensor_eigens_get(tensor_field_tensor(tf, i));
    if(max_lambda < tensor_eigenvalue(eigs, 0)) {
      max_lambda = tensor_eigenvalue(eigs, 0);
    }
  }

  // Add an appropriate margin
  for(int i = 0; i < dim; i++) {
    min[i] -= tensor_scale * max_lambda;
    max[i] += tensor_scale * max_lambda;
  }

  field_list_valid = false;
}

tensor_field_t *TensorCanvas::getTensorField()
{
  return tf;
}

void TensorCanvas::setLineGraph(line_graph_t *new_G)
{
  //G = line_graph_copy(new_G);
  G = new_G;
  field_list_valid = false;
}

line_graph_t *TensorCanvas::getLineGraph()
{
  return G;
}

void TensorCanvas::setSpherePrecision(int p)
{
  sphere_precision = p;
  initGlyphLists();
}

int TensorCanvas::getSpherePrecision()
{
  return sphere_precision;
}

void TensorCanvas::drawCircle()
{
  glBegin(GL_LINE_STRIP);
  for(int i = 0; i < sphere_precision; i++) {
    float a = (2 * M_PI * i) / (sphere_precision - 1);
    float x = cos(a), y = sin(a);

    glVertex2f(x, y);
  }
  glEnd();
}

void TensorCanvas::drawSphere()
{
  gluSphere(quad, 1.0, sphere_precision, sphere_precision);
}

void TensorCanvas::initGlyphLists()
{
  if(!use_display_lists) {
    return;
  }

  glNewList(circleList(), GL_COMPILE);
  drawCircle();
  glEndList();

  glNewList(sphereList(), GL_COMPILE);
  drawSphere();
  glEndList();
}

static int sphere_count = 0;

void TensorCanvas::drawTensor(point_t p, tensor_t *T)
{

  int dim = tensor_dim(T);
  assert(dim == GetDimensionality());

  // Perform eigendecomposition
  tensor_eigens_t *eigs = tensor_eigens_get(T);
  if(tensor_eigenvalue(eigs, 2) * tensor_scale < 1e-4) {
    return;
  }

  glPushMatrix();

  if(dim == 2) {
    // Translate to point center
    glt::translate(p[0], p[1], 0.0);

    // Rotate first eigenvector to x
    point_t v1 = tensor_eigenvector(eigs, 0);
    coord_t a = atan2(v1[1], v1[0]);
    glt::rotate(180.0 / M_PI * a, 0.0, 0.0, 1.0);

    // Scale by eigenvalues
    glt::scale(tensor_eigenvalue(eigs, 0), tensor_eigenvalue(eigs, 1), 1.0);

    // Global tensor scaling
    glScalef(tensor_scale / max_lambda, tensor_scale / max_lambda, 1.0);

    // Draw!
    if(use_display_lists) {
      glCallList(circleList());
    }
    else {
      drawCircle();
    }
  }
  else {
    assert(dim == 3);

    // Translate to point center
    glt::translate(p[0], p[1], p[2]);

    //printf("p = "); point_printf(3, p); putc('\n', stdout);
    //printf("min = "); point_printf(3, min); putc('\n', stdout);
    // printf("max = "); point_printf(3, max); putc('\n', stdout);
    //printf("Ap = %f, %f, %f\n",
    //	   (float)((p[0] - min[0]) / (max[0] - min[0])),
    //	   (float)((p[1] - min[1]) / (max[1] - min[1])),
    //	   (float)((p[2] - min[2]) / (max[2] - min[2])));

    //coord_t prod[3];
    //GLMatrix::get(GL_MODELVIEW).mult(prod, p);

    //printf("gl = "); point_printf(3, prod); putc('\n', stdout);

    //glt::scale(1e-1, 1e-1, 1e-1);

    ///*
    // Rotate first eigenvector to x
    coord_t v1[4];
    point_assign(3, v1, tensor_eigenvector(eigs, 0));
    v1[3] = 1;

    // Rotate into x-y plane
    coord_t a1 = atan2(v1[2], v1[1]);
    glt::rotate(180.0 / M_PI * a1, 1.0, 0.0, 0.0);

    coord_t prod[4];
    GLMatrix A = GLMatrix::rotation(a1, 1.0, 0.0, 0.0);
    A.mult(prod, v1);

    // Rotate to x-axis
    coord_t a2 = atan2(prod[1], prod[0]);
    glt::rotate(180.0 / M_PI * a2, 0.0, 0.0, 1.0);
    A *= GLMatrix::rotation(a2, 0, -sin(a1), cos(a1));

    // Rotate second eigenvector to y
    coord_t v2[4];
    point_assign(3, v2, tensor_eigenvector(eigs, 1));
    v2[3] = 1;
    A.mult(prod, v2);
   
    glt::rotate(180.0 / M_PI * atan2(prod[2], prod[1]), 1.0, 0.0, 0.0);

    // Scale by eigenvalues
    glt::scale(tensor_eigenvalue(eigs, 0),
               tensor_eigenvalue(eigs, 1),
               tensor_eigenvalue(eigs, 2));
    

    // Global tensor scaling
    glScalef(tensor_scale / max_lambda, tensor_scale / max_lambda, tensor_scale / max_lambda);

    //printf("scale * eig1 = %f\n", (float)(tensor_scale * tensor_eigenvalue(eigs, 0)));
    //printf("lambda = %f, %f, %f\n",
    //	   (float)tensor_eigenvalue(eigs, 0),
    //	   (float)tensor_eigenvalue(eigs, 1),
    //	   (float)tensor_eigenvalue(eigs, 2));
    //*/

    // Draw!
    sphere_count++;
    if(use_display_lists) {
      glCallList(sphereList());
    }
    else {
      drawSphere();
    }
  }


  glPopMatrix();
}

void TensorCanvas::fieldSpace()
{  
  // Move to space of the tensor field 
  int dim = tensor_field_dim(tf);

  coord_t s = 0;
  for(int i = 0; i < dim; i++) {
    if(s < max[i] - min[i]) {
      s = max[i] - min[i];
    }
  }
  s = 1.8 / s;

  if(dim == 2) {
    glTranslatef(-0.9, -0.9, 0.0);
    glt::scale(s, s, 1.0);
    glt::translate(-min[0], -min[1], 0.0);
  }
  else {
    assert(dim == 3);

    printf("min = %f, %f, %f\n", (float)min[0], (float)min[1], (float)min[2]);
    printf("max = %f, %f, %f\n", (float)max[0], (float)max[1], (float)max[2]);

    glTranslatef(-0.9, -0.9, -0.9);
    glt::scale(s, s, s);
    glt::translate(-min[0], -min[1], -min[2]);
  }
}

void TensorCanvas::drawField()
{
  int dim;

  printf("Drawing field...\n");

  if(tf == NULL) {
    return;
  }

  dim = tensor_field_dim(tf);
  if(dim == 2) {
    glDisable(GL_LIGHTING);
  }
  else {
    InitMaterials();
  }

  glColor3f(1.0, 1.0, 1.0);

  // Draw the tensors themselves
  int n = tensor_field_count(tf);
  for(int i = 0; i < n; i++) {
    drawTensor(tensor_field_point(tf, i), tensor_field_tensor(tf, i));

    if(n > 1000 && i % (n / 100) == 0) {
      printf("%.1f\r", (100.0 * i) / n);
      fflush(stdout);
    }
  }
  printf("sphere_count = %d\n", sphere_count);
}

void TensorCanvas::drawGraph()
{
  if(G == NULL) {
    return;
  }

  // Draw lines
  int dim = tensor_field_dim(tf);

  if(dim == 2) {
    glColor3f(1.0, 0.0, 0.0);
  }
  else {
    assert(dim == 3);

    // Set up color
    static const GLfloat front_mat_shininess[1] = {60.0f};
    static const GLfloat front_mat_specular[4] = {1.0f, 0.0f, 0.0f, 1.0f};
    static const GLfloat front_mat_diffuse[4] = {1.0f, 0.0f, 0.0f, 1.0f};

    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);
  }

  glBegin(GL_LINES);

  int n = line_graph_vertex_count(G);
  for(int i = 0; i < n; i++) {
    int d = line_graph_edge_pair_count(G, i);

    for(int j = 0; j < d; j++) {
      edge_pair_t *ep = line_graph_edge_pair(G, i, j);

      point_t p1 = tensor_field_point(tf, ep->neighbors[0]);
      point_t p2 = tensor_field_point(tf, ep->common_vertex);
      point_t p3 = tensor_field_point(tf, ep->neighbors[1]);

      if(dim == 2) {
	glt::vertex2(p1[0], p1[1]);
	glt::vertex2(p2[0], p2[1]);

	glt::vertex2(p2[0], p2[1]);
	glt::vertex2(p3[0], p3[1]);
      }
      else {
	glt::vertex3(p1[0], p1[1], p1[2]);
	glt::vertex3(p2[0], p2[1], p2[2]);

	glt::vertex3(p2[0], p2[1], p2[2]);
	glt::vertex3(p3[0], p3[1], p3[2]);
      }
    }
  }

  glEnd();
}

void TensorCanvas::drawPaths()
{
  if(paths == NULL) {
    return;
  }

  printf("Drawing paths...\n");
  
  // Draw lines
  int dim = bundle_dim(paths);

  if(dim == 2) {
    glColor3f(0.0, 0.0, 1.0);
  }
  else {
    assert(dim == 3);

    // Set up color
    static const GLfloat front_mat_shininess[1] = {60.0f};
    static const GLfloat front_mat_specular[4] = {0.0f, 0.0f, 1.0f, 1.0f};
    static const GLfloat front_mat_diffuse[4] = {0.0f, 0.0f, 1.0f, 1.0f};

    glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);
  }

  int n = bundle_count(paths);
  for(int i = 0; i < n; i++) {
    glBegin(GL_LINE_STRIP);

    path_t *path = bundle_get(paths, i);

    int length = path_length(path);
    for(int j = 0; j < length; j++) {
      point_t p = path_get(path, j);

      if(dim == 2) {
	glt::vertex2(p[0], p[1]);
      }
      else {
	assert(dim == 3);

	glt::vertex3(p[0], p[1], p[2]);
      }
    }

    glEnd();
  }
}

void TensorCanvas::draw()
{
  if(field_list_valid) {
    glCallList(fieldList());
  }
  else {
    if(use_display_lists) {
      glNewList(fieldList(), GL_COMPILE_AND_EXECUTE);
    }

    glPushMatrix();
    fieldSpace();

    if(draw_tensor_glyphs) {
      drawField();
    }
    drawGraph();
    drawPaths();

    glPopMatrix();

    if(use_display_lists) {
      glEndList();
      field_list_valid = true;
    }
  }
}
