#include "tensor_canvas.hh"

extern "C" {
#include "interface/tensor_field_io.h"
#include "interface/line_graph_io.h"
#include "interface/bundle_io.h"
#include "core/nifti.h"
}

#include <wx/cmdline.h>

class Frame: public wxFrame
{
public:
  Frame(const wxString &title, const wxPoint &pos, const wxSize &size,
        tensor_field_t *tf, line_graph_t *G, bool draw_glyphs = true,
	bundle_t *paths = NULL);

  void OnQuit(wxCommandEvent &event);
  
  DECLARE_EVENT_TABLE()

private:
  ModelCanvas *canvas;
};

Frame::Frame(const wxString &title, const wxPoint &pos, const wxSize &size,
             tensor_field_t *tf, line_graph_t *G, bool draw_glyphs,
	     bundle_t *paths) :
  wxFrame(NULL, -1, title, pos, size)
{
  canvas = new TensorCanvas(this,
                            wxDefaultPosition, wxDefaultSize, _("TensorCanvas"),
                            tf, G,
                            draw_glyphs,
			    paths);  
}

void Frame::OnQuit(wxCommandEvent &event)
{
  Close(true);
}

BEGIN_EVENT_TABLE(Frame, wxFrame)
END_EVENT_TABLE()

class App: public wxApp
{
public:
  App();
  virtual ~App();

protected:
  virtual bool OnInit();
  virtual void OnInitCmdLine(wxCmdLineParser &);
  virtual bool OnCmdLineParsed(wxCmdLineParser &);

private:
  tensor_field_t *tf;
  line_graph_t *G;
  bundle_t *paths;
  bool draw_glyphs;

};

App::App()
{
  tf = NULL;
  G = NULL;
  paths = NULL;
}

App::~App()
{
  tensor_field_destroy(tf);
  line_graph_destroy(G);
}

bool App::OnInit()
{
  if (!wxApp::OnInit())
    return false;

  srand(time(NULL));

  wxString title = _("Tensor World");
  Frame *frame = new Frame(title, wxPoint(50, 50), wxSize(640, 640), tf, G, draw_glyphs, paths);
  
  frame->Show(true);
  SetTopWindow(frame);

  return true;
}

void App::OnInitCmdLine(wxCmdLineParser& parser)
{
  parser.SetSwitchChars(_("-"));

  parser.AddParam(_("tensors"),
                  wxCMD_LINE_VAL_STRING,
                  wxCMD_LINE_OPTION_MANDATORY);

  parser.AddParam(_("graph"),
                  wxCMD_LINE_VAL_STRING,
                  wxCMD_LINE_PARAM_OPTIONAL);

  parser.AddSwitch(_("glyphs"), wxEmptyString,
                   _("Whether to display tensor glyphs."));

  parser.AddSwitch(_("h"), _("help"),
                   _("Display command-line usage help."),
                   wxCMD_LINE_OPTION_HELP);

  parser.AddOption(_("paths"), wxEmptyString,
		   _("Display paths."));
}
 
bool App::OnCmdLineParsed(wxCmdLineParser& parser)
{

  // Tensors
  wxString wx_tf_filename = parser.GetParam(0);

  assert(wx_tf_filename.Length() < 4096);
  char tf_filename[4096 + 1];
  strncpy(tf_filename, wx_tf_filename.mb_str(), 4096);
  tf_filename[4096] = '\0';

  if(nifti_filename(tf_filename)) {
    nifti_load_options_t nifti_opts = nifti_load_options_default;
    nifti_opts.element_order = 'L';
    
    tf = nifti_load_tensor_field(tf_filename, &nifti_opts, NULL);
  }
  else {
    tf = tensor_field_from_file_name(tf_filename);
  }

  printf("Tensor Count = %d\n", tensor_field_count(tf));

  // Graph
  if(parser.GetParamCount() > 1) {
    wxString G_file = parser.GetParam(1);

    G = line_graph_read_file_name(G_file.mb_str());
  }

  // Glyphs
  draw_glyphs = parser.Found(_("glyphs"));

  // Paths
  wxString wx_paths_filename;
  if(parser.Found(_("paths"), &wx_paths_filename)) {
    assert(wx_paths_filename.Length() < 4096);
    char paths_filename[4096 + 1];
    strncpy(paths_filename, wx_paths_filename.mb_str(), 4096);
    paths_filename[4096] = '\0';

    paths = bundle_read_file_name(paths_filename);
  }

  return true;
}

IMPLEMENT_APP(App)
