#include "line_graph_io.h"
#include "tensor_field_io.h"
#include "weight_file.h"

#include "weights.h"

#include "cli.h"

static bool verbose = false;

static line_graph_t *G = NULL;
static char out_path[PATH_MAX + 1];

static double threshold = 0.5;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_threshold(void *closure_data, const argdef_t *def, const char *value);

static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

  {'t', "threshold", 0,
   "number", ARG_VALUE_ONE, NULL,
   "The value at which to threshold the line graph.",
	 &set_threshold},

	{0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
	{"in", "Input weighted line graph.", ARG_REQUIRED,
	 &set_in_graph},

	{"out", "Output thresholded line graph.", ARG_REQUIRED,
	 &set_out_path},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "threshold",
  "Thresholding a line graph by its weights.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	line_graph_t *out;

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	out = line_graph_threshold(G, threshold);

	// Output
	line_graph_write_file_name(out_path, out);

	return 0;
}

static float float_arg(const argdef_t *def, const char *value)
{
	int s;
	float v;

	s = sscanf(value, "%g", &v);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return v;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int set_threshold(void *closure_data, const argdef_t *def, const char *value)
{
	threshold = float_arg(def, value);
	return 0;
}

static int set_in_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	G = line_graph_read_file_name(filename);
	if(G == NULL) {
		fprintf(stderr, "Error reading line graph file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
