#include "tensor_field_io.h"
#include "nifti.h"
#include "affine.h"

#include "cli.h"
#include "cli_util.h"

// Set up options, input/output
static bool verbose = false;
static tensor_field_t *tf = NULL;
static char out_path[PATH_MAX + 1];
static affine_t *transform = NULL;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int scale(void *closure_data, const argdef_t *def, const char *value);

static int set_in(void *closure_data, const paramdef_t *def, const char *filename);
static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

	// Optimization options
	{'\0', "scale", 0,
	 "x,y,z", ARG_VALUE_ONE, NULL,
	 "An x-y-z scaling.",
	 &scale},

	{0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"in", "Input tensor field.", ARG_REQUIRED,
	 &set_in},

	{"out", "File to output the transformed tensor field.", ARG_REQUIRED,
	 &set_out_path},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "transform-tensor-field",
  "Apply an affine transform to the points of a tensor field.",
  "Curvature-LP",
  "0.1",
  COPYRIGHT
};

static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	tensor_field_t *out;
	int i, n;
	point_t p;

	transform = affine_create(3);
	affine_identity(transform);

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Print out arguments
	if(verbose) {
		printf("\nApplying transformation:\n");
		affine_print(transform);
	}

	// Transform
	assert(tensor_field_dim(tf) == 3);
	n = tensor_field_count(tf);
	for(i = 0; i < n; i++) {
		p = tensor_field_point(tf, i);

		affine_transform(transform, p, p);
	}
	
	if(verbose) {
		printf("Writing tensor field file...\n");
	}
	tensor_field_write_file_name(out_path, tf);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static int scale(void *closure_data, const argdef_t *def, const char *value)
{
	coord_t s[3];
	int n;

	n = sscanf(value, "%lg,%lg,%lg", s, s+1, s+2);
	if(n != 3) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	affine_post_scale(transform, s);

	return 0;
}


static int set_in(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_out_path(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
