#include "bundle_io.h"

#include "cli.h"

static bool verbose = false;
static bundle_t *paths = NULL;
static char out_path[PATH_MAX+1];

static double radius = 1.0;
static int count = 10;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);
static int set_radius(void *closure_data, const argdef_t *def, const char *value);
static int set_count(void *closure_data, const argdef_t *def, const char *value);
static int add_bundle(void *closure_data, const argdef_t *def, const char *value);

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information.",
	 &set_verbose},

	{'p', "paths", ARG_MANY,
	 "path-file", ARG_VALUE_ONE, NULL,
	 "Add new core paths.",
	 &add_bundle},

	{'r', "radius", 0,
	 "distance", ARG_VALUE_ONE, NULL,
	 "Set the radius of each tube.",
	 &set_radius},

	{'N', "count", 0,
	 "integer", ARG_VALUE_ONE, NULL,
	 "Set the number of paths per tube.",
	 &set_count},

	{0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"out-file", "Output paths file.", ARG_REQUIRED,
	 &set_out_file},

	{0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "tubes",
  "Create tubes of paths.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};


static int argc;
static char **argv;

int main(int _argc, char **_argv)
{
	bundle_t *tubes, *tube;
	path_t *path;
	int i, n;
	
	srand(time(NULL));

  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

	// Output job
	if(paths == NULL) {
		printf("No paths given\n");
		return 0;
	}

	n = bundle_count(paths);
	if(verbose) {
		printf("Generating %d tubes:\n", n);
		printf("  radius = %lf\n", radius);
		printf("  count = %d\n", count);
	}

	tubes = bundle_create(bundle_dim(paths), 0);
	for(i = 0; i < n; i++) {
		path = bundle_get(paths, i);

		tube = bundle_create_uniform(path, radius, count);
		tubes = bundle_merge(tubes, tube);
		bundle_destroy(tube);
	}

	// Output
	if(verbose) {
		printf("Writing output...\n");
	}
	bundle_write_file_name(out_path, tubes);

	bundle_destroy(tubes);
	bundle_destroy(paths);

	return 0;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = true;
	return 0;
}

static float float_arg(const argdef_t *def, const char *value)
{
	int s;
	float v;

	s = sscanf(value, "%g", &v);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return v;
}


static unsigned int unsigned_int_arg(const argdef_t *def, const char *value)
{
	int s;
	unsigned int u;

	s = sscanf(value, "%u", &u);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->value, value);
		exit(EX_USAGE);
	}

	return u;
}

static int set_radius(void *closure_data, const argdef_t *def, const char *value)
{
	radius = float_arg(def, value);
	return 0;
}

static int set_count(void *closure_data, const argdef_t *def, const char *value)
{
	count = unsigned_int_arg(def, value);
	return 0;
}

static int add_bundle(void *closure_data, const argdef_t *def, const char *filename)
{
	bundle_t *new_bundle;

	new_bundle = bundle_read_file_name(filename);
	if(new_bundle == NULL) {
		fprintf(stderr, "Error reading input path file \"%s\"\n", filename);
		exit(EX_NOINPUT);
	}

	if(paths == NULL) {
		paths = new_bundle;
	}
	else {
		paths = bundle_merge(paths, new_bundle);
		bundle_destroy(new_bundle);
	}

	return 0;
}

static int set_out_file(void *closure_data, const paramdef_t *def, const char *filename)
{
	strncpy(out_path, filename, PATH_MAX);
	out_path[PATH_MAX] = '\0';

	return 0;
}
