#include "line_graph_io.h"
#include "tensor_field_io.h"
#include "mask_io.h"
#include "nifti.h"

#include "neighborhood.h"

#include "cli.h"

int verbose = 0;
static tensor_field_t *tf = NULL;
static char out_graph_path[PATH_MAX + 1];
static double neighborhood;

static int set_verbose(void *closure_data, const argdef_t *def, const char *value);

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename);
static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename);
static int set_neighborhood(void *closure_data, const paramdef_t *def, const char *filename);

// Set up CLI
static const argdef_t arguments[] = {
	// General options
  {'v', "verbose", 0,
   NULL, ARG_VALUE_OPTIONAL, NULL,
   "Whether to print status information during weight calculation and optimization.",
	 &set_verbose},

  {0}
};

static const arggroup_t groups[] = {
	{0}
};

static const paramdef_t params[] = {
  {"tensor-field", "File containing the tensor field.", ARG_REQUIRED,
	 &set_tensor_field},
	
	{"neighborhood", "The maximum distance within which two tensors are potentially connected.", ARG_REQUIRED,
	 &set_neighborhood},

	{"line-graph", "File to output the weighted line graph.", ARG_REQUIRED,
	 &set_line_graph},

  {0}
};

static const cli_t cli = {
  arguments,
  groups,
  params,
  "Unweighted Graph",
  "Building an unweighted line graph between tensors.",
  "Curvature-LP",
  "0.1",
  "Maxwell Collins"
};

static int argc;
static char **argv;

static point_t get_point(void *closure_data, int i)
{
	return tensor_field_point((tensor_field_t *)closure_data, i);
}

int main(int _argc, char **_argv)
{
  int dim, n;
	line_graph_t *G;
	
  // Handle arguments
	argc = _argc;
	argv = _argv;

	cli_argparse(&cli, argc, argv, NULL);

  // Create line graph
  dim = tensor_field_dim(tf);
  n = tensor_field_count(tf);
  G = line_graph_create_from_neighborhoods(dim, n, &get_point, tf, neighborhood);

	if(G == NULL) {
		fprintf(stderr, "Error in generating line graph.\n");
		return 1;
	}
	else {
		// Output
		line_graph_write_file_name(out_graph_path, G);
	}

	line_graph_destroy(G);
	tensor_field_destroy(tf);

  return 0;
}

static float float_arg(const paramdef_t *def, const char *value)
{
	int s;
	float v;

	s = sscanf(value, "%g", &v);
	if(s != 1) {
		cli_error(&cli, argc, argv, "Invalid value for %s: \"%s\"",
							def->name, value);
		exit(EX_USAGE);
	}

	return v;
}

static int set_verbose(void *closure_data, const argdef_t *def, const char *value)
{
	verbose = 1;
	return 0;
}

static int set_tensor_field(void *closure_data, const paramdef_t *def, const char *filename)
{
  // Parse input tensor field
  tf = tensor_field_from_file_name(filename);
  if(tf == NULL) {
		fprintf(stderr, "Error reading tensor field file\n");
		exit(EX_NOINPUT);
  }

	return 0;
}

static int set_neighborhood(void *closure_data, const paramdef_t *def, const char *arg)
{
	neighborhood = float_arg(def, arg);
	return 0;
}

static int set_line_graph(void *closure_data, const paramdef_t *def, const char *filename)
{
	// Copy to output file
	strncpy(out_graph_path, filename, PATH_MAX);
	out_graph_path[PATH_MAX] = '\0';

	return 0;
}
