#include "glmatrix.hh"

GLMatrix GLMatrix::get(GLenum which)
{
  GLMatrix M;
  GLenum getconst;

  switch(which) {
  case GL_MODELVIEW:
    getconst = GL_MODELVIEW_MATRIX;
    break;
  case GL_PROJECTION:
    getconst = GL_PROJECTION_MATRIX;
    break;
  case GL_TEXTURE:
    getconst = GL_TEXTURE_MATRIX;
    break;
  case GL_COLOR:
    getconst = GL_COLOR_MATRIX;
    break;
  default:
    return M;
  }

  glGetDoublev(getconst, M.data);

  return M;
}

void GLMatrix::set(GLenum which, const GLMatrix &M)
{
  glMatrixMode(which);
  glLoadMatrixd(M.data);
}

const GLMatrix GLMatrix::identity =
  {{1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1}};

GLMatrix GLMatrix::rotation(double angle, double x, double y, double z)
{
  // Trig
  double s = sin(angle), c = cos(angle);

  // Normalize
  double mag = sqrt(x*x + y*y + z*z);
  x /= mag; y /= mag; z /= mag;

  // Construct matrix in column-major format
  // As in OpenGL documentation for glRotate
  GLMatrix r;

  r.data[0]  = x*x*(1-c) + c;
  r.data[1]  = y*x*(1-c) + z*s;
  r.data[2]  = x*z*(1-c) - y*s;
  r.data[3]  = 0;

  r.data[4]  = x*y*(1-c) - z*s;
  r.data[5]  = y*y*(1-c) + c;
  r.data[6]  = y*z*(1-c) + x*s;
  r.data[7]  = 0;

  r.data[8]  = x*z*(1-c) + y*s;
  r.data[9]  = y*z*(1-c) - x*s;
  r.data[10] = z*z*(1-c) + c;
  r.data[11] = 0;

  r.data[12] = 0;
  r.data[13] = 0;
  r.data[14] = 0;
  r.data[15] = 1;

  return r;
}

GLMatrix GLMatrix::translation(double x, double y, double z)
{
  GLMatrix t = identity;

  t.data[12] = x;
  t.data[13] = y;
  t.data[14] = z;

  return t;
}

double *GLMatrix::elt(int i, int j)
{
  return &data[4*j+i];
}

const double *GLMatrix::elt(int i, int j) const
{
  return &data[4*j+i];
}

void GLMatrix::mult(double *prod, const double *v) const
{
  for(int i = 0; i < 4; i++) {
    prod[i] = 0;
  }

  for(int j = 0; j < 4; j++) {
    for(int i = 0; i < 4; i++) {
      prod[i] += data[4*j+i] * v[j];
    }
  }
}

template<class T>
static inline void swap(T &a, T &b)
{
  T s;

  s = a;
  a = b;
  b = s;
}

static void printmat(double A[4][5])
{
  for(int i = 0; i < 4; i++) {
    for(int j = 0; j < 5; j++) {
      cout<<A[i][j]<<", ";
    }
    cout<<endl;
  }
  cout<<endl;
}

static void printvec(double *x)
{
  for(int i = 0; i < 4; i++) {
    cout<<x[i]<<", ";
  }
  cout<<endl<<endl;
}

void GLMatrix::solve(double *dst, const double *b) const
{
  double A[4][5];
  double x[4];

  // Copy into A matrix
  for(int j = 0; j < 4; j++) {
    for(int i = 0; i < 4; i++) {
      A[i][j] = *elt(i, j);
    }
  }
  for(int i = 0; i < 4; i++) {
    A[i][4] = b[i];
  }


  // Perform gaussian elimination with pivoting
  const int n = 4, m = 5;
  int perm[4] = {1, 2, 3, 4};

  for(int i = 0; i < n; i++) {
    // Pivot
    int pivot = i;
    for(int k = i+1; k < n; k++) {
      if(fabs(A[k][i]) > fabs(A[pivot][i])) {
        pivot = k;
      }
    }

    if(A[pivot][i] == 0) {
      return;
    }

    // Swap rows
    for(int j = 0; j < m; j++) {
      swap(A[i][j], A[pivot][j]);
    }
    swap(perm[pivot], perm[i]);

    // Elimination
    for(int j = i+1; j < m; j++) {
      A[i][j] /= A[i][i];
    }
    A[i][i] = 1;

    for(int k = i+1; k < n; k++) {
      for(int j = i+1; j < m; j++) {
        A[k][j] -= A[k][i] * A[i][j];
      }
      A[k][i] = 0;
    }
  }

  // Back-substitution
  for(int i = n-1; i >= 0; i--) {
    x[i] = A[i][m-1];
    for(int j = i+1; j < n; j++) {
      x[i] -= A[i][j] * x[j];
    }
  }

  // Reverse pivoting permutation
  for(int i = 0; i < n; i++) {
    dst[perm[i]] = x[i];
  }

  //printvec(dst);
}

GLMatrix &GLMatrix::operator *= (const GLMatrix &B)
{
  double A[4][4];
  
  // Copy original value
  for(int j = 0; j < 4; j++) {
    for(int i = 0; i < 4; i++) {
      A[i][j] = *elt(i, j);
    }
  }

  for(int i = 0; i < 4; i++) {
    for(int j = 0; j < 4; j++) {
      double S = 0;

      for(int k = 0; k < 4; k++) {
        S += A[i][k] * *B.elt(k, j);
      }

      *elt(i, j) = S;
    }
  }
}

GLMatrix operator * (const GLMatrix &A, const GLMatrix &B)
{
  GLMatrix prod;

  for(int i = 0; i < 4; i++) {
    for(int j = 0; j < 4; j++) {
      double S = 0;

      for(int k = 0; k < 4; k++) {
        S += *A.elt(i, k) * *B.elt(k, j);
      }

      *prod.elt(i, j) = S;
    }
  }

  return prod;
}

std::ostream &operator << (std::ostream &stream, const GLMatrix &M)
{
  const char *prefix = "[";

  for(int i = 0; i < 4; i++) {
    stream<<prefix;
    prefix = " ";

    stream<<(*M.elt(i, 0));
    for(int j = 1; j < 4; j++) {
      stream<<", "<<(*M.elt(i, j));
    }

    if(i == 3) {
      stream<<"]";
    }
    else {
      stream<<";";
    }
    stream<<std::endl;
  }

  return stream;
}
