#include "model_canvas.hh"
#include "glmatrix.hh"

#define KEY_ROTATE_INCREMENT (5.0 * M_PI/180.0)
#define MOUSE_ROTATE_COEFF (0.5 * M_PI / 180.0)
#define MOUSE_ZOOM_COEFF (1.1)
#define MOUSE_TRANSLATE_COEFF (0.1)


// loosely based on isosurf example from wxWidgets

static int gl_attributes[] = {
  WX_GL_RGBA,
  WX_GL_DOUBLEBUFFER,
  0};

static int style = wxFULL_REPAINT_ON_RESIZE | wxBORDER_NONE;

ModelCanvas::ModelCanvas(
  wxWindow *parent,
  const wxPoint &pos,
  const wxSize &size,
  const wxString &name,
  int dimensionality,
  bool take_focus,
  bool left_rotates, bool right_rotates) :
  
  wxGLCanvas(parent, wxID_ANY, pos, size, style, name, gl_attributes),

  left_dragging(false),
  right_dragging(false),
  left_button_rotates(left_rotates), right_button_rotates(right_rotates)
{
  parent->Show(true);
  SetCurrent();

  // Set up environment
  glEnable(GL_DEPTH_TEST);

  SetDimensionality(dimensionality);

  // Initially grab keyboard focus
  if(take_focus) {
    SetFocus();
  }
}

ModelCanvas::~ModelCanvas()
{
}

////////////////////////// Event Handlers //////////////////////////

void ModelCanvas::OnPaint(wxPaintEvent &event)
{
  SetCurrent();

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Perform drawing
  draw();

  // Finish drawing
  glFlush();
  SwapBuffers();
}

void ModelCanvas::OnSize(wxSizeEvent &event)
{
  // pass to parent class
  wxGLCanvas::OnSize(event);

  // get new window size
  int w, h;
  GetClientSize(&w, &h);

  // set the OpenGL viewport
  SetCurrent();
  glViewport(0, 0, (GLint)w, (GLint)h);
}

void ModelCanvas::rotate(double increment, const double *axis)
{
  SetCurrent();

  double translation[3];
  GLMatrix M = GLMatrix::get(GL_MODELVIEW);

  // Save translation
  for(int i = 0; i < 3; i++) {
    translation[i] = *M.elt(i, 3);
  }

  // Perform premultiply rotation
  GLMatrix R = GLMatrix::rotation(increment, axis[0], axis[1], axis[2]);
  M = R * M;

  // Restore translation
  for(int i = 0; i < 3; i++) {
    *M.elt(i, 3) = translation[i];
  }

  GLMatrix::set(GL_MODELVIEW, M);

  Refresh();
}

void ModelCanvas::zoom(double factor)
{
  SetCurrent();

  if(dimensionality == 2) {
    glScalef(factor, factor, 0.0);
  }
  else if(dimensionality == 3) {
    GLMatrix M = GLMatrix::get(GL_MODELVIEW);
    (*M.elt(2, 3)) *= factor;
    GLMatrix::set(GL_MODELVIEW, M);
  }

  Refresh();
}

void ModelCanvas::translate(double x, double y, double z)
{
  SetCurrent();

  GLMatrix M = GLMatrix::get(GL_MODELVIEW);

  GLMatrix T = GLMatrix::translation(x, y, z);
  M = T * M;

  GLMatrix::set(GL_MODELVIEW, M);
  
  Refresh();
}

static const double e1[4] = {1, 0, 0, 0};
static const double e2[4] = {0, 1, 0, 0};

void ModelCanvas::OnKeyDown(wxKeyEvent& event)
{
  if(dimensionality == 3) {
    switch(event.GetKeyCode()) {
    case WXK_LEFT:
      rotate(-KEY_ROTATE_INCREMENT, e2);
      break;

    case WXK_RIGHT:
      rotate(KEY_ROTATE_INCREMENT, e2);
      break;

    case WXK_UP:
      rotate(KEY_ROTATE_INCREMENT, e1);
      break;

    case WXK_DOWN:
      rotate(-KEY_ROTATE_INCREMENT, e1);
      break;

    default:
      event.Skip();
    }
  }
  else {
    event.Skip();
  }
}

void ModelCanvas::OnMouseLeftDown(wxMouseEvent &event)
{
  // Initialize dragging
  if(!right_dragging) {
    left_dragging = true;

    last_x = event.GetX();
    last_y = event.GetY();
  }

  // Grab keyboard focus
  SetFocus();
}

void ModelCanvas::OnMouseLeftUp(wxMouseEvent &event)
{
  left_dragging = false;
}

void ModelCanvas::OnMouseRightDown(wxMouseEvent &event)
{
  // Initialize dragging
  if(!left_dragging) {
    right_dragging = true;

    last_x = event.GetX();
    last_y = event.GetY();
  }

  // Grab keyboard focus
  SetFocus();
}

void ModelCanvas::OnMouseRightUp(wxMouseEvent &event)
{
  right_dragging = false;
}

void ModelCanvas::OnMouseMove(wxMouseEvent &event)
{
  if(!(left_dragging || right_dragging)) {
    return;
  }

  if(dimensionality == 2) {
    int width, height;
    GetSize(&width, &height);

    double last_angle = atan2(last_y - height/2.0, last_x - width/2.0);
    double angle = atan2(event.GetY() - height/2.0, event.GetX() - width/2.0);

    if(left_dragging && left_button_rotates
       || right_dragging && right_button_rotates) {
      
      double axis[3];

      axis[0] = 0;
      axis[1] = 0;
      axis[2] = -1;

      rotate((angle - last_angle), axis);
    }
  }
  else if(dimensionality == 3) {
    // Determine movement
    double x, y;

    x = event.GetX() - last_x;
    y = -(event.GetY() - last_y);

    if(left_dragging && left_button_rotates
       || right_dragging && right_button_rotates) {
      // Set up rotation
      double axis[3];
      double mag;

      axis[0] = -y;
      axis[1] = x;
      axis[2] = 0;

      // Perform rotation
      mag = sqrt(x*x + y*y);
      rotate(MOUSE_ROTATE_COEFF * mag, axis);
    }
  }

  last_x = event.GetX();
  last_y = event.GetY();
}

void ModelCanvas::OnMouseWheel(wxMouseEvent &event)
{
  double rotation = (double)event.GetWheelRotation()/event.GetWheelDelta();

  zoom(pow(MOUSE_ZOOM_COEFF, rotation));
}

void ModelCanvas::OnEraseBackground(wxEraseEvent &event)
{
  // no-op, OpenGL follows different drawing model
}

BEGIN_EVENT_TABLE(ModelCanvas, wxGLCanvas)
  EVT_SIZE(ModelCanvas::OnSize)
  EVT_PAINT(ModelCanvas::OnPaint)
  EVT_ERASE_BACKGROUND(ModelCanvas::OnEraseBackground)

  EVT_KEY_DOWN(ModelCanvas::OnKeyDown)

  EVT_LEFT_DOWN(ModelCanvas::OnMouseLeftDown)
  EVT_LEFT_UP(ModelCanvas::OnMouseLeftUp)
  EVT_RIGHT_DOWN(ModelCanvas::OnMouseRightDown)
  EVT_RIGHT_UP(ModelCanvas::OnMouseRightUp)
  EVT_MOTION(ModelCanvas::OnMouseMove)

  EVT_MOUSEWHEEL(ModelCanvas::OnMouseWheel)
END_EVENT_TABLE()

int ModelCanvas::GetDimensionality()
{
  return dimensionality;
}

void ModelCanvas::SetDimensionality(int dim)
{
  // Reset any ongoing rotations
  left_dragging = right_dragging = false;

  switch(dim) {

  case 2:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1,1, -1,1, 0.0, 1e100);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    dimensionality = 2;
    break;

  case 3:
  default:
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glFrustum(-1.0, 1.0, -1.0, 1.0, 5.0, 32.0);
    glFrustum(-0.5, 0.5, -0.5, 0.5, 5.0, 32.0);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Initial translation
    glTranslatef(0, 0, -7.0);
    zoom_translation = -7.0;

    dimensionality = 3;
    break;
  }
}

////////////////////// Placeholder Drawing ///////////////////
#define ANGLES (16)
#define STEPS (8)
#define WIRE_STEPS (128)

void ModelCanvas::draw()
{
  // Draw axes
  glColor3f(1.0, 1.0, 1.0);

  glBegin(GL_LINES);

  glVertex3f(0.0, 0.0, 0.0);
  glVertex3f(1.0, 0.0, 0.0);

  glVertex3f(0.0, 0.0, 0.0);
  glVertex3f(0.0, 1.0, 0.0);

  glVertex3f(0.0, 0.0, 0.0);
  glVertex3f(0.0, 0.0, 1.0);

  glEnd();

  // Draw wireframe paraboloid
  glColor3f(1.0, 0.0, 0.0);
  for(int i = 0; i < ANGLES; i++) {
    glBegin(GL_LINE_STRIP);

    // Find the desired angle
    float a = ((2*M_PI)/ANGLES) * i;
    float s = sin(a), c = cos(a);

    for(int j = 0; j < WIRE_STEPS; j++) {
      float r = (double)j/WIRE_STEPS;
      float x = r*c, y = r*s, z = r*r;

      glVertex3f(x, y, z);
    }

    glEnd();
  }

  for(int i = 0; i < STEPS; i++) {
    glBegin(GL_LINE_STRIP);

    float r = (double)(i + 1)/STEPS;
    float z = r*r;

    for(int j = 0; j < WIRE_STEPS; j++) {
      float a = (2*M_PI)*(double)j/(WIRE_STEPS-1);
      float s = sin(a), c = cos(a);
      float x = r*c, y = r*s;

      glVertex3f(x, y, z);
    }

    glEnd();
  }
}
