#include "TC.h"

#include "util.h"
#include "integral.h"
#include "spline_segment.h"
#include "cross_product.h"

///////////////////// TC integral closure /////////////////////

typedef struct TC_closure TC_closure_t;
struct TC_closure {
	double power;
	spline_t *spline;
};

static double curvature_closure(void *closure_data, double t)
{
	TC_closure_t *state;

	state = (TC_closure_t *)closure_data;

	return pow(spline_curvature(state->spline, t), state->power);
}


double TC(spline_t *spline, double power, int integration_intervals)
{
	TC_closure_t state;

  state.power = power;
	state.spline = spline;

  return integrate_composite_simpsons(
    &curvature_closure,
    0, spline->length-1,
    integration_intervals,
    &state);
}


///////////////////// TC gradient integral closure /////////////////////
typedef struct TC_grad_closure TC_grad_closure_t;
struct TC_grad_closure {
	segment_t segment;
  double power;
};

// Find the derivative of C'_i(t) * C''_j(t)
// a term that appears in cross products
static coord_t dcrossterm(int i, int j, param_t t,
                          point_t u0, point_t dp, point_t ddp)
{
  coord_t d;

  d  = ((1-3*t)*(1-t)*u0[i]) * ddp[j];
  d += dp[i] * ((6*t - 4) * u0[j]);

  return d;
}

static double magnitude_gradient_closure(void *closure_voidp, double t)
{
  // Unpack closure
  TC_grad_closure_t *closure = (TC_grad_closure_t *)closure_voidp;

  segment_t *seg = &closure->segment;
	coord_t alpha = (coord_t)closure->power;
  int dim = seg->dim;

  // Calculation variables
  point_t p0, m0, p1, m1;

  const crossprod_def_t *crossprod;
	coord_t g, gt, ndp, da_dp, da_ddp, sgn;
  coord_t dnumerator;
  coord_t crossprod_elt, da_crossprod_elt;
  length_t crossprod_norm;

	coord_t dp[dim], ddp[dim], u0[dim];

  coord_t norm_dp, squared_norm_dp;

	int i, j, k;

  // Unpack knot points
  p0 = seg->p0;
  m0 = seg->m0;
  p1 = seg->p1;
  m1 = seg->m1;
	unit_vector(dim, m0, u0);

	// Calculate C'(t), C''(t) for our current a
	spline_segment_derivative(seg, t, dp);
	spline_segment_derivative2(seg, t, ddp);

	squared_norm_dp = point_squared_l2_norm(dim, dp);
	norm_dp = sqrt(squared_norm_dp);

	// da(||C'(t) x C''(t)||) term in numerator
	// this verified
  if(dim == 2) {
    dnumerator  = dcrossterm(1, 2, (param_t)t, u0, dp, ddp);
    dnumerator -= dcrossterm(2, 1, (param_t)t, u0, dp, ddp);

    crossprod_norm = dp[0] * ddp[1] - dp[1]*ddp[0];

    if(crossprod_norm < 0) {
      dnumerator *= -1;
    }

    crossprod_norm = fabs(crossprod_norm);
  }
  else if(dim == 3) {
    dnumerator = 0;
    crossprod_norm = 0;

    for(i = 0; i < 3; i++) {
      // Calculate an element of the cross product
      // and it's derivative
      j = (i + 1) % 3;
      k = (i + 2) % 3;

      crossprod_elt    = dp[j] * ddp[k];
      da_crossprod_elt = dcrossterm(j, k, (param_t)t, u0, dp, ddp);

      j = (i + 2) % 3;
      k = (i + 1) % 3;

      crossprod_elt   -= dp[j] * ddp[k];
      da_crossprod_elt = dcrossterm(j, k, (param_t)t, u0, dp, ddp);

      // Add its component in the norm
      crossprod_norm += square(crossprod_elt);
      dnumerator += 2 * crossprod_elt * da_crossprod_elt;
    }

		if(crossprod_norm == 0) {
			//printf("t = %lf\n", t);
			return 0;
		}

    crossprod_norm = sqrt(crossprod_norm);
    dnumerator = 0.5 / crossprod_norm * dnumerator;
	}
  else {
		UNSUPPORTED_DIMENSIONALITY(dim);
  }

	// ||C'(t)||^3 * da(||C'(t) * C''(t)||) term in numerator
	gt = dnumerator * cube(norm_dp);

	// da(||C'(t)||^3) = 1.5 * ||C'(t)|| * da(||C'(t)||^2)
	// this section verified
	da_dp = 0;
	for(i = 0; i < dim; i++) {
		da_dp += 2 * dp[i] * (1 - 3*t)*(1-t) * u0[i]; // TODO: express as dotprod?
	}
	da_dp *= 1.5 * norm_dp;

	// Add ||C'(t) x C''(t)|| * da(||C'(t)||^3)
  gt -= crossprod_norm * da_dp;

	// ||C'(t)||^6 denominator
	gt /= cube(squared_norm_dp);

	// Chain rule term for ^\alpha
	gt *= alpha * pow(spline_derivatives_curvature(dim, dp, ddp), alpha-1);

	return gt;
}

double TC_magnitude_gradient(
  spline_t *spline, int i, double power,
  int integration_intervals)
{
	double g;
  TC_grad_closure_t state;
  coord_t backm[2][spline->dim];

	state.power = power;

	g = 0;

	// Forward
	if(i != spline->length - 1) {
		spline_segment_get(&state.segment, spline, i);

		g += integrate_composite_simpsons(&magnitude_gradient_closure, 0, 1, integration_intervals, &state);
	}

	// Backward
	if(i != 0) {
		// Just take the segment going in the opposite direction
		spline_segment_get_backwards(&state.segment, spline, i, backm[0], backm[1]);

		g += integrate_composite_simpsons(&magnitude_gradient_closure, 0, 1, integration_intervals, &state);
	}

	if(g != g) {
		printf("g = %lf\n", g);
		exit(1);
	}

	return g;
}
