#include "arc_length.h"

#include "spline_segment.h"
#include "integral.h"

typedef struct arc_length_closure arc_length_closure_t;
struct arc_length_closure
{
  double power;
  spline_t *spline;
};

static double arc_length_t(void *closure_data, double t)
{
  arc_length_closure_t *closure;
  spline_t *spline;
  double power;
  speed_t speed;
  
  // Unpack closure
  closure = (arc_length_closure_t *)closure_data;
  power = closure->power;
  spline = closure->spline;

  // Caluclate speed^alpha
  return pow(spline_speed(spline, t), power);
}

double arc_length(spline_t *spline, double power, int integration_intervals)
{
  arc_length_closure_t closure;

  closure.power = power;
  closure.spline = spline;

  return integrate_composite_simpsons(
    &arc_length_t,
    0, spline->length-1,
    integration_intervals,
    &closure);
}

typedef struct arc_length_grad_closure arc_length_grad_closure_t;
struct arc_length_grad_closure
{
  double power;
  segment_t segment;
};

double magnitude_gradient_t(void *data, double t)
{
  // Unpack closure
  arc_length_grad_closure_t *closure = (arc_length_grad_closure_t *)data;
  double p = closure->power;
  segment_t *seg = &closure->segment;

  int dim = seg->dim;

  // Derivative calculation
  coord_t dp[dim], u0[dim];
  length_t ndp;
  double da_dp;

  int i;

  // Unit vector of first part
  unit_vector(dim, seg->m0, u0);

  // Get derivative
  spline_segment_derivative(seg, t, dp);
  ndp = point_squared_l2_norm(dim, dp);

  // Calculate magnitude derivative
  da_dp = 0;
  for(i = 0; i < dim; i++) {
		da_dp += 2 * dp[i] * (1 - 3*t)*(1-t) * u0[i]; // TODO: as dotprod?
  }
  da_dp *= p/2 * pow(ndp, p/2-1);

  //printf("da_dp = %2f\n", da_dp);

  return da_dp;
}

double arc_length_magnitude_gradient(
  spline_t *spline, int i, double power,
  int integration_intervals)
{
  arc_length_grad_closure_t closure;
	double g;
  coord_t backm[2][spline->dim];

  closure.power = power;

  g = 0;

	// Forward
	if(i != spline->length - 1) {
		spline_segment_get(&closure.segment, spline, i);

		g += integrate_composite_simpsons(&magnitude_gradient_t, 0, 1, integration_intervals, &closure);
	}

	// Backward
	if(i != 0) {
		// Just take the segment going in the opposite direction
		spline_segment_get_backwards(&closure.segment, spline, i, backm[0], backm[1]);

		g += integrate_composite_simpsons(&magnitude_gradient_t, 0, 1, integration_intervals, &closure);
	}
  
  return g;
}
