#include "min_spline.h"
#include "util.h"
#include "timing.h"

#include <gsl/gsl_multimin.h>

static void print_error(minimize_options_t *opts, int err, const char *msg)
{
	if(opts->verbose) {
		fputs(msg, stderr);
		fputc('\n', stderr);
	}
}

const minimize_options_t minimize_options_default = {
  1024, // max_iter
  1e-1, // tol
  {1024,}, // start precision
  0, // verbose
	&print_error, // error
	NULL // user_data
};

//////////// Picking optimum derivative constraint magnitudes /////////////

static int magnitudes_from_gsl_vector(spline_t *spline, const gsl_vector *v)
{
	int i, dim;
	length_t mag;
  coord_t *m;

  assert(v->size == spline->length);

  dim = spline->dim;

  for(i = 0; i < spline->length; i++) {
    mag = (length_t)gsl_vector_get(v, i);
    m = spline_m(spline, i);

    point_mul(dim, m, mag);
  }

	return 0;
}

static void assign_derivatives(spline_t *dst, spline_t *src)
{
	int i, dim;

	assert(dst->length == src->length);
  assert(dst->dim == src->dim);
 
  dim = dst->dim;

	for(i = 0; i < src->length; i++) {
    point_assign(dim, spline_m(dst, i), spline_m(src, i));
	}
}

static void assign_gsl_vector(gsl_vector *dst, int dim, double *src)
{
	int i;

	assert(dst->size == dim);

	for(i = 0; i < dim; i++) {
		gsl_vector_set(dst, i, src[i]);
	}
}


typedef struct {
	spline_t *spline;
	void *closure_data;

	precision_t e;
	double (*objective)(void *closure_data, spline_t *s, precision_t e);
	void (*gradient)(void *closure_data, spline_t *s, int dim, double *grad, precision_t e);

} state_t;

static spline_t *spline_setup(state_t *state, const gsl_vector *v)
{
	spline_t *s;

	// Make a local copy of the spline
	s = spline_copy(state->spline);

	// Set magnitudes
	magnitudes_from_gsl_vector(s, v);

	return s;
}

static double min_objective(const gsl_vector *v, void *params)
{
	state_t *state;
	spline_t *s;
	double f;

	state = (state_t *)params;
	s = spline_setup(state, v);

  f = (*state->objective)(state->closure_data, s, state->e);

	spline_destroy(s);

	return f;
}

static void min_gradient(const gsl_vector *v, void *params, gsl_vector *g)
{
	state_t *state;
	spline_t *s;
	int dim;
	double *grad;

	state = (state_t *)params;
	s = spline_setup(state, v);

	dim = g->size;
	grad = malloc(dim * sizeof(double));

	(*state->gradient)(state->closure_data, s, dim, grad, state->e);

	assign_gsl_vector(g, dim, grad);

	free(grad);
	spline_destroy(s);
}

static void min_fdf(const gsl_vector *v, void *params, double *f, gsl_vector *g)
{
	state_t *state;
	spline_t *s;
	int dim;
	double *grad;
	
	state = (state_t *)params;
	s = spline_setup(state, v);

	dim = g->size;
	grad = malloc(dim * sizeof(double));

  *f = (*state->objective)(state->closure_data, s, state->e);
	(*state->gradient)(state->closure_data, s, dim, grad, state->e);

	assign_gsl_vector(g, dim, grad);

	free(grad);
	spline_destroy(s);
}

typedef struct opt_prob opt_prob_t;
struct opt_prob
{
	double (*objective)(void *closure_data, spline_t *s, precision_t e);
	void (*gradient)(void *closure_data, spline_t *s, int dim, double *grad, precision_t e);
	int (*refine)(void *closure_data, precision_t *e);

	void *closure_data;
	minimize_options_t *opts;
};

static double optimize(opt_prob_t *prob, spline_t *s, precision_t e)
{
	minimize_options_t *opts;
  int i, j, status;

  gsl_vector *start;

	state_t state;
  gsl_multimin_function_fdf func;
  gsl_multimin_fdfminimizer *min;

  double f;
	char msg[256];

  opts = prob->opts;

  // Set up minimization problem
  start = gsl_vector_alloc(s->length);
  for(i = 0; i < s->length; i++) {
    gsl_vector_set(start, i, point_l2_norm(s->dim, spline_m(s, i)));
  }

  // Normalize spline vectors
  for(i = 0; i < s->length; i++) {
    unit_vector(s->dim, spline_m(s, i), spline_m(s, i));
  }

  // Set up function
	state.spline = s;
	state.closure_data = prob->closure_data;
	state.objective = prob->objective;
	state.gradient = prob->gradient;
	state.e = opts->start_precision;

  func.n = s->length;
  func.f = &min_objective;
	func.df = &min_gradient;
	func.fdf = &min_fdf;
  func.params = &state;

  // Set up minimizer
  min = gsl_multimin_fdfminimizer_alloc(
    gsl_multimin_fdfminimizer_steepest_descent, // TODO: tweak
    //gsl_multimin_fdfminimizer_conjugate_fr,
		//gsl_multimin_fdfminimizer_vector_bfgs2,
		s->length);
  gsl_multimin_fdfminimizer_set(min, &func, start, 1e-2, opts->tol);

  // Optimization
	i = 0;
	do {
    status = gsl_multimin_fdfminimizer_iterate(min);

		if(status) {
			if(opts->error != NULL) {
				(*opts->error)(opts, status, gsl_strerror(status));
			}
      break;
		}

    status = gsl_multimin_test_gradient(min->gradient, opts->tol);

		i++;
	} while(status == GSL_CONTINUE && i < opts->max_iter);

  // Retrieve optimum
  f = min->f;
	magnitudes_from_gsl_vector(s, min->x);

  // Cleanup
  gsl_vector_free(start);
  gsl_multimin_fdfminimizer_free(min);

  // Return result
	if(status == GSL_SUCCESS) {
    if(opts->verbose) {
      printf("Completed minimization in %d iterations\n", i);
      printf("Objective value = %2f\n\n", f);
    }

		return f;
	}
	else {
		if(opts->error != NULL) {
			snprintf(msg, 256, "Did not complete minimization: TC = %2f, i = %d", f, i);
			(*opts->error)(opts, status, msg);
		}

		return f;
	}
}

static double spline_optimize_magnitudes_sign(
	void *closure_data,
  spline_t *s)
{
	opt_prob_t *prob;
  precision_t e;
  double f, old_f, diff;

	prob = (opt_prob_t *)closure_data;

  e = prob->opts->start_precision;
  f = old_f = optimize(prob, s, e);

  while((*prob->refine)(prob->closure_data, &e)) {
    f = optimize(prob, s, e);
    diff = fabs(old_f - f);
    old_f = f;

    // Quit when we're no longer seeing improvement
    if(diff / fabs(old_f) < 1e-8) {
      return f;
    }
  }

  return f;
}

static double spline_optimize_signs(
  spline_t *s,
  double (*objective)(void *closure_data, spline_t *s), void *closure_data)
{
	double obj, start_obj, best_obj, sign;
	spline_t *spline, *best_spline;
	uint64_t signvec, j;

	start_obj = best_obj = (*objective)(closure_data, s);
	spline = spline_copy(s);
	best_spline = spline_copy(s);

	// Consider every possible sign vector
	assert(spline->length < 64);
	for(signvec = 0; signvec < (1 << (uint64_t)spline->length); signvec++) {

		// Construct a spline in which we multiply each derivative constraint
		// by the sign vector
		for(j = 0; j < (uint64_t)spline->length; j++) {
			if((signvec & (1 << j)) == 0) {
				sign = 1;
			}
			else {
				sign = -1;
			}

			point_mul(spline->dim, spline_m(spline, j), sign);
		}
		// Find best magnitude in this direction
		obj = (*objective)(closure_data, spline);

		if(obj == obj && obj < best_obj) {
			assign_derivatives(best_spline, spline);
			best_obj = obj;
		}

		// Reset magnitudes
		assign_derivatives(spline, s);
	}

	// Choose this spline if it improves on the original
	if(best_obj < start_obj) {
		assign_derivatives(s, best_spline);
	}
  else {
    best_obj = start_obj;
  }

	spline_destroy(spline);
	spline_destroy(best_spline);

  return best_obj;
}

double spline_optimize_magnitudes(
  spline_t *s,
	double (*objective)(void *closure_data, spline_t *s, precision_t e),
	void (*gradient)(void *closure_data, spline_t *s, int dim, double *grad, precision_t e),
	int (*refine)(void *closure_data, precision_t *e),
	void *closure_data,
	minimize_options_t *opts)
{
	struct timeval timer;
	double obj;
	opt_prob_t prob;

	timer_start(&timer);

	prob.objective = objective;
	prob.gradient = gradient;
	prob.refine = refine;

	prob.closure_data = closure_data;
	prob.opts = opts;

	obj = spline_optimize_signs(s, &spline_optimize_magnitudes_sign, &prob);
	
	if(opts->verbose) {
		printf("Time: ");
		time_print(stdout, timer_interval(&timer));
		printf("\n");
	}

  return obj;
}
