#include "spline.h"
#include "spline_segment.h"
#include "util.h"
#include "cross_product.h"
#include "integral.h"


////////////////// Constructor/Destructors //////////////////

spline_t *spline_create(unsigned int dim, unsigned int length)
{
  size_t s;
  intptr_t p;
  spline_t *spline;

  // Allocate
  s = length * (2 * dim) * sizeof(coord_t) + sizeof(spline_t);
  spline = (spline_t *)malloc(s);

  // Fill structure
  spline->dim = dim;
  spline->length = length;

  p = (intptr_t)spline;
  p += sizeof(spline_t);

  spline->data = (coord_t *)p;

  return spline;
}

spline_t *spline_copy(const spline_t *s)
{
  spline_t *copy;
  int i;

  copy = spline_create(s->dim, s->length);
  memcpy(copy->data, s->data, 2 * s->dim * s->length * sizeof(coord_t));

  return copy;
}


void spline_destroy(spline_t *spline)
{
  free(spline);
}

///////////////////// Finding Anchor Points ////////////////////

void spline_anchor(spline_t *s, int i, point_t p, point_t m)
{
  point_assign(s->dim, p, spline_p(s, i));
  point_assign(s->dim, m, spline_m(s, i));
}

void spline_segment_get(segment_t *seg, spline_t *s, int i)
{
	assert(i >= 0 && i < s->length - 1);

  seg->dim = s->dim;

  seg->p0 = spline_p(s, i);
  seg->m0 = spline_m(s, i);

  seg->p1 = spline_p(s, i+1);
  seg->m1 = spline_m(s, i+1);
}

void spline_segment_get_backwards(segment_t *seg, spline_t *s, int i, point_t m0, point_t m1)
{
	assert(i >= 1 && i < s->length);
  
  seg->dim = s->dim;

  seg->p0 = spline_p(s, i);
  seg->p1 = spline_p(s, i-1);

  seg->m0 = m0;
  point_assign(seg->dim, m0, spline_m(s, i));
  point_mul(seg->dim, m0, -1);

  seg->m1 = m1;
  point_assign(seg->dim, m1, spline_m(s, i-1));
  point_mul(seg->dim, m1, -1);
}

static void spline_anchors(
  spline_t *s, param_t *t,
	segment_t *segment)
{
  int i;

  // Find segment
  i = (int)(*t);
  if(*t != 0) { // TODO: huh?
    if(*t == floor(*t)) {
      *t = 1;
    }
    else {
      *t = *t - floor(*t);
    }
  }

  assert(i <= s->length);

  if(i == s->length-1) {
    i--;
  }

  // Fill anchor points
  spline_segment_get(segment, s, i);
}

//////////////////////////// C(t) /////////////////////

void spline_segment_point(
	segment_t *seg,
	param_t t, point_t p_t)
{
  coord_t h00, h10, h01, h11;
  int i;
	
  h00 = (1 + 2*t) * square(1 - t);
  h10 = t * square(1-t);
  h01 = square(t) * (3 - 2*t);
  h11 = square(t) * (t - 1);

  //h10 *= 3;
  //h11 *= 3;

  for(i = 0; i < seg->dim; i++) {
    p_t[i] = h00 * seg->p0[i]
           + h10 * seg->m0[i]
           + h01 * seg->p1[i]
           + h11 * seg->m1[i];
  }
}

void spline_point(spline_t *s, param_t t, point_t p_t)
{
	segment_t seg;

  spline_anchors(s, &t, &seg);
	spline_segment_point(&seg, t, p_t);
}

//////////////////////////// C'(t) /////////////////////

void spline_segment_derivative(
	segment_t *seg,
	param_t t, point_t dp_t)
{
  coord_t dh00, dh10, dh01, dh11;
  int i;

  dh00 = 2*((1-t) - (1 + 2*t)) * (1-t);
  dh10 = (1 - 3*t)*(1-t);
  dh01 = 2*t*((3-2*t) - t);
  dh11 = t*(3*t - 2);

  //dh10 *= 3;
  //dh11 *= 3;

  for(i = 0; i < seg->dim; i++) {
    dp_t[i] = dh00 * seg->p0[i]
            + dh10 * seg->m0[i]
            + dh01 * seg->p1[i]
            + dh11 * seg->m1[i];
  }
}

void spline_derivative(spline_t *s, param_t t, point_t dp_t)
{
	segment_t seg;

  spline_anchors(s, &t, &seg);
	spline_segment_derivative(&seg, t, dp_t);
}

//////////////////////////// C''(t) /////////////////////

void spline_segment_derivative2(
	segment_t *seg,
	param_t t, point_t ddp_t)
{
  coord_t ddh00, ddh10, ddh01, ddh11;
  int i;

  ddh00 = 12*t - 6;
  ddh10 = 6*t - 4;
  ddh01 = -12*t + 6;
  ddh11 = 6*t - 2;

  //ddh10 *= 3;
  //ddh11 *= 3;

  for(i = 0; i < seg->dim; i++) {
    ddp_t[i] = ddh00 * seg->p0[i]
             + ddh10 * seg->m0[i]
             + ddh01 * seg->p1[i]
             + ddh11 * seg->m1[i];
  }
}

void spline_derivative2(spline_t *s, param_t t, point_t ddp_t)
{
	segment_t seg;

  spline_anchors(s, &t, &seg);
	spline_segment_derivative2(&seg, t, ddp_t);
}

//////////////////////////// \kappa(t) /////////////////////

coord_t spline_derivatives_curvature(int dim, point_t dp, point_t ddp)
{
  coord_t numerator, denominator, squared_speed;
  int i;

  // Curvature numerator
  if(dim == 2) {
    numerator = fabs(dp[0] * ddp[1] - dp[1]*ddp[0]);
  }
  else {
		numerator = crossprod3d_norm(dp, ddp);
  }

  // Curvature denominator
  squared_speed = 0;
  for(i = 0; i < dim; i++) {
    squared_speed += square(dp[i]);
  }

  denominator = pow(squared_speed, 1.5);

	return numerator / denominator;
}

curvature_t spline_segment_curvature(segment_t *seg, param_t t)
{
  int dim = seg->dim;
  coord_t dp[dim], ddp[dim];

  spline_segment_derivative(seg, t, dp);
  spline_segment_derivative2(seg, t, ddp);

	return spline_derivatives_curvature(dim, dp, ddp);
}

curvature_t spline_curvature(spline_t *s, param_t t)
{
	segment_t seg;

  spline_anchors(s, &t, &seg);
	return spline_segment_curvature(&seg, t);
}

//////////////////////// Curvature ////////////////////////
speed_t spline_speed(spline_t *s, param_t t)
{
  int dim = s->dim;
  coord_t dp[3];

  spline_derivative(s, t, dp);
  return (speed_t)point_l2_norm(dim, dp);
}

//////////////////////// Calculate ////////////////////////

void spline_calculate(spline_t *s, int N, coord_t *path)
{
  coord_t *p;
  param_t t;
  int i, gaps, dim;

  dim = s->dim;

  for(i = 0; i < N; i++) {
    p = path + i*dim;
    t = (s->length-1) * i / (param_t)(N-1);

    spline_point(s, t, p);
  }
}

path_t *spline_path(spline_t *s, int N)
{
  point_t p;
  int i, n;
  param_t t;
  path_t *path;

  path = path_create(s->dim, N);

  for(i = 0; i < N; i++) {
    p = path_get(path, i);
    t = (s->length-1) * i / (param_t)(N-1);

    spline_point(s, t, p);
  }

  return path;
}

//////////////////////// Total Curvature ////////////////////////

static double squared_curvature_closure(void *spline, param_t t)
{
	return square(spline_curvature((spline_t *)spline, t));
}

curvature_t spline_total_curvature(spline_t *s, double epsilon, int max_evals)
{
  double I;

  I = integrate_adaptive_simpsons(
    &squared_curvature_closure,
    0, s->length-1,
    epsilon, max_evals, s);

	return (curvature_t)I;
}

static double spline_speed_closure(void *p, param_t t)
{
  spline_t *s = (spline_t *)p;
	coord_t dp[s->dim];

	spline_derivative(s, t, dp);
	return (double)point_l2_norm(s->dim, dp);
}

length_t spline_arc_length(spline_t *spline, double epsilon, int max_evals)
{
  double I;

  I = integrate_adaptive_simpsons(
    &spline_speed_closure,
    0, spline->length-1,
    epsilon, max_evals,
    spline);

  return (length_t)I;
}

//////////////////////////// I/O ///////////////////////

void spline_print(spline_t *s)
{
	int i;

	for(i = 0; i < s->length; i++) {
		printf("p_%d = ", i);
		point_printf(s->dim, spline_p(s, i));

		printf("  m_%d = ", i);
		point_printf(s->dim, spline_m(s, i));

		fputc('\n', stdout);
	}
}
