#include "trk.h"
#include "tract_file_struct.h"

#include "util.h"
#include "io.h"

// TODO: issues of endianness

struct trk
{
	tract_file_t f;
  trk_hdr_t hdr;

	int count_stale;
};

static path_t *read_tract(tract_file_t *f);
static int write_tract(tract_file_t *f, path_t *tract);
static int flush(tract_file_t *f);

static const tract_file_vtable_t trk_vtable = {
  &read_tract,
  &write_tract,

  &flush,
  NULL,
  NULL
};

tract_file_t *trk_open_read(const char *filename)
{
	trk_t *trk;
	FILE *fp;
	size_t n;

	// Allocate
	trk = (trk_t *)malloc(sizeof(trk_t));
	if(trk == NULL) {
		return NULL;
	}

	// Open file
	fp = fopen(filename, "rb");
	if(fp == NULL) {
		free(trk);
		return NULL;
	}

	// Read header
  n = fread(&trk->hdr, sizeof(trk_hdr_t), 1, fp);
	if(n != 1) {
		fclose(fp);
		free(trk);
		return NULL;
	}
	trk->count_stale = 0;

	// Initialize parent
	tract_file_init(&trk->f, &trk_vtable, fp);
	trk->f.count = trk->hdr.n_count;

	return &trk->f;
}

tract_file_t *trk_open_write(const char *filename, trk_hdr_t *hdr)
{
	trk_t *trk;
	FILE *fp;
	size_t n;

	// Allocate
	trk = (trk_t *)malloc(sizeof(trk_t));
	if(trk == NULL) {
		return NULL;
	}

	// Open file
	fp = fopen(filename, "wb+");
	if(fp == NULL) {
		free(trk);
		return NULL;
	}

	// Write header
	trk->hdr = *hdr;
	trk->hdr.n_count = 0;

  n = fwrite(hdr, sizeof(trk_hdr_t), 1, fp);
	if(n != 1) {
		fclose(fp);
		free(trk);
		return NULL;
	}
	trk->count_stale = 0;

	// Initialize parent
	tract_file_init(&trk->f, &trk_vtable, fp);
	trk->f.count = 0;


	return &trk->f;
}

//////////////////////////// I/O //////////////////////////

static path_t *read_tract(tract_file_t *f)
{
  trk_t *trk;
	FILE *fp;
  int n, m, d, i, j;
  point_t p;
	float q[3];
	path_t *tract;

	trk = (trk_t *)f;
	fp = f->fp;

  // Get sizes
  n = read_int32(fp);

	// Allocate
	tract = path_create(3, n);

	// Read points
	for(i = 0; i < n; i++) {
		p = path_get(tract, i);
		read_floats(fp, 3, q);

		for(j = 0; j < 3; j++) {
			p[j] = q[j];
		}
	}

  return tract;
}


static int write_tract(tract_file_t *f, path_t *tract)
{
	trk_t *trk;
  int i, j, d, n, m;
	point_t p;
  float q[3];
	FILE *fp;

	trk = (trk_t *)f;
	fp = f->fp;

  n = path_length(tract);

  // Write length
  fwrite(&n, sizeof(int), 1, fp);

  // Write points
  for(i = 0; i < n; i++) {
    p = path_get(tract, i);
		for(j = 0; j < 3; j++) {
			q[j] = p[j];
		}
    fwrite(q, sizeof(float), 3, fp);
  }

  // Check write success
  // TODO: upon failure?
  if(ferror(fp)) {
    return 1;
  }

  // Add to our internal tract count
  trk->hdr.n_count++;
	trk->count_stale = 1;

	//printf("%10d %3d\n", trk->hdr.n_count, n);

  return 0;
}

static int flush(tract_file_t *f)
{
	trk_t *trk;
	FILE *fp;
	off_t old_offset;
	int n;
	intptr_t p;
	int s;

	trk = (trk_t *)f;
	fp = f->fp;

	if(!trk->count_stale) {
		return 0;
	}

	// Update header
	s = tract_file_count(f, &trk->hdr.n_count);
	if(s != 0) {
		return s;
	}

	// Seek to n_count value
	old_offset = ftello(fp);

	p = (intptr_t)&trk->hdr.n_count - (intptr_t)&trk->hdr;
	n = fseek(fp, p, SEEK_SET);
	if(n != 0) {
		return 1;
	}

	// Write new n_count value
	n = fwrite(&trk->hdr.n_count, sizeof(trk->hdr.n_count), 1, fp);
	if(n != 1) {
		return 1;
	}

	// Restore old offset
	n = fseeko(fp, old_offset, SEEK_SET);
	if(n != 0) {
		return 1;
	}

	return 0;
}


///////////////////////// header ///////////////////////

void trk_hdr_init(trk_hdr_t *hdr)
{
	int i;

	strcpy(hdr->id_string,"TRACK");

	for(i = 0; i < 3; i++) {
		hdr->dim[i] = 0;
		hdr->voxel_size[i] = 1.0;
		hdr->origin[i] = 0;
	}

	hdr->n_scalars=0;
	hdr->n_properties=0;

	strcpy(hdr->voxel_order,"RPI");
	strcpy(hdr->pad2,"RAS");

	hdr->image_orientation_patient[0]=1;
	hdr->image_orientation_patient[1]=0;
	hdr->image_orientation_patient[2]=0;
	hdr->image_orientation_patient[3]=0;
	hdr->image_orientation_patient[4]=1;
	hdr->image_orientation_patient[5]=0;

	hdr->invert_x=0;
	hdr->invert_y=0;
	hdr->invert_z=0;

	hdr->swap_xy=0;
	hdr->swap_yz=0;
	hdr->swap_zx=0;

	hdr->n_count=0;

	hdr->version=1;
	hdr->hdr_size=1000;

	assert(sizeof(trk_hdr_t) == 1000);
}

trk_hdr_t *trk_get_hdr(tract_file_t *f)
{
	trk_t *trk;

	// Dynamic cast
	if(f->vtable != &trk_vtable) {
		return NULL;
	}
	trk = (trk_t *)f;

  return &trk->hdr;
}


static const char * const TRK_HDR_FORMAT = "\
Data dims: %d %d %d\n\
Voxel dims: %f %f %f\n\
";

void trk_hdr_print(FILE *fp, trk_hdr_t *hdr)
{
	fprintf(fp, TRK_HDR_FORMAT,
					(int)hdr->dim[0], (int)hdr->dim[1], (int)hdr->dim[2],
					(float)hdr->voxel_size[0],
					(float)hdr->voxel_size[1],
					(float)hdr->voxel_size[2]);
}
