#include "affine.h"
#include "util.h"

#include <gsl/gsl_linalg.h>
#include <f2c.h>

typedef doublereal LU_elt_t;
typedef integer LU_perm_t;

struct affine
{
  int dim;
  int LU_updated;
};

static inline coord_t *affine_data(affine_t *A)
{
  intptr_t p;

  p = (intptr_t)A;
  p += sizeof(affine_t);

  return (coord_t *)p;
}

static inline LU_elt_t *affine_LU(affine_t *A)
{
  coord_t *data;
  intptr_t p;
  
  data = affine_data(A);
  p = (intptr_t)data;
  p += sizeof(coord_t) * isquare(A->dim + 1);

  return (LU_elt_t *)p;
}

static inline LU_perm_t *affine_LU_perm(affine_t *A)
{
  LU_elt_t *LU;
  intptr_t p;

  LU = affine_LU(A);
  p = (intptr_t)LU;
  p += sizeof(LU_elt_t) * isquare(A->dim + 1);

  return (LU_perm_t *)p;
}

size_t affine_size(int dim)
{
  return sizeof(affine_t)
    + sizeof(coord_t) * isquare(dim + 1)
    + sizeof(LU_elt_t) * isquare(dim + 1)
    + sizeof(LU_perm_t) * (dim + 1);
}

static inline void affine_invalidate_LU(affine_t *A)
{
  A->LU_updated = 0;
}

//////////////////////// Constructor/destructor ////////////////////////
affine_t *affine_create(int dim)
{
  void *p;

  p = malloc(affine_size(dim));
  if(p == NULL) {
    return NULL;
  }

  return affine_create_at(dim, p);
}

affine_t *affine_create_at(int dim, void *location)
{
  affine_t *A;

  A = (affine_t *)location;
  A->dim = dim;
  A->LU_updated = 0;

  return A;
}

affine_t *affine_copy(affine_t *src)
{
  affine_t *dst;
  
  dst = affine_create(src->dim);
  affine_assign(dst, src);

  return dst;
}

void affine_destroy(affine_t *A)
{
  affine_invalidate_LU(A);
  free(A);
}

void affine_assign(affine_t *dst, affine_t *src)
{
  assert(src->dim == dst->dim);

  affine_invalidate_LU(dst);
  memcpy(dst, src, affine_size(dst->dim));
}


//////////////////////// Accesssors ////////////////////////
int affine_dim(affine_t *A)
{
  return A->dim;
}

static inline coord_t *affine_matrix(affine_t *A, int i, int j)
{
  return affine_data(A) + (A->dim + 1) * i + j;
}

coord_t affine_matrix_get(affine_t *A, int i, int j)
{
  return *affine_matrix(A, i, j);
}

void affine_matrix_set(affine_t *A, int i, int j, coord_t v)
{
  *affine_matrix(A, i, j) = v;
  affine_invalidate_LU(A);
}

gsl_matrix *affine_as_gsl_matrix(affine_t *A)
{
  gsl_matrix *G;
	int i, j, n;
	
	n = A->dim + 1;
  G = gsl_matrix_alloc(n, n);

	for(i = 0; i < n; i++) {
		for(j = 0; j < n; j++) {
			gsl_matrix_set(G, i, j, affine_matrix_get(A, i, j));
		}
	}
	
	return G;
}

//////////////////////// Standard Transforms ////////////////////////

void affine_identity(affine_t *A)
{
  int i, j, dim;

  dim = A->dim;
  for(i = 0; i < dim + 1; i++) {
    for(j = 0; j < dim + 1; j++) {
      *affine_matrix(A, i, j) = (i == j ? 1.0 : 0.0);
    }
  }

  affine_invalidate_LU(A);
}

void affine_rotation2(affine_t *A, angle_t angle)
{
  int i;
  coord_t z[3], c, s;
  int dim;
  coord_t *M;

  dim = A->dim;
  if(dim == 2) {
    c = cos(angle);
    s = sin(angle);

    M = affine_data(A);

    M[0] = c;
    M[1] = -s;
    M[2] = 0;

    M[3] = s;
    M[4] = c;
    M[5] = 0;

    M[6] = 0;
    M[7] = 0;
    M[8] = 1;
  }
  else {
    assert(dim == 3);

    z[0] = z[1] = 0;
    z[2] = 1;

    affine_rotation3(A, angle, z);
  }

  affine_invalidate_LU(A);
}

void affine_rotation3(affine_t *A, angle_t angle, point_t axis)
{
  length_t mag;
  coord_t x, y, z, s, c;
  coord_t *M;
  
  assert(A->dim == 3);

  // Trig
  s = sin(angle);
  c = cos(angle);

  // Grab normalized vector
  mag = point_l2_norm(3, axis);
  x = axis[0] / mag;
  y = axis[1] / mag;
  z = axis[2] / mag;
  
  // Fill rotation matrix
  M = affine_data(A);

  M[0]  = x*x*(1-c) + c;
  M[1]  = y*x*(1-c) + z*s;
  M[2]  = x*z*(1-c) - y*s;
  M[3]  = 0;

  M[4]  = x*y*(1-c) - z*s;
  M[5]  = y*y*(1-c) + c;
  M[6]  = y*z*(1-c) + x*s;
  M[7]  = 0;

  M[8]  = x*z*(1-c) + y*s;
  M[9]  = y*z*(1-c) - x*s;
  M[10] = z*z*(1-c) + c;
  M[11] = 0;

  M[12] = 0;
  M[13] = 0;
  M[14] = 0;
  M[15] = 1;

  affine_invalidate_LU(A);
}

void affine_translation(affine_t *A, point_t translation)
{
  int i, dim;

  affine_identity(A);

  dim = A->dim;
  for(i = 0; i < dim; i++) {
    *affine_matrix(A, i, dim) = translation[i];
  }

  affine_invalidate_LU(A);
}

void affine_scale(affine_t *A, point_t scale)
{
  int i, dim;

  affine_identity(A);

  dim = A->dim;
  for(i = 0; i < dim; i++) {
    *affine_matrix(A, i, i) = scale[i];
  }

  affine_invalidate_LU(A);
}

//////////////////////// Transforming ////////////////////////

void affine_transform(affine_t *A, point_t pdst, point_t psrc)
{
  int i, j, k, dim = A->dim;
  coord_t *M, dst[dim+1], src[dim+1];

  M = affine_data(A);

  // Move to homogenous coordinates
  point_assign(dim, src, psrc);
  src[dim] = 1;

  // Perform multiplication
  k = 0;
  for(i = 0; i < dim + 1; i++) {
    dst[i] = 0;

    for(j = 0; j < dim + 1; j++) {
      dst[i] += M[k++] * src[j];
    }
  }
 
  // Move from homogenous coordinates
  point_assign(dim, pdst, dst);
  for(i = 0; i < dim; i++) {
    pdst[i] /= dst[dim];
  }
}

void affine_compose(affine_t *dst, affine_t *A, affine_t *B)
{
  int i, j, k, dim;
  coord_t S;
 
  dim = dst->dim;
  assert(A->dim == dim && B->dim == dim);

  for(i = 0; i < dim + 1; i++) {
    for(j = 0; j < dim + 1; j++) {
      S = 0;
      
      for(k = 0; k < dim + 1; k++) {
        S += (*affine_matrix(A, i, k)) * (*affine_matrix(B, k, j));
      }

      *affine_matrix(dst, i, j) = S;
    }
  }

  affine_invalidate_LU(dst);
}

void affine_inverse_transform_gsl(affine_t *A, point_t dst, point_t src)
{
	// TODO: may want to speed up
	int i, j;

	gsl_matrix *G;
	gsl_vector *b, *x;
	gsl_permutation *p;
	int signum[A->dim + 1];

	G = affine_as_gsl_matrix(A);
	b = gsl_vector_alloc(A->dim + 1);
	x = gsl_vector_alloc(A->dim + 1);
	p = gsl_permutation_alloc(A->dim + 1);

	gsl_linalg_LU_decomp(G, p, signum);

	for(i = 0; i < A->dim; i++) {
		gsl_vector_set(b, i, src[i]);
	}
	gsl_vector_set(b, A->dim, 1);

	gsl_linalg_LU_solve(G, p, b, x);

	for(i = 0; i < A->dim; i++) {
		dst[i] = gsl_vector_get(x, i) / gsl_vector_get(x, A->dim);
	}

	gsl_vector_free(x);
	gsl_vector_free(b);
	gsl_matrix_free(G);
	gsl_permutation_free(p);
}  

// LAPACK routines
extern int dgetrf_(
  integer *m, integer *n,
  doublereal *A, integer *ldA,
  integer *piv, integer *info);

extern int dgetrs_(
  char *trans,
  integer *n, integer *nrhs,
  doublereal *A, integer *lda,
  integer *ipiv,
  doublereal *b, integer *ldb,
  integer *info);

extern int dgesv_(
  integer *n, integer *nrhs,
	doublereal *A, integer *lda,
	integer *ipiv,
	doublereal *b, integer *ldb,
	integer *info);


void affine_inverse_transform(affine_t *A, point_t dst, point_t src)
{
  int i, j, dim;
  coord_t *data;
	LU_elt_t *LU;
  LU_perm_t *perm;
  integer N, nrhs, info;
  LU_elt_t x[4];
  char trans;

  // Grab data from A
  dim = A->dim;
  assert(dim == 2 || dim == 3);
  LU = affine_LU(A);
  perm = affine_LU_perm(A);
  N = (integer)dim + 1;

  // Find LU factorization
  if(A->LU_updated == 0) {
    data = affine_data(A);

		// Calculate transpose
		for(i = 0; i < dim + 1; i++) {
			for(j = 0; j < dim + 1; j++) {
				LU[(dim + 1) * j + i] = (LU_elt_t)data[(dim + 1) * i + j];
			}
		}
   
    dgetrf_(&N, &N, LU, &N, perm, &info);

    A->LU_updated = 1;
  }

  // Copy src into homogenous coordinates
  for(i = 0; i < dim; i++) {
    x[i] = (LU_elt_t)src[i];
  }
  x[dim] = 1;

  // Solve the system from the LU factorization
  trans = 'N';
  nrhs = 1;

  dgetrs_(&trans, &N, &nrhs, LU, &N, perm, x, &N, &info);

  // Copy from homogenous coordinates to dst
  for(i = 0; i < dim; i++) {
    dst[i] = (coord_t)(x[i] / x[dim]);
  }
}

//////////////////////// I/O ////////////////////////

static const char * const DEFAULT_COORD_FMT = "% .2f";

void affine_print(affine_t *A)
{
  affine_fprint_fmt(stdout, A, DEFAULT_COORD_FMT);
}

void affine_fprint_fmt(FILE *fp, affine_t *A, const char *coord_fmt)
{
  int i, j, k, dim;
  coord_t *M;

  dim = A->dim;
  M = affine_data(A);

  k = 0;
  for(i = 0; i < dim + 1; i++) {
    fputc(i == 0 ? '[' : ' ', fp);
    fprintf(fp, coord_fmt, (float)M[k++]);

    for(j = 1; j < dim + 1; j++) {
      fputc(',', fp);
      fputc(' ', fp);
      
      fprintf(fp, coord_fmt, (float)M[k++]);
    }

    fputc(i == dim ? ']' : ';', fp);
    fputc('\n', fp);
  }

  assert(k == isquare(dim + 1));
}

//////////////////////// Composition shortcuts ////////////////////////

#define PREAMBLE                                          \
  int dim = dst->dim;                                     \
  affine_t *A, *B;                                        \
  char Adata[affine_size(dim)], Bdata[affine_size(dim)];  \
                                                          \
  A = affine_create_at(dim, Adata);                       \
  B = affine_create_at(dim, Bdata);

void affine_pre_rotate2(affine_t *dst, angle_t angle)
{
  PREAMBLE;

  affine_rotation2(A, angle);
  affine_assign(B, dst);

  affine_compose(dst, A, B);
}

void affine_post_rotate2(affine_t *dst, angle_t angle)
{
  PREAMBLE;

  affine_assign(A, dst);
  affine_rotation2(B, angle);

  affine_compose(dst, A, B);
}

void affine_pre_rotate3(affine_t *dst, angle_t angle, point_t axis)
{
  PREAMBLE;

  affine_rotation3(A, angle, axis);
  affine_assign(B, dst);

  affine_compose(dst, A, B);
}

void affine_post_rotate3(affine_t *dst, angle_t angle, point_t axis)
{
  PREAMBLE;

  affine_assign(A, dst);
  affine_rotation3(B, angle, axis);

  affine_compose(dst, A, B);
}

void affine_pre_translate(affine_t *dst, point_t translation)
{
  PREAMBLE;

  affine_translation(A, translation);
  affine_assign(B, dst);

  affine_compose(dst, A, B);
}

void affine_post_translate(affine_t *dst, point_t translation)
{
  PREAMBLE;

  affine_assign(A, dst);
  affine_translation(B, translation);

  affine_compose(dst, A, B);
}

void affine_pre_scale(affine_t *dst, point_t scale)
{
  PREAMBLE;

  affine_scale(A, scale);
  affine_assign(B, dst);

  affine_compose(dst, A, B);
}

void affine_post_scale(affine_t *dst, point_t scale)
{
  PREAMBLE;

  affine_assign(A, dst);
  affine_scale(B, scale);

  affine_compose(dst, A, B);
}
