#include "cli.h"
#include "build_info.h"

static const argdef_t helpdef = {
  'h', "help",
  ARG_MANY,
  NULL, ARG_VALUE_OPTIONAL, NULL,
  "Print this usage message and exit.",
  NULL
};

static const char * const help_args[] = {"-h", "--help", "-?", "--usage", NULL};
static int is_help_arg(char *arg)
{
  const char * const *s;

  for(s = help_args; *s != NULL; s++) {
    if(strcmp(*s, arg) == 0) {
      return 1;
    }
  }

  return 0;
}

static int print_version(const cli_t *cli)
{
  struct tm time;
  char timebuf[128];
	int i;

  // Build time
  localtime_r(&BUILD_INFO.time, &time);
  strftime(timebuf, 128, "%c", &time);
  timebuf[127] = '\0';

	// TODO: additional build information (flags and such)
  printf("%s\n", cli->name);
  printf("%s %s\n", cli->package, cli->version);
  printf("(c) %d %s\n", 1900 + time.tm_year, cli->copyright); // TODO: use real time
  printf("\n");

  printf("Build info:\n");
  printf("  Built: %s\n\n", timebuf);
  if(BUILD_INFO.svn) {
    printf("  SVN:\n");
    printf("    Repository UUID: %s\n", BUILD_INFO.svn_uuid);
    printf("    Path: %s\n", BUILD_INFO.svn_path);
    printf("    Revision: %d\n\n", BUILD_INFO.svn_revision);
  }
	if(BUILD_INFO.nopts > 0) {
		printf("  Options:\n");
		for(i = 0; i < BUILD_INFO.nopts; i++) {
			printf("    %s = %s\n", BUILD_INFO.opts[2*i], BUILD_INFO.opts[2*i+1]);
		}
	}

	exit(0);

	return 0;
}

static const argdef_t versiondef = {
  '\0', "version",
  ARG_MANY,
  NULL, ARG_VALUE_OPTIONAL, NULL,
  "Print version and package information and exit.",
  NULL
};

static int last_def(const argdef_t *def)
{
  return def->short_arg == '\0' && def->long_arg == NULL;
}

static int def_count(const argdef_t *defs)
{
  int n;

	if(defs == NULL) {
		return 0;
	}

  for(n = 0; !last_def(defs + n); n++);

  return n;
}

static int last_param(const paramdef_t *def)
{
	return def->name == NULL;
}

static int param_count(const paramdef_t *defs)
{
	int n;

	if(defs == NULL) {
		return 0;
	}

	for(n = 0; !last_param(defs + n); n++);

	return n;
}

static int shortdef(const argdef_t *defs, char arg)
{
  int i;

  for(i = 0; !last_def(defs + i); i++) {
    if(defs[i].short_arg != '\0' && defs[i].short_arg == arg) {
      return i;
    }
  }

  return -1;
}

static int fulldef(const argdef_t *defs, char *arg)
{
  int i;

  for(i = 0; !last_def(defs + i); i++) {
    if(defs[i].long_arg != NULL && strcmp(defs[i].long_arg, arg) == 0) {
      return i;
    }
  }

  return -1;
}

static int def(const argdef_t *defs, char *arg)
{
  int i;
  char brief, *full;

  if(arg[0] == '-') {
    if(arg[1] == '-') {
      return fulldef(defs, arg + 2);
    }
    else if(arg[1] != '\0' && arg[2] == '\0') {
      return shortdef(defs, arg[1]);
    }
    else {
      return -1;
    }
  }
  else {
    return -1;
  }
}


static void deferror(int argc, char *argv[])
{
  assert(argc > 0);

  fprintf(stderr, "%s: invalid argument definitions\n", argv[0]);
  exit(EX_USAGE);
}

static void call_handler(const cli_t *cli, int argc, char *argv[],
												 void *closure, int i, int flag, int j, char *arg)
{
  int s;
	const argdef_t *def;
	const paramdef_t *paramdef;

	s = 0;
	if(flag) {
		def = &cli->argdefs[j];

		if(def->handler != NULL) {
			s = (*def->handler)(closure, def, arg);
		}
		if(s != 0) {
			cli_error(cli, argc, argv,
								"handler error for option \"%s\"", argv[i]); 
		}
	}
	else {
		paramdef = &cli->paramdefs[j];

		if(paramdef->handler != NULL) {
			s = (*paramdef->handler)(closure, paramdef, arg);
		}
		if(s != 0) {
			cli_error(cli, argc, argv,
								"handler error for parameter \"%s\"", paramdef->name);
		}
	}

}

int cli_argparse(const cli_t *cli, int argc, char *argv[], void *closure)
{
  int i, j, paramstart, n, s;
  int *counts;
  char *arg;
	const argdef_t *defs;
	const paramdef_t *paramdef;
  
	defs = cli->argdefs;
  n = def_count(defs);
  counts = (int *)calloc(n, sizeof(int));

  i = 1;
	j = 0;
  while(i < argc) {
		// Handle special arguments
    if(is_help_arg(argv[i])) {
      cli_usage(stderr, cli, argc, argv);
      exit(0);
    }
		else if(strcmp("--version", argv[i]) == 0) {
			print_version(cli);
		} 

    // Grab the next argument off the top
    if(argv[i][0] == '-') {
      j = def(defs, argv[i]);
      if(j == -1) {
				cli_error(cli, argc, argv, "unrecognized argument: \"%s\"", argv[i]);
      }

      if((defs[j].flags & ARG_MANY) == 0 && counts[j] != 0) {
				if(defs[j].short_arg == '\0') {
					cli_error(cli, argc, argv,
										"option \"--%s\" appears more than once", defs[j].long_arg);
				}
				else if(defs[j].long_arg == NULL) {
					cli_error(cli, argc, argv,
										"option \"-%c\"  appears more than once", defs[j].short_arg);
				}
				else {
					cli_error(cli, argc, argv,
										"option \"-%c\" or \"--%s\" appears more than once",
									 defs[j].short_arg, defs[j].long_arg);
				}
      }

      counts[j]++;
    }
    else {
			break;
    }
    i++;

    // Grab the argument
    if(defs[j].value == NULL) {
      call_handler(cli, argc, argv, closure, i-1, 1, j, NULL);
      continue;
    }

    if(defs[j].value_count == ARG_VALUE_OPTIONAL) {
      if(i == argc || argv[i][0] == '-') {
				arg = NULL;
      }
      else {
				arg = argv[i];
				i++;
      }

      call_handler(cli, argc, argv, closure, i-2, 1, j, arg);
    }
    else if(defs[j].value_count == ARG_VALUE_ONE) {
      if(i == argc || argv[i][0] == '-') {
				cli_error(cli, argc, argv,
									"missing %s for option \"%s\"", defs[j].value, argv[i-1]);
      } 
      else {
				arg = argv[i];
				i++;

				call_handler(cli, argc, argv, closure, i-2, 1, j, arg);
      }
    }
    else {
      deferror(argc, argv);
    }
  }

  // Check required arguments are present
  for(j = 0; j < n; j++) {
		if(counts[j] == 0) {
			if((defs[j].flags & ARG_REQUIRED) != 0) {
				if(defs[j].default_value != NULL) {

					if(defs[j].handler != NULL) {
						s = (*defs[j].handler)(closure, &defs[j], defs[j].default_value);
						if(s != 0) {
							if(defs[j].long_arg == NULL) {
								cli_error(cli, argc, argv,
													"handler error for default \"-%c\"", defs[j].short_arg);
							}
							else {
								cli_error(cli, argc, argv,
													"handler error for default \"--%s\"", defs[j].long_arg);
							} 
						}
					}

				}
				else {
					if(defs[j].long_arg == NULL) {
						cli_error(cli, argc, argv,
											"missing required argument \"-%c\"", defs[j].short_arg);
					}
					else {
						cli_error(cli, argc, argv,
											"missing required argument \"--%s\"", defs[j].long_arg);
					}
				}
			}
		}
  }

  free(counts);

	paramstart = i;
	n = param_count(cli->paramdefs);

	// Read non-flag parameters
	for(; i < argc; i++) {

		// Figure out what parameter we're on
		if(n > 0 && i - paramstart < n) {
			j = i - paramstart;
		}
		else if(cli->paramdefs[n - 1].flags & ARG_MANY) {
			j = n - 1;
		}
		else {
			cli_error(cli, argc, argv, "unexpected parameter \"%s\"", argv[i]);
		}

		paramdef = cli->paramdefs + j;

		// Call the handler
		call_handler(cli, argc, argv, closure, i, 0, j, argv[i]);
	}
	
	// Check that remaining parameters are optional
	for(; i - paramstart < n; i++) {
		paramdef = cli->paramdefs + i - paramstart;

		if(paramdef->flags & ARG_REQUIRED) {
			cli_error(cli, argc, argv, "missing required parameter \"%s\"", paramdef->name);
		}
	}

	return 0;
}


static int format_args(const char *format)
{
  int n;

  n = 0;
  for(; *format != '\0'; format++) {
    if(format[0] == '%' && format[1] != '%') {
      n++;
    }
  }

  return n;
}

void cli_error(const cli_t *cli, int argc, char *argv[],
							 const char *format, ...)
{
  va_list format_args;

  assert(argc > 0);

  va_start(format_args, format);

  // Print beginning message
  fprintf(stderr, "%s: ", argv[0]);
  vfprintf(stderr, format, format_args);
  fputc('\n', stderr);

  // Print usage help
  fprintf(stderr, "Run `%s --help` for more information.\n", argv[0]);

  va_end(format_args);

  exit(EX_USAGE);
}

static const char *skipspace(const char *s)
{
  for(; isspace(*s) && *s != '\0'; s++);
  return s;
}

static int wordlen(const char *s1)
{
  const char *s2;

  for(s2 = s1; !isspace(*s2) && *s2 != '\0'; s2++);

  return s2 - s1;
}

static void print_description(FILE *f, const char *s, int tab, int width, int overflow)
{
  int i, n;
  int line;

  line = tab;

  while(*s != '\0') {
    s = skipspace(s);
    n = wordlen(s);

    // Overflow to new line if necessary
    if(line + n >= width) {
      overflow = 1;
      fputc('\n', f);
      for(i = 0; i < tab; i++) {
				fputc(' ', f);
      }
      line = tab;
    }

    // Add the word
    for(i = 0; i < n; i++) {
      fputc(s[i], f);
    }
    fputc(' ', f);
    line += n + 1;

    s += n;
  }
  fputc('\n', f);

  if(overflow) {
    fputc('\n', f);
  }
}

#define TAB_STOP (8)
#define COLUMN_STOP (24)

static void print_spaces(FILE *f, int n)
{
	int i;
 
	for(i = 0; i < n; i++) {
		fputc(' ', f);
	}
}

static void print_usage_description(FILE *f, int label_len, const char *description, int width)
{
	if(label_len >= COLUMN_STOP) {
		fputc('\n', f);
		print_spaces(f, TAB_STOP);

		print_description(f, description, TAB_STOP, width, 1);
	}
	else {
		print_spaces(f, COLUMN_STOP - label_len);

		print_description(f, description, COLUMN_STOP, width, 1);
	}
}

static void argusage(FILE *f, const argdef_t *def, int width)
{
  int n;

  // Print the argument
  if(def->short_arg == '\0') {
    n = fprintf(f, "--%s", def->long_arg);
  }
  else if(def->long_arg == NULL) {
    n = fprintf(f, "-%c", def->short_arg);
  }
  else {
    n = fprintf(f, "-%c, --%s", def->short_arg, def->long_arg);
  }

  // Print the value
  if(def->value != NULL) {
    if(def->value_count == ARG_VALUE_OPTIONAL) {
      n += fprintf(f, " [%s]", def->value);
    }
    else {
      n += fprintf(f, " %s", def->value);
    }
  }

	print_usage_description(f, n, def->description, width);
}

static void paramusage(FILE *f, const paramdef_t *def, int width)
{
	int n;

	n = fprintf(f, "%s: ", def->name);
	print_usage_description(f, n, def->description, width);
}

void cli_usage(FILE *f, const cli_t *cli, int argc, char *argv[])
{
  int width, i;
  const argdef_t *def;
	const arggroup_t *group;
	const paramdef_t *param;
	int required;

  // hmmmm... TODO
  width = 80;

  // Usage
	fprintf(f, "Usage: %s [options]", argv[0]);
	if(cli->paramdefs != NULL) {
		for(param = cli->paramdefs; param->name != NULL; param++) {
			required = param->flags & ARG_REQUIRED;

			if(required) {
				fprintf(f, " %s", param->name);
			}
			else {
				fprintf(f, " [%s]", param->name);
			}
		}
	}
	fputs("\n\n", f);

	// CLI description
	print_description(f, cli->description, 0, width, 0);

	// Parameters
  if(cli->paramdefs != NULL) {
		fputs("\nParameters:\n", f);
		for(param = cli->paramdefs; param->name != NULL; param++) {
			paramusage(f, param, width);
		}
	}

	// Flags
  fputs("\nOptions:\n", f);

  // Section for each options
  for(def = cli->argdefs, i = 0, group = cli->arggroups;
			!last_def(def);
			def++, i++) {

		// See if we're at a group start
		if(group != NULL && group->name != NULL && group->index == i) {
			fprintf(f, "\n%s:\n", group->name);
			group++;
		}

		// Print argument definition
    argusage(f, def, width);
  }

	fprintf(f, "\nSpecial options:\n");
  argusage(f, &helpdef, width);
  argusage(f, &versiondef, width);

  fputc('\n', f);
  fflush(f);
}
