#ifndef __ARGS_H__
#define __ARGS_H__

#define ARG_REQUIRED (1)
#define ARG_MANY (2)

/**
 * The number of times a value may be provided for a given argument
 */
enum argvaluecount {
  ARG_VALUE_ONE,
  ARG_VALUE_OPTIONAL
};

typedef struct argdef argdef_t;

/**
 * Defines a single flag argument.
 */
struct argdef {
  char short_arg;
  const char *long_arg;

  uint64_t flags;

  const char *value;
  enum argvaluecount value_count;
	const char *default_value;

  const char *description;

  int (*handler)(void *closure_data, const argdef_t *def, const char *argument);
};

typedef struct arggroup arggroup_t;

/**
 * Defines a group of related arguments.
 */
struct arggroup {
	/**
	 * The name of this group.
	 */
	const char *name;

	/**
	 * Index into argdefs array where group starts.
	 */
	int index;
};

typedef struct paramdef paramdef_t;

/**
 * Defines a single non-flag parameter.
 */
struct paramdef {
	/**
	 * A short name for usage.
	 */
	const char *name;

	/**
	 * A long description for help text.
	 */
	const char *description;

	/**
	 * Flags giving frequency information.
	 */
  uint64_t flags;

	/**
	 * Function called upon reading this parameter from argv.
	 */
	int (*handler)(void *closure_data, const paramdef_t *def, const char *argument);
};

typedef struct cli cli_t;

/**
 * Defines a command-line interface.
 */
struct cli {
	/**
	 * List of flag arguments.  Terminated with an argdef with all zeroes.
	 */
	const argdef_t *argdefs;

	/**
	 * List of groups of flag arguments.  Terminated with a group of all zeroes.
	 *
	 * Groups should be ordered by start index.
	 *
	 */
	const arggroup_t *arggroups;

	/**
	 * List of parameters.
	 */
	const paramdef_t *paramdefs;

	/**
	 * Name of the program.  Printed in usage.
	 */
	const char *name;
	
	/**
	 * Short description of the command-line utility.  Printed in usage.
	 */
	const char *description;

	/**
	 * Name of the package to which this program belongs.
	 */
	const char *package;

	/**
	 * Version string.
	 */
	const char *version;

	/**
	 * Copyright string.
	 */
	const char *copyright;
};

void cli_error(const cli_t *cli, int argc, char *argv[],
							 const char *format, ...);

void cli_usage(FILE *stream, const cli_t *cli, int argc, char *argv[]);

int cli_argparse(const cli_t *cli, int argc, char *argv[], void *closure_data);

#endif // #ifndef __ARGS_H__
