#include "integral.h"

// Adaptive Simpson's
// Bradie page 511

static double adapt(
  double I_a_b, double f_a, double f_mid, double f_b,
  double a, double b,
  double (*f)(void *user_data, double x), void *user_data,
  double epsilon, int *evals, unsigned int max_evals)
{
	double mid, f_a_mid, f_mid_b;
	double I_a_mid, I_mid_b, I_a_b_2;

	// Find higher-order simpson's estimate
	mid = (a + b)/2;
	f_a_mid = (*f)(user_data, (a + mid) / 2);
  f_mid_b = (*f)(user_data, (mid + b) / 2);
	(*evals) += 2;

  I_a_mid = (mid - a) * (f_a + 4 * f_a_mid + f_mid)/6;
	I_mid_b = (b - mid) * (f_mid + 4 * f_mid_b + f_b)/6;

	I_a_b_2 = I_a_mid + I_mid_b;

	// If necessary, divide region and recalculate for smaller
	if(fabs(I_a_b_2 - I_a_b) > epsilon && max_evals >= 6) {
		I_a_mid = adapt(I_a_mid, f_a, f_a_mid, f_mid, a, mid, f, user_data, epsilon/2, evals, (max_evals - 2)/2);
		I_mid_b = adapt(I_mid_b, f_mid, f_mid_b, f_b, mid, b, f, user_data, epsilon/2, evals, (max_evals - 2)/2);

		I_a_b_2 = I_a_mid + I_mid_b;
	}

	return I_a_b_2;
}

double integrate_adaptive_simpsons(
  double (*f)(void *user_data, double x),
	double a, double b,
	double epsilon, unsigned int max_evals,
	void *user_data)
{
	double mid;
	double f_a, f_mid, f_b, I_a_b;
	int evals;

	f_a = (*f)(user_data, a);
	f_mid = (*f)(user_data, (a + b) / 2);
	f_b = (*f)(user_data, b);

	evals = 3;

	I_a_b = (b - a) * (f_a + 4*f_mid + f_b)/6;

	I_a_b = adapt(I_a_b, f_a, f_mid, f_b, a, b, f, user_data, epsilon, &evals, max_evals);

	//printf("epsilon = %2g, evals = %d\n", epsilon, evals);

	return I_a_b;
}

double integrate_composite_simpsons(
  double (*f)(void *user_data, double x),
	double a, double b,
	int N,
	void *user_data)
{
	double I, even, odd, h;
	int i;

	h = (b - a) / N;

	// Calculate integral by composite Simpson's rule

	// Even sum
	even = 0;
	for(i = 0; i < N/2 - 1; i++) {
		even += (*f)(user_data, a + (2*i) * h);
	}

	// Odd sum
	odd = 0;
	for(i = 0; i < N/2; i++) {
		odd += (*f)(user_data, a + (2*i - 1) * h);
	}

	// Endpoints
	I = (*f)(user_data, a) + 2*even + 4*odd + (*f)(user_data, b);
	I *= h/3;

	return I;
}
