#include "kd_tree.h"
#include "util.h"

////////////////////// DEBUG CODE //////////////////////

#define NDEBUG_MALLOC
#define NDEBUG_KD_TREE

#ifndef NDEBUG_MALLOC
int kd_tree_count = 0;
int kd_tree_entry_count = 0;
int kd_tree_point_count = 0;
#endif

#ifdef NDEBUG
#define NDEBUG_KD_TREE
#endif

#ifdef NDEBUG_KD_TREE

#define VERIFY_TREE(t)
#define verify_entry_membership(e) 1

#else

#define VERIFY_TREE2(t, file, line) verify_whole_tree(t, file, line)
#define VERIFY_TREE(t) VERIFY_TREE2(t, __FILE__, __LINE__)

#define TREE_ASSERT_FORMAT "\
error: assert(%s) at %s:%d\n\
       (in verify_tree at %s:%d)\n"

#define tree_assert2(expr, vfile, vline)                                \
  if(!(expr)) {                                                         \
    fprintf(stderr, TREE_ASSERT_FORMAT,                                 \
             #expr, file, line, vfile, vline);                          \
    exit(1);                                                            \
  }

#define tree_assert(expr) tree_assert2(expr, __FILE__, __LINE__)

static void verify_tree(spatial_tree_t *t, const char *file, int line)
{
  size_t bin_size;
  spatial_tree_entry_t *e;
  size_t subtree_count, s;

  tree_assert(t != NULL);

  // Check subtree count
  subtree_count = 0;
  for(s = 0; s < 2; s++) {
    if(t->subtrees[s] != NULL) {
      subtree_count++;
    }
  }
  tree_assert(t->subtree_count == subtree_count);

  /*
  printf("t = 0x%x\n", t);
  printf("  t->subtree_count = %d\n", t->subtree_count);
  printf("  t->bin_size = %d\n", t->bin_size);
  printf("  t->split_dim = %d\n", t->split_dim);
  printf("  t->split = %2f\n", t->split);
  printf("\n");
  */

  if(subtree_count != 0) {
    // Split node

    // Bin should be empty
    tree_assert(t->bin_size == 0);
    tree_assert(t->bin == NULL);
    
    // Ensure split dim is in the allowable range 
    tree_assert(t->split_dim >= 0);
    tree_assert(t->split_dim < t->opts->dim);

    for(s = 0; s < 2; s++) {
      if(t->subtrees[s] != NULL) {
        // Check bidirectionality of relationships
        tree_assert(t->subtrees[s]->parent == t);

        // opts should be the same throughout the tree
        tree_assert(t->subtrees[s]->opts == t->opts);

        // Continue down to children
        verify_tree(t->subtrees[s], file, line);
      }
    }
  }
  else {
    // Leaf node

    // Only a singleton tree can have an empty leaf
    if(t->parent != NULL) {
      tree_assert(t->bin_size > 0);
      tree_assert(t->bin != NULL);
    }

    // Doesn't pretend to be a split node
    tree_assert(t->split_dim == -1);

    tree_assert(t->subtrees[0] == NULL);
    tree_assert(t->subtrees[1] == NULL);

    // Check bin_size count
    e = t->bin;
    for(bin_size = t->bin_size; bin_size != 0; bin_size--) {
      tree_assert(e != NULL);
      e = e->next;
    }
    tree_assert(bin_size == 0);
    tree_assert(e == NULL);

    // Check doubly-linked list
    if(t->bin != NULL) {
      tree_assert(t->bin->prev == NULL);
      for(e = t->bin; e != NULL; e = e->next) {
        if(e->next != NULL) {
          tree_assert(e->next->prev == e);
        }
        if(e->prev != NULL) {
          tree_assert(e->prev->next == e);
        }

        tree_assert(e->t == t);
        tree_assert(e->p != NULL);
      }
    }
  }
}

static spatial_tree_t *root(spatial_tree_t *t)
{
  while(t->parent != NULL) {
    t = t->parent;
  }

  return t;
}

static void verify_whole_tree(spatial_tree_t *t, const char *file, int line)
{
	//printf("Verify tree\n");

  tree_assert(t != NULL);

  verify_tree(root(t), file, line);
}

// check that e is in e->t
static int verify_entry_membership(spatial_tree_entry_t *e)
{
  assert(e != NULL);

  spatial_tree_entry_t *e_search;

  for(e_search = e->t->bin; e_search != NULL; e_search = e_search->next) {
    if(e_search == e) {
      return 1;
    }
  }

  return 0;
}

#endif

////////////////////////// Real Algorithms ////////////////////////

spatial_tree_t *spatial_tree_create(spatial_tree_options_t *opts)
{
  assert(opts != NULL);

  spatial_tree_t *t;

  t = (spatial_tree_t *)malloc(sizeof(spatial_tree_t));
#ifndef NDEBUG_MALLOC
	assert(t != NULL);
	kd_tree_count++;
#endif

  if(t != NULL) {
    t->bin_size = 0;
    t->bin = NULL;

    t->split_dim = -1;

    t->parent = NULL;
    t->subtree_count = 0;
    t->subtrees[0] = NULL;
    t->subtrees[1] = NULL;

    t->opts = opts;
  }

  VERIFY_TREE(t);

  return t;
}

static void spatial_tree_destroy_bin(spatial_tree_t *t)
{
	spatial_tree_entry_t *n, *n_next;

	n = t->bin;
	while(n != NULL) {
		n_next = n->next;

		free(n->p);
		free(n);
#ifndef NDEBUG_MALLOC
		kd_tree_point_count--;
		kd_tree_entry_count--;
#endif

		n = n_next;
	}
}

void spatial_tree_destroy(spatial_tree_t *t)
{
	int s;

	if(t->subtrees == 0) {
		spatial_tree_destroy_bin(t);
		free(t);
	}
	else {
		for(s = 0; s < 2; s++) {
			if(t->subtrees[s] != NULL) {
				spatial_tree_destroy(t->subtrees[s]);
			}
		}
		free(t);
	}

#ifndef NDEBUG_MALLOC
		kd_tree_count--;
#endif
}

static spatial_tree_t *subtree(spatial_tree_t *t, double *p)
{
  assert(p != NULL);

  if(p[t->split_dim] < t->split) {
    return t->subtrees[0];
  }
  else {
    return t->subtrees[1];
  }
}

// Get the node in the spatial tree containing the given point
static spatial_tree_t *get(spatial_tree_t *t, double *p)
{
  spatial_tree_t *s;

  if(t->split_dim == -1) {
    return t;
  }
  else {
    s = subtree(t, p);

    if(s != NULL) {
      return get(s, p);
    }
    else {
      return t;
    }
  }
}

static void bin_add(spatial_tree_t *t, spatial_tree_entry_t *e)
{
  e->t = t;
  e->prev = NULL;
  e->next = t->bin;

  if(e->next != NULL) {
    e->next->prev = e;
  }

  t->bin_size++;
  t->bin = e;
}

inline static spatial_tree_t *new_leaf(spatial_tree_t *parent)
{
  spatial_tree_t *leaf;

  leaf = spatial_tree_create(parent->opts);
  if(leaf != NULL) {
    leaf->parent = parent;
  }

  return leaf;
}


// Splits a leaf node along the mean in the dimension of greatest variance
//   returns: the new leaf that new_p belongs in
static spatial_tree_t *split(spatial_tree_t *t, double *new_p)
{
  assert(new_p != NULL);

  size_t d, dim = t->opts->dim;

  double mean[dim];
  double var[dim];

  spatial_tree_entry_t *e, *e_next;

  size_t max_var_d;

  // Calculate statistics of bin contents
  for(d = 0; d < dim; d++) {
    mean[d] = new_p[d];
  }
  for(e = t->bin; e != NULL; e = e->next) {
    for(d = 0; d < dim; d++) {
      mean[d] += e->p[d];
    }
  }
  for(d = 0; d < dim; d++) {
    mean[d] /= (t->bin_size + 1);
  }
  
  for(e = t->bin; e != NULL; e = e->next) {
    for(d = 0; d < dim; d++) {
      var[d] += square(e->p[d] - mean[d]);
    }
  }

  // Find best split
  t->split_dim = 0;
  for(d = 1; d < dim; d++) {
    if(var[t->split_dim] < var[d]) {
      t->split_dim = d;
    }
  }

  if(var[t->split_dim] != 0) {
    t->split = mean[t->split_dim];
  }
  else {
    // All the points are the same, we cannot split this tree
    return t;
  }

  // Create two leafs as subtrees
  if((t->subtrees[0] = new_leaf(t)) == NULL
     || (t->subtrees[1] = new_leaf(t)) == NULL) {

    // Split failed due to out-of-memory
    t->split = -1;
    free(t->subtrees[0]);
		t->subtrees[0] = NULL;
    return t;
  }
  t->subtree_count = 2;

  // Sort each from t into the right new leaf
  e = t->bin;
  while(e != NULL) {
    e_next = e->next;

    bin_add(subtree(t, e->p), e);

    e = e_next;
  }

  // Clear our bin
  t->bin_size = 0;
  t->bin = NULL;

  // Return the correct subtree for new_p
  return subtree(t, new_p);
}

spatial_tree_entry_t *spatial_tree_add(spatial_tree_t *t, double *p, void *data)
{
  VERIFY_TREE(t);

  spatial_tree_entry_t *e;
  size_t d, dim;

  dim = t->opts->dim;

  // Initialize a bin entry
  e = (spatial_tree_entry_t *)malloc(sizeof(spatial_tree_entry_t));
#ifndef NDEBUG_MALLOC
	assert(e != NULL);
	kd_tree_entry_count++;
#endif
  if(e == NULL) {
    return NULL;
  }

  e->p = (double *)malloc(dim * sizeof(double));
#ifndef NDEBUG_MALLOC
	assert(e->p != NULL);
	kd_tree_point_count++;
#endif
  if(e->p == NULL) {
    free(e);
    return NULL;
  }
  for(d = 0; d < dim; d++) {
    e->p[d] = p[d];
  }
  e->data = data;

  // Descend to the best-fit leaf
  t = get(t, p);

	if(t->subtree_count == 0) {
		// We're in a leaf node
		// Split if necessary
		if(t->bin_size >= t->opts->max_bin_size) {
			t = split(t, p);
		}
	}
	else {
		// We're at a split node, but the subtree we want to visit is gone
		t->subtree_count++;

		if(p[t->split_dim] < t->split) {
			assert(t->subtrees[0] == NULL);

			t = t->subtrees[0] = new_leaf(t);
		}
		else {
			assert(t->subtrees[1] == NULL);

			t = t->subtrees[1] = new_leaf(t);
		}
	}

  // Insert
  bin_add(t, e);

  VERIFY_TREE(t);
  return e;
}

static void remove_empty_leaf(spatial_tree_t *t)
{
  assert(t->subtree_count == 0);
  assert(t->bin_size == 0);

  spatial_tree_t *parent;

  parent = t->parent;

	if(parent == NULL) {
		t->split_dim = -1; // reset root node
	}
  else {
    parent->subtree_count--;
    if(parent->subtrees[0] == t) {
      parent->subtrees[0] = NULL;
    }
    else {
      parent->subtrees[1] = NULL;
    }

		free(t);
#ifndef NDEBUG_MALLOC
		kd_tree_count--;
#endif

		if(parent->subtree_count == 0) {
			remove_empty_leaf(parent);
		}
  }
}

void spatial_tree_remove(spatial_tree_entry_t *e)
{
  assert(verify_entry_membership(e));
  VERIFY_TREE(e->t);

  // Remove self from bin
  e->t->bin_size--;

  if(e->prev != NULL) {
    e->prev->next = e->next;
  }
  else {
    e->t->bin = e->next;
  }

  if(e->next != NULL) {
    e->next->prev = e->prev;
  }

  // Remove node if bin is now empty
  if(e->t->bin_size == 0) {
    remove_empty_leaf(e->t);
  }

  // Free memory
  free(e->p);
  free(e);
#ifndef NDEBUG_MALLOC
	kd_tree_point_count--;
	kd_tree_entry_count--;
#endif
}



static spatial_tree_t *any_leaf(spatial_tree_t *t)
{
  VERIFY_TREE(t);

  if(t->subtree_count == 0) {
    return t;
  }
  else if(t->subtrees[0] != NULL) {
    return any_leaf(t->subtrees[0]);
  }
  else {
    assert(t->subtrees[1] != NULL);

    return any_leaf(t->subtrees[1]);
  }
}

static double l2dist(int d, double *p, double *q)
{
	double dist;
	int i;

	dist = 0;
	for(i = 0; i < d; i++) {
		dist += square(p[i] - q[i]);
	}

	return sqrt(dist);
}

typedef struct {
	double *p;

	spatial_tree_entry_t **nn;
	double *dist;
	int k, dim;

} nn_state_t;

// Looks at a single entry, adding it as a NN if appropriate
static void nn_consider_entry(
	spatial_tree_entry_t *e, double *p,
	int dim, int k, double *dist, spatial_tree_entry_t **nn)
{
	double d;
	int i, j;

	d = l2dist(dim, e->p, p);

	for(i = 0; i < k; i++) {
		if(nn[i] == NULL || d < dist[i]) {

			// sorted insert
			for(j = k - 1; j > i; j--) {
				nn[j] = nn[j-1];
				dist[j] = dist[j-1];
			}

			nn[i] = e;
			dist[i] = d;

			break;
		}
	}
}

static void nn_consider_leaf(spatial_tree_t *t,
  double *p, int dim, int k, double *dist, spatial_tree_entry_t **nn)
{
	spatial_tree_entry_t *e;

	assert(t->subtree_count == 0);

	for(e = t->bin; e != NULL; e = e->next) {
		nn_consider_entry(e, p, dim, k, dist, nn);
	}
}

static void nn_real(spatial_tree_t *t, nn_state_t *s, spatial_tree_t *exclude)
{
	double search_dist;
	int i;
	spatial_tree_t *sub;
	int searched[2];

	if(t->subtree_count == 0) {
		nn_consider_leaf(t, s->p, s->dim, s->k, s->dist, s->nn);
	}
	else if(t->split_dim != -1) {
		search_dist = s->dist[s->k-1];

		for(i = 0; i < 2; i++) {
			sub = t->subtrees[i];

			if(sub == NULL || sub == exclude) {
				continue;
			}

			if(search_dist < 0
				 || (i == 0 && s->p[t->split_dim] - search_dist <= t->split)
         || (i == 1 && s->p[t->split_dim] + search_dist >= t->split)) {

				nn_real(sub, s, t);
			}
		}
	}

	// Pass it up to parent for wider search
	if(t->parent != NULL && t->parent != exclude) {
		nn_real(t->parent, s, t);
	}
}

void spatial_tree_nearest_neighbor(
  spatial_tree_t *t, double *p,
	int k, spatial_tree_entry_t **nn)
{
	nn_state_t s;
	double dist[k];
	int i;

	VERIFY_TREE(t);

	// Initialize state
	s.p = p;
	s.nn = nn;
	s.dist = dist;
	s.k = k;
	s.dim = t->opts->dim;

	for(i = 0; i < k; i++) {
		nn[i] = NULL;
		dist[i] = -1;
	}

	// Grab a leaf
	t = any_leaf(get(t, p));
	nn_real(t, &s, NULL);
}

void spatial_tree_neighborhood(
  spatial_tree_t *t, double *p, double distance,
	void (*add_neighbor)(void *closure_data, spatial_tree_entry_t *neighbor),
	void *closure_data)
{
	int i;
	spatial_tree_t *sub;
	spatial_tree_entry_t *e;

	VERIFY_TREE(t);

	if(t->subtree_count == 0) {
		// It's a leaf, just look at the bin

		for(e = t->bin; e != NULL; e = e->next) {
			if(l2dist(t->opts->dim, e->p, p) <= distance) {
				(*add_neighbor)(closure_data, e);
			}
		}
	}
	else if(t->split_dim != -1) {
		// It's a split node, potentially search the subtrees
		
		for(i = 0; i < 2; i++) {
			sub = t->subtrees[i];

			if(sub == NULL) {
				continue;
			}

			if((i == 0 && p[t->split_dim] - distance <= t->split)
				 || (i == 1 && p[t->split_dim] + distance >= t->split)) {

				// Recurse
				spatial_tree_neighborhood(sub, p, distance, add_neighbor, closure_data);
			}
		}
	}
}
