#include "bzip.h"
#include "internal.h"

#include <bzlib.h>

typedef struct {
  mnemosyne_t base;
	bool write;
	FILE *fp;
	BZFILE *bzp;

	int err;

} mnemosyne_bzip2_t;

static const mnemosyne_functions_t bzip2_functions;

static inline mnemosyne_bzip2_t *cast(mnemosyne_t *mn)
{
	assert(mn->v == &bzip2_functions);
	return (mnemosyne_bzip2_t *)mn;
}

mnemosyne_t *mnemosyne_open_bzip2(const char *filename, const char *mode,
																	const mnemosyne_options_t *opts)
{
	mnemosyne_bzip2_t *bmn;
	int err;

	// Initialize base class
	bmn = (mnemosyne_bzip2_t *)malloc(sizeof(mnemosyne_bzip2_t));
	bmn->base.v = &bzip2_functions;
	bmn->base.opts = *opts;

	// Initialize state
	bmn->err = BZ_OK;

	// Determine mode
	if(mode[0] == 'w' || mode[1] == 'r') {
		bmn->write = true;
	}
	else if(mode[0] == 'r') {
		bmn->write = false;
	}
	else {
		free(bmn);
		return NULL;
	}

	// Open file
	bmn->fp = fopen(filename, mode);
	if(bmn->write) {
		bmn->bzp = BZ2_bzWriteOpen(&err, bmn->fp, 9, 0, 0);
	}
	else {
		bmn->bzp = BZ2_bzReadOpen(&err, bmn->fp, 0, 0, NULL, 0);
	}

	if(err != BZ_OK) {
		fclose(bmn->fp);
		free(bmn);
		return NULL;
	}

	assert((intptr_t)&bmn->base == (intptr_t)bmn);
	return &bmn->base;
}


static int bzip2_close(mnemosyne_t *mn)
{
	mnemosyne_bzip2_t *bmn;

	bmn = cast(mn);

	if(bmn->write) {
		BZ2_bzWriteClose(&bmn->err, bmn->bzp, 0, NULL, NULL);
	}
	else {
		BZ2_bzReadClose(&bmn->err, bmn->bzp);
	}
	fclose(bmn->fp);

	free(bmn);

	return bmn->err;
}

static bool bzip2_eof(mnemosyne_t *mn)
{
	mnemosyne_bzip2_t *bmn;

	bmn = cast(mn);

	return bmn->err == BZ_STREAM_END;
}

static size_t bzip2_write(mnemosyne_t *mn, const void *buffer, size_t size, size_t nitems)
{
	mnemosyne_bzip2_t *bmn;
	int err;

	bmn = cast(mn);

	if(!bmn->write || bmn->err != BZ_OK) {
		return 0;
	}

	BZ2_bzWrite(&bmn->err, bmn->bzp, (void *)buffer, size * nitems);

	if(bmn->err != BZ_OK) {
		return 0;
	}

	return size * nitems;
}

static size_t bzip2_read(mnemosyne_t *mn, void *buffer, size_t size, size_t nitems)
{
	mnemosyne_bzip2_t *bmn;
	int n;

	bmn = cast(mn);

	if(bmn->write || bmn->err != BZ_OK) {
		return 0;
	}

	n = BZ2_bzRead(&bmn->err, bmn->bzp, buffer, size * nitems);

	return (size_t)n;
}

static const mnemosyne_functions_t bzip2_functions = {
	.close = &bzip2_close,
	.eof = &bzip2_eof,
	.write = &bzip2_write,
	.read = &bzip2_read,
	.printf = NULL
};
