#include "mnemosyne.h"
#include "internal.h"

#include "plain.h"
#include "gzip.h"
#include "bzip.h"

//////////////////////////// Extension Sleuthing /////////////////////

static const char *ext(const char *filename)
{
	const char *s;

	for(s = filename; *s != '\0'; s++);
	for(; *s != '.' && s != filename; s++);

	if(*s != '.') {
		return NULL;
	}
	s++;

	return s;
}

typedef struct {
  const char *string;
  int type;
} compression_ext_t;

static const compression_ext_t exts[] = {
  {"gz", MNEMOSYNE_COMPRESS_GZIP},
  {"bz2", MNEMOSYNE_COMPRESS_BZIP2},
  {NULL, 0}
};

int mnemosyne_compression_extension(const char *filename)
{
	const char *e;
	const compression_ext_t *f;

	e = ext(filename);
	if(e == NULL) {
		return MNEMOSYNE_COMPRESS_NONE;
	}

	for(f = exts; f->string != NULL; f++) {
		if(strcmp(f->string, e) == 0) {
			return f->type;
		}
	}

	return MNEMOSYNE_COMPRESS_NONE;
}

const char *mnemosyne_compression_suffix(const mnemosyne_options_t *opts)
{
	switch(opts->compress) {

	case MNEMOSYNE_COMPRESS_NONE:
		return "";

	case MNEMOSYNE_COMPRESS_GZIP:
		return ".gz";

	case MNEMOSYNE_COMPRESS_BZIP2:
		return ".bz2";

	default:
		return NULL;
	}
}

//////////////////////////// File handling /////////////////////
const mnemosyne_options_t mnemosyne_default_options = {
  .compress = MNEMOSYNE_COMPRESS_NONE,
};


const mnemosyne_options_t *mnemosyne_options(const mnemosyne_t *mn)
{
	return &mn->opts;
}

mnemosyne_t *mnemosyne_open(
  const char *filename,
	const char *mode,
	const mnemosyne_options_t *opts)
{
	// Open file
	switch(opts->compress) {
	case MNEMOSYNE_COMPRESS_NONE:
		return mnemosyne_open_plain(filename, mode, opts);

	case MNEMOSYNE_COMPRESS_GZIP:
		return mnemosyne_open_gzip(filename, mode, opts);

	case MNEMOSYNE_COMPRESS_BZIP2:
		return mnemosyne_open_bzip2(filename, mode, opts);

	default:
		return NULL;
	}

}

int mnemosyne_close(mnemosyne_t *mn)
{
	return (*mn->v->close)(mn);
}

bool mnemosyne_eof(mnemosyne_t *mn)
{
	return mneof(mn);
}

int mnemosyne_err(mnemosyne_t *mn)
{
	return (*mn->v->err)(mn);
}

size_t mnemosyne_write(mnemosyne_t *mn, const void *buffer, size_t size, size_t nitems)
{
	return mnwrite(mn, buffer, size, nitems);
}

size_t mnemosyne_read(mnemosyne_t *mn, void *buffer, size_t size, size_t nitems)
{
	return mnread(mn, buffer, size, nitems);
}

size_t mnemosyne_printf(mnemosyne_t *mn, const char *format, ...)
{
	va_list args;
	size_t n;
	int m;
	char *s;

	if(mn->v->printf != NULL) {
		// Pass to overriding function
		va_start(args, format);
		n = mnvprintf(mn, format, args);
		va_end(args);

		return n;
	}
  else {
		// Find formatted string length
		va_start(args, format);
		m = vsnprintf(NULL, 0, format, args);
		va_end(args);

		// Allocate
		s = (char *)malloc(m + 1);

		// Format
		va_start(args, format);
		m = vsnprintf(s, m+1, format, args);
		va_end(args);
		
		// Write
		mnemosyne_write(mn, s, 1, (size_t)m);

		// Cleanup
		free(s);

		return (size_t)m;
	}
}


//////////////////////////// Structured I/O //////////////////////////////
int16_t mnemosyne_read_int16(mnemosyne_t *mn)
{
	int16_t i;
	
	mnread(mn, &i, 2, 1);

	return i;
}

void mnemosyne_write_int16(mnemosyne_t *mn, int16_t i)
{
	mnwrite(mn, &i, 2, 1);
}

uint16_t mnemosyne_read_uint16(mnemosyne_t *mn)
{
	uint16_t i;
	
	mnread(mn, &i, 2, 1);

	return i;
}

void mnemosyne_write_uint16(mnemosyne_t *mn, uint16_t i)
{
	mnwrite(mn, &i, 2, 1);
}

int32_t mnemosyne_read_int32(mnemosyne_t *mn)
{
	int32_t i;
	
	mnread(mn, &i, 4, 1);

	return i;
}

void mnemosyne_write_int32(mnemosyne_t *mn, int32_t i)
{
	mnwrite(mn, &i, 4, 1);
}

uint32_t mnemosyne_read_uint32(mnemosyne_t *mn)
{
	uint32_t i;
	
	mnread(mn, &i, 4, 1);

	return i;
}

void mnemosyne_write_uint32(mnemosyne_t *mn, uint32_t i)
{
	mnwrite(mn, &i, 4, 1);
}

int64_t mnemosyne_read_int64(mnemosyne_t *mn)
{
	int64_t i;
	
	mnread(mn, &i, 8, 1);

	return i;
}

void mnemosyne_write_int64(mnemosyne_t *mn, int64_t i)
{
	mnwrite(mn, &i, 8, 1);
}

uint64_t mnemosyne_read_uint64(mnemosyne_t *mn)
{
	uint64_t i;
	
	mnread(mn, &i, 8, 1);

	return i;
}

void mnemosyne_write_uint64(mnemosyne_t *mn, uint64_t i)
{
	mnwrite(mn, &i, 8, 1);
}

float mnemosyne_read_float(mnemosyne_t *mn)
{
	float f;

	mnread(mn, &f, sizeof(float), 1);

	return f;
}

void mnemosyne_write_float(mnemosyne_t *mn, float f)
{
	mnwrite(mn, &f, sizeof(float), 1);
}

void mnemosyne_read_floats(mnemosyne_t *mn, size_t n, float *f)
{
	mnread(mn, f, sizeof(float), n);
}

void mnemosyne_write_floats(mnemosyne_t *mn, size_t n, const float *f)
{
	mnwrite(mn, f, sizeof(float), n);
}

double mnemosyne_read_double(mnemosyne_t *mn)
{
	double d;

	mnread(mn, &d, sizeof(double), 1);

	return d;
}

void mnemosyne_write_double(mnemosyne_t *mn, double d)
{
	mnwrite(mn, &d, sizeof(double), 1);
}

void mnemosyne_read_doubles(mnemosyne_t *mn, size_t n, double *d)
{
	mnread(mn, d, sizeof(double), n);
}

void mnemosyne_write_doubles(mnemosyne_t *mn, size_t n, const double *d)
{
	mnwrite(mn, d, sizeof(double), n);
}
