#ifndef __MNEMOSYNE_H__
#define __MNEMOSYNE_H__

typedef struct mnemosyne mnemosyne_t;

/////////////////////////////// Options ///////////////////////////////

#define MNEMOSYNE_COMPRESS_NONE (0)
#define MNEMOSYNE_COMPRESS_GZIP (1)
#define MNEMOSYNE_COMPRESS_BZIP2 (2)

typedef struct {
  int compress;

} mnemosyne_options_t;

extern const mnemosyne_options_t mnemosyne_default_options;


/**
 * Get the type of compression used from the extension.
 *
 * Recognizes the following extensions:
 *   - .gz
 *   - .bz2
 * Returns MNEMOSYNE_COMPRESS_NONE upon an unrecognized extension.
 */
int mnemosyne_compression_extension(const char *filename);

/**
 * Return the suffix used to denote the given type of compression.
 *
 * @note Returns empty string "" for no compression, NULL for
 *       unrecognized.
 */
const char *mnemosyne_compression_suffix(const mnemosyne_options_t *);

const mnemosyne_options_t *mnemosyne_options(const mnemosyne_t *);

/////////////////////////// File Management ///////////////////////////
mnemosyne_t *mnemosyne_open(const char *filename, const char *mode,
														const mnemosyne_options_t *options);
int mnemosyne_close(mnemosyne_t *);

bool mnemosyne_eof(mnemosyne_t *);
int mnemosyne_err(mnemosyne_t *);

////////////////////////////// Basic I/O //////////////////////////////
size_t mnemosyne_write(mnemosyne_t *, const void *buffer, size_t size, size_t nitems);
size_t mnemosyne_read(mnemosyne_t *, void *buffer, size_t size, size_t nitems);

size_t mnemosyne_printf(mnemosyne_t *, const char *format, ...) __attribute__ ((format (printf, 2, 3)));

//size_t mnemosyne_scanf(mnemosyne_t *, const char *format, ...) __attribute__ ((format (scanf, 2, 3)));

//////////////////////////// Structured I/O //////////////////////////////

int16_t mnemosyne_read_int16(mnemosyne_t *);
void mnemosyne_write_int16(mnemosyne_t *, int16_t i);

uint16_t mnemosyne_read_uint16(mnemosyne_t *);
void mnemosyne_write_uint16(mnemosyne_t *, uint16_t i);

int32_t mnemosyne_read_int32(mnemosyne_t *);
void mnemosyne_write_int32(mnemosyne_t *, int32_t i);

uint32_t mnemosyne_read_uint32(mnemosyne_t *);
void mnemosyne_write_uint32(mnemosyne_t *, uint32_t i);

int64_t mnemosyne_read_int64(mnemosyne_t *);
void mnemosyne_write_int64(mnemosyne_t *, int64_t i);

uint64_t mnemosyne_read_uint64(mnemosyne_t *);
void mnemosyne_write_uint64(mnemosyne_t *, uint64_t i);

float mnemosyne_read_float(mnemosyne_t *);
void mnemosyne_write_float(mnemosyne_t *, float f);

void mnemosyne_read_floats(mnemosyne_t *, size_t n, float *f);
void mnemosyne_write_floats(mnemosyne_t *, size_t n, const float *f);

double mnemosyne_read_double(mnemosyne_t *);
void mnemosyne_write_double(mnemosyne_t *, double d);

void mnemosyne_read_doubles(mnemosyne_t *, size_t n, double *d);
void mnemosyne_write_doubles(mnemosyne_t *, size_t n, const double *d);

#endif // #ifndef __MNEMOSYNE_H__
