#include "path.h"

struct path {
	int dim, length;
};

static inline point_t path_data(path_t *path)
{
	intptr_t p;

	p = (intptr_t)path;
	p += sizeof(path_t);

	return (point_t)p;
}

path_t *path_create(int dim, int length)
{
	size_t s;
	path_t *path;

	s = sizeof(path_t) + sizeof(coord_t) * dim * length;
	path = (path_t *)malloc(s);

	path->dim = dim;
	path->length = length;

	return path;
}

path_t *path_copy(path_t *src)
{
	int dim, n;
	path_t *dst;
	
	dim = src->dim;
	n = src->length;

	dst = path_create(dim, n);
	memcpy(path_data(dst), path_data(src), sizeof(coord_t) * dim * n);

	return dst;
}

void path_destroy(path_t *path)
{
	free(path);
}

path_t *path_resize(path_t *path, int new_length)
{
	size_t s;

	s = sizeof(path_t) + sizeof(coord_t) * path->dim * new_length;
	path = (path_t *)realloc(path, s);
	path->length = new_length;

	return path;
}

path_t *path_append(path_t *path, point_t p)
{
	int n, dim;

	dim = path->dim;
	n = path->length;

	// Resize
	path = path_resize(path, n + 1);

	// Append
	point_assign(dim, path_get(path, n), p);

	return path;
}

// Accessors
point_t path_get(path_t *path, int i)
{
	return path_data(path) + path->dim * i;
}

int path_length(path_t *path)
{
	return path->length;
}

int path_dim(path_t *path)
{
	return path->dim;
}

// Modifiers
void path_zero(path_t *path)
{
	int i, n;
	coord_t *data;

	data = path_data(path);
	n = path->dim * path->length;

	for(i = 0; i < n; i++) {
		data[i] = 0;
	}
}

void path_reverse(path_t *path)
{
	int i, j, n, dim;
	point_t p, q;
	coord_t swap;

	dim = path->dim;
	n = path->length;

	for(i = 0; i < n/2; i++) {
		p = path_get(path, i);
		q = path_get(path, n - 1 - i);
		
		for(j = 0; j < dim; j++) {
			swap = p[j];
			p[j] = q[j];
			q[j] = swap;
		}
	}
}

path_t *path_add(path_t *start, path_t *end)
{
	int dim, n, m;
	path_t *dst;
	
	assert(start->dim == end->dim);

	dim = start->dim;
	n = start->length;
	m = end->length;

	dst = path_create(dim, n + m);
	memcpy(path_data(dst), path_data(start), sizeof(coord_t) * dim * n);
	memcpy(path_get(dst, n), path_data(end), sizeof(coord_t) * dim * m);

	return dst;
}

// Calculations
length_t path_l2_length(path_t *path)
{
	length_t l;
	int i, dim, n;
	point_t p, q;
	
	dim = path->dim;
	n = path->length;
	l = 0;
	for(i = 0; i < n - 1; i++) {
		p = path_get(path, i);
		q = path_get(path, i+1);

		l += point_l2_dist(dim, p, q);
	}

	return l;
}

void path_tangent(path_t *path, int i, point_t d)
{
	point_t p, q;

	if(i == 0) {
		p = path_get(path, 0);
		q = path_get(path, 1);
		point_sub(path->dim, d, q, p);
	}
	else if(i == path->length - 1) {
		p = path_get(path, path->length - 2);
		q = path_get(path, path->length - 1);
		point_sub(path->dim, d, q, p);
	}
	else {
		p = path_get(path, i - 1);
		q = path_get(path, i + 1);
		point_sub(path->dim, d, q, p);
	}
}

void path_normal(path_t *path, int i, point_t dd)
{
	point_t p[3];
	int j, start;
	int dim;

	dim = path->dim;

	if(path->length < 3) {
		point_zero(dim, dd);
		return;
	}
	
	// Figure out where our difference starts
	if(i == 0) {
		start = 0;
	}
	else if(i == path->length - 1) {
		start = path->length - 3;
	}
	else {
		start = i - 1;
	}

	for(j = 0; j < 3; j++) {
		p[j] = path_get(path, start + j);
	}
	
	// Calculate derivative approximation
	for(j = 0; j < dim; j++) {
		dd[j] = p[0][j] - 2 * p[1][j] + p[2][j];
	}
}
