#include "point.h"
#include "util.h"

length_t point_l1_norm(int dim, point_t p)
{
  length_t mag;
  int d;

  mag = 0;
  for(d = 0; d < dim; d++) {
    mag += fabs(p[d]);
  }

  return mag;
}

length_t point_l1_dist(int dim, point_t p, point_t q)
{
  coord_t diff[dim];

  point_sub(dim, diff, p, q);
  return point_l1_norm(dim, diff);
}

length_t point_squared_l2_norm(int dim, point_t p)
{
  length_t mag;
  int d;

  mag = 0;
  for(d = 0; d < dim; d++) {
    mag += square(p[d]);
  }

  return mag;
}

length_t point_l2_norm(int dim, point_t p)
{
  return sqrt(point_squared_l2_norm(dim, p));
}

length_t point_l2_dist(int dim, point_t p, point_t q)
{
  coord_t diff[dim];

  point_sub(dim, diff, p, q);
  return point_l2_norm(dim, diff);
}

length_t point_linfty_norm(int dim, point_t p)
{
	length_t norm;
	int i;

	norm = 0;
	for(i = 0; i < dim; i++) {
		norm = MAX(norm, fabs(p[i]));
	}

	return norm;
}

length_t point_linfty_dist(int dim, point_t p, point_t q)
{
	coord_t diff[dim];

  point_sub(dim, diff, p, q);
  return point_linfty_norm(dim, diff);
}

coord_t point_dot_product(int dim, point_t p, point_t q)
{
  coord_t dot;
  int d;

  dot = 0;
  for(d = 0; d < dim; d++) {
    dot += p[d] * q[d];
  }

  return dot;
}

void point_projection(int dim, point_t from, point_t to, point_t proj)
{
  length_t mag;
  coord_t unit[dim];

  unit_vector(dim, to, unit);

  mag = point_dot_product(dim, from, unit);

  point_mul(dim, unit, mag);
  point_assign(dim, proj, unit);
}


void unit_vector(int dim, point_t v, point_t unit)
{
  length_t mag;
  int d;

  mag = point_l2_norm(dim, v);
  for(d = 0; d < dim; d++) {
    unit[d] = v[d] / mag;
  }
}

///////////////////// I/O /////////////////////

static const char * const DEFAULT_COORD_FMT = "% .2f";

void point_printf(int dim, point_t p)
{
  point_fprintf_fmt(stdout, dim, p, DEFAULT_COORD_FMT);
}

void point_fprintf(FILE *f, int dim, point_t p)
{
  point_fprintf_fmt(f, dim, p, DEFAULT_COORD_FMT);
}

int point_snprintf(size_t buflen, char *buf, int dim, point_t p)
{
  return point_snprintf_fmt(buflen, buf, dim, p, DEFAULT_COORD_FMT);
}

void point_printf_fmt(int dim, point_t p, const char *coord_fmt)
{
  point_fprintf_fmt(stdout, dim, p, coord_fmt);
}

void point_fprintf_fmt(FILE *fp, int dim, point_t p, const char *coord_fmt)
{
  int i;

  fputc('(', fp);
  for(i = 0; i < dim; i++) {
    if(i != 0) {
      fputc(',', fp);
      fputc(' ', fp);
    }

    fprintf(fp, coord_fmt, (float)p[i]);
  }

  fputc(')', fp);
}

int point_snprintf_fmt(
  size_t buflen, char *buf,
  int dim, point_t p,
  const char *coord_fmt)
{
  int i, c;

  if(buflen == 0) {
    return 0;
  }

  buflen--;

  // (
  c = 0;
  if(c < buflen) {
    buf[c++] = '(';
  }

  for(i = 0; i < dim && c < buflen; i++) {
    if(i != 0) {
      if(c < buflen) {
        buf[c++] = ',';
      }
      if(c < buflen) {
        buf[c++] = ' ';
      }
    }

    if(c < buflen) {
      c += snprintf(buf + c, buflen - c, coord_fmt, (float)p[i]);
    }
  }

  // )
  if(c < buflen) {
    buf[c++] = ')';
  }

  buf[c] = '\0';

  return c;
}
